#ifndef _LINUX_NOOKS_H
#define _LINUX_NOOKS_H

#include <linux/config.h>
#include <asm/page.h>
#include <linux/nooks-trap.h>

#define pte_pageindex(x)  ((unsigned long)(((x).pte_low >> PAGE_SHIFT)))
#define pte_pageprot(x)	  ((unsigned long)(((x).pte_low & ~_PAGE_CHG_MASK)))
#if 0
void nooks_update_pte(pte_t * pte, unsigned long vaddr);
void nooks_update_pte_phys(pte_t * pte, unsigned long vaddr);
void nooks_update_pmd(pmd_t * pte, unsigned long vaddr);
#else
void nooks_update_pte(void * pte, unsigned long vaddr);
void nooks_update_pte_phys(void * pte, unsigned long vaddr);
void nooks_update_pmd(void * pmd, unsigned long vaddr);
#endif

struct nook;
struct nook_kernel_stack_frame;
struct nook_stack_frame;

void *
nooks_context_switch_in(struct nook * nook, int callreturn);

struct nook * 
nooks_context_switch_out(int callback);

unsigned long 
nooks_get_free_pages(unsigned int gfp_mask, unsigned int order);

void 
nooks_free_pages(unsigned long addr, unsigned int order);

unsigned long 
nooks_mm_get_free_pages(struct nook * nook, unsigned int gfp_mask, unsigned int order);

void 
nooks_mm_free_pages(struct nook * nook, unsigned long addr, unsigned int order);

struct page;
void nooks_set_page_slab (struct page * page, int gfp_order, void * cachep, void * slabp);

struct pt_regs;

int
nooks_fault(int fault, struct pt_regs * pt_regs);

struct tss_struct;
int
nooks_stack_fault(struct tss_struct * tss);

struct sk_buff;

void
nooks_mm_free_skb(struct nook * nook, unsigned int order, struct sk_buff * skb);

#define NOOKS_READONLY 0
#define NOOKS_WRITEABLE 1
int
nooks_mm_update_pagetable(struct nook * nook,
			  int writeable, 
			  unsigned long address, 
			  unsigned long size);


#define __nooks_call_return(type, res) \
do { \
	return (type) (res); \
} while (0)

/* XXX - _foo needs to be __foo, while __NR_bar could be _NR_bar. */
#define _nooks_call0(type,name) \
type name(void) \
{ \
long __res; \
__asm__ volatile ("int $0x81" \
	: "=a" (__res) \
	: "0" (__NR_##name)); \
__nooks_call_return(type,__res); \
}

#define __NR_nooks_dummy_trap1  1
#define __NR_nooks_dummy_trap2  2
#define __NR_nooks_dummy_trap3  3
#define NR_nooks_calls 4


struct nook * 
nooks_find_nook(char * name);

int
nooks_get_user_ext(int *ret_gu, int *val_gu, void *ptr, size_t size);
int
nooks_put_user_check(unsigned long x, void * ptr, size_t size);

//
// Routines for referencing and dereferencing nooks for multithreaded
// safety
//


void nooks_enter_nook(struct nook * nook);
void nooks_exit_nook(void);
void nooks_dereference_nook(struct nook * nook);

#ifdef CONFIG_NOOKS_SHADOW
int nooks_is_shadow(struct nook *);
#endif

#ifndef _LINUX_MODULE_H
struct module;
#endif // _LINUX_MODULE_H

int nooks_init_nook_module(struct nook * nook, struct module * module);
int nooks_unregister_module(struct nook * nook, char * module_name);

int nooks_verify_waitqueue(struct nook * nook, void * waitqueue);


unsigned long nooks_dummy_trap1(void);
unsigned long nooks_dummy_trap2(void);
unsigned long nooks_illegal_inst_trap(void);
void nooks_do_interrupt(unsigned int irq);

//
// When compiling modules with nooks, use the trap for illegal instructions
//

//#if defined(CONFIG_NOOKS) && defined(__KERNEL__) && defined(MODULE)
//#define NOOKS_ILLEGAL_INST() nooks_illegal_inst_trap()
//#else
#define NOOKS_ILLEGAL_INST() 
//#endif // defined(CONFIG_NOOKS) && defined(__KERNEL__) && defined(MODULE)

#include <linux/wait.h>
struct __wait_queue;
int nooks_wake_up_common(struct __wait_queue * q, 
			 unsigned int mode, 
			 int nr_exclusive, 
			 const int sync);


struct task_struct;
int nooks_try_to_wake_up(struct task_struct * p, int synchronous);

void set_current_comment(char * comment);

#define TASK_IN_NOOK(_t_) (((_t_)->nook_state.curr.flags & NOOK_STATE_IN_NOOK) != 0)
#define IN_NOOK() (TASK_IN_NOOK(get_current()))

#define TASK_IN_SHADOW_RECOVERY(_t_) (((_t_)->nook_state.curr.flags & NOOK_STATE_SHADOW_RECOVERY) != 0)
#define IN_SHADOW_RECOVERY() (TASK_IN_SHADOW_RECOVERY(get_current())) //FIXME Asim

#ifdef CONFIG_NOOKS_NOPROT
#define INTERRUPT_IN_NOOK(_regs_) IN_NOOK()
#else
#define INTERRUPT_IN_NOOK(_regs_) ((_regs_)->nook_reg_pt != 0)
#endif // CONFIG_NOOKS_NOPROT

#endif /* _LINUX_NOOKS_H */

