#ifndef _LINUX_NOOKS_MM_H
#define _LINUX_NOOKS_MM_H


static inline pte_t pte_mklocal(pte_t pte)	{ (pte).pte_low &= ~_PAGE_GLOBAL; return pte; }

#define NOOKS_PAGE_KERNEL_RO __pgprot(__PAGE_KERNEL_RO)
#define NOOKS_PAGE_KERNEL __pgprot(__PAGE_KERNEL)


extern struct mm_struct nooks_mm_struct;

void * nooks_local_kmalloc(size_t size, int flags);
    
void * nooks_local_kmem_cache_alloc(kmem_cache_t * cachep, int flags);
    
kmem_cache_t *nooks_local_kmem_cache_create(const char * name, 
					    size_t size, 
					    size_t offset, 
					    unsigned long flags,
					    void (*ctor)(void *, 
							 kmem_cache_t *, 
							 unsigned long),
					    void (*dtor)(void *, 
							 kmem_cache_t *, 
							 unsigned long));

int nooks_local_kmem_cache_destroy(kmem_cache_t * cachep);
    
void nooks_local_kmem_cache_free(kmem_cache_t * cachep, void *p);

void nooks_local_kfree(const void * addr);

void * nooks_mm_kmalloc (pnook_t nook, size_t size, int flags);

void nooks_mm_kfree (pnook_t nook, const void *objp);

void * nooks_mm_vmalloc (pnook_t nook, 
			 unsigned long size, 
			 int flags, 
			 pgprot_t prot);

void nooks_mm_vfree (pnook_t nook,  void *objp);


int
nooks_mm_init_nook(pnook_t nook);

void
nooks_mm_cleanup_nook(pnook_t nook);

kmem_cache_t *
nooks_mm_kmem_cache_create(const char * name, 
			   size_t size, 
			   size_t offset, 
			   unsigned long flags,
			   void (*ctor)(void *, kmem_cache_t *, 
					unsigned long),
			   void (*dtor)(void *, kmem_cache_t *, 
					unsigned long));

int
nooks_mm_kmem_cache_destroy(kmem_cache_t * cache);

int
nooks_mm_release_pages(pnook_t nook, unsigned long addr, unsigned long size);

void 
nooks_mm_free_pages(pnook_t nook, unsigned long addr, unsigned int order);


unsigned long 
nooks_mm_get_free_pages(pnook_t nook, 
			unsigned int gfp_mask, 
			unsigned int order);


int
nooks_mm_free_page_map(pnook_t nook);

int
nooks_add_resource_elem(pnook_t nook, 
			nooks_resource_type_t restype, 
			void * resource,
			void * value);

int
nooks_del_resource_elem(pnook_t nook, 
			nooks_resource_type_t restype, 
			void * resource,
			void ** value);

int
nooks_find_resource_elem(pnook_t nook, 
			 nooks_resource_type_t restype, 
			 void * resource,
			 void ** value);



int
nooks_del_resource_elem_reverse(pnook_t nook, 
				nooks_resource_type_t restype,
				void * value,
				void ** resource);
int
nooks_find_resource_elem_reverse(pnook_t nook, 
				 nooks_resource_type_t restype,
				 void * value,
				 void ** resource);

#define NOOKS_MM_SKB_POOL_SIZE 256

struct sk_buff *
nooks_mm_alloc_skb(unsigned int order, int gfp_mask);

struct sk_buff *
nooks_alloc_skb(pnook_t nook, unsigned int order);

int
nooks_mm_move_skb_to_nook(pnook_t nook, struct sk_buff * skb);

int
nooks_mm_move_skb_to_kernel(pnook_t nook, struct sk_buff * skb);

void
nooks_mm_free_skb_pool(pnook_t nook);

#define NOOKS_MM_KMALLOC_OVERHEAD 8

void * nooks_dbg_kmalloc(size_t size, int flags);

void nooks_dbg_kfree(const void * ptr);

#define nooks_compare_mm(_o_, _t_, _f_) 
// #define nooks_compare_mm(_o_, _t_, _f_) nooks_compare_mm_dbg(_o_, _t_, _f_)

int
nooks_compare_mm_dbg(struct mm_struct * mm_one, 
			 struct mm_struct * mm_two, 
			 int flags);

void 
nooks_dump_mm(struct mm_struct * mm);

void 
nooks_protect_mm(struct mm_struct * mm, 
		 int start, 
		 int count);

void
nooks_protect_pt_mm(struct mm_struct * mm);

void 
nooks_use_other_pgd(struct mm_struct * mm_dest, 
		    struct mm_struct * mm_src, 
		    int start, 
		    int count);

extern int nooks_read_only;
extern int nooks_private_pt;

unsigned long
nooks_mm_get_pte_flags(pnook_t nook, void * vaddr);

unsigned long
nooks_mm_get_pte(pnook_t nook, void * vaddr);


void
nooks_mm_init_lookaside(pnooks_lookaside_list_t list,
			int highwater,
			int size,
			int flags);

void *
nooks_mm_lookaside_alloc(pnooks_lookaside_list_t list);

void
nooks_mm_lookaside_free(pnooks_lookaside_list_t list, void * ptr);

void
nooks_mm_free_lookaside(pnook_t nook, pnooks_lookaside_list_t list);

void *
nooks_kw_malloc(size_t size, int flags);

void
nooks_kw_free(void * ptr);

int
nooks_mm_add_pages(pnook_t nook, 
		   unsigned long address,
		   unsigned int gfp_mask, 
		   unsigned long size);
int
nooks_mm_release_pages(pnook_t nook, unsigned long addr, unsigned long size);


void
nooks_update_addr_prot(struct mm_struct * mm,
		       void * pte_addr);

void
nooks_update_pmd_set_prot(struct mm_struct * mm,
			  pmd_t * pmd, 
			  unsigned long address);


#endif /* _LINUX_NOOKS_MM_H */
