/*
 * nooks-dwrappers.c -- wrappers for driver functions
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  No warranty 
 * is attached; * we cannot take responsibility for errors or 
 * fitness for use.
 *
 */

 
#include "nooks-i.h"
#include "nooks-dwrappers.h"
#include "nooks-iwrappers.h"
#include "nooks-recover.h"
#include "nooks-workers.h"
#include <linux/module.h>
#define NOOKS_USE_DRIVER_DEV

#if 0
#ifdef NOOKS_USE_DRIVER_DEV
#define NOOK_DWRAPPER_GET_NETDEV(_dev_) ((struct net_device *) dev)
#else
#define NOOK_DWRAPPER_GET_NETDEV(_dev_) ((struct net_device *) dev)->priv
#endif
#endif

#define NOOK_DWRAPPER_GET_NETDEV(_dev_) ((struct net_device *) dev)

/*
 * struct_cpy(x,y), copy structure *x into (matching structure) *y.
 *
 * We get link-time errors if the structure sizes do not match.
 * There is no runtime overhead, it's all optimized away at
 * compile time.
 */
extern void __struct_cpy_bug (void);

#define struct_cpy(x,y)                         \
({                                              \
        if (sizeof(*(x)) != sizeof(*(y)))       \
                __struct_cpy_bug;               \
        memcpy(x, y, sizeof(*(x)));             \
})

/*
int
nooks_sync_file(pnook_t nook,
                  struct file * file,
                  struct file ** nooks_file)
{
  int result = 0;

  //
  // Find the file
  //

  if (*nooks_file == NULL) {
    if (GET_NOOK_ARG(NOOKS_FILP_ARG) == (unsigned long) file) {
      *nooks_file = (struct file *) GET_NOOK_ARG(NOOKS_NOOK_FILP_ARG);
    } else {

      result = nooks_find_resource_elem(GET_CURRENT_NOOK(),
                                        file_e,
                                        file,
                                        (void **) nooks_file);
      if (result < 0) {
              NOOKS_FAULT(resource);
                return(result);
      }
    }
  }

  // Copy the file to the nook
  //

  result = nooks_copy_file(nook, *nooks_file, file, 1);

  return(result);
}
*/
  
/**********************************************************
 *
 * Module operations wrappers
 *
 *********************************************************/

int NOOKS_DWRAPPER(mops_init)(void)
{ 
  int result;
  NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
  printk ("Calling init in dwrapper.\n");
  result = NOOKS_DRIVER_CALL0(mops.init);
  NOOKS_LEAVE_DRIVER();

  NOOKS_DWRAPPER_EXIT(); 
  return(result);
}

void NOOKS_DWRAPPER(mops_cleanup)(void)
{ 
  NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
  NOOKS_DRIVER_CALL0(mops.exit);
  NOOKS_LEAVE_DRIVER();

  NOOKS_DWRAPPER_EXIT(); 
  return;
}

/*
int NOOKS_DWRAPPER(mops_can_unload)(void)
{ 
  int result;

  NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
  result = NOOKS_DRIVER_CALL0(mops.can_unload);
  NOOKS_LEAVE_DRIVER();

  NOOKS_DWRAPPER_EXIT(); 
  return(result);
}
*/
struct module_operations NOOKS_DWRAPPER(mops) = {
  init: NOOKS_DWRAPPER(mops_init),
  exit: NOOKS_DWRAPPER(mops_cleanup)
  //,
  //can_unload: NOOKS_DWRAPPER(mops_can_unload)
};

/**********************************************************
 *
 * ProcFs operations wrappers
 *
 *********************************************************/

int NOOKS_DWRAPPER(procops_read_proc)(char *page, 
			      char **start, 
			      off_t off,
			      int count, 
			      int *eof, 
			      void *data)
{
	int result = 0;
	int * nooks_eof;
	char ** nooks_start;

	char * nooks_page = NULL;

	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_DWRAPPER_ENTER();

	/* 1. CREATE NOOK RESOURCE */
	// Allocate a page that the nook can scribble on

	nooks_page = (char *) nooks_mm_get_free_pages(nook, GFP_KERNEL, 0);
	if (nooks_page != NULL) {
		// allocate space for the start address and eof
		//
		nooks_start = nooks_dwrapper_stack_alloc(2*sizeof(void *));
		if (nooks_start != NULL) {
			*nooks_start = *start;
			nooks_eof = (int *) (nooks_start + 1);
			*nooks_eof = *eof;

	/* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
			result = NOOKS_DRIVER_CALL6(procops.read_proc,
					nooks_page, nooks_start, off,
					count, nooks_eof, data);
  NOOKS_LEAVE_DRIVER();



	/* 3. COPY from the nooks instance to the kernel structures */
			if (result > 0) {
				memcpy(page, nooks_page, result);
			}
			*eof = *nooks_eof;
			*start = *nooks_start;

	/* 4. RELEASE NOOK RESOURCE */
			nooks_dwrapper_stack_release();
		}
		nooks_mm_free_pages(nook, (unsigned long) nooks_page, 0);
	}

	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}


int NOOKS_DWRAPPER(procops_write_proc)(struct file *file, 
			       const char *buffer,
			       unsigned long count, 
			       void *data)
{
  int result;
  NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
  result = NOOKS_DRIVER_CALL4(procops.write_proc, 
			      file,
			      buffer,
			      count,
			      data);
  NOOKS_LEAVE_DRIVER();

  NOOKS_DWRAPPER_EXIT(); 
  return(result);
}

int NOOKS_DWRAPPER(procops_get_info)(char * buffer, 
			     char ** start, 
			     off_t off, 
			     int len)
{
  int result;
  NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
  result = NOOKS_DRIVER_CALL4(procops.get_info,
			      buffer,
			      start,
			      off,
			      len);
  NOOKS_LEAVE_DRIVER();

  NOOKS_DWRAPPER_EXIT(); 
  return(result);
}

struct procfs_operations NOOKS_DWRAPPER(procops) = {
  read_proc: NOOKS_DWRAPPER(procops_read_proc),
  write_proc: NOOKS_DWRAPPER(procops_write_proc),
  get_info: NOOKS_DWRAPPER(procops_get_info)
};

/**********************************************************
 *
 * net device operations wrappers
 *
 *********************************************************/

int NOOKS_DWRAPPER(netops_init)(struct net_device *dev)
{
#ifndef NOOKS_USE_DRIVER_DEV
	pnook_driver_t driver = GET_CURRENT_NOOK_DRIVER();
#endif
	int result;
	pnook_t nook = GET_CURRENT_NOOK();
	
	/* function parameter */
	struct net_device temp_device;
	struct net_device * nook_dev = NULL;

	NOOKS_DWRAPPER_ENTER();

	printk ("dwrapper: Init called.\n");
	
  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL1(netops.init, dev);
  NOOKS_LEAVE_DRIVER();
	printk ("dwrapper:Done with init.\n");
	
 Cleanup:
	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

struct net_device_stats * NOOKS_DWRAPPER(netops_get_stats)(struct net_device *dev)
{
	int result;
	pnook_t nook = GET_CURRENT_NOOK();

	if (nook == NULL)
	{
		printk ("Fatal: NOOK is NULL");
		return -ENOMEM;
	}
	else {;}
		//printk ("Nook obtained is %s for device %p.\n", nook->name, dev);
	
	/* function parameter */
	struct net_device * nook_dev;
	
	/* return parameter */
	struct net_device_stats * temp_stats;
	struct net_device_stats * kernel_stats = NULL;

	NOOKS_DWRAPPER_ENTER();
  
	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */
	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);

		/* find the kernel instance of the stats structure */
	result = nooks_find_resource_elem(nook, netdev_stats_e, dev,
					  (void **) &kernel_stats);
	if (result < 0) {
		NOOKS_FAULT(resource);
		goto Cleanup;
	}

	/* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
	kernel_stats = (struct net_device_stats *)
			NOOKS_DRIVER_CALL1(netops.get_stats,nook_dev);
//	temp_stats = (struct net_device_stats *)
//			NOOKS_DRIVER_CALL1(netops.get_stats, nook_dev);
  NOOKS_LEAVE_DRIVER();
	goto Cleanup;
	if (IS_ERR(temp_stats)) {
		kernel_stats = NULL;
		goto Cleanup;
	}

	/* 3 COPY from the nooks instance to the kernel structure */ 
	result = nooks_memcpy((void *)kernel_stats, temp_stats,
				sizeof(struct net_device_stats));
	if (result < 0) {
		PERROR("  failed to copy netdevice from %p\n",nook_dev);
		goto Cleanup;
	}
	
	/* 4. CHECK */
	
Cleanup:
	NOOKS_DWRAPPER_EXIT(); 
	return((struct net_device_stats *)kernel_stats);
}

struct iw_statistics * NOOKS_DWRAPPER(netops_get_wireless_stats)(struct net_device *dev)
{
	struct iw_statistics * result;
  
	/* function parameter */
	struct net_device * nook_dev;
   
	NOOKS_DWRAPPER_ENTER();  

	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */
	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);

	/* 2. DRIVER CALL */

  NOOKS_ENTER_DRIVER();
	result = (struct iw_statistics *) NOOKS_DRIVER_CALL1(netops.get_wireless_stats, nook_dev);
  NOOKS_LEAVE_DRIVER();

	if (IS_ERR(result)) {
		result = NULL;
	}

	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}



void NOOKS_DWRAPPER(netops_uninit)(struct net_device *dev)
{
	/* function parameter */
	struct net_device * nook_dev;
   
	NOOKS_DWRAPPER_ENTER();  

	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */
	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);

	/* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
	NOOKS_DRIVER_CALL1(netops.uninit, nook_dev);
  NOOKS_LEAVE_DRIVER();

	NOOKS_DWRAPPER_EXIT(); 
}

void NOOKS_DWRAPPER(netops_destructor)(struct net_device *dev)
{
	/* function parameter */
	struct net_device * nook_dev;
   
	NOOKS_DWRAPPER_ENTER();  

	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */
	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);

	/* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
	NOOKS_DRIVER_CALL1(netops.destructor, nook_dev);
  NOOKS_LEAVE_DRIVER();

	NOOKS_DWRAPPER_EXIT(); 
}

int NOOKS_DWRAPPER(netops_open)(struct net_device *dev)
{
	int result;

 	/* function parameter */
	struct net_device temp_device;
	struct net_device * nook_dev;
   
	NOOKS_DWRAPPER_ENTER();  

	printk ("In netops_open dwrapper.\n");

	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */
	
	//nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);

	/* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL1(netops.open, dev);
  NOOKS_LEAVE_DRIVER();

	printk("dwrappers: done with netops_open.\n");

	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

int NOOKS_DWRAPPER(netops_stop)(struct net_device *dev)
{
	int result;

 	/* return parameter */
	struct net_device * nook_dev;

	NOOKS_DWRAPPER_ENTER();
  
	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */
	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);

	/* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL1(netops.stop, nook_dev);
  NOOKS_LEAVE_DRIVER();

	// TODO

	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

int NOOKS_DWRAPPER(netops_hard_start_xmit) (struct sk_buff *skb,
					    struct net_device *dev)
{
	int result = 0;
	pnook_t nook = GET_CURRENT_NOOK();

	/* return parameter */
	struct net_device * nook_dev;

	//printk ("In hard start xmit.\n");
	result = nooks_mm_move_skb_to_nook(nook, skb);
	if (result < 0) 
		printk ("hard_start_xmit: moving skb to nook fails.\n");
	/* 2. DRIVER CALL */
	result = NOOKS_DRIVER_CALL2(netops.hard_start_xmit, skb, dev);

	if (result < 0) 
		nooks_mm_move_skb_to_kernel(nook, skb);
	
	return(result);
}

/**********************************************************
 *
 * Function: netops_hard_header
 * 
 * Type:
 * 
 * Notes: need to unprotect skb->data, and copy skb header
 * to allow pointers to change.
 *
 *********************************************************/

int NOOKS_DWRAPPER(netops_hard_header) (struct sk_buff *skb,
					struct net_device *dev,
					unsigned short type,
					void *daddr,
					void *saddr,
					unsigned len)
{
	int result = 0;
	/* function parameter */
	struct net_device * nook_dev;

	NOOKS_DWRAPPER_ENTER();
  
	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */
	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);

	/* 2. DRIVER CALL */
	printk ("dwrapper: netops_hard_header.\n");
  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL6(netops.hard_header, 
			skb,
			nook_dev,
			type,
			daddr,
			saddr,
			len);
  NOOKS_LEAVE_DRIVER();

	printk ("dwrapper: Exit netops_hard_header.\n");
	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

/* skb is used READ-ONLY */
int NOOKS_DWRAPPER(netops_rebuild_header)(struct sk_buff *skb)
{
  int result;
  NOOKS_DWRAPPER_ENTER();
  
  NOOKS_ENTER_DRIVER();
  result = NOOKS_DRIVER_CALL1(netops.rebuild_header, skb);
  NOOKS_LEAVE_DRIVER();
  
  NOOKS_DWRAPPER_EXIT(); 
  return(result);
}

void NOOKS_DWRAPPER(netops_set_multicast_list)(struct net_device *dev)
{
	/* function parameter */
	struct net_device * nook_dev;
	
	NOOKS_DWRAPPER_ENTER();
  
	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */
	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);

	/* 2. DRIVER CALL */

	//printk ("dwrapper: set_multicast_list start.\n");
  NOOKS_ENTER_DRIVER();
	NOOKS_DRIVER_CALL1(netops.set_multicast_list, dev);
  NOOKS_LEAVE_DRIVER();
	//printk ("dwrapper: set_multicast_list end.\n");
	NOOKS_DWRAPPER_EXIT(); 
}

int NOOKS_DWRAPPER(netops_set_mac_address)(struct net_device *dev,
					   void *addr)
{
	int result;

	/* function parameter */
	struct net_device temp_device;
	struct net_device * nook_dev;
	
	NOOKS_DWRAPPER_ENTER();
  
	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */

	printk ("dwrapper: netops_set_mac_address.\n");	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);

	/* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL2(netops.set_mac_address, dev, addr);
  NOOKS_LEAVE_DRIVER();

	printk ("dwrapper exit: netops_set_mac_address.\n");
	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

int NOOKS_DWRAPPER(netops_do_ioctl)(struct net_device *dev,
				    struct ifreq *ifr, 
				    int cmd)
{
	int result;
	
	/* function parameter */
	struct net_device * nook_dev;
	struct ifreq * nook_ifreq;
	
	NOOKS_DWRAPPER_ENTER();
 	
	printk ("In netops_do_ioctl.\n"); 
	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */
	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);


	nook_ifreq = nooks_dwrapper_stack_alloc(sizeof(struct ifreq));
	if (nook_ifreq != NULL) {
	  nooks_memcpy_out(nook_ifreq, ifr, sizeof(struct ifreq));
	  /* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
	  result = NOOKS_DRIVER_CALL3(netops.do_ioctl,
				      nook_dev, 
				      nook_ifreq, 
				      cmd);
  NOOKS_LEAVE_DRIVER();

	  nooks_memcpy(ifr, nook_ifreq, sizeof(struct ifreq));
	  nooks_dwrapper_stack_release();
	} else {
	  result = -ENOMEM;
	}

	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

int NOOKS_DWRAPPER(netops_set_config)(struct net_device *dev,
				      struct ifmap *map)
{
	int result;

	/* function parameter */
	struct net_device * nook_dev;
	
	NOOKS_DWRAPPER_ENTER();
  
	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */

	printk ("dwrapper:setconfig called.\n");	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);

	/* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL2(netops.set_config, nook_dev, map);
  NOOKS_LEAVE_DRIVER();
	printk ("dwrapper:End of set_config.\n");


	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

int NOOKS_DWRAPPER(netops_hard_header_cache)(struct neighbour *neigh,
					     struct hh_cache *hh)
{
  int result;
  NOOKS_DWRAPPER_ENTER();
  
  printk ("dwrapper enter: netops_hard_header_cache.\n");
  NOOKS_ENTER_DRIVER();
  result = NOOKS_DRIVER_CALL2(netops.hard_header_cache, neigh, hh);
  NOOKS_LEAVE_DRIVER();

  printk ("dwrapper exit: netops_hard_header_cache.\n");
  NOOKS_DWRAPPER_EXIT(); 
  return(result);
}

void NOOKS_DWRAPPER(netops_header_cache_update)(struct hh_cache *hh,
						struct net_device *dev,
						unsigned char *  haddr)
{
	/* function parameter */
	struct net_device * nook_dev;
	
	NOOKS_DWRAPPER_ENTER();
  
	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */
	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);

	printk ("dwrapper enter: netops_header_cache_update.\n");
	/* 2. DRIVER CALL */
	
  NOOKS_ENTER_DRIVER();
	NOOKS_DRIVER_CALL3(netops.header_cache_update, 
			hh,
			nook_dev, 
			haddr);
  NOOKS_LEAVE_DRIVER();
	
	printk ("dwrapper exit: netops_header_cache_update.\n");
	NOOKS_DWRAPPER_EXIT(); 
}

int NOOKS_DWRAPPER(netops_change_mtu)(struct net_device *dev, int new_mtu)
{
	int result;
	int nook_mtu;

	/* function parameter */
	struct net_device * nook_dev;
	
	NOOKS_DWRAPPER_ENTER();
  
	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */
	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);

	/* 2. DRIVER CALL */
	
  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL2(netops.change_mtu, nook_dev, new_mtu);
  NOOKS_LEAVE_DRIVER();


	/* 3 COPY AND CHECK */
	if (nooks_memcpy_out(&nook_mtu, &nook_dev->mtu, sizeof(int)) >= 0) {
		if (nook_mtu < new_mtu) {
			dev->mtu = nook_mtu;
		} else {
			dev->mtu = new_mtu;
		}
	}
	if (dev->mtu >= 65535) {
		dev->mtu = 65535;
	}

	// this may not always be right, but we will assume that it is
	//

	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

void NOOKS_DWRAPPER(netops_tx_timeout) (struct net_device *dev)
{
	//unsigned long nook_trans_start;

	/* function parameter */
	struct net_device * nook_dev;
	
	NOOKS_DWRAPPER_ENTER();
  
	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */
	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);

	/* 2. DRIVER CALL */
	
  NOOKS_ENTER_DRIVER();
	NOOKS_DRIVER_CALL1(netops.tx_timeout, nook_dev);
  NOOKS_LEAVE_DRIVER();

#if 0
	// TODO
	nooks_memcpy(&nook_trans_start, &(nook_dev)->trans_start, sizeof(unsigned long));
	if (nook_trans_start > jiffies) {
		dev->trans_start = jiffies;
	} else {
		dev->trans_start = nook_trans_start;
	}
#endif
	NOOKS_DWRAPPER_EXIT(); 
}

int NOOKS_DWRAPPER(netops_hard_header_parse)(struct sk_buff *skb,
					     unsigned char *haddr)
{
  int result;
  NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
  result = NOOKS_DRIVER_CALL2(netops.hard_header_parse,
			      skb,
			      haddr);
  NOOKS_LEAVE_DRIVER();

  // TODO result
  NOOKS_DWRAPPER_EXIT(); 
  return(result);
}

int NOOKS_DWRAPPER(netops_neigh_setup)(struct net_device *dev, 
				       struct neigh_parms * parms)
{
	int result;

	/* function parameter */
	struct net_device * nook_dev;
	
	NOOKS_DWRAPPER_ENTER();
  
	/* 1. LOOKUP RESOURCE */
		/* find the nooks instance of the net device */
	
	nook_dev = NOOK_DWRAPPER_GET_NETDEV(dev);
	printk ("dwrapper enter: netops_neigh_setup.\n");
	/* 2. DRIVER CALL */
	
  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL2(netops.neigh_setup, nook_dev, parms);
  NOOKS_LEAVE_DRIVER();
	printk ("dwrapper exit: netops_neigh_setup.\n");
	// TODO result

	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

int NOOKS_DWRAPPER(e1000_clean)(struct net_device *poll_dev, int *budget)
{
	int result;

	NOOKS_DWRAPPER_ENTER();
	NOOKS_ENTER_DRIVER();
        result = NOOKS_DRIVER_CALL2(netops.e1000_clean, poll_dev, budget);
	NOOKS_LEAVE_DRIVER();
	NOOKS_DWRAPPER_EXIT();
	return (result);
}



struct netdev_operations NOOKS_DWRAPPER(netdev_ops) = {
  init: NOOKS_DWRAPPER(netops_init),
  get_stats:  NOOKS_DWRAPPER(netops_get_stats),
  get_wireless_stats: NOOKS_DWRAPPER(netops_get_wireless_stats),
  uninit: NOOKS_DWRAPPER(netops_uninit),
  destructor: NOOKS_DWRAPPER(netops_destructor),
  open: NOOKS_DWRAPPER(netops_open),
  stop: NOOKS_DWRAPPER(netops_stop),
  hard_start_xmit: NOOKS_DWRAPPER(netops_hard_start_xmit),
  hard_header: NOOKS_DWRAPPER(netops_hard_header),
  rebuild_header: NOOKS_DWRAPPER(netops_rebuild_header),
  set_multicast_list: NOOKS_DWRAPPER(netops_set_multicast_list),
  set_mac_address: NOOKS_DWRAPPER(netops_set_mac_address),
  do_ioctl: NOOKS_DWRAPPER(netops_do_ioctl),
  set_config: NOOKS_DWRAPPER(netops_set_config),
  hard_header_cache: NOOKS_DWRAPPER(netops_hard_header_cache),
  header_cache_update: NOOKS_DWRAPPER(netops_header_cache_update),
  change_mtu: NOOKS_DWRAPPER(netops_change_mtu),
  tx_timeout: NOOKS_DWRAPPER(netops_tx_timeout),
  hard_header_parse: NOOKS_DWRAPPER(netops_hard_header_parse),
  neigh_setup: NOOKS_DWRAPPER(netops_neigh_setup),
  //accept_fastpath: NOOKS_DWRAPPER(netops_accept_fastpath)
  e1000_clean: NOOKS_DWRAPPER(e1000_clean), 
};

/**********************************************************
 *
 * irq operations wrappers
 *
 *********************************************************/

void NOOKS_DWRAPPER(irqops_handler)(int irq, 
				    void * dev_id, 
				    struct pt_regs * regs)
{
  NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
  NOOKS_DRIVER_CALL3(irqops.handler,
		     irq, 
		     dev_id, 
		     regs);
  NOOKS_LEAVE_DRIVER();

  NOOKS_DWRAPPER_EXIT(); 
}

struct irq_operations NOOKS_DWRAPPER(irq_ops) = {
  handler: NOOKS_DWRAPPER(irqops_handler)
};

/**********************************************************
 *
 * pci operations wrappers
 *
 *********************************************************/
#if 0
int NOOKS_DWRAPPER(pciops_prepare)(struct pci_dev *dev)
{
	int result;
	NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL1(pciops.prepare, dev);
  NOOKS_LEAVE_DRIVER();

	// TODO result
	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

int NOOKS_DWRAPPER(pciops_activate)(struct pci_dev *dev)
{
	int result;
	NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL1(pciops.activate, dev);
  NOOKS_LEAVE_DRIVER();

	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

int NOOKS_DWRAPPER(pciops_deactivate)(struct pci_dev *dev)
{
	int result;
	NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL1(pciops.deactivate, dev);
  NOOKS_LEAVE_DRIVER();

	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}
#endif

int  NOOKS_DWRAPPER(pciops_probe)(struct pci_dev *dev, 
		const struct pci_device_id *id)
{
	pnook_t nook = GET_CURRENT_NOOK();
	int result = 0;
	NOOKS_DWRAPPER_ENTER();


	/* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
			result = NOOKS_DRIVER_CALL2(pciops.probe, dev, id);
  NOOKS_LEAVE_DRIVER();


	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

void NOOKS_DWRAPPER(pciops_shutdown) (struct pci_dev *dev)
{
 /* FIXME Empty wrapper fill in late when writing dops -- Currently just a copy of remove. */
 pnook_t nook = GET_CURRENT_NOOK();
 int result = 0;
 NOOKS_DWRAPPER_ENTER();

 printk ("In pciops_shutdown.\n");
        /* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
                NOOKS_DRIVER_CALL1(pciops.shutdown, dev);
  NOOKS_LEAVE_DRIVER();

        /* 4. RELEASE NOOK RESOURCE */
                //nooks_del_resource_elem(nook, pcidev_e, dev, NULL);
                //nooks_kw_free(l_dev);

        NOOKS_DWRAPPER_EXIT();

}

void NOOKS_DWRAPPER(pciops_remove) (struct pci_dev *dev)
{
	pnook_t nook = GET_CURRENT_NOOK();
	int result = 0;
	static freed_device = NULL;
	/* function parameter */
	NOOKS_DWRAPPER_ENTER();
#if NOOKS_MIGRATE
	printk ("MODIFIED AND BROKEN: In pciops_remove.\n");
	pci_disable_device(dev);
	return;
#endif
	/* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
	if (freed_device != dev)	{ //BUGBUGBUG Dirty hack to preven device from being freed twice.
		NOOKS_DRIVER_CALL1(pciops.remove, dev);
		freed_device = dev;
	}
	
  NOOKS_LEAVE_DRIVER();


	NOOKS_DWRAPPER_EXIT(); 
}

int  NOOKS_DWRAPPER(pciops_suspend)(struct pci_dev *dev, u32 state)
{
	int result;
	NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL2(pciops.suspend, dev, state);
  NOOKS_LEAVE_DRIVER();

	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

int  NOOKS_DWRAPPER(pciops_resume) (struct pci_dev *dev)
{
  int result;
  NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
  result = NOOKS_DRIVER_CALL1(pciops.resume, dev);
  NOOKS_LEAVE_DRIVER();

  NOOKS_DWRAPPER_EXIT(); 
  return(result);
}


int  NOOKS_DWRAPPER(pciops_enable_wake) (struct pci_dev *dev, 
					 u32 state, 
					 int enable)
{
  int result;
  NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
  result = NOOKS_DRIVER_CALL3(pciops.enable_wake, 
			      dev,
			      state,
			      enable);
  NOOKS_LEAVE_DRIVER();

  NOOKS_DWRAPPER_EXIT(); 
  return(result);
}


struct pci_operations NOOKS_DWRAPPER(pci_ops) = {
	//prepare: NOOKS_DWRAPPER(pciops_prepare),
	//activate: NOOKS_DWRAPPER(pciops_activate),
	//deactivate: NOOKS_DWRAPPER(pciops_deactivate),
	remove: NOOKS_DWRAPPER(pciops_remove),
	shutdown:NOOKS_DWRAPPER(pciops_shutdown),
	suspend: NOOKS_DWRAPPER(pciops_suspend),
	resume: NOOKS_DWRAPPER(pciops_resume),
	enable_wake: NOOKS_DWRAPPER(pciops_enable_wake)
};

/**********************************************************
 *
 * callback operations wrappers
 *
 *********************************************************/

int NOOKS_DWRAPPER(callbackops_handler)(unsigned long arg)
{
	int result;
	void * driver_data;
	pnook_driver_t driver = GET_CURRENT_NOOK_DRIVER();
	
	pnooks_tasklet_t tasklet = NULL;
	
	NOOKS_DWRAPPER_ENTER();

	/* 1. LOOKUP NOOK RESOURCE */
	driver_data = driver->data;
	if (driver_data == NOOKS_CALLBACK_TASKLET) {

		tasklet = list_entry(arg, nooks_tasklet_t, t);
		//    tasklet = (pnooks_tasklet_struct_t) arg;
		arg = tasklet->data;
		// Copy out the state to the nooks tasklet
		nooks_memcpy_out(&tasklet->nt->state, 
				 &tasklet->t.state, 
				 sizeof(unsigned long));
	}

	/* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL1(callbackops.handler,arg);
  NOOKS_LEAVE_DRIVER();


	// Delete the driver or tasklet after the call is done.
	//
	// BUGBUG: possible race condition if nook is shutdown before thread
	// returns.
	//

	if (driver_data == (void *) NOOKS_CALLBACK_THREAD) {
		nooks_delayed_free_driver(driver);
	} else if (driver_data == (void *) NOOKS_CALLBACK_TASKLET) {
		// Let the tasklet dangle - it might get used again
	}

	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}

int NOOKS_DWRAPPER(callbackops_pm_handler)(struct pm_dev *dev, pm_request_t rqst, void *data)
{
  int result;
  struct pm_dev * nook_dev = 0;
  pnook_t nook = GET_CURRENT_NOOK();

	
  NOOKS_DWRAPPER_ENTER();
  
  result = nooks_find_resource_elem(nook,
				    pm_dev_e,
				    dev,
				    (void **) &nook_dev);
  if (result < 0) {
    goto Cleanup;
  }

  NOOKS_ENTER_DRIVER();
  result = NOOKS_DRIVER_CALL3(callbackops.pm_handler,nook_dev, rqst, data);
  NOOKS_LEAVE_DRIVER();

 Cleanup:
  NOOKS_DWRAPPER_EXIT(); 
  return(result);
}

struct callback_operations  NOOKS_DWRAPPER(callback_ops) = {
  handler: NOOKS_DWRAPPER(callbackops_handler),
  pm_handler: NOOKS_DWRAPPER(callbackops_pm_handler)
};


/**********************************************************
 *
 * timer operations wrappers
 *
 *********************************************************/

void NOOKS_DWRAPPER(timerops_handler)(unsigned long arg)
{
  pnook_driver_t driver = NULL;
  pnook_t nook = GET_CURRENT_NOOK();
  struct timer_list * nook_timer = (struct timer_list *) arg;
  int result;

  NOOKS_DWRAPPER_ENTER();
  
  // The argument is the nook timer.
  //
  // BUGBUG: Unlink the timer
  //
  result = nooks_del_resource_elem(nook, timer_e, nook_timer, (void **) &driver);
  if (result >= 0 && nook_timer->data) {


    // Release the driver as well when we return

    nooks_delayed_free_driver(driver);
    // BUGBUG: unsafe derefernece of timer data. Put in driver instead?

    NOOKS_ENTER_DRIVER();
    NOOKS_DRIVER_CALL1(timerops.handler,nook_timer->data);
    NOOKS_LEAVE_DRIVER();

    // Driver data is the kernel timer structure

    if (driver->need_delete) {
      nooks_kfree(driver->data);
      driver->data = NULL;
    }
  } else {
    PERROR("Failed to find timer %p for nook %p for del_timer\n",
	   data1, nook);
    NOOKS_FAULT(parameter);
  }

  NOOKS_DWRAPPER_EXIT(); 
}

struct timer_operations  NOOKS_DWRAPPER(timer_ops) = {
  handler: NOOKS_DWRAPPER(timerops_handler)
};


/**********************************************************
 *
 * sysctl operations wrappers
 *
 *********************************************************/


int NOOKS_DWRAPPER(sysctlops_strategy)(ctl_table *ctl, int *name, 
		int nlen, void *oldval, 
		size_t *oldlenp, void *newval, 
		size_t newlen, void **context)
{
	int result;
	pnook_driver_t driver = (pnook_driver_t) ctl->data;
	ctl_table * table = (ctl_table *) driver->data;

	NOOKS_DWRAPPER_ENTER();

  NOOKS_ENTER_DRIVER();
	result = NOOKS_DRIVER_CALL8(sysctlops.strategy, table, name,
			nlen, oldval, oldlenp,
			newval, newlen, context);
  NOOKS_LEAVE_DRIVER();

	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}


int NOOKS_DWRAPPER(sysctlops_proc_handler)(ctl_table *ctl, int write, 
		struct file * filp,
		void *buffer, size_t *lenp, loff_t *ppos)
{
	int result = 0;
	size_t * l_lenp = NULL;
	struct file * l_filp = NULL;
	loff_t l_ppos;
	pnook_driver_t driver = (pnook_driver_t) ctl->data;
	ctl_table * table = (ctl_table *) driver->data;

	NOOKS_DWRAPPER_ENTER();

	/* 1. CREATE NOOK RESOURCE */
	l_filp = nooks_dwrapper_stack_alloc(sizeof(struct file) + sizeof(size_t));
	if (l_filp != NULL) {
		result = nooks_memcpy_out(l_filp, filp, sizeof(struct file));
		if (result < 0) {
			goto Cleanup;
		}
		l_lenp = (size_t *) (l_filp + 1);
		result = nooks_memcpy_out(l_lenp, lenp, sizeof(size_t));
		if (result < 0) {
			goto Cleanup;
		}

	/* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
		result = NOOKS_DRIVER_CALL6(sysctlops.proc_handler, table, write,
				l_filp, buffer, l_lenp, ppos);
  NOOKS_LEAVE_DRIVER();


	/* 3. COPY from the nooks instance to the kernel structures */
		if (nooks_memcpy(lenp, l_lenp, sizeof(size_t)) < 0) {
			if (result == 0) {
				result = -EFAULT;
				goto Cleanup;
			}
		}
		if (nooks_memcpy(&filp->f_pos, &l_filp->f_pos, sizeof(loff_t)) < 0) {
			if (result == 0) {
				result = -EFAULT;
				goto Cleanup;
			}
		}
	}

Cleanup:
	if (l_filp != NULL) {
		nooks_dwrapper_stack_release();
	}

	NOOKS_DWRAPPER_EXIT(); 
	return(result);

}


struct sysctl_operations NOOKS_DWRAPPER(sysctl_ops) = {
strategy: NOOKS_DWRAPPER(sysctlops_strategy),
proc_handler: NOOKS_DWRAPPER(sysctlops_proc_handler)
};


/**********************************************************
 *
 * actor operations wrappers
 *
 *********************************************************/

int 
NOOKS_DWRAPPER(actorops_read)(read_descriptor_t * desc, 
		struct page * page, 
		unsigned long offset, 
		unsigned long size)
{
	int result = -ENOMEM;
	read_descriptor_t * nook_desc = NULL;
	NOOKS_DWRAPPER_ENTER();

	/* 1. CREATE NOOK RESOURCE */
	nook_desc = nooks_dwrapper_stack_alloc(sizeof(read_descriptor_t));
	if (nook_desc != NULL) {
		struct_cpy(nook_desc, desc);

	/* 2. DRIVER CALL */
  NOOKS_ENTER_DRIVER();
		result = NOOKS_DRIVER_CALL4(actorops.read, nook_desc, page, offset, size);
  NOOKS_LEAVE_DRIVER();

	/* 3. COPY from the nooks instance to the kernel structures */
		struct_cpy(desc, nook_desc);

		nooks_dwrapper_stack_release();
	}

	NOOKS_DWRAPPER_EXIT(); 
	return(result);
}


struct actor_operations NOOKS_DWRAPPER(actor_ops) = {
read: NOOKS_DWRAPPER(actorops_read),
};

#if 0
#include "nooks-dwrappers-muthu.c"


#endif


#ifdef USE_GENERATED_FILES
#include "nooks-dwrap-c-gen.h"
#endif
