/*
 * nooks-iwrapper-gen_c.h -- iwrapper structure allocation function
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  No warranty
 * is attached; * we cannot take responsibility for errors or
 * fitness for use.
 */

#include <asm/pgtable.h>

#define NOOKS_IWRAPPER_GET_ITF(_x_) \
   case netops_e: return nooks_iwrapper_get_itf_netops_e(_x_);\
   case mops_e: return nooks_iwrapper_get_itf_mops_e(_x_);\
   case actorops_e: return nooks_iwrapper_get_itf_actorops_e(_x_);\
   case procops_e: return nooks_iwrapper_get_itf_procops_e(_x_);\
   case pciops_e: return nooks_iwrapper_get_itf_pciops_e(_x_);\
   case notifierops_e: return nooks_iwrapper_get_itf_notifierops_e(_x_);\
   case callbackops_e: return nooks_iwrapper_get_itf_callbackops_e(_x_);\
   case timerops_e: return nooks_iwrapper_get_itf_timerops_e(_x_);\
   case sysctlops_e: return nooks_iwrapper_get_itf_sysctlops_e(_x_);\
   case irqops_e: return nooks_iwrapper_get_itf_irqops_e(_x_);\
   default: return NULL;

/**********************************************************
 * netdev_operations wrappers allocation
 *********************************************************/

nooks_iwrapper_itf
    nooks_iwrapper_get_itf_netops_e(pnook_driver_t driver)
{
  struct iwrapper_netdev_operations *itf;
  unsigned long addr, addr_end;
  NOOKS_ENTER();
  itf = (struct iwrapper_netdev_operations*)nooks_kmalloc(sizeof(struct iwrapper_netdev_operations), NOOKS_MEM_FLAGS);

  if(!itf) {
    NOOKS_FAULT(enomem);
    return NULL;
  }
 
  addr = (unsigned long) itf;
  addr_end = (unsigned long) itf + (unsigned long) sizeof(struct iwrapper_netdev_operations); 
 
  /* Make the memory allocated executable. */
  for (; addr < addr_end; addr += PAGE_SIZE)
    change_page_attr(virt_to_page(addr), 1, PAGE_KERNEL_EXEC);

  printk ("Making the pages executable.\n");
  global_flush_tlb();
  
  spin_lock(&nooks_iwrapper_lock);

  nooks_iwrapper_set(&(itf->hard_header_cache), driver, 2, 0,
      NOOKS_DWRAPPER(netops_hard_header_cache), -EFAULT);
  nooks_iwrapper_set(&(itf->open), driver, 1, 0,
      NOOKS_DWRAPPER(netops_open), -EFAULT);
  nooks_iwrapper_set(&(itf->hard_header), driver, 6, 0,
      NOOKS_DWRAPPER(netops_hard_header), -EFAULT);
  nooks_iwrapper_set(&(itf->init), driver, 1, 0,
      NOOKS_DWRAPPER(netops_init), -EFAULT);
  nooks_iwrapper_set(&(itf->do_ioctl), driver, 3, 0,
      NOOKS_DWRAPPER(netops_do_ioctl), -EFAULT);
  nooks_iwrapper_set(&(itf->hard_header_parse), driver, 2, 0,
      NOOKS_DWRAPPER(netops_hard_header_parse), -EFAULT);
  nooks_iwrapper_set(&(itf->stop), driver, 1, 0,
      NOOKS_DWRAPPER(netops_stop), -EFAULT);
  nooks_iwrapper_set(&(itf->get_wireless_stats), driver, 1, 0,
      NOOKS_DWRAPPER(netops_get_wireless_stats), 0);
  nooks_iwrapper_set(&(itf->header_cache_update), driver, 3, 1,
      NOOKS_DWRAPPER(netops_header_cache_update), -EFAULT);
  nooks_iwrapper_set(&(itf->uninit), driver, 1, 1,
      NOOKS_DWRAPPER(netops_uninit), -EFAULT);
  nooks_iwrapper_set(&(itf->tx_timeout), driver, 1, 1,
      NOOKS_DWRAPPER(netops_tx_timeout), -EFAULT);
  nooks_iwrapper_set(&(itf->set_multicast_list), driver, 1, 1,
      NOOKS_DWRAPPER(netops_set_multicast_list), -EFAULT);
  nooks_iwrapper_set(&(itf->destructor), driver, 1, 1,
      NOOKS_DWRAPPER(netops_destructor), -EFAULT);
  nooks_iwrapper_set(&(itf->hard_start_xmit), driver, 2, 0,
      NOOKS_DWRAPPER(netops_hard_start_xmit), -EFAULT);
  nooks_iwrapper_set(&(itf->set_mac_address), driver, 2, 0,
      NOOKS_DWRAPPER(netops_set_mac_address), -EFAULT);
  nooks_iwrapper_set(&(itf->neigh_setup), driver, 2, 0,
      NOOKS_DWRAPPER(netops_neigh_setup), -EFAULT);
  //nooks_iwrapper_set(&(itf->accept_fastpath), driver, 2, 0,
   //   NOOKS_DWRAPPER(netops_accept_fastpath), -EFAULT);
  nooks_iwrapper_set(&(itf->rebuild_header), driver, 1, 0,
      NOOKS_DWRAPPER(netops_rebuild_header), -EFAULT);
  nooks_iwrapper_set(&(itf->get_stats), driver, 1, 0,
      NOOKS_DWRAPPER(netops_get_stats), 0);
  nooks_iwrapper_set(&(itf->set_config), driver, 2, 0,
      NOOKS_DWRAPPER(netops_set_config), -EFAULT);
  nooks_iwrapper_set(&(itf->change_mtu), driver, 2, 0,
      NOOKS_DWRAPPER(netops_change_mtu), -EFAULT);

  spin_unlock(&nooks_iwrapper_lock);
  NOOKS_EXIT();
  return (nooks_iwrapper_itf)itf;
}

/**********************************************************
 * module_operations wrappers allocation
 *********************************************************/

nooks_iwrapper_itf
    nooks_iwrapper_get_itf_mops_e(pnook_driver_t driver)
{
  struct iwrapper_module_operations *itf;
  NOOKS_ENTER();
  unsigned long addr = 0;
  unsigned long addr_end = 0;
  itf = (struct iwrapper_module_operations*)nooks_kmalloc(sizeof(struct iwrapper_module_operations), NOOKS_MEM_FLAGS);

  if(!itf) {
    NOOKS_FAULT(enomem);
    return NULL;
  }

  spin_lock(&nooks_iwrapper_lock);
  addr = (unsigned long) itf;
  addr_end = (unsigned long) itf + (unsigned long) sizeof(struct iwrapper_module_operations);

  /* Make the memory allocated executable.*/ 
  for (; addr < addr_end; addr += PAGE_SIZE)
    change_page_attr(virt_to_page(addr), 1, PAGE_KERNEL_EXEC);

  printk ("Making the pages executable.\n");
  global_flush_tlb();
 
  nooks_iwrapper_set(&(itf->init), driver, 0, 0,
      NOOKS_DWRAPPER(mops_init), -EFAULT);
  nooks_iwrapper_set(&(itf->exit), driver, 0, 1,
      NOOKS_DWRAPPER(mops_cleanup), -EFAULT);
  /* nooks_iwrapper_set(&(itf->can_unload), driver, 0, 0,
      NOOKS_DWRAPPER(mops_can_unload), -EFAULT);
 FIXME can_unload is unsafe now not a pointer. */ 
  spin_unlock(&nooks_iwrapper_lock);
  NOOKS_EXIT();
  return (nooks_iwrapper_itf)itf;
}


/**********************************************************
 * actor_operations wrappers allocation
 *********************************************************/

nooks_iwrapper_itf
    nooks_iwrapper_get_itf_actorops_e(pnook_driver_t driver)
{
  struct iwrapper_actor_operations *itf;
  NOOKS_ENTER();
  itf = (struct iwrapper_actor_operations*)nooks_kmalloc(sizeof(struct iwrapper_actor_operations), NOOKS_MEM_FLAGS);

  if(!itf) {
    NOOKS_FAULT(enomem);
    return NULL;
  }
  spin_lock(&nooks_iwrapper_lock);

  nooks_iwrapper_set(&(itf->read), driver, 4, 0,
      NOOKS_DWRAPPER(actorops_read), -EFAULT);

  spin_unlock(&nooks_iwrapper_lock);
  NOOKS_EXIT();
  return (nooks_iwrapper_itf)itf;
}

/**********************************************************
 * procfs_operations wrappers allocation
 *********************************************************/

nooks_iwrapper_itf
    nooks_iwrapper_get_itf_procops_e(pnook_driver_t driver)
{
  struct iwrapper_procfs_operations *itf;
  NOOKS_ENTER();
  itf = (struct iwrapper_procfs_operations*)nooks_kmalloc(sizeof(struct iwrapper_procfs_operations), NOOKS_MEM_FLAGS);

  if(!itf) {
    NOOKS_FAULT(enomem);
    return NULL;
  }
  spin_lock(&nooks_iwrapper_lock);

  nooks_iwrapper_set(&(itf->write_proc), driver, 4, 0,
      NOOKS_DWRAPPER(procops_write_proc), -EFAULT);
  nooks_iwrapper_set(&(itf->read_proc), driver, 6, 0,
      NOOKS_DWRAPPER(procops_read_proc), -EFAULT);
  nooks_iwrapper_set(&(itf->get_info), driver, 4, 0,
      NOOKS_DWRAPPER(procops_get_info), -EFAULT);

  spin_unlock(&nooks_iwrapper_lock);
  NOOKS_EXIT();
  return (nooks_iwrapper_itf)itf;
}


/**********************************************************
 * pci_operations wrappers allocation
 *********************************************************/

nooks_iwrapper_itf
    nooks_iwrapper_get_itf_pciops_e(pnook_driver_t driver)
{
  struct iwrapper_pci_operations *itf;
  unsigned long addr = 0;
  unsigned long addr_end = 0;
  NOOKS_ENTER();
  itf = (struct iwrapper_pci_operations*)nooks_kmalloc(sizeof(struct iwrapper_pci_operations), NOOKS_MEM_FLAGS);

  if(!itf) {
    NOOKS_FAULT(enomem);
    return NULL;
  }

  addr = (unsigned long) itf;
  addr_end = (unsigned long) itf + (unsigned long) sizeof(struct iwrapper_pci_operations);

  /* Make the memory allocated executable. */
  for (; addr < addr_end; addr += PAGE_SIZE)
    change_page_attr(virt_to_page(addr), 1, PAGE_KERNEL_EXEC);

  printk ("Making the pages executable.\n");
  global_flush_tlb();

  spin_lock(&nooks_iwrapper_lock);

 // nooks_iwrapper_set(&(itf->prepare), driver, 1, 0,
  //    NOOKS_DWRAPPER(pciops_prepare), -EFAULT);
  nooks_iwrapper_set(&(itf->remove), driver, 1, 1,
      NOOKS_DWRAPPER(pciops_remove), -EFAULT);
  nooks_iwrapper_set(&(itf->enable_wake), driver, 3, 0,
      NOOKS_DWRAPPER(pciops_enable_wake), -EFAULT);
  //nooks_iwrapper_set(&(itf->save_state), driver, 2, 0,
   //   NOOKS_DWRAPPER(pciops_save_state), -EFAULT);
  nooks_iwrapper_set(&(itf->shutdown), driver, 1, 1,   
	NOOKS_DWRAPPER(pciops_shutdown), -EFAULT);
//  nooks_iwrapper_set(&(itf->activate), driver, 1, 0,
//      NOOKS_DWRAPPER(pciops_activate), -EFAULT);
  nooks_iwrapper_set(&(itf->resume), driver, 1, 0,
      NOOKS_DWRAPPER(pciops_resume), -EFAULT);
  nooks_iwrapper_set(&(itf->suspend), driver, 2, 0,
      NOOKS_DWRAPPER(pciops_suspend), -EFAULT);
//  nooks_iwrapper_set(&(itf->deactivate), driver, 1, 0,
//      NOOKS_DWRAPPER(pciops_deactivate), -EFAULT);
  nooks_iwrapper_set(&(itf->probe), driver, 2, 0,
      NOOKS_DWRAPPER(pciops_probe), -EFAULT);

  spin_unlock(&nooks_iwrapper_lock);
  NOOKS_EXIT();
  return (nooks_iwrapper_itf)itf;
}

/**********************************************************
 * notifier_operations wrappers allocation
 *********************************************************/

nooks_iwrapper_itf
    nooks_iwrapper_get_itf_notifierops_e(pnook_driver_t driver)
{
  struct iwrapper_notifier_operations *itf;
  NOOKS_ENTER();
  itf = (struct iwrapper_notifier_operations*)nooks_kmalloc(sizeof(struct iwrapper_notifier_operations), NOOKS_MEM_FLAGS);

  if(!itf) {
    NOOKS_FAULT(enomem);
    return NULL;
  }
  spin_lock(&nooks_iwrapper_lock);

  nooks_iwrapper_set(&(itf->notifier_call), driver, 3, 0,
      NOOKS_DWRAPPER(notifierops_notifier_call), -EFAULT);

  spin_unlock(&nooks_iwrapper_lock);
  NOOKS_EXIT();
  return (nooks_iwrapper_itf)itf;
}


/**********************************************************
 * callback_operations wrappers allocation
 *********************************************************/

nooks_iwrapper_itf
    nooks_iwrapper_get_itf_callbackops_e(pnook_driver_t driver)
{
  struct iwrapper_callback_operations *itf;
  NOOKS_ENTER();
  itf = (struct iwrapper_callback_operations*)nooks_kmalloc(sizeof(struct iwrapper_callback_operations), NOOKS_MEM_FLAGS);

  if(!itf) {
    NOOKS_FAULT(enomem);
    return NULL;
  }
  spin_lock(&nooks_iwrapper_lock);

  nooks_iwrapper_set(&(itf->handler), driver, 1, 0,
      NOOKS_DWRAPPER(callbackops_handler), -EFAULT);
  nooks_iwrapper_set(&(itf->pm_handler), driver, 3, 0,
      NOOKS_DWRAPPER(callbackops_pm_handler), -EFAULT);

  spin_unlock(&nooks_iwrapper_lock);
  NOOKS_EXIT();
  return (nooks_iwrapper_itf)itf;
}

/**********************************************************
 * timer_operations wrappers allocation
 *********************************************************/

nooks_iwrapper_itf
    nooks_iwrapper_get_itf_timerops_e(pnook_driver_t driver)
{
  struct iwrapper_timer_operations *itf;
  NOOKS_ENTER();
  itf = (struct iwrapper_timer_operations*)nooks_kmalloc(sizeof(struct iwrapper_timer_operations), NOOKS_MEM_FLAGS);

  if(!itf) {
    NOOKS_FAULT(enomem);
    return NULL;
  }
  spin_lock(&nooks_iwrapper_lock);

  nooks_iwrapper_set(&(itf->handler), driver, 1, 1,
      NOOKS_DWRAPPER(timerops_handler), -EFAULT);

  spin_unlock(&nooks_iwrapper_lock);
  NOOKS_EXIT();
  return (nooks_iwrapper_itf)itf;
}

/**********************************************************
 * sysctl_operations wrappers allocation
 *********************************************************/

nooks_iwrapper_itf
    nooks_iwrapper_get_itf_sysctlops_e(pnook_driver_t driver)
{
  struct iwrapper_sysctl_operations *itf;
  NOOKS_ENTER();
  itf = (struct iwrapper_sysctl_operations*)nooks_kmalloc(sizeof(struct iwrapper_sysctl_operations), NOOKS_MEM_FLAGS);

  if(!itf) {
    NOOKS_FAULT(enomem);
    return NULL;
  }
  spin_lock(&nooks_iwrapper_lock);

  nooks_iwrapper_set(&(itf->strategy), driver, 8, 0,
      NOOKS_DWRAPPER(sysctlops_strategy), -EFAULT);
  nooks_iwrapper_set(&(itf->proc_handler), driver, 6, 0,
      NOOKS_DWRAPPER(sysctlops_proc_handler), -EFAULT);

  spin_unlock(&nooks_iwrapper_lock);
  NOOKS_EXIT();
  return (nooks_iwrapper_itf)itf;
}

/**********************************************************
 * irq_operations wrappers allocation
 *********************************************************/

nooks_iwrapper_itf
    nooks_iwrapper_get_itf_irqops_e(pnook_driver_t driver)
{
  struct iwrapper_irq_operations *itf;
  NOOKS_ENTER();
  unsigned long addr = 0;
  unsigned long addr_end = 0;
  itf = (struct iwrapper_irq_operations*)nooks_kmalloc(sizeof(struct iwrapper_irq_operations), NOOKS_MEM_FLAGS);

  if(!itf) {
    NOOKS_FAULT(enomem);
    return NULL;
  }

  addr = (unsigned long) itf;
  addr_end = (unsigned long) itf + (unsigned long) sizeof(struct iwrapper_irq_operations);

  /* Make the memory allocated executable. */
  for (; addr < addr_end; addr += PAGE_SIZE)
    change_page_attr(virt_to_page(addr), 1, PAGE_KERNEL_EXEC);

  printk ("Making the pages executable.\n");
  global_flush_tlb();

  spin_lock(&nooks_iwrapper_lock);

  nooks_iwrapper_set(&(itf->handler), driver, 3, 1,
      NOOKS_DWRAPPER(irqops_handler), -EFAULT);

  spin_unlock(&nooks_iwrapper_lock);
  NOOKS_EXIT();
  return (nooks_iwrapper_itf)itf;
}

