/*
 * nooks-recover.c -- fault handling and recovery code for nooks
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  
 * No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

 
#include "nooks-i.h"
#include "nooks-recover.h"
#include "nooks-wrappers.h"
#include "nooks-iwrappers.h"
#include "nooks-nwrappers.h"
#include <linux/file.h> // for fput()
#include <linux/kmod.h> // for fput()
#include <asm/dma.h> // for free_dma()
#include <linux/buffer_head.h>
#include <linux/proc_fs.h>

int nooks_do_recovery = 1;
int nooks_always_poll = 0;
int error_code = 0;

extern struct list_head vfsmntlist;
extern struct vfsmount *ide_device;
extern struct super_block *current_sb;
extern void kill_locked_buf(struct super_block *s);


extern unsigned int d_hash_mask;
extern unsigned int d_hash_shift;
extern struct list_head *dentry_hashtable;
extern struct list_head vfsmntlist;
int do_umount(struct vfsmount *mnt, int flags);
extern struct semaphore mount_sem;

//#include <asm/unistd.h>
//static inline _syscall2(long,umount2,char *, name, int, flags);

long __sys_umount(char * name, int flags);

#ifdef USE_GENERATED_FILES
#include "nooks-reco-c-gen.h"
#else
nooks_driver_recovery_fn * nooks_driver_recovery[] = {
  nooks_recover_generic,
  nooks_recover_mops,
  nooks_recover_fops,
  //nooks_recover_bdops,
  //nooks_recover_vmops,
  //nooks_recover_blkops,
  nooks_recover_procops,
  nooks_recover_netops,
  nooks_recover_irqops,
  nooks_recover_pciops,
  nooks_recover_callbackops,
  nooks_recover_timerops,
  //nooks_recover_inodeops,
  nooks_recover_sysctlops,
  nooks_recover_actorops,
  nooks_recover_asops,
  //nooks_recover_dentryops,
  //nooks_recover_superops,
  //nooks_recover_getblockops,
  //nooks_recover_fsops,
  nooks_recover_notifierops,
  NULL
};
#endif
nooks_resource_recovery_fn * nooks_resource_recovery[] = {
  nooks_res_recover_generic,
  nooks_res_recover_fasync,
  nooks_res_recover_waitqueue,
  nooks_res_recover_waitqueue_sync,
  nooks_res_recover_devfs_handle,
  nooks_res_recover_netdev,
  nooks_res_recover_netdev_stats,
  nooks_res_recover_netdev_nook,
  nooks_res_recover_timer,
  nooks_res_recover_pci_driver,
  nooks_res_recover_pci_memory,
  nooks_res_recover_worker,  //10
  nooks_res_recover_proc_entry,
  nooks_res_recover_tasklet,
  nooks_res_recover_skbuff_k,
  nooks_res_recover_skbuff_n,
  nooks_res_recover_area_map,
  nooks_res_recover_pcidev,
  nooks_res_recover_irq_dev,
  nooks_res_recover_region,
  nooks_res_recover_ctl_table,
  nooks_res_recover_ctl_head,
  nooks_res_recover_socket,
  nooks_res_recover_waitqueuehead, //20
  nooks_res_recover_sock,
  nooks_res_recover_locked_sock,
  nooks_res_recover_tcp_open_req,
  nooks_res_recover_file,
  nooks_res_recover_actor_driver,
  nooks_res_recover_fs,
  nooks_res_recover_superblock,
  nooks_res_recover_inode,
  nooks_res_recover_inode_ops,
  nooks_res_recover_dentry,
  nooks_res_recover_dentry_ops,   //30
  nooks_res_recover_buffer_head,
  nooks_res_recover_super_ops,
  nooks_res_recover_file_ops,
  nooks_res_recover_iomap,
  nooks_res_recover_as_ops,
  nooks_res_recover_address_space,
  nooks_res_recover_getblock_driver,
  nooks_res_recover_buffer_data,
  nooks_res_recover_dma,
  nooks_res_recover_vm_ops,
  nooks_res_recover_misc_device,
  nooks_res_recover_mmap,
  nooks_res_recover_module_data,
  nooks_res_recover_irqaction,
  nooks_res_recover_ioctl,
  nooks_res_recover_pm_dev,
  nooks_res_recover_pciregion,
  nooks_res_recover_notifier,
  /* Added By Muthu */

  nooks_res_recover_dummy1,
  nooks_res_recover_dummy2,
  nooks_res_recover_gendisk,
  nooks_res_recover_sector_info,	
  nooks_res_recover_bd_ops, 
  nooks_res_recover_major, 

  /******************/
  nooks_res_recover_vmalloc,
  nooks_res_recover_bssstart,
  nooks_res_recover_bssend,
  nooks_res_recover_kmap,
  nooks_res_recover_dmamap,
  nooks_res_recover_dmadevice,
  nooks_res_recover_dmadir,
  nooks_res_recover_dmaalloc,
  nooks_res_recover_dmamappage,
  nooks_res_recover_dmamapsingle,
	NULL
};


int nooks_res_recover_dmaalloc(pnook_t nook, void * key, void * data, int pass)
{
  //TBD
  return(0);
}



int nooks_res_recover_dmadir(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}



int nooks_res_recover_dmadevice(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}


int nooks_res_recover_dmamap(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}


int nooks_res_recover_bssend(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_bssstart(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}



/**********************************************************
 *
 * nooks_fault_alert_usermode
 *
 * Alert user mode fault handler to recover a driver (class)
 *
 * We pass into the driver what types of drivers were present
 *  (type / name / major or minor numbers)
 *
 *********************************************************/
void 
nooks_fault_alert_usermode(void * u_nook)
{
#define NOOK_MAX_AGENT_ARGS 40
  pnook_t nook = (pnook_t) u_nook;
  char *argv[NOOK_MAX_AGENT_ARGS + 1], *envp[3], nook_name[NOOK_MAX_NAME_LEN];
  char line[8];
  char * page = NULL;
  char * where;
  struct list_head * next;
  struct list_head * spare;
  pnook_driver_t driver = NULL;
  int i, j;
  int result = 0;
  int remaining = PAGE_SIZE;

  
  page = (char *) __get_free_page(GFP_KERNEL);
  if (page == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }
;
  //
  // Add all registered devices in the nook to the arguments
  //
  i = 0;
  argv[i++] = nooks_agent_path;
  sprintf(nook_name, "%s", nook->name);
  argv[i++] = nook_name;

  switch (nook->recovery_state.type) {
	  case FAULT_PARAMETER: 
		  argv[i++] = "P";
		  argv[i++] = (char *) nook->recovery_state.file;
		  sprintf(line, "%d", nook->recovery_state.line);
		  argv[i++] = line;
		  break;
	  case FAULT_RESOURCE: 
		  argv[i++] = "R";
		  argv[i++] = (char *) nook->recovery_state.file;
		  sprintf(line, "%d", nook->recovery_state.line);
		  argv[i++] = line;
		  break;
	  case FAULT_ENOMEM: 
		  argv[i++] = "M";
		  argv[i++] = (char *) nook->recovery_state.file;
		  sprintf(line, "%d", nook->recovery_state.line);
		  argv[i++] = line;
		  break;
	  case FAULT_STACK: 
		  argv[i++] = "S";
		  sprintf(line, "%d", error_code);
		  argv[i++] = line;
		  break;
	  default:
		  argv[i++] = "D";
		  sprintf(line, "%d", error_code);
		  argv[i++] = line;
  }

  where = page;
  list_for_each_safe(next, spare, &nook->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_nook);
    
    if ((nooks_iwrapper_flags[driver->driver_type] & 
	 NOOKS_IWRAPPER_REGISTERED_ITF) != 0) {
      if (i == NOOK_MAX_AGENT_ARGS) {
	result = -ENOMEM;
	goto Cleanup;
      }
#if 0 // Not interesting for net devices.
      if (driver->driver_type == fsops_e) {

	result = nooks_get_fs_mounts(nook, 
				     driver, 
				     argv, 
				     &i, 
				     NOOK_MAX_AGENT_ARGS,
				     &where, 
				     &remaining);
	if (result < 0) {
	  goto Cleanup;
	}
      } 
      else if (driver->driver_type == bdops_e)
      {
	result = nooks_get_ide_mounts(nook, 
				     driver, 
				     argv, 
				     &i, 
				     NOOK_MAX_AGENT_ARGS,
				     &where, 
				     &remaining);
	if (result < 0) {
	  goto Cleanup;
	}

      }	
      else 
#endif
      {
	//
	// Add it to the arguments
	//
	
	result = snprintf(where, remaining, "%s:%s:%d", 
			  nooks_iwrapper_names[driver->driver_type],
			  ((nooks_iwrapper_flags[driver->driver_type] & 
			    NOOKS_IWRAPPER_NAMED) != 0) ? driver->name : "",
			  ((nooks_iwrapper_flags[driver->driver_type] & 
			    NOOKS_IWRAPPER_NUMBERED) != 0) ? driver->major : -1);
	if (result < 0) {
	  result = -ENOMEM;
	  goto Cleanup;
	}
	argv[i++] = where;
	where += result;
	*where = '\0';
	where ++;
	remaining -= (result + 1);
      }
    }
  }  

    
  argv[i] = 0;

  PDEBUG2("Printing the nooks_alert_fault_handler string:\n\n");	

  for(j=0; j<=i; j++)
  	PDEBUG("%s ", argv[j]); 

  PDEBUG2("\n\n"); 
  //
  // Build the environment
  //
  i = 0;

  /* minimal command environment */

  envp [i++] = "HOME=/";
  envp [i++] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";
  envp [i] = 0;

  //NOTE We run with wait=0 to mirror behaviour of Nooks 2.4
  result = call_usermodehelper(argv [0], argv, envp, 0);
 
  Cleanup:
  if (page != NULL) {
    free_page((unsigned long) page);
  }

}

/*void 
nooks_fault_alert_usermode(void * u_nook)
{
#define NOOK_MAX_AGENT_ARGS 40
  pnook_t nook = (pnook_t) u_nook;
  char *argv[NOOK_MAX_AGENT_ARGS + 1], *envp[3], nook_name[NOOK_MAX_NAME_LEN];
  char line[8];
  char * page = NULL;
  char * where;
  struct list_head * next;
  struct list_head * spare;
  pnook_driver_t driver = NULL;
  int i;
  int result = 0;
  int remaining = PAGE_SIZE;

  
  page = (char *) __get_free_page(GFP_KERNEL);
  if (page == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }
;
  //
  // Add all registered devices in the nook to the arguments
  //
  i = 0;
  argv[i++] = nooks_agent_path;
  sprintf(nook_name, "%s", nook->name);
  argv[i++] = nook_name;

  switch (nook->recovery_state.type) {
	  case FAULT_PARAMETER: 
		  argv[i++] = "P";
		  argv[i++] = (char *) nook->recovery_state.file;
		  sprintf(line, "%d", nook->recovery_state.line);
		  argv[i++] = line;
		  break;
	  case FAULT_RESOURCE: 
		  argv[i++] = "R";
		  argv[i++] = (char *) nook->recovery_state.file;
		  sprintf(line, "%d", nook->recovery_state.line);
		  argv[i++] = line;
		  break;
	  case FAULT_ENOMEM: 
		  argv[i++] = "M";
		  argv[i++] = (char *) nook->recovery_state.file;
		  sprintf(line, "%d", nook->recovery_state.line);
		  argv[i++] = line;
		  break;
	  case FAULT_STACK: 
		  argv[i++] = "S";
		  sprintf(line, "%d", error_code);
		  argv[i++] = line;
		  break;
	  default:
		  argv[i++] = "D";
		  sprintf(line, "%d", error_code);
		  argv[i++] = line;
  }

  where = page;
  list_for_each_safe(next, spare, &nook->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_nook);
    
    if ((nooks_iwrapper_flags[driver->driver_type] & 
	 NOOKS_IWRAPPER_REGISTERED_ITF) != 0) {
      if (i == NOOK_MAX_AGENT_ARGS) {
	result = -ENOMEM;
	goto Cleanup;
      }

      if (driver->driver_type == fsops_e) {

	result = nooks_get_fs_mounts(nook, 
				     driver, 
				     argv, 
				     &i, 
				     NOOK_MAX_AGENT_ARGS,
				     &where, 
				     &remaining);
	if (result < 0) {
	  goto Cleanup;
	}
      } else {
	//
	// Add it to the arguments
	//
	
	result = snprintf(where, remaining, "%s:%s:%d", 
			  nooks_iwrapper_names[driver->driver_type],
			  ((nooks_iwrapper_flags[driver->driver_type] & 
			    NOOKS_IWRAPPER_NAMED) != 0) ? driver->name : "",
			  ((nooks_iwrapper_flags[driver->driver_type] & 
			    NOOKS_IWRAPPER_NUMBERED) != 0) ? driver->major : -1);
	if (result < 0) {
	  result = -ENOMEM;
	  goto Cleanup;
	}
	argv[i++] = where;
	where += result;
	*where = '\0';
	where ++;
	remaining -= (result + 1);
      }
    }
  }  

    
  argv[i] = 0;
  
  //
  // Build the environment
  //
  i = 0;


  envp [i++] = "HOME=/";
  envp [i++] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";
  envp [i] = 0;

  
  result = call_usermodehelper(argv [0], argv, envp);
 Cleanup:
  if (page != NULL) {
    free_page((unsigned long) page);
  }

}*/

void
nooks_fault_start_recovery(void * u_nook)
{
  pnook_t nook = (pnook_t) u_nook;
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
  nook->waiting_time = 0;
#endif
  if (nook->recovery_state.info.recover != NULL) {
    nooks_recovery_fn * recovery_fn = nook->recovery_state.info.recover;

    //
    // Clear the recovery function so we don't loop if this causes a fault
    //

    if (!replug_can_recover_again(nook))
	    nook->recovery_state.info.recover = NULL;
    recovery_fn(nook);
  } else {
    nooks_fault_alert_usermode(nook);
  }
  
}


/**********************************************************
 *
 * nooks_free_resource_hash
 *
 * Frees a hash of resources
 *
 *********************************************************/

void
nooks_free_resource_hash(pnook_t nook, pnooks_hash_table_t table, int pass)
{
  int result;
  int retries = 5;
  int again;

  //
  // Traverse the active skbs and free them now
  //
  //printk ("Freeing in nooks_free_resource_hash for nook.\n");
  do {
    nooks_hash_enum_t context = {0};
    pnooks_hash_entry_t entry = NULL;
    again = 0;
    //printk ("Freeing for table. %p.\n", table);
    while (nooks_enumerate_in_hash(table,
				   &context,
				   &entry) == 0) {
     
	 
      if (entry->entry_type < last_restype_e) {
	//printk ("Entry type is %d in table %p.\n", entry->entry_type, table);
	if ((entry->entry_type == 16)) // || (entry->entry_type == 6) || (entry->entry_type == 7) )	
	{
		printk ("16 found detouring.\n");
	}
	else
	{
		if (entry->entry_type == pciregion_e)
		//	printk ("pciregion found.\n");
		//if ((entry->entry_type == 6) || (entry->entry_type == 7))
		//	printk ("6/7 found.\n");
		result = nook->recovery_state.info.resource_recovery[entry->entry_type](nook, 
											entry->key,
											entry->data,
											pass);
	
		//
		// If this returns anything but EAGAIN, we can free the resource
		//
	
	
		if ((pass > 1) || (result != -EAGAIN)) {
	  	nooks_delete_entry_from_hash(table,
					       entry);
		}
		if (result == -ERESTART) {
		  again = 1;
		  retries--;
		  break;
		} //if
	} //else
      } //outer if
    } //while loop 
  } while (again && retries);
    
}

//
// BUGBUG: hack for pcnet32 driver
//
unsigned long nooks_recovering_interrupts[NR_IRQS];

void
nooks_set_recovering_interrupt(unsigned int irq)
{
  nooks_recovering_interrupts[irq] = 1;
}

void
nooks_clear_recovering_interrupt(unsigned int irq)
{
  nooks_recovering_interrupts[irq] = 0;
}

/**********************************************************
 *
 * nooks_disable_drivers
 *
 * called when fault is initially detected
 * should to whatever is necessary to prevent driver being
 *   called - e.g. disable interrupts for that device
 *
 *********************************************************/

void
nooks_disable_drivers(pnook_t nook)
{
  struct list_head * next, *spare;
  pnook_driver_t driver = NULL;
  unsigned long flags;

  spin_lock_irqsave(&nook_list_lock, flags);
  list_for_each_safe(next, spare, &nook->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_nook);

    if (driver->driver_type < last_ops_e) {
      driver->shadow_state = shadow_state_faulted;
      nook->recovery_state.info.driver_recovery[driver->driver_type]
	(nook, 
	 driver, 
	 NOOKS_RECOVERY_STOP_PASS);
            
    }
    
  }

  spin_unlock_irqrestore(&nook_list_lock, flags);

}


void
nooks_fault_handler(pnook_t nook)
{

  nooks_disable_drivers(nook);
  nook->state = nook_state_faulted;
 
 /*FIXME  TBD 
  INIT_TQUEUE(&nook->recovery_state.recovery_task, 
	      nooks_fault_start_recovery, 
	      nook);
 
  schedule_task(&nook->recovery_state.recovery_task);
  */
  nooks_fault_alert_usermode(nook);  
}


/**********************************************************
 *
 * Name: nooks_fault
 *
 * Synopsis: Receives indication that a nook has faulted.
 * Marks the nook as dead, and triggers recovery. At interrupt
 * level will queue a softirq to handle recovery.
 *
 * Assumptions: is called with valid current->nook_state
 *
 * Arguments:
 *
 *    fault:   error code for the fault, such as -EFAULT
 *    pt_regs: registers for trapping faults (to enable restart in kernel code)
 *
 * return:
 *    0 if exception was handled, and can return
 *    < 0 if caller should continue processing
 *
 *********************************************************/

extern void __struct_cpy_bug (void);
int
nooks_fault(int fault, struct pt_regs * pt_regs)
{
/*
 * We get link-time errors if the structure sizes do not match.
 * There is no runtime overhead, it's all optimized away at
 * compile time.
 */

#define struct_cpy(x,y)                         \
({                                              \
        if (sizeof(*(x)) != sizeof(*(y)))       \
                printk("Fatal:ERROR");               \
        memcpy(x, y, sizeof(*(x)));             \
})


  if (nooks_do_recovery) {
    struct task_struct * task;
    pnook_t nook;
    task = current;
    nook = GET_TASK_CURRENT_NOOK(task);

    error_code = fault;
#ifdef NOOKS_DEBUG2    
    //    BUG();
#endif // NOOKS_DEBUG2


    struct_cpy(&nook->recovery_state.regs, pt_regs);
    
    //
    // Set the instruction pointer to be the driver fault recover code
    //
    // FIXME pt_regs->eip = (long) &nooks_driver_fault; // nooks_recovery_address;
    
    //
    // Set the stack pointer to be the kernel stack - we don't want to rely
    // on the nooks stack any more.
    // BUGBUG: maybe not
    //  pt_regs->esp = GET_NOOK_STATE()->thread_state.trap_state.kernel_stack;
    
    //
    // Clear the nooks page table nook stack, so they don't get recovered
    // BUGBUG how about just changing the page table?

#ifndef CONFIG_NOOKS_NOPROT    
    // pt_regs->nook_reg_pt = 0; FIXME
#endif

    // pt_regs->nook_reg_esp = 0;
    
    //
    // Reset the nook state so we won't come through here again
    //

    nooks_fault_handler(nook);

#if 0
    nooks_disable_drivers(nook);
    nook->state = nook_state_faulted;

    //
    // Queue the recovery task (asynchronously)
    //

    INIT_TQUEUE(&nook->recovery_state.recovery_task, 
		nooks_fault_alert_usermode, 
		nook);
    schedule_task(&nook->recovery_state.recovery_task);
#endif

    return(0);
  } else {
    return(-EFAULT);
  }
}




/* CA_DBG */
int nooks_parameter_fault(const char * file, const int line) {

	pnook_t nook = GET_CURRENT_NOOK();
	//nooks_fault(-EFAULT, &(GET_CURRENT_NOOK()->recovery_state.regs));

	nook->recovery_state.file = file;
	nook->recovery_state.line = line;
	nook->recovery_state.type = FAULT_PARAMETER;
	
	nooks_fault_handler(nook);
	return 0;
}

int nooks_enomem_fault(const char * file, const int line) {

	pnook_t nook = GET_CURRENT_NOOK();
	//nooks_fault(-EFAULT, &(GET_CURRENT_NOOK()->recovery_state.regs));

	nook->recovery_state.file = file;
	nook->recovery_state.line = line;
	nook->recovery_state.type = FAULT_ENOMEM;
	
	nooks_fault_handler(nook);
	return 0;
}

int nooks_resource_fault(const char * file, const int line) {

	pnook_t nook = GET_CURRENT_NOOK();
	//nooks_fault(-EFAULT, &(GET_CURRENT_NOOK()->recovery_state.regs));

	nook->recovery_state.file = file;
	nook->recovery_state.line = line;
	nook->recovery_state.type = FAULT_RESOURCE;
	
	nooks_fault_handler(nook);
	return 0;
}

int nooks_deadlock_fault(const char * file, const int line) {

	pnook_t nook = GET_CURRENT_NOOK();
	//nooks_fault(-EFAULT, &(GET_CURRENT_NOOK()->recovery_state.regs));

	nook->recovery_state.file = file;
	nook->recovery_state.line = line;
	nook->recovery_state.type = FAULT_DEADLOCK;
	
	nooks_fault_handler(nook);
	return 0;
}

#ifndef CONFIG_X86_NO_TSS //FIXME
int
nooks_stack_fault(struct tss_struct * tss)
{

  struct task_struct * task;
  pnook_t nook;
  task = current;
  nook = GET_TASK_CURRENT_NOOK(task);
    
  //
  // Set the instruction pointer to be the driver fault recover code
  //
  //FIXME tss->eip = (unsigned long) &nooks_driver_fault; // nooks_recovery_address;
  tss->esp = tss->esp0 = (unsigned long) task->nook_state.thread_state.trap_state.kernel_stack;
  tss->__cr3 = __pa(init_mm.pgd);
  tss->es		= __USER_DS;
  tss->cs		= __KERNEL_CS;
  tss->ss		= __KERNEL_DS;
  tss->ds		= __USER_DS;
  tss->ss = tss->ss0    = __KERNEL_DS;
  tss->eflags           = 0x82;
    
    
  //
  // Reset the nook state so we won't come through here again
  //
  nook->recovery_state.type = FAULT_STACK;
	 
  nooks_fault_handler(nook);

  return(0);
}
#endif
/**********************************************************
 *
 * Function: nooks_unwind_task
 * 
 * Type: 
 * 
 * Notes: Sends a task a signal to abort, and waits for it to
 * return.
 *
 *********************************************************/

int 
nooks_unwind_task(struct task_struct * task,
		  pnook_t nook)
{
  struct siginfo info;
  unsigned long flags;
  
  info.si_signo = SIGCRASH;
  info.si_errno = -EINTR; // BUGBUG: should be better
  info.si_code = SI_KERNEL;
  info.si_pid = current->pid;
  info.si_uid = current->uid;

  //
  // Make sure the task can receive SIGCRASH
  //
  // BUGBUG: this is probably unsafe
  //
  spin_lock_irqsave(&task->sighand->siglock, flags);
  sigdelset(&task->blocked, SIGCRASH);
  recalc_sigpending_nooks(task);
  spin_unlock_irqrestore(&task->sighand->siglock, flags);

  send_sig_info(SIGCRASH, &info, task);
  return(0);
}

/**********************************************************
 *
 * Function: nooks_check_task_history
 * 
 * Type: checks whether a task is executing on behalf of a nook
 * 
 * Notes: return 1 if the nook is in the history, zero if not
 *
 *********************************************************/

int
nooks_check_task_history(struct task_struct * task,
			 pnook_t nook)
{
  pnook_thread_state_t thread_state = &(task->nook_state.thread_state);

  for (; thread_state->kernel_frame != NULL; 
       thread_state = &thread_state->kernel_frame->thread_state) {
    if (thread_state->nook == nook) {
      return(1);
    }
  }   
  return(0);
}



int
nooks_reset_nook_state(pnook_t nook)
{
  int result = 0;
  //
  // Delete the kernel work list
  //
  nooks_delete_work_list(nook);

  //printk ("CLEANUP: Reinitializing nook. nooks_reset_nook_state.\n");
  //
  // Reinitialize the locals
  //
  spin_lock_init(&nook->locals->lock);
  INIT_LIST_HEAD(&nook->locals->free_work_list);
    INIT_LIST_HEAD(&nook->locals->cache_list);
  skb_queue_head_init(&nook->locals->skb_pool);
  skb_queue_head_init(&nook->locals->skb_rx_list);
  INIT_LIST_HEAD(&nook->locals->skb_free_list);  

  //
  // BUGBUG: what about memory? Should we free memory or reuse cache, which
  // may have become corrupt? For now we reuse cache
  //

  //
  // Initialize the user work list (let heap cleanup recover the entries)
  //
  nooks_mm_init_lookaside(&nook->locals->work_pool,
			  64,
			  sizeof(nooks_work_item_t),
			  NOOKS_LOOKASIDE_NOOK);

  nooks_mm_init_lookaside(&nook->locals->elem_pool,
			  20,
			  sizeof(nooks_list_elem_t),
			  NOOKS_LOOKASIDE_NOOK);

  nooks_mm_init_lookaside(&nook->locals->page_pool,
			  20,
			  0, 
			  NOOKS_LOOKASIDE_NOOK_PAGE);
  
  return(result);
}

void nook_print_drivers(pnook_t nook)
{
  struct list_head * next;
  struct list_head * spare;
  pnook_driver_t driver = NULL;

  printk("***************************\n");
  printk("***Printing Nook Drivers***\n");
  list_for_each_safe(next, spare, &nook->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_nook);
    if(driver)
    {
        if (driver->driver_type < last_ops_e)
                if(driver && driver->name && driver->major)
                        printk("Name: %s, Major: %d, Type: %d\n", driver->name,
                                                        driver->major, driver->driver_type);
                else
                        printk("Arbit driver\n");
    }
    else
    {
        printk("NO DRIVER!\n");
    }

  }

}


/**********************************************************
 *
 * nooks_do_unload_nook
 *
 * do the work of releasing all resources that can be 
 * released before the nook is deleted
 *
 * should set nook-state to zombie afterwards
 *
 * The things cleaned up are:
 *   drivers are deleted / unregistered
 *   work items
 *   skb pool
 *   stack pool
 *   work pool
 *   resources:
 *     fasync
 *     waitqueue
 *     devfs_handle
 *     netdev
 *     netdev_nook
 *     timer
 *     pci_driver
 *     worker
 *     proc_entry
 *     tasklet
 *     skbuff_k //
 *     area_map
 *     pcidev
 *    
 * Then, the list of tasks is scanned for tasks executing
 * within the nook. These tasks are signalled (if possible)
 *   
 *   
 * 
 * The things that are not cleaned up
 *   virtual memory pages
 *   socket buffers
 *   pci_memory
 *   skbuff_n
 *
 *********************************************************/
long
nooks_do_unload_nook(pnook_t nook)
{
  struct list_head * next;
  struct list_head * spare;
  pnook_driver_t driver = NULL;
  //  unsigned long flags;

  int result = 0;
  int signalled_tasks = 0;
  int retries = 0;
  struct task_struct * task = NULL;
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
  struct timeval tv1, tv2;
#endif

  //printk ("CLEANUP:Performing nooks_do_unload_nook.\n");
  //  nooks_hash_enum_t context = {0};
  //  pnooks_hash_entry_t entry = NULL;

  //
  // disable drivers first, so we don't get interrupts while recovering
  //

	/***********************/

	//nooks_fault_alert_usermode(nook);

	/**********************/


  
  if (nook->state == nook_state_running) {
    nooks_disable_drivers(nook);
  }

  nook->state = nook_state_cleanup;
  
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
#define nooks_DIFFTIME(x,y)  (long)((x.tv_sec - y.tv_sec)*1000 + ((x.tv_usec - y.tv_usec)/1000))
  do_gettimeofday(&tv1);
  signalled_tasks = (nook->waiting_time != 0);
  while(signalled_tasks) {
	signalled_tasks = 0;
	for_each_process(task) {
		if (nooks_check_task_history(task, nook))
			signalled_tasks++;
	}
	if (nook->waiting_time > 0) {
		do_gettimeofday(&tv2);
		if (nooks_DIFFTIME(tv1, tv2) > nook->waiting_time)
			signalled_tasks = 0;
	}
  }
  
#endif
  
  //
  // Step through the tasks in the system and make sure that this nook
  // isn't executing, and won't execute in the future.
  //
  
  do {
    signalled_tasks = 0;
    read_lock(&tasklist_lock);
    for_each_process(task) { 
      if (nooks_check_task_history(task, nook)) {
	nooks_unwind_task(task, nook);
	signalled_tasks++;
      }
    }
    read_unlock(&tasklist_lock);
    if (signalled_tasks != 0) {
      wait_queue_head_t dummy;
      init_waitqueue_head(&dummy);
      (void)interruptible_sleep_on_timeout(&dummy,HZ/2);	
      
      retries++;
    }
  } while ((signalled_tasks != 0) && 
	   (retries <= NOOKS_TASK_ABORT_RETRY_COUNT));

  if (signalled_tasks != 0) {
    PERROR("Failed to wake up all sleeping tasks\n");
  }

  //
  // Step through the drivers and call the recovery function for each one
  //
  
  // spin_lock_irqsave(&nook_list_lock, flags);
  list_for_each_safe(next, spare, &nook->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_nook);
    
    if (driver->driver_type < last_ops_e) {
      result = nook->recovery_state.info.driver_recovery[driver->driver_type]
	(nook, 
	 driver, 
	 NOOKS_RECOVERY_UNLOAD_PASS1);
      
      //
      // If this returns zero, we can free the driver
      //
      
      if (result == 0) {
	nooks_free_driver(driver);
      }
      
    }
  }  

  //
  // Traverse the resources and free them now
  //

  nooks_free_resource_hash(nook,
			   &nook->resources.resource_list.hash,
			   NOOKS_RECOVERY_UNLOAD_PASS2);
  
  printk("Printing Nook Drivers:\n"); 
  list_for_each_safe(next, spare, &nook->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_nook);
    if(driver)
    {
    	if (driver->driver_type < last_ops_e) 
    		if(driver && driver->name && driver->major)
			printk("Name: %s, Major: %d, Type: %d\n", driver->name, 
							driver->major, driver->driver_type); 
		else
			printk("Arbit driver\n");
    }
    else
    {
	printk("NO DRIVER!\n");
    }

  }

 
  list_for_each_safe(next, spare, &nook->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_nook);
    
    if (driver->driver_type < last_ops_e) {
      printk ("Calling unload pass for device %p.\n", driver->data);
      result = nook->recovery_state.info.driver_recovery[driver->driver_type]
	(nook, 
	 driver, 
	 NOOKS_RECOVERY_UNLOAD_PASS2);
      
      //
      // If this returns zero, we can free the driver
      //
      
      if (result == 0) {
	printk ("Freeing driver for device %p.\n",driver->data);
	nooks_free_driver(driver);
      }
      
    }
  }  
  
  return(0);
}


/**********************************************************
 *
 * nooks_clean_nook
 *
 * Release resources prior to deleting a nook
 *
 *********************************************************/

int
nooks_clean_nook(pnook_t nook)
{
  struct list_head * next;
  struct list_head * spare;
  pnook_driver_t driver = NULL;
  //  nooks_hash_enum_t context = {0};
  //  pnooks_hash_entry_t entry = NULL;


  // spin_lock_irqsave(&nook_list_lock, flags);
  list_for_each_safe(next, spare, &nook->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_nook);
    
    if (driver->driver_type < last_ops_e) {
      nook->recovery_state.info.driver_recovery[driver->driver_type](nook, driver, NOOKS_RECOVERY_CLEANUP_PASS);
      
      //nooks_driver_recovery[driver->driver_type](nook, driver, 2);
      //
      // If this returns zero, we can free the driver
      //
      printk ("Freeing the driver %s in nooks_clean_nook.\n", driver->name);     
      nooks_free_driver(driver);
      
    }
  }  

  //
  // Traverse the resources and free them now
  //

  nooks_free_resource_hash(nook,
			    &nook->resources.resource_list.hash,
			    NOOKS_RECOVERY_CLEANUP_PASS);




  //
  // Free the stack pool, which uses the area map
  //

  nooks_mm_free_lookaside(nook, &nook->resources.stack_pool);

  //
  // Free the skbuff pools
  //
 
  nooks_mm_free_skb_pool(nook);

  //
  // Traverse the active skbs and free them now
  //

  nooks_free_resource_hash(nook,
			    &nook->resources.active_skbs,
			    NOOKS_RECOVERY_CLEANUP_PASS);


  //
  // Free the memory in use by the nook
  //

  nooks_mm_free_page_map(nook);


  //
  // Free the vm area map
  //

  nooks_free_resource_hash(nook,
			    &nook->resources.area_map,
			    NOOKS_RECOVERY_CLEANUP_PASS);


  //
  // Free the caches
  //

  nooks_mm_cleanup_nook(nook);

  //
  // Free the work list
  //

  nooks_delete_work_list(nook);

  //
  // Free our lookaside lists
  //

  nooks_mm_free_lookaside(nook, &nook->resources.work_pool);
  nooks_mm_free_lookaside(nook, &nook->resources.map_pool);
  nooks_mm_free_lookaside(nook, &nook->resources.hash_pool);


  return(0);
}


unsigned long nooks_interrupt_poll_mask[NR_IRQS/32];
struct timer_list nooks_interrupt_poll_timer;
#define NOOKS_NEXT_IRQ_POLL() (jiffies + HZ/10)

/**********************************************************
 *
 * nooks_interrupt_poll
 *
 * 
 *
 *********************************************************/
void
nooks_interrupt_poll(unsigned long arg)
{
  unsigned int i;
  int did_irq = 0;

  for (i = 0; i < NR_IRQS/32; i++) {
    unsigned long mask = nooks_interrupt_poll_mask[i];
    int shift = 0;
    int next_shift;
    while ((next_shift = ffs(mask)) != 0) {
      shift += next_shift - 1;
      //nooks_do_interrupt(shift + i * (NR_IRQS/32)); FIXME
      did_irq = 1;
      shift ++;
      mask = mask >> next_shift;
    }
  }
  //
  // If there are any interrupts still to mask, continue
  // polling
  //
  if (did_irq) {
    nooks_interrupt_poll_timer.expires = NOOKS_NEXT_IRQ_POLL();
    add_timer(&nooks_interrupt_poll_timer);
  }
}


void 
nooks_poll_device(unsigned int irq)
{
  nooks_interrupt_poll_mask[irq / 32] |= (1 << (irq - (irq/32)));
  disable_irq_nosync(irq);
  mod_timer(&nooks_interrupt_poll_timer, NOOKS_NEXT_IRQ_POLL());
}

void 
nooks_unpoll_device(unsigned int irq, int from_cleanup)
{
  if (nooks_interrupt_poll_mask[irq / 32] == (1 << (irq - (irq/32)))) {
    nooks_interrupt_poll_mask[irq / 32] &= ~(1 << (irq - (irq/32)));

    enable_irq(irq); 
  }


#if 0
  //  if (!from_cleanup) {
    if (nooks_interrupt_poll_mask[irq / 32] == (1 << (irq - (irq/32)))) {
      irq_desc_t * desc = irq_desc + irq;
      unsigned long flags;
      void * dev_id = nooks_interrupt_devids[irq];
      
      local_irq_save(flags);
      nooks_interrupt_devids[irq] = NULL;

      nooks_interrupt_poll_mask[irq / 32] &= ~(1 << (irq - (irq/32)));

      if ((desc->status & IRQ_DISABLED) == 0) {
	disable_irq_nosync(irq);
      }
      free_irq(irq, dev_id);
      enable_irq(irq); 
      local_irq_restore(flags);

      //      if ((desc->status & IRQ_DISABLED) == 0) {
      //	disable_irq_nosync(irq);
      //      }
    }
    //  } else {
    //    irq_desc_t * desc = irq_desc + irq;
    //
    //    nooks_interrupt_poll_mask[irq / 32] &= ~(1 << (irq - (irq/32)));
    //
    //    if ((desc->status & IRQ_DISABLED) == 0) {
    //      disable_irq_nosync(irq);
    //    }
    //    enable_irq(irq); 
    //  }
#endif
}



/**********************************************************
 *
 * nooks_init_recovery
 *
 * Set up data structures for recovery.
 *
 *********************************************************/

int 
nooks_init_recovery(void)
{
  int i; 
  for (i = 0; i < NR_IRQS/32; i++) {
    nooks_interrupt_poll_mask[i] = 0;
  }

  init_timer(&nooks_interrupt_poll_timer);
  nooks_interrupt_poll_timer.data = 0;
  nooks_interrupt_poll_timer.function = nooks_interrupt_poll;
  return(0);
}

/*******************************************************************
 *
 * per-driver recovery functions
 *
 *******************************************************************/


int nooks_recover_generic(pnook_t nook, pnook_driver_t driver, int pass)
{
  BUG();
  return(0);
}


int nooks_recover_mops(pnook_t nook, pnook_driver_t driver, int pass)
{
  struct module * module, * mod;
  int found = 0;

  printk ("In nooks_recover_mops for module."); 
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS2) {
    return(-EAGAIN);
  }

  /*if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }*/

  module = (struct module *) driver->data;
  printk ("%s.\n", module->name);
  //
  // See if the module still exists
  //
  lock_kernel();

  mod = (struct module *) find_module(module->name);
  if (mod != NULL)
	found = 1; 
  unlock_kernel(); 
  
  //
  // BUGBUG: this may leave area maps corrupt
  //
  if (!found) {
    return(0);
  }

  //
  // Set the function pointers in the module to be NULL, so we won't
  // be called
  //

  module->exit = NULL;
  module->unsafe = 0;

  //
  // mark the module writeable pages as read-only
  //
  // if (mod_member_present(module, writeable_end) && NOT PRESENT IN 2.6  FIXME 
  /*
  if (module->writeable_end != 0) {
    nooks_mm_update_pagetable(nook,
			      NOOKS_READONLY, 
			      (unsigned long) module->writeable_start, 
			      (module->writeable_end - module->writeable_start 
			       + PAGE_SIZE - 1) & PAGE_MASK);
  }
  */
  //
  // Mark the module as unused, so it can be unloaded
  //
  /*    
  while (__MOD_IN_USE(module)) {
    __MOD_DEC_USE_COUNT(module);
  }
  TBD */ 

  return(0);
}

int nooks_recover_procops(pnook_t nook, pnook_driver_t driver, int pass)
{
  struct proc_dir_entry * entry;

  if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }

  //
  // There are two pieces of state:
  // - a driver
  // - a resource for nook->kernel
  //
  // The driver may not be allocated, because that code may not have executed.
  //

  //
  // Delete the nook entry
  //

  entry = (struct proc_dir_entry *) driver->data;

  nooks_del_resource_elem_reverse(nook, proc_entry_e, entry, NULL);


  __remove_proc_entry(entry);
  driver->data = NULL;
  return(0);
}

int nooks_recover_netops(pnook_t nook, pnook_driver_t driver, int pass)
{
  struct net_device * device;
  device = (struct net_device *) driver->data;

  if (device == NULL) //Nothing to do
	return 0;
  if (pass == NOOKS_RECOVERY_STOP_PASS) {
      //
      // For net devices, stop the queue so no packets will be send.
      //
      netif_stop_queue(device);
      return(-EAGAIN);

  }    
  //
  // Just unregister the device - don't bother cleaning up callbacks,
  // because those will be deleted separately
  //


  unregister_netdev(device);  

  //
  // This was allocated with init_etherdev or kmalloc, so don't 
  // use nooks_kfree
  //

  nooks_kfree(device);
  
  return(0);
}

int nooks_recover_irqops(pnook_t nook, pnook_driver_t driver, int pass)
{
  unsigned int irq = (int) driver->data;


  if (pass == NOOKS_RECOVERY_STOP_PASS) {
    int result = 0;
    /*    
	  void * dev_id = NULL;
    
	  result =  nooks_find_resource_elem(nook, 
	  irq_dev_e, 
	  (void *) irq, 
	  &dev_id); 
    */
    if (result >= 0) {
      irq_desc_t * desc = irq_desc + irq;

      //
      // BUGBUG: can't do this from inside an interrupt handler
      //
      //
      disable_irq_nosync(irq);
      // If there are any other actions, turn on polling
      //
      if (desc->action != NULL) {
	nooks_poll_device((int) irq);
      }
      //
      // BUGBUG: bogus hack for pcnet32
      //
      
      nooks_set_recovering_interrupt(irq);
    }
    return(-EAGAIN);
      
  } else if (pass == NOOKS_RECOVERY_UNLOAD_PASS1) {
    int result;
    void * dev_id = NULL;
    
    result =  nooks_del_resource_elem(nook, 
				      irq_dev_e, 
				      (void *) irq, 
				      &dev_id);

    free_irq(irq, dev_id);
    //
    // Undo our prior disable
    //
    enable_irq(irq);
    return(-EAGAIN);

  } else if (pass == NOOKS_RECOVERY_CLEANUP_PASS) {
    //    result =  nooks_del_resource_elem(nook, 
    //    				      irq_dev_e, 
    //    				      (void *) irq, 
    //    				      &dev_id);
    //    if (result >= 0) {
    //      free_irq(irq, dev_id);
    //    }
    
    //
    // Re-enable the irq (disabled by disable drivers)
    //
    
    //
    // BUGBUG: For i810 driver
    //
    
    if (nooks_always_poll) {
      if (irq != 10) {
	nooks_unpoll_device(irq, 1);
	nooks_clear_recovering_interrupt(irq);
      }
    } else {
	nooks_unpoll_device(irq, 1);
	nooks_clear_recovering_interrupt(irq);
    }
    return(0);
  } else {
    return(-EAGAIN);
  }
  
}

int nooks_recover_pciops(pnook_t nook, pnook_driver_t driver, int pass)
{
  struct pci_driver * kernel_drv = NULL;

  printk("In recover pciops.\n");

  if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }

  nooks_del_resource_elem_reverse(nook, 
				  pci_driver_e, 
				  driver, 
				  NULL);
  kernel_drv = (struct pci_driver *) driver->data;

  if (kernel_drv == NOOKS_PCIOPS_ISAPNP) {
    //
    // BUGBUG: free kernel_ids from nooks-wrappers.c
    //
    return(0);
  }

  
  //
  // Freeing the driver will release any interfaces it is using
  //
  
  pci_unregister_driver(kernel_drv);

  if (kernel_drv->id_table != NULL) {
    nooks_kfree(kernel_drv->id_table);
  }

  nooks_kfree(kernel_drv);
  return(0);
}

int nooks_recover_callbackops(pnook_t nook, pnook_driver_t driver, int pass)
{
  //
  // BUGBUG: unclear what to do here - we have threads and tasklets. 
  // We don't even know how to get rid of a tasklet, or a thread.
  //
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }
  if (driver->data == NOOKS_CALLBACK_THREAD) {
  } else if (driver->data == NOOKS_CALLBACK_TASKLET) {
    struct tasklet_struct * tasklet = NULL;

    //
    // Find all the tasklets in the list that use this driver, and delete
    // them.
    //

    while (nooks_del_resource_elem_reverse(nook, tasklet_e, driver, (void **) &tasklet) == 0) {
      tasklet_kill(tasklet);
      nooks_kfree(tasklet);
      tasklet = NULL;
    }
  }

  return(0);
}

int nooks_recover_timerops(pnook_t nook, pnook_driver_t driver, int pass)
{
  nooks_del_resource_elem_reverse(nook, timer_e, driver, NULL);

  if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }
  if (del_timer((struct timer_list *) driver->data) >= 0) {
    nooks_kfree(driver->data);
  } 

  return(0);
}

int nooks_recover_inodeops(pnook_t nook, pnook_driver_t driver, int pass)
{
  //
  // Free it on the second pass to let inodes be freed first
  //

  if (pass < NOOKS_RECOVERY_CLEANUP_PASS) {
    return(-EAGAIN);
  } else {
    return(0);
  }
}

int nooks_recover_sysctlops(pnook_t nook, pnook_driver_t driver, int pass)
{
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS2) {
    return(-EAGAIN);
  } else {
    return(0);
  }
}

int nooks_recover_actorops(pnook_t nook, pnook_driver_t driver, int pass)
{
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }
  nooks_del_resource_elem_reverse(nook, actor_driver_e, driver, NULL);
  return(0);

}


int nooks_recover_notifierops(pnook_t nook, pnook_driver_t driver, int pass)
{
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS2) {
    return(-EAGAIN);
  }

  unregister_reboot_notifier((struct notifier_block *) driver->data);
  nooks_kfree(driver->data);

  return(0);
}


/*******************************************************************
 *
 * per-resource recovery functions
 *
 *******************************************************************/


int nooks_res_recover_generic(pnook_t nook, void * key, void * data, int pass)
{
  BUG();
  return(-EAGAIN);
}

int nooks_res_recover_fasync(pnook_t nook, void * key, void * data, int pass)
{
  return(-EAGAIN);
}

int nooks_res_recover_waitqueue(pnook_t nook, void * key, void * data, int pass)
{
  //
  // This is a wait_queue_head entry that points at a kernel. We need to wake
  // up anybody waiting on this wait queue.
  //
  // key = kernel wait queue entry
  // data =  driver waitqueue entry
  //
  // BUGBUG: do we use the key or the data? Only safe this is to key
  //
  wake_up_all((wait_queue_head_t *) key);

  //
  // BUGBUG: who does the remove wait queue here? I think we leave this for later
  //
  
  //   nooks_kfree(key);

  return(0);
}

int nooks_res_recover_waitqueue_sync(pnook_t nook, void * key, void * data, int pass)
{
  //
  // This is a wait_queue entry that points at a kernel. We need to wake
  // up anybody waiting on this wait queue.
  //
  // key = driver wait queue entry
  // data =  kernel waitqueue entry
  //
  // BUGBUG: do we use the key or the data? Only safe this is to key
  //
  
  nooks_kfree(data);

  return(0);
}

int nooks_res_recover_devfs_handle(pnook_t nook, void * key, void * data, int pass)
{
  // 
  // devfs drivers are cleaned up by the driver recovery code, and there
  // is no way to clean up directories
  //

  return(0);
}

int nooks_res_recover_netdev(pnook_t nook, void * key, void * data, int pass)
{
  //
  // This should be hanlded by the driver recovery
  //
  return(-0);
}

int nooks_res_recover_netdev_stats(pnook_t nook, void * key, void * data, int pass)
{
  //
  // This should be hanlded by the driver recovery
  //
  if (data != NULL) {
	  nooks_kfree(data);
  }
  return(-0);
}

int nooks_res_recover_netdev_nook(pnook_t nook, void * key, void * data, int pass)
{
  //
  // This should be hanlded by the driver recovery
  //
  return(0);
}

int nooks_res_recover_timer(pnook_t nook, void * key, void * data, int pass)
{
  //
  // Handled by driver recovery
  //

  return(0);
}

int nooks_res_recover_pci_driver(pnook_t nook, void * key, void * data, int pass)
{
  //
  // Handled by driver recovery
  //
  return(0);
}

int nooks_res_recover_pci_memory(pnook_t nook, void * key, void * data, int pass)
{
  if (pass >= NOOKS_RECOVERY_CLEANUP_PASS) {
    nooks_mm_free_pages(nook, 
			(unsigned long) key, 
			get_order((unsigned long) data));
    return(0);
  }

  return(-EAGAIN);
}

int nooks_res_recover_worker(pnook_t nook, void * key, void * data, int pass)
{
  //
  // The worker will be cleand up automatically, so do nothing
  //
  return(0);
}

int nooks_res_recover_proc_entry(pnook_t nook, void * key, void * data, int pass)
{

  struct proc_dir_entry * entry = (struct proc_dir_entry *) data;

  //
  // This should have been cleaned up by the driver. If it still exists,
  // then remove the proc entry. We can get here if the proc entry is a 
  // directory, in which case we need to be careful about deleting
  // them in order. If only we could record them in the order created...
  //

  
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS2) {
    return(-EAGAIN);
  }

  //
  // We need to remove any children before we remove this.
  //
  {
    struct proc_dir_entry *p;
    struct proc_dir_entry *n = NULL;
    int result = 0;
    
  retry:
    for (p = entry->subdir; p; p = p->next ) {
      //
      // If this child exists in the nook, call recursively to delete
      // the child
      //

      if (nooks_find_resource_elem_reverse(nook,
					   proc_entry_e,
					   p,
					   (void *)n) >= 0) {
	result = nooks_res_recover_proc_entry(nook,
					      n,
					      p,
					      pass);
	//
	// If the delete was successful, retry walking the children again
	//
	if (result >= 0) {
	  nooks_del_resource_elem_reverse(nook, proc_entry_e, p, NULL);
	  goto retry;
	}
      }
      
    }
  }
  __remove_proc_entry(entry);


  return(0);
}

int nooks_res_recover_tasklet(pnook_t nook, void * key, void * data, int pass)
{
  //
  // Should be handled by driver cleanup
  //

  return(0);
}

int nooks_res_recover_skbuff_k(pnook_t nook, void * key, void * data, int pass)
{
  struct sk_buff * skb = (struct sk_buff *) key;
  
  //
  // We don't have to do anything - these are sk_buffs owned by the kernel.
  // Just free the socket buffer
  //

  /* skbuffs no longer maintain list information. FIXME */
  //if (skb->list == NULL) {
  //  kfree_skb(skb);
  //}
  return(0);
}

int nooks_res_recover_skbuff_n(pnook_t nook, void * key, void * data, int pass)
{
  //
  // These are 
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS2) {
    return(-EAGAIN);
  } else {
    unsigned long size = (unsigned long) data;
    struct sk_buff * skb = (struct sk_buff *) key;
    //
    // Release the memory
    //
    /* nooks_mm_update_pagetable(nook, 
			      NOOKS_READONLY, 
			      (unsigned long) skb, 
			      size);
    */
    nooks_mm_free_pages(nook, (unsigned long) skb, nooks_log2pages(size));
  }
  return(0);
}

int nooks_res_recover_area_map(pnook_t nook, void * key, void * data, int pass)
{
  unsigned long address = (unsigned long) key;
  unsigned long size = (unsigned long) data;

  if (pass < NOOKS_RECOVERY_UNLOAD_PASS2) {
    return(-EAGAIN);
  }
  //
  // Make the range unreadable
  //
  /* nooks_mm_update_pagetable(nook, 
			    NOOKS_READONLY, 
			    address, 
			    size);
  */
  //
  // Free the range
  //

  vfree((void *) address);
  return(0);

}

int nooks_res_recover_pcidev(pnook_t nook, void * key, void * data, int pass)
{
  //
  // This is just a mapping - do nothing
  //

  return(0);
}

int nooks_res_recover_irq_dev(pnook_t nook, void * key, void * data, int pass)
{
  //
  // Should be handled by driver recovery
  //
  return(-EAGAIN);
}

int nooks_res_recover_region(pnook_t nook, void * key, void * data, int pass)
{
  pnooks_resource_t resource = (pnooks_resource_t) data;
  __release_region(resource->parent, resource->start, resource->n);

  nooks_kfree(resource->name);
  nooks_kfree(resource);
 

  return(0);
}


int nooks_res_recover_ctl_table(pnook_t nook, void * key, void * data, int pass)
{
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS2) {
    //
    // Do nothing - the drivers are freed already by the pass that 
    // frees drivers first.
    //
    return(-EAGAIN);
  } else {
    //
    // If we still exist in a later pass, go ahead and clean up
    //
    return(0);
  }
}

int nooks_res_recover_ctl_head(pnook_t nook, void * key, void * data, int pass)
{
  struct ctl_table_header * header = (struct ctl_table_header *) key;
  ctl_table * table = (ctl_table *) data;
  int i;

  unregister_sysctl_table(header);
  for (i = 0; table[i].ctl_name != 0; i++) {
    nooks_free_ctl_table(nook, &table[i]);
  }
  nooks_kfree(table);

  return(-ERESTART);
}


int nooks_res_recover_socket(pnook_t nook, void * key, void * data, int pass)
{
  struct socket * sock = (struct socket *) data;
  sock_release(sock);
  return(0);
}

int nooks_res_recover_waitqueuehead(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_sock(pnook_t nook, void * key, void * data, int pass)
{
  //
  // From tcp_v4_lookup_listener
  //
  sock_put((struct sock *) data);
  return(0);
}

int nooks_res_recover_locked_sock(pnook_t nook, void * key, void * data, int pass)
{
  release_sock((struct sock *) data);
  return(0);
}


int nooks_res_recover_tcp_open_req(pnook_t nook, void * key, void * data, int pass)
{
  //tcp_openreq_free((struct open_request *) data);
  return(0);
}


int nooks_res_recover_file(pnook_t nook, void * key, void * data, int pass)
{
  fput((struct file *) key);
  return(0);
}

int nooks_res_recover_actor_driver(pnook_t nook, void * key, void * data, int pass)
{
  //
  // the driver will be freed on its own
  //
  return(0);
}

int nooks_res_recover_fs(pnook_t nook, void * key, void * data, int pass)
{

  return(0);
}



int nooks_res_recover_superblock(pnook_t nook, void * key, void * data, int pass)
{
#if 0
  int i;
  struct super_block * super = (struct super_block *) key;
  struct list_head * next;
  //
  // Walk the list of files in the superblock, and close them (mark as
  // invalid) and nuke the dentries / inodes
  //

  file_list_lock();
  list_for_each(next, &super->s_files) {
    struct file * f;
    f = list_entry(next, struct file, f_list);
    f->f_op = &empty_fops;
    if (f->f_dentry != NULL) {
      f->f_dentry->d_op = &NOOKS_NWRAPPER(dentryops);
      if (f->f_dentry->d_inode != NULL) {
	f->f_dentry->d_inode->i_op = &empty_iops;
	if (f->f_dentry->d_inode->i_mapping != NULL) {
	  f->f_dentry->d_inode->i_mapping->a_ops = &empty_aops;
	}
      }
    }
  }
  file_list_unlock();


  spin_lock(&dcache_lock);
  for (i = 0; i <= d_hash_mask; i++) {
    list_for_each(next, (dentry_hashtable + i)) {
      struct list_head * next2;
      struct dentry * d = list_entry(next, struct dentry, d_hash);
      if (d->d_sb != super) {
	continue;
      }
      d->d_op = &NOOKS_NWRAPPER(dentryops);
      
      list_for_each(next2, &d->d_alias) {
	struct inode * i = list_entry(next2, struct inode, i_dentry);
	if (i->i_sb != super) {
	  continue;
	}
	i->i_op = &empty_iops;
	if (i->i_mapping != NULL) {
	  i->i_mapping->a_ops = &empty_aops;
	}
      }
    }
  }
  spin_unlock(&dcache_lock);
  
#endif
  return(0);
}

int nooks_res_recover_inode(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_inode_ops(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_dentry(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_dentry_ops(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_buffer_head(pnook_t nook, void * key, void * data, int pass)
{
  struct buffer_head * head = (struct buffer_head *) data;
  __brelse(head);
  return(0);
}

int nooks_res_recover_super_ops(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_file_ops(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_iomap(pnook_t nook, void * key, void * data, int pass)
{
  printk ("Not freeing for iounmap since we do it pre-migration.\n");
  //nooks_mm_release_pages(nook, (unsigned long) key, (unsigned long) data);
  //iounmap(key);

  return(0);
}

int nooks_res_recover_as_ops(pnook_t nook, void * key, void * data, int pass)
{

  return(0);
}

int nooks_res_recover_address_space(pnook_t nook, void * key, void * data, int pass)
{

  return(0);
}


int nooks_res_recover_getblock_driver(pnook_t nook, void * key, void * data, int pass)
{

  return(0);
}

int nooks_res_recover_buffer_data(pnook_t nook, void * key, void * data, int pass)
{

  return(0);
}


int nooks_res_recover_dma(pnook_t nook, void * key, void * data, int pass)
{
  free_dma((unsigned int)key);

  return(0);
}

int nooks_res_recover_vm_ops(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_misc_device(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_mmap(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_module_data(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_irqaction(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_ioctl(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_kmap(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}


int nooks_res_recover_pm_dev(pnook_t nook, void * key, void * data, int pass)
{
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }
  //pm_unregister((struct pm_dev *) key);
  return(0);
}


int nooks_res_recover_pciregion(pnook_t nook, void * key, void * data, int pass)
{
  struct pci_dev *dev = (struct pci_dev *) key;

  if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }
  printk ("Releasing PCI memory.\n");
  pci_release_regions(dev);

  return(0);
}

int nooks_res_recover_dmamappage(pnook_t nook, void * key, void * data, int pass)
{
  int result = 0;
  struct device *dev = NULL;
  short direction = 0;

  if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }

  result =
       nooks_find_resource_elem(nook, dma_device_e, key,
                                     (void **)&dev);

  result =
       nooks_find_resource_elem(nook, dma_direction_e, key,
                                     (void **)direction);

  if (result < 0) {
                PERROR("Failed to find dev for dma device/direction  operations.\n", dev);
                goto Cleanup;
   }
 
  if (dev == NULL)
	goto Cleanup;

  printk ("Releasing dma_map_page memory.\n");
  dma_unmap_page(dev, (dma_addr_t) key, (size_t) data, (enum dma_data_direction) direction); 

Cleanup:
  return(0);
}

int nooks_res_recover_dmamapsingle(pnook_t nook, void * key, void * data, int pass)
{
  int result = 0;
  struct device *dev = NULL;
  short direction = 0;

  if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }

  result =
       nooks_find_resource_elem(nook, dma_device_e, key,
                                     (void **)&dev);

  result =
       nooks_find_resource_elem(nook, dma_direction_e, key,
                                     (void **)direction);

  if (result < 0) {
                PERROR("Failed to find dev for dma device/direction  operations.\n", dev);
                goto Cleanup;
   }

  if (dev == NULL)
        goto Cleanup;

  printk ("Releasing dma_map_single memory.\n");
  dma_unmap_single(dev, (dma_addr_t) key, (size_t) data, (enum dma_data_direction) direction);

Cleanup:
  return(0);
}

int nooks_res_recover_vmalloc(pnook_t nook, void * key, void * data, int pass)
{
  int result = 0;

  printk ("Releasing vmalloc'ed memory.\n");
  vfree(key);

Cleanup:
  return(0);
}



int nooks_res_recover_notifier(pnook_t nook, void * key, void * data, int pass)
{
  //
  // nothing to do - handled by notifierops
  //
  return(0);
}


/***********************/


//void __remove_proc_entry(struct proc_dir_entry *entry)
//{
//	return;
//}

int nooks_res_recover_gendisk(pnook_t nook, void * key, void * data, int pass)
{
	
	int major,  result = 0;
	struct gendisk *gp=data;	

  	if(pass<NOOKS_RECOVERY_UNLOAD_PASS1)
		return -EAGAIN;
	

  	printk("Entered recovery for gendisk\n");	
	
	//result = nooks_del_resource_elem(nook, gendisk_e, nook_gp, 
	//					(void **) &gp);

	major=gp->major;

	printk("Deleting gendisk\n");


	if (result >= 0) {

		del_gendisk(gp);


		if(gp->driverfs_dev)
			nooks_kfree(gp->driverfs_dev);	

		if(gp->holder_dir)
			nooks_kfree(gp->holder_dir);

		if (gp->slave_dir)
			nooks_kfree(gp->slave_dir);

		if (gp->fops)
			nooks_kfree(gp->fops);
		
		if (gp->queue)
			nooks_kfree(gp->queue);
	
		//nooks_kfree(gp->part);
		nooks_kfree(gp);

	} else {

		PERROR ("Failed to find gendisk %p for nook %p for del_gendisk\n", data1, nook);
		return (-EAGAIN);
	}

  	printk("Exiting recovery for gendisk\n");	
	

  	return(0);

}

int nooks_res_recover_sector_info(pnook_t nook, void * key, 
				void * data, int pass)
{
  if(pass < NOOKS_RECOVERY_UNLOAD_PASS2)
	return -EAGAIN;

  nooks_kfree(key);

  return(0);
}

int nooks_res_recover_dummy1(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_dummy2(pnook_t nook, void * key, void * data, int pass)
{
  return(0);
}

int nooks_res_recover_bd_ops(pnook_t nook, void * key, 
				void *data, int pass)
{
	return (0);

}

 
int nooks_res_recover_major(pnook_t nook, void * key, 
				void *data, int pass)
{
  if (pass < NOOKS_RECOVERY_CLEANUP_PASS) {
    return (-EAGAIN);
  } else {
    return(0);
  }

}

/***********************/


