/*
 * nooks-wrappers.c -- Wrappers for system functions escpecially
 * for the e1000 driver. This to implement direct device access
 * VM migration support.
 * 2008 - Asim Kadav
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.
 * No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

#include "nooks-i.h"
#include "nooks-wrappers.h"
#include "nooks-iwrappers.h"
#include "nooks-workers.h"
#include "nooks-recover.h"
#include <../shadow/netshadow.h>
#include <net/tcp.h>
#include <linux/ctype.h>
#include <linux/rtnetlink.h>	// rtnl_lock()
#include <linux/ethtool.h>

struct nooks_dma_info_list nooks_dma_info;
static int is_dma_init = 0;
struct tasklet_struct nooks_w_bh_task_vec[32];
struct mm_struct nooks_w_init_mm;
struct resource nooks_w_iomem_resource;
struct softnet_data *nooks_w_softnet_data;
struct resource nooks_w_ioport_resource;
struct list_head nooks_w_pci_devices;
struct page *nooks_w_mem_map;
unsigned long nooks_last_wrapper_local;
unsigned long volatile nooks_w_jiffies;
unsigned long nooks_w___per_cpu_offset[NR_CPUS];
struct pci_dev *nooks_pci_device = NULL;
struct pci_driver *nooks_pci_driver = NULL;
void * nooks_ioremap_addr[2];
static int nooks_ioremap_ctr = 0;
static int nooks_device_shadow = 0;
struct net_device * nooks_net_device = NULL;
int shadow_migrated = 0;
int shadow_inject_fault = 0;

#define NOOKS_INTERFACE "eth1"
#define NOOKS_DEBUG
/* Initialize the wrappers.
 */

int nooks_wrappers_init(void)
{
	int cpu_id;
	nooks_w_init_mm = init_mm;
	nooks_w_iomem_resource = iomem_resource;
	nooks_w_ioport_resource = ioport_resource;
	nooks_w_mem_map = mem_map;
#ifdef CONFIG_NOOKS_STATS
	memcpy(nooks_w_ms_ops_count, ms_ops_count,
	       MS_OPS_NUM * sizeof(unsigned long));
#endif				// CONFIG_NOOKS_STATS
	cpu_id = smp_processor_id();
	nooks_w_softnet_data = &per_cpu(softnet_data, cpu_id);
	for (cpu_id = 0; cpuid < NR_CPUS; cpu_id++)
		nooks_w___per_cpu_offset[cpu_id] = __per_cpu_offset[cpu_id];

	nooks_w_pci_devices = pci_devices;

	//nooks_w_HYPERVISOR_shared_info = (shared_info_t *) empty_zero_page;
	return (0);

}

/* This is for hot-plugging of network devices. */
void NOOKS_KWRAPPER(netif_device_attach) (struct net_device * dev) {
	NOOKS_KWRAPPER_ENTER();
	NOOKS_SWRAPPER(netif_device_attach) (dev);
	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(netif_device_attach) (struct net_device * dev) {
	NOOKS_WRAPPER_ENTER();

	NOOKS_KERNEL_CALL1(nooks_k_netif_device_attach, dev);
	NOOKS_WRAPPER_EXIT();
	return;
}

/*FIXME */
struct sk_buff *NOOKS_WRAPPER(__netdev_alloc_skb) (struct net_device * dev,
						   unsigned int length,
						   gfp_t gfp_mask) {
	struct sk_buff *skb;
	pnook_t nook = GET_CURRENT_NOOK();

	if (nook == NULL)	{
		PDEBUG ("WARN(netdev_alloc_skb): Nook not found - tracking for nook1.\n");	
		nook = nooks_find_nook_internal("nook1");
		SET_CURRENT_NOOK(nook);
	}

	NOOKS_WRAPPER_ENTER();
	skb = nooks_w___alloc_skb(length + NET_SKB_PAD, gfp_mask, 0);
	if (skb == NULL)
		PDEBUG("Error allocating skb in __netdev_alloc_skb.");
	goto Cleanup;

	if (likely(skb)) {
		skb_reserve(skb, NET_SKB_PAD);
		skb->dev = dev;
	}

      Cleanup:
	NOOKS_WRAPPER_EXIT();
	return (skb);
}

/* This can't  be caught by the kfree semantics. 
void NOOKS_KWRAPPER (free_netdev) (struct net_device * dev)
{
  NOOKS_KWRAPPER_ENTER ();
  NOOKS_SWRAPPER (free_netdev) (dev);
  NOOKS_KWRAPPER_EXIT ();
  return;
}
*/

void nooks_w_free_netdev(struct net_device *dev)
{
	NOOKS_KWRAPPER_ENTER ();
	NOOKS_SWRAPPER (free_netdev) (dev);
	NOOKS_KWRAPPER_EXIT ();
	return;
}

void NOOKS_KWRAPPER(netif_device_detach) (struct net_device * dev) {
	NOOKS_KWRAPPER_ENTER();

	NOOKS_SWRAPPER(netif_device_detach) (dev);

	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(netif_device_detach) (struct net_device * dev) {
	NOOKS_WRAPPER_ENTER();

	NOOKS_KERNEL_CALL1(nooks_k_netif_device_detach, dev);
	NOOKS_WRAPPER_EXIT();
	return;
}

/* BigTBD  */
dma_addr_t NOOKS_KWRAPPER(dma_map_page) (struct device * dev,
					 struct page * page,
					 unsigned long offset, size_t size,
					 enum dma_data_direction direction) {

	 return(NOOKS_SWRAPPER(dma_map_page) (dev, page, offset, size, direction));
}

dma_addr_t NOOKS_WRAPPER(dma_map_page) (struct device * dev, struct page * page,
					unsigned long offset, size_t size,
					enum dma_data_direction direction) {
	dma_addr_t result = NULL;
	result =
	    NOOKS_KERNEL_CALL5(nooks_k_dma_map_page, dev, page, offset, size,
			       direction);
	//store_dma_info(dev, page, offset, size, direction);
	return (result);
}

/* This needs to be similar to pci_alloc consistent in 2.4. There have 
 * been changes in the kernel since then (additional search for pages
 * from a bitmap. Also, there are additional xen modifications done 
 * over 2.6 that needs to be taken care of.
 * 
 * Allocate some uncached, unbuffered memory for a device for performing 
 * DMA. This function allocates pages, and will return the CPU-viewed 
 * address, and sets handle to be the device-viewed address.  
 * The device here is not pci device but a separate device?
 */ 
void *NOOKS_KWRAPPER(dma_alloc_coherent) (struct device * dev, size_t size,
					  dma_addr_t * dma_handle, gfp_t flag) {



	return (dma_alloc_coherent(dev, size, dma_handle,flag));

}

void *NOOKS_WRAPPER(dma_alloc_coherent) (struct device * hwdev,
					 size_t size,
					 dma_addr_t * dma_handle, gfp_t gfp) {
	void *ret;
#if 0
	gfp &= ~(__GFP_DMA | __GFP_HIGHMEM);
	if (hwdev == NULL || *(hwdev->dma_mask) != 0xffffffff) {
		gfp |= GFP_DMA;
	}
#endif
	ret =
	    (void *)NOOKS_KERNEL_CALL4(nooks_k_dma_alloc_coherent,
				       hwdev, size, dma_handle, gfp);
#if 0
	if (ret != NULL) {
		memset(ret, 0, size);
		*dma_handle = virt_to_bus(ret);
	}
#endif
	NOOKS_WRAPPER_EXIT();
	return (ret);
}

/*TBD  -- We don't do anything for now, later we may need to do 
 * fix this if irqs cannt be restored safely.
 */
void NOOKS_KWRAPPER(_spin_unlock_irqrestore) (spinlock_t * lock,
					      unsigned long flags) {
	NOOKS_KWRAPPER_ENTER();
	NOOKS_SWRAPPER(_spin_unlock_irqrestore) (lock, flags);
	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(_spin_unlock_irqrestore) (spinlock_t * lock,
					     unsigned long flags) {
	NOOKS_KERNEL_CALL2(nooks_k__spin_unlock_irqrestore, lock, flags);
	return;
}

void NOOKS_KWRAPPER(pci_release_regions) (struct pci_dev * dev) {
	int result = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_KWRAPPER_ENTER();

	NOOKS_SWRAPPER(pci_release_regions) (dev);
	/* Clear associated pci_region. */
	result =  nooks_del_resource_elem(nook, pciregion_e,
					    dev, NULL);
  	if (result < 0)
		printk ("Fatal: Error in pci_release_regions(WRAPPER) object tracking mechanism.\n");
	NOOKS_KWRAPPER_EXIT();
}

void NOOKS_WRAPPER(pci_release_regions) (struct pci_dev * dev) {
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL1(nooks_k_pci_release_regions, dev);
	NOOKS_WRAPPER_EXIT();
}

void NOOKS_WRAPPER(dump_stack) (void) {
	dump_stack();
}

/* No wrappers entry/exit calls here - probably to avoid upsetting the
 * purpose of the function.
*/

void NOOKS_WRAPPER(__const_udelay) (unsigned long usecs) {
	__const_udelay(usecs);
	return;
}

void NOOKS_KWRAPPER(pci_unregister_driver) (struct pci_driver * pci_driver) {
	int result = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	struct pci_driver *kernel_dev = NULL;

	NOOKS_KWRAPPER_ENTER();

	NOOKS_SWRAPPER(pci_unregister_driver) (pci_driver);
	printk ("pci_unregister_driver returns.\n");
        result =
            nooks_del_resource_elem(nook, pci_driver_e, pci_driver,
                                    (void **)&driver);

        if (result < 0) {
                PERROR(" Failed to find pci driver for cleanup: %p\n",
                       pci_driver);
                goto Cleanup;
        }

	//Freeing the driver will release any interfaces it is using.

	//if (driver !=NULL)
	//	nooks_free_driver(driver);

      Cleanup:
	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(pci_unregister_driver) (struct pci_driver * arg0) {
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL1(nooks_k_pci_unregister_driver, arg0);
	NOOKS_WRAPPER_EXIT();
	return;
}

/* The network interface drivers use the netif_carrier_on() and 
 * netif_carrier_off() functions to update their interface card 
 * status.
 */

void NOOKS_KWRAPPER(netif_carrier_off) (struct net_device * dev) {

	NOOKS_KWRAPPER_ENTER();
	NOOKS_SWRAPPER(netif_carrier_off) (dev);
	NOOKS_KWRAPPER_EXIT();
}

void NOOKS_WRAPPER(netif_carrier_off) (struct net_device * dev) {
	NOOKS_KERNEL_CALL1(nooks_k_netif_carrier_off, dev);
	return;
}

/* Are there any changes required here? TBD  - relook. */
void NOOKS_KWRAPPER(kunmap_atomic) (void *kvaddr, enum km_type type) {
	unsigned long pgP;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_KWRAPPER_ENTER();
	NOOKS_SWRAPPER(kunmap_atomic) (kvaddr, type);
	if (nooks_del_from_hash
	    (&nook->resources.area_map, kmap_e, (void *)kvaddr,
	     (void **)&pgP) < 0)
		PDEBUG("Object tracking failed in kunmap_atomic.\n");

	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(kunmap_atomic) (void *kvaddr, enum km_type type) {
	NOOKS_KERNEL_CALL2(nooks_k_kunmap_atomic, kvaddr, type);
	return;
}

void *nooks_w_kmalloc(size_t size, int flags)
{
	void *result;
	printk("In nooks_w_kmalloc.");
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_WRAPPER_ENTER();

	//result = nooks_mm_kmalloc(nook, size, flags);
	result = kmalloc(size, flags);
	if (result == NULL)
		printk ("FATAL:kmalloc failed for size %d.\n", size);

	//nooks_add_resource_elem(nook, kmalloc_e, result, size);
	
	if (nooks_add_to_hash(&nook->resources.area_map,
        	                     area_map_e,
                	             result,
                        	     (void *) size) <  0)
		printk ("Object tracker in nooks_w_kmalloc is having problems.\n");



	NOOKS_WRAPPER_EXIT();

	return (result);
}

/* This is similar to the Nooks 2.4 version. We write wrapper for alloc_netdev
 * the function which is called by alloc_etherdev.
 */

/*Forward Declerations.*/
void nooks_w_ether_setup(struct net_device *dev);
struct net_device *NOOKS_WRAPPER(alloc_netdev) (int sizeof_priv,
						const char *name,
						void (*setup) (struct net_device
							       *));

struct net_device *NOOKS_WRAPPER(alloc_etherdev) (int sizeof_priv) {

	return nooks_w_alloc_netdev(sizeof_priv, "eth%d", ether_setup);
}

struct net_device *NOOKS_WRAPPER(alloc_netdev) (int sizeof_priv,
						const char *name,
						void (*setup) (struct net_device
							       *)) {

	struct net_device *dev = NULL;
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
  	static int dev_count = 0;	
	void *p;
	int alloc_size;

	NOOKS_WRAPPER_ENTER();

	printk ("In the nooks wrapper alloc_netdev with dev_count %d and name %s.\n", dev_count, name);
	if (strcmp(name, "sn%d") == 0)
	 	driver = shadow_find_driver(nook, netops_e, dev_count);
  	else
	{
#if NOOKS_MIGRATE
		if (NULL != nooks_net_device)
			return nooks_net_device;			
#endif
		driver = shadow_find_driver(nook, netops_e, 1);	
	}

	if (driver != NULL)	{	
		printk ("In alloc_netdev: Device exists as %p for device %d(%s).\n", driver->data, dev_count, dev->name);
		if (strcmp(name, "sn%d") == 0)	
		{
			dev_count ++;
			if (dev_count == 2)
                	        dev_count = 0;
		}
		//return (struct net_device *) driver->data; //may need to be commented for forcedeth
	}

	
	printk ("No shadow found.\n");
	
	/* Ensure 32-byte alignment of both the device and private area. */
	alloc_size = (sizeof(*dev) + NETDEV_ALIGN_CONST) & ~NETDEV_ALIGN_CONST;
	alloc_size += sizeof_priv + NETDEV_ALIGN_CONST;

	p = nooks_w_kmalloc(alloc_size, GFP_KERNEL);
	//p = kzalloc(alloc_size, GFP_KERNEL);  // We never free the netdevice - no need to keep track.
	if (!p) {
		PERROR(KERN_ERR "alloc_dev: Unable to allocate device.\n");
		goto Cleanup;
	}

	dev = (struct net_device *)
	    (((long)p + NETDEV_ALIGN_CONST) & ~NETDEV_ALIGN_CONST);
	dev->padded = (char *)dev - (char *)p;

	memset(dev, 0, alloc_size);

	if (sizeof_priv != 0) {
		dev->priv = (char *)dev + ((sizeof(struct net_device) +
					    NETDEV_ALIGN_CONST) &
					   ~NETDEV_ALIGN_CONST);
	}

	setup(dev);
	strcpy(dev->name, name);

      Cleanup:
	printk ("alloc_netdev returns with device name as %s.\n",dev->name);
	NOOKS_WRAPPER_EXIT();
	return (dev);

}

pci_power_t NOOKS_KWRAPPER(pci_choose_state) (struct pci_dev * dev,
					      pm_message_t state) {
	pci_power_t result = NULL;
	NOOKS_KWRAPPER_ENTER();
	result = NOOKS_SWRAPPER(pci_choose_state) (dev, state);
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

pci_power_t NOOKS_WRAPPER(pci_choose_state) (struct pci_dev * dev,
					     pm_message_t state) {
	pci_power_t result;
	result =
	    (pci_power_t) NOOKS_KERNEL_CALL2(nooks_k_pci_choose_state,
					     dev, state);
	return (result);
}

/* This wrapper is not required. Is this because this function is never
 * called in the kernel domain (or called by the kernel?

int NOOKS_KWRAPPER(pci_set_power_state)(struct pci_dev *dev, pci_power_t state)
{
  int result;
  NOOKS_KWRAPPER_ENTER();
  result = NOOKS_SWRAPPER(pci_set_power_state)(dev, state);
  NOOKS_KWRAPPER_EXIT();
  return(result);
}
*/

/* The wrapper does nothing here TBD */
int NOOKS_WRAPPER(pci_set_power_state) (struct pci_dev * dev, pci_power_t state) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = pci_set_power_state(dev, state); //Need SWRAPPER here TBD FIXME
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* iounmap() - This function unmaps a I/O address range that was 
 * mapped into the kernel address range by the ioremap function.
 * iomap_e is a resource type. 
 *
 */
void NOOKS_KWRAPPER(iounmap) (volatile void *addr) {
	pnook_t nook = GET_CURRENT_NOOK();
	unsigned long size = 0;
	NOOKS_SWRAPPER(iounmap) (addr);
	if (nooks_del_resource_elem(nook, iomap_e, addr, (void **)&size) < 0) {
		printk("Fatal:Problems with object tracker in iounmap(WRAPPER).\n");
	}
}

/* Just call the kernel iounmap function. */
void NOOKS_WRAPPER(iounmap) (volatile void *addr) {
	NOOKS_KERNEL_CALL1(nooks_k_iounmap, addr);
	return;
}

void NOOKS_KWRAPPER(netif_carrier_on) (struct net_device * dev) {

	NOOKS_KWRAPPER_ENTER();
	NOOKS_SWRAPPER(netif_carrier_on) (dev);
	NOOKS_KWRAPPER_EXIT();
}

void NOOKS_WRAPPER(netif_carrier_on) (struct net_device * dev) {
	NOOKS_KERNEL_CALL1(nooks_k_netif_carrier_on, dev);
	return;
}

unsigned long NOOKS_KWRAPPER(schedule_generic) (long state, long timeout) {
	unsigned long ret = 0;
	NOOKS_KWRAPPER_ENTER();
	if (state != -1) {
		current->state = state;
	}
	if (timeout != -1) {
		ret = NOOKS_SWRAPPER(schedule_timeout) (timeout);
	} else {
		NOOKS_SWRAPPER(schedule) ();
	}
	NOOKS_KWRAPPER_EXIT();
	return ret;
}

signed long NOOKS_WRAPPER(schedule_timeout_interruptible) (signed long timeout) {
	signed long result;
	result =
	    (signed long)NOOKS_KERNEL_CALL2(nooks_k_schedule_generic,
					    TASK_INTERRUPTIBLE, timeout);
	return (result);
}

void NOOKS_WRAPPER(msleep) (unsigned int msecs) {
	NOOKS_SWRAPPER(msleep) (msecs);
}

unsigned long NOOKS_WRAPPER(msleep_interruptible) (unsigned int msecs) {
	unsigned long result;
	result = (unsigned long)NOOKS_SWRAPPER(msleep_interruptible) (msecs);
	return result;
}

/*TBD Will need a relook */
int NOOKS_KWRAPPER(dev_close) (struct net_device * dev) {
	int result;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_KWRAPPER_ENTER();
	result = NOOKS_SWRAPPER(dev_close) (dev);

	if (result >= 0)
		if (nooks_del_resource_elem(nook, netdev_e, dev, NULL)
		    <= 0)
			PDEBUG("Object tracking failed at dev_close.\n");

	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(dev_close) (struct net_device * dev) {
	int result;
	result = (int)NOOKS_KERNEL_CALL1(nooks_k_dev_close, dev);
	return (result);
}

/* Implemented as is from the 2.4 verison with very minor changes.
 */
int NOOKS_KWRAPPER(pci_request_regions) (struct pci_dev * dev, const char *name) {

	int result = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	char *kernel_name = NULL;
	NOOKS_KWRAPPER_ENTER();

	//printk ("DMA WARN:Allocating PCI Memory(pci_request_regions).\n");	
	result = NOOKS_SWRAPPER(pci_request_regions)(dev, name);
	if (result >= 0)	{
		result =
	    		nooks_add_resource_elem(nook, pciregion_e,
				    		dev, NULL);
	
		if (result < 0) 
			printk ("Fatal: Object Tracking for pci_request_regions(WRAPPER) failed.\n");
		
	}	
	
      Cleanup:
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(pci_request_regions) (struct pci_dev * arg0, const char *arg1) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result =
	    (int)NOOKS_KERNEL_CALL2(nooks_k_pci_request_regions, arg0, arg1);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

resource_size_t __iobase = 0;
resource_size_t __lenarg = 0;
struct resource * __resp = NULL;
/* Implemented as is from the 2.4 verison with very minor changes.
*/ 
struct resource * NOOKS_KWRAPPER(__request_region) (struct resource * rp, resource_size_t arg0, resource_size_t arg1, const char * arg2) {

	struct resource * result = NULL;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_KWRAPPER_ENTER();

	//printk ("DMA WARN:Allocating PCI Memory(pci_request_regions).\n");	
	result = __request_region(rp, arg0, arg1, arg2);
        __iobase = arg0;
	__lenarg = arg1;
	__resp = rp;	
	printk ("Result %d: SAving %d and %d and %p.\n", result, __iobase, __lenarg, __resp);	
      Cleanup:
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

struct resource * NOOKS_WRAPPER(__request_region) (struct resource * rp, resource_size_t arg0, resource_size_t arg1, const char *arg2) {
	//return __request_region(rp, arg0, arg1, arg2);
	struct resource * result;
	NOOKS_WRAPPER_ENTER();
	result =
            NOOKS_KERNEL_CALL4(nooks_k___request_region, rp, arg0, arg1, arg2);
        NOOKS_WRAPPER_EXIT();
        return (result);
}



int NOOKS_WRAPPER(pci_bus_write_config_dword) (struct pci_bus * bus,
					       unsigned int devfn,
					       int where, u32 val) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = (int)pci_bus_write_config_dword(bus, devfn, where, val);

	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* The function eth_type_trans is used to determine the packet's 
 * protocol ID. No WRAPPER required here.
 */

__be16 NOOKS_WRAPPER(eth_type_trans) (struct sk_buff * skb,
				      struct net_device * dev) {
	__be16 result;
	NOOKS_WRAPPER_ENTER();
	result = eth_type_trans(skb, dev);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

#if 0
struct net_device *(dev_from_nooks_dev) (struct net_device * dev) {
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	struct net_device *kernel_dev = NULL;
	int result = 0;
	result = 
	  nooks_find_resource_elem(nook, netdev_nook_e, dev,
				     (void **)&driver);

	if (result < 0) {
		PERROR("Failed to find netdev for ethtool operations.\n", dev);
		goto Cleanup;
	}

	kernel_dev = (struct net_device *)driver->data;
	return (kernel_dev);

      Cleanup:
	return NULL;
}
#endif

u32 NOOKS_WRAPPER(ethtool_op_get_link) (struct net_device * dev) {
	u32 result;
	NOOKS_WRAPPER_ENTER();
	result = ethtool_op_get_link(dev);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(ethtool_op_get_perm_addr) (struct net_device * dev,
					     struct ethtool_perm_addr * addr,
					     u8 * data) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = ethtool_op_get_perm_addr(dev, addr, data);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

u32 NOOKS_WRAPPER(ethtool_op_get_sg) (struct net_device * dev) {
	u32 result;
	NOOKS_WRAPPER_ENTER();
	result = ethtool_op_get_sg(dev);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

u32 NOOKS_WRAPPER(ethtool_op_get_tso) (struct net_device * dev) {
	u32 result;
	NOOKS_WRAPPER_ENTER();
	result = ethtool_op_get_tso(dev);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(ethtool_op_set_sg) (struct net_device * dev, u32 data) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = ethtool_op_set_sg(dev, data);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* TBD Will need a relook. */
int NOOKS_KWRAPPER(dev_open) (struct net_device * dev) {
	int result;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_KWRAPPER_ENTER();
	result = NOOKS_SWRAPPER(dev_open) (dev);

	if (result >= 0)
		if ((nooks_add_resource_elem(nook, netdev_e, dev, NULL)
		     < 0))
			PDEBUG
			    ("Fatal:Object tracker failed to track net_device at dev_open(WRAPPER).");


	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(dev_open) (struct net_device * dev) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = (int)NOOKS_KERNEL_CALL1(nooks_k_dev_open, dev);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* Written from the 2.4 version. */
int NOOKS_WRAPPER(pci_set_dma_mask) (struct pci_dev * dev, u64 mask) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = pci_set_dma_mask(dev, mask);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* TBD */
int NOOKS_WRAPPER(pci_bus_read_config_word) (struct pci_bus * bus,
					     unsigned int devfn,
					     int where, u16 * val) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = (int)pci_bus_read_config_word(bus, devfn, where, val);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* TBD -- May need a relook */
void *NOOKS_KWRAPPER(kmap_atomic) (struct page * page, enum km_type type) {
	pnook_t nook = GET_CURRENT_NOOK();
	void *result;
	NOOKS_KWRAPPER_ENTER();

	result = NOOKS_SWRAPPER(kmap_atomic) (page, type);

	if (nooks_add_to_hash
	    (&nook->resources.area_map, kmap_e, (void *)result,
	     (void *)page) < 0) {
		printk	
		    ("Failed to track kmap_atomic pointer by the object tracker.\n");
	}
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

void *NOOKS_WRAPPER(kmap_atomic) (struct page * page, enum km_type type) {
	void *result;
	result = (void *)NOOKS_KERNEL_CALL2(nooks_k_kmap_atomic, page, type);
	return (result);
}

void NOOKS_WRAPPER(__udelay) (unsigned long usecs) {
	udelay(usecs);
	return;
}

/* Free an interrupt. */
void NOOKS_KWRAPPER(free_irq) (unsigned int irq, void *dev_id) {
	pnook_driver_t driver = NULL;
	void *kernel_dev_id = NULL;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_KWRAPPER_ENTER();

	//Find the driver to cleanup.

	driver = nooks_find_driver_in_nook(NOOKS_IRQ_TO_MAJOR(irq), nook);
	if (driver == NULL)
		PERROR(" ERROR: no driver %d in nook %s, function %s\n",
		       NOOKS_IRQ_TO_MAJOR(irq), nook->name, __FUNCTION__);
	else {
		int result;
		result =
		    nooks_find_resource_elem(nook, irq_dev_e,
					     driver->data, &kernel_dev_id);
		if ((result >= 0) && (kernel_dev_id == dev_id)) {
			nooks_del_resource_elem(nook, irq_dev_e,
						driver->data, &kernel_dev_id);
			NOOKS_SWRAPPER(free_irq) (irq, dev_id);
			//Freeing the driver will release any interface it is using.
			nooks_free_driver(driver);
		}
	}
	NOOKS_KWRAPPER_EXIT();
}

void NOOKS_WRAPPER(free_irq) (unsigned int arg0, void *arg1) {
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL2(nooks_k_free_irq, arg0, arg1);
	NOOKS_WRAPPER_EXIT();
	return;
}

/*
 *  We allocate local memory for the kmalloc call. 
 */

void *NOOKS_WRAPPER(__kmalloc) (size_t size, gfp_t flags) {
	void *result;
	NOOKS_WRAPPER_ENTER();

	result = __kmalloc(size, flags);
/*
	result = nooks_mm_kmalloc(GET_CURRENT_NOOK(),
#ifdef NOOKS_KEEP_STATS
				  size + NOOKS_MM_KMALLOC_OVERHEAD,
#else				// NOOKS_KEEP_STATS
				  size,
#endif				// NOOKS_KEEP_STATS
				  flags);
#ifdef NOOKS_KEEP_STATS
	if (result != NULL) {
		NOOKS_STATS_ADD(NOOKS_OPS_W_KMALLOC_BYTES, size);
		*(size_t *) result = size;
		result = (u8 *) result + NOOKS_MM_KMALLOC_OVERHEAD;
	}
#endif
*/
	NOOKS_WRAPPER_EXIT();
	return (result);
}

	/* Nothing to be done. */
void NOOKS_WRAPPER(dma_sync_single_for_device) (struct device * dev,
						dma_addr_t dma_handle,
						size_t size,
						enum dma_data_direction
						direction) {
	NOOKS_KERNEL_CALL4(dma_sync_single_for_device, dev, dma_handle,
			   size, direction);
	return;
}


void NOOKS_KWRAPPER(pci_set_master) (struct pci_dev * dev) {
	NOOKS_SWRAPPER(pci_set_master) (dev);

}

void NOOKS_WRAPPER(pci_set_master) (struct pci_dev * dev) {
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL1(nooks_k_pci_set_master, dev);
	NOOKS_WRAPPER_EXIT();
}

int NOOKS_WRAPPER(netpoll_trap) (void) {
	int result;
	result = (int)NOOKS_KERNEL_CALL0(netpoll_trap);
	return (result);
}

int NOOKS_WRAPPER(_spin_trylock) (spinlock_t * lock) {
	int result;
	result = (int)NOOKS_KERNEL_CALL1(_spin_trylock, lock);
	return (result);
}

/* unregister_netdev() shuts down a device interface and removes it
 * from the kernel tables. This code is coming from the 2.4 version 
 * as-is.
 */
/* We do not create any extra drivers/devices so no need to delete them.
 * We just free the driver and netdev_stats.
 */

void NOOKS_KWRAPPER(unregister_netdev) (struct net_device * dev) {
	int result = 0;
	pnook_driver_t driver = NULL;
	pnook_driver_t nook_rec = NULL;
	pnook_t nook = GET_CURRENT_NOOK();
	struct net_device_stats *kernel_stats = NULL;
	short in_nook = 0;
	printk("Entered unregister_netdev.\n");
	if (nook == NULL) {
		printk("Fatal: Nook is null.\n");
	} else
		printk("In nook %s", nook->name);
	if (IN_NOOK()) {
		in_nook = 1;
		EXIT_NOOK();
	}

	NOOKS_KWRAPPER_ENTER();

	/* We call unregister_netdev first because the hash enteries
	 * may be used by the dwrappers.
	 */
	NOOKS_SWRAPPER(unregister_netdev) (dev);
	result =
	    nooks_del_resource_elem(nook, netdev_stats_e, dev,
				    (void **)&kernel_stats);

	if (result < 0) {
		PERROR(" failed to find netdev_stats for dev %p\n", dev);
		NOOKS_FAULT(resource);
		result = -EFAULT;
		goto Cleanup;
	}

	if (kernel_stats != NULL)
		nooks_kfree(kernel_stats);

	result =
	    nooks_del_resource_elem(nook, netdev_nook_e, dev, (void **)&driver);

	if (driver)
		nooks_free_driver(driver);

	printk("Unregister_netdev completes fine.\n");
      Cleanup:
	NOOKS_KWRAPPER_EXIT();
	if (in_nook)
		ENTER_NOOK();

	return;
}

void NOOKS_WRAPPER(unregister_netdev) (struct net_device * dev) {
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL1(nooks_k_unregister_netdev, dev);
	NOOKS_WRAPPER_EXIT();
	return;
}

/* Just directly call the associated kernel function. */
void NOOKS_WRAPPER(disable_irq) (unsigned int irq) {
	NOOKS_KERNEL_CALL1(disable_irq, irq);
	return;
}

/* Request for an irq.
 */

extern unsigned long nooks_recovering_interrupts[NR_IRQS];

int NOOKS_KWRAPPER(request_irq) (unsigned int irq,
				 irqreturn_t(*handler) (int, void *,
							struct pt_regs
							*),
				 unsigned long irqflags,
				 const char *devname, void *dev_id) {

	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	int result = 0, shadow_driver = 0;
	char *kernel_name = NULL;

	/* When dev is opened for recovery, nook is not set (e1000). This is a 
	 * a workaround hack and ideally one should do a SET_CURRENT_NOOK at
	 * shadow_netdev_recovery.
	 */
        if (nook == NULL)       {
                PDEBUG ("WARN(request_irq): Nook not found - tracking for nook1.\n");
                nook = nooks_find_nook_internal("nook1");
                SET_CURRENT_NOOK(nook);
        }

	NOOKS_KWRAPPER_ENTER();

	printk ("In request_irq(nooks wrapper).\n");
	result = nooks_get_driver_string(devname, &kernel_name);
	strcpy (devname, kernel_name);
	if (result < 0)
		goto Cleanup;

	if (IN_SHADOW_RECOVERY())
		driver = shadow_find_driver2(nook, irqops_e, irq, kernel_name);
	else
		driver = NULL;


	if (driver != NULL)
		shadow_driver = 1;
	else {
		//Allocate a new driver object.
		driver =
		    nooks_create_driver(nook, irqops_e,
					NOOKS_IRQ_TO_MAJOR(irq), kernel_name);
		if (driver == NULL) {
			PERROR
			    (" Failed to reate driver %d for request irq\n",
			     NOOKS_IRQ_TO_MAJOR(irq));
			result = -ENOMEM;
			goto err1;
		}

		driver->data = (void *)irq;
	}

	nooks_iwrapper_copy_irqops(handler, &driver->operations.irqops,
				   driver->iwrapper_operations.irqops,
				   &driver->wrapper_operations.irqops);
	if (!shadow_driver)
		shadow_add_sirqops(driver, (void *)irq);


	result =
	    request_irq 		(irq,
					 driver->wrapper_operations.irqops.
					 handler, irqflags, driver->name,
					 dev_id);


	
	if (result < 0)
		goto err2;

	// Re-enable the irq (disabled by disable drivers)
	//
	result = nooks_add_resource_elem(nook, irq_dev_e, (void *)irq, dev_id);
#if 0
	if (nooks_recovering_interrupts[irq] == 0) {
		nooks_unpoll_device(irq, 0);

	} else {
		nooks_poll_device(irq);
	}
	enable_irq(irq);
#endif

	goto Cleanup;
      err2:
	nooks_free_driver(driver);
      err1:
	nooks_free_driver_string(kernel_name);
Cleanup:
	kernel_name = NULL;
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(request_irq) (unsigned int arg0,
				irqreturn_t(*handler) (int, void *,
						       struct pt_regs
						       *),
				unsigned long arg1, const char *arg2,
				void *arg3) {
	int result;
	result =
	    (int)NOOKS_KERNEL_CALL5(nooks_k_request_irq, arg0, handler,
				    arg1, arg2, arg3);
	return (result);
}

int NOOKS_WRAPPER(pci_enable_wake) (struct pci_dev * dev,
				    pci_power_t state, int enable) {
	int result;
	result = (int)pci_enable_wake(dev, state, enable);
	return (result);
}

/* Use the nooks_w_dev_kfree_skb_irq from 2.4 implementation
 * to implement this function. This function can be called
 * within or out of interrupt context. This has been  
 * implemented without  deferred function using work queues.
 */
void NOOKS_WRAPPER(dev_kfree_skb_any) (struct sk_buff * skb) {
	pnook_t nook = GET_CURRENT_NOOK();
	if (nook == NULL)       {
                //printk ("WARN(dev_kfree_skb_any): Nook not found - tracking for nook1.\n");
		SET_CURRENT_NOOK(nooks_find_nook_internal("nook1"));
        }
	nooks_mm_move_skb_to_kernel(nook, skb);
	dev_kfree_skb_any(skb);
}

int NOOKS_KWRAPPER(pci_set_mwi) (struct pci_dev * dev) {
	int result;
	NOOKS_KWRAPPER_ENTER();
	result = NOOKS_SWRAPPER(pci_set_mwi) (dev);
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(pci_set_mwi) (struct pci_dev * dev) {
	int result;
	result = (int)NOOKS_KERNEL_CALL1(nooks_k_pci_set_mwi, dev);
	return (result);
}

/* May need a relook - TBD */
int NOOKS_WRAPPER(___pskb_trim) (struct sk_buff * skb, unsigned int len) {
	int result;
	pnook_t nook = GET_CURRENT_NOOK();

	NOOKS_WRAPPER_ENTER();

	nooks_mm_move_skb_to_kernel(nook, skb);

	result = (int)NOOKS_KERNEL_CALL2(___pskb_trim, skb, len);

	if (skb != NULL)
		nooks_mm_move_skb_to_nook(nook, skb);

	NOOKS_WRAPPER_EXIT();
	return (result);
}

void NOOKS_KWRAPPER(dma_unmap_single) (struct device * dev,
				       dma_addr_t dma_addr, size_t size,
				       enum dma_data_direction direction) {
	NOOKS_SWRAPPER(dma_unmap_single) (dev, dma_addr, size, direction);
	del_dma_info (dev, dma_addr, size, direction);
	return;
}

void NOOKS_WRAPPER(dma_unmap_single) (struct device * dev,
				      dma_addr_t dma_addr, size_t size,
				      enum dma_data_direction direction) {
	NOOKS_KERNEL_CALL4(nooks_k_dma_unmap_single, dev, dma_addr,
			   size, direction);
	return;
}

/* Force an event channel callback. 
void NOOKS_WRAPPER(force_evtchn_callback) (void) {
	NOOKS_KERNEL_CALL0(force_evtchn_callback);
	return;
}
*/
void NOOKS_KWRAPPER(dma_sync_single_for_cpu) (struct device * dev,
					      dma_addr_t dma_handle,
					      size_t size,
					      enum dma_data_direction
					      direction) {
	NOOKS_KWRAPPER_ENTER();
	NOOKS_SWRAPPER(dma_sync_single_for_cpu) (dev, dma_handle, size,
						 direction);
	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(dma_sync_single_for_cpu) (struct device * dev,
					     dma_addr_t dma_handle,
					     size_t size,
					     enum dma_data_direction
					     direction) {
	NOOKS_KERNEL_CALL4(nooks_k_dma_sync_single_for_cpu, dev,
			   dma_handle, size, direction);
	return;
}

/* This is helper function for kernel parameters. */

int NOOKS_WRAPPER(param_array_set) (const char *val, struct kernel_param * kp) {
	int result;
	result = (int)NOOKS_KERNEL_CALL2(param_array_set, val, kp);
	return (result);
}

/* Flush the work-queue. */
void NOOKS_WRAPPER(flush_scheduled_work) (void) {
	NOOKS_KERNEL_CALL0(flush_scheduled_work);
	return;
}

int NOOKS_WRAPPER(param_array_get) (char *buffer, struct kernel_param * kp) {
	int result;
	result = (int)NOOKS_KERNEL_CALL2(param_array_get, buffer, kp);
	return (result);
}

/* Process security function. */
int NOOKS_WRAPPER(capable) (int cap) {
	int result;
	result = (int)NOOKS_KERNEL_CALL1(capable, cap);
	return (result);
}

/* TBD -- This may need a relook as we do not know if such trading
 * of skb works correctly.  
 */

unsigned char *NOOKS_WRAPPER(__pskb_pull_tail) (struct sk_buff * skb, int delta) {
	unsigned char *result = NULL;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_WRAPPER_ENTER();
	
	nooks_mm_move_skb_to_kernel(nook, skb);

	result =
	    (unsigned char *)NOOKS_KERNEL_CALL2(__pskb_pull_tail, skb, delta);
	if (result == NULL)
		PDEBUG("__pskb_pull_tail wrapper call failed.\n");

	/* Play safe */
	if (skb != NULL) 
		nooks_mm_move_skb_to_nook(nook, skb);
	else
		PDEBUG
		    ("__pskb_pull_tail wrapper call failed while moving skb to nook. \n");
      Cleanup:
	NOOKS_WRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(netif_rx) (struct sk_buff * skb) {
	int rc = -1;
	int result = -1;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_WRAPPER_ENTER();

        if (nook == NULL)       {
                //printk ("WARN(kfree_skb): Nook not found - tracking for nook1.\n");
                nook = nooks_find_nook_internal("nook1");
                SET_CURRENT_NOOK(nook);
        }

	
	rc = nooks_mm_move_skb_to_kernel(nook, skb);

	/* If the skb is already in kernel, above procedure call will fail.
         */
	if (rc < 0) {
		//printk("WARN:Error moving skb to kernel in nooks_w_netif_rx(%d).\n", rc);
		goto SecondChance;
	}

SecondChance : 
	result = NOOKS_KERNEL_CALL1(netif_rx, skb);

	if (result != NET_RX_DROP)
		rc = nooks_mm_move_skb_to_nook(nook, skb);

	if (rc < 0) {
		PDEBUG("Error moving skb to nook in nooks_w_netif_rx.\n");
		goto Cleanup;
	}

      Cleanup:
	NOOKS_WRAPPER_EXIT();
	return result;
}

void nooks_w___shadow_recover(void)	{

	shadow_ioctl("nook1", 10000, 0);
}



/* Implementing from the 2.4 version. Needs to be re-implemented to take 
 * care of the migration of task queues to work queues. Code cleanup
 * pending. 
 */

int NOOKS_KWRAPPER(register_netdev) (struct net_device * dev) {
	int result = 0, shadow_driver = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	char *kernel_name = NULL;
	struct net_device_stats *kernel_stats = NULL;
	static int driver_count = 0;
	short in_nook = 0;

	if (nook == NULL)
		printk("Fatal nook is null.\n");
	if (IN_NOOK()) {
		in_nook = 1;
		EXIT_NOOK();
	}

	NOOKS_KWRAPPER_ENTER();

	//printk("In register_netdev - searching for shadow with driver_count %d.\n", driver_count);
	
	if (strcmp(dev->name, "sn%d") == 0)
		driver = shadow_find_driver(nook, netops_e, driver_count);
	else
#if NOOKS_MIGRATE
	{
		if (NULL != nooks_net_device)	{
			//printk ("Shadow found in register_netdev.\n");
			dev = nooks_net_device;
			strcpy(dev->name, "eth1"); //BUGBUGBUG e1000 clobbers the name when its not its business.
			//NOOKS_SWRAPPER(nooks_net_device);
			// A new driver is to be created  in the new nook if we are migrating to a different device.

#ifdef NOOKS_NEW_DEVICE
			33
			result = nooks_get_driver_string(dev->name, &kernel_name);
			driver =
                         nooks_create_driver(nook, netops_e,
                                        NOOK_DRIVER_NO_MAJOR, kernel_name);
			driver->data = dev;
			if (driver->nook == NULL)
         	               	driver->nook = nook;
			result = nooks_add_resource_elem(nook, netdev_nook_e, dev, driver);
#endif
			return 0;
		}
	}
#else	
	{
		driver = shadow_find_driver(nook, netops_e, 1);
		if (driver != NULL)	{
			dev = driver->data;
			strcpy(dev->name, "eth1");
#if NOOKS_MIGRATE
			strcpy(dev->name, "eth1"); //BUGBUGBUG e1000 clobbers the name when its not its business.
#endif
		}
	}
#endif
	

	//nook_print_drivers(nook);

	if (driver != NULL)  {
		shadow_driver = 1;
		/* Reuse the old dev pointer. */
		dev = driver->data; //This is already seen by alloc_netdev
		printk ("Shadow found and dev is %p(%s) for driver %d.\n", dev, dev->name, driver_count);
		if (strcmp(dev->name, "sn") == 0)	
		{
			driver_count++;
			if (driver_count == 2)
				driver_count = 0;
		}
		
	}
	else	{
		result = nooks_get_driver_string(dev->name, &kernel_name);
		if (result < 0) {
        	        printk("nooks_get_driver_string failed.");
                	goto Cleanup;
        	}	

		/* Now create the driver. */
		printk ("No shadow found.\n");
		driver =
		    nooks_create_driver(nook, netops_e,
					NOOK_DRIVER_NO_MAJOR, kernel_name);

		if (driver == NULL) {

			printk
			    ("Nooks(register_netdev): Failed to create driver for register_netdev.\n");
			result = -ENOMEM;
			goto err3;
		}

		if (driver->nook == NULL)
			driver->nook = nook;

		strcpy(dev->name, kernel_name);
		printk(" NOOKS(register_netdev): Registering netdevice %s, %s.\n",
	       		kernel_name, dev->name);

	
		driver->data = dev;
		nooks_net_device = dev;	
		//} // end of else loop
		
		nooks_iwrapper_copy_netops(dev,
				   &driver->operations.netops,
				   driver->iwrapper_operations.netops, dev);
		
		printk("Copy netops complete for driver %s.\n", driver->name);

	} //end of else

        //driver->nook = GET_CURRENT_NOOK();

	result = nooks_add_resource_elem(nook, netdev_nook_e, dev, driver);	
	if (result < 0)
		goto err1;
	
	kernel_stats =
	    nooks_kmalloc(sizeof(struct net_device_stats), GFP_KERNEL);
	if (kernel_stats == NULL)
		goto err3;

	result = nooks_add_resource_elem(nook, netdev_stats_e, dev, kernel_stats);	
	
	
	if (result < 0)
		goto err2;

	//dev->priv = dev; e1000 uses this priv pointer - do not use it.!!!
#ifdef CONFIG_NOOKS_SHADOW
	if (!shadow_driver)	{
	   result = shadow_add_snetops(driver, dev);
	   printk ("NOOKS: Shadow netops finished with result %d.\n", result);
	}
#endif
	/*Execute the original code */
	dev = (struct net_device *) driver->data; 
	printk ("Calling register_netdev with dev %p.\n",dev);
	result = NOOKS_SWRAPPER(register_netdev)(dev);
	printk(" register_netdev returns %d for device %s.\n", result, dev->name);


	//printk("**NOOKS**:Register netdev completed successfully.\n");
	goto Cleanup;

	/* 6. Cleanup */
      err5:
	nooks_del_resource_elem(nook, netdev_stats_e,
				dev, (void **)&kernel_stats);
      err2:
	nooks_kfree(kernel_stats);
      err1:
	 nooks_del_resource_elem(nook,netdev_nook_e, dev, (void **)&driver);
      err3:
	nooks_free_driver_string(kernel_name);
      Cleanup:

	NOOKS_KWRAPPER_EXIT();
	if (in_nook)
		ENTER_NOOK();

	driver->nook = GET_CURRENT_NOOK();

	return result;
}

int NOOKS_WRAPPER(register_netdev) (struct net_device * dev) {
	int result;
	result = (int)NOOKS_KERNEL_CALL1(nooks_k_register_netdev, dev);
	return (result);
}

void NOOKS_KWRAPPER(vfree) (void *addr) {
	pnook_t nook = GET_CURRENT_NOOK();
	size_t size = 0;
	NOOKS_KWRAPPER_ENTER();
	vfree(addr);
	nooks_find_resource_elem(nook, vmalloc_e, addr,
                                     (void **)&size);
	NOOKS_KWRAPPER_EXIT();
}

void nooks_w_vfree(void *addr)
{
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL1(nooks_k_vfree, addr);
	NOOKS_WRAPPER_EXIT();
	return;
}

int NOOKS_WRAPPER(pskb_expand_head) (struct sk_buff * skb,
                                     int nhead, int ntail, gfp_t gfp_mask) {
        int result = 0;
	struct sk_buff *skb2;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_WRAPPER_ENTER();
	nooks_mm_move_skb_to_kernel(nook, skb);
	result = NOOKS_SWRAPPER(pskb_expand_head) (skb,
						nhead, ntail, gfp_mask);
	nooks_mm_move_skb_to_nook(nook, skb);	
	NOOKS_WRAPPER_EXIT();
	return result;
}

void NOOKS_WRAPPER(enable_irq) (unsigned int irq) {
	NOOKS_KERNEL_CALL1(enable_irq, irq);
}

void NOOKS_KWRAPPER(pci_disable_device) (struct pci_dev * dev) {
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_KWRAPPER_ENTER();

	NOOKS_SWRAPPER(pci_disable_device) (dev);
	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(pci_disable_device) (struct pci_dev * dev) {
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL1(nooks_k_pci_disable_device, dev);
	NOOKS_WRAPPER_EXIT();
	return;
}

/* The following function behaves like del_timer except on SMPs. 
 * besides deactivating the timer it also makes sure the handler 
 * has finished executing on other CPUs.
 */
int NOOKS_KWRAPPER(del_timer_sync) (struct timer_list * nook_timer) {
	int result;
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;

	NOOKS_KWRAPPER_ENTER();

	result =
	    nooks_del_resource_elem(nook, timer_e, nook_timer,
				    (void **)&driver);
	if (result <0)
		printk ("Bug in object tracking at del_timer_sync.\n");
	NOOKS_SWRAPPER(del_timer) (nook_timer);
	if (driver != NULL)
		nooks_free_driver(driver);

	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(del_timer_sync) (struct timer_list * timer) {
	int result;
	result = NOOKS_KERNEL_CALL1(nooks_k_del_timer_sync, timer);
	return (result);
}

/* Remove access by the nook to an SKB so it can be safely
 * used in the kernel and appropriately free it.
 */
void NOOKS_WRAPPER(kfree_skb) (struct sk_buff * skb) {
	int result = 0;
	NOOKS_WRAPPER_ENTER();  
	pnook_t nook = GET_CURRENT_NOOK();

       if (nook == NULL)       {
                //printk ("WARN(kfree_skb): Nook not found - tracking for nook1.\n");
                nook = nooks_find_nook_internal("nook1");
                SET_CURRENT_NOOK(nook);
        }

	result = nooks_mm_move_skb_to_kernel(nook, skb);
	/* Better to leak than to panic. */
	if (result < 0)
	{
		printk ("Fatal: Error deleting nooks information in kfree_skb.\n");
	}
	else
		kfree_skb(skb);

	NOOKS_WRAPPER_EXIT();
	return;
}

/*Implemented from the 2.4 version. */
void nooks_w_skb_over_panic(struct sk_buff *skb, int len, void *here)
{
	NOOKS_WRAPPER_ENTER();
	skb_over_panic(skb, len, here);
	NOOKS_WRAPPER_EXIT();
}

/* BigTBD  */
void NOOKS_KWRAPPER(dma_unmap_page) (struct device * dev,
				     dma_addr_t dma_address,
				     size_t size,
				     enum dma_data_direction direction) {
	NOOKS_SWRAPPER(dma_unmap_page) (dev, dma_address, size, direction);
	del_dma_info (dev, dma_address, size, direction);
	return;
}

void NOOKS_WRAPPER(dma_unmap_page) (struct device * dev,
				    dma_addr_t dma_address,
				    size_t size,
				    enum dma_data_direction direction) {
	NOOKS_KERNEL_CALL4(nooks_k_dma_unmap_page, dev,
			   dma_address, size, direction);
	return;
}

/* ioremap - Map a range of device address to the kernel address. */
void *NOOKS_KWRAPPER(__ioremap) (unsigned long offset,
				 unsigned long size, unsigned long flags) {
	pnook_t nook = GET_CURRENT_NOOK();
	void *result;
	NOOKS_KWRAPPER_ENTER();
	result = NOOKS_SWRAPPER(__ioremap) (offset, size, flags);

    /* FAULT INJECTION */  
    if (shadow_inject_fault == 1)
        result = nook;


	if (result != NULL) {
		if (nooks_add_resource_elem
		    (nook, iomap_e, result, (void *)size) < 0)
			printk ("Fatal:Error in object tracking at ioremap(WRAPPER).\n"); 
		
		}
	
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

void *NOOKS_WRAPPER(__ioremap) (unsigned long offset,
				unsigned long size, unsigned long flags) {
	void *result;
	NOOKS_WRAPPER_ENTER();
	result =
	    (void *)NOOKS_KERNEL_CALL3(nooks_k___ioremap, offset, size, flags);

	printk ("ioremap allocated address %p.\n", result);
	if (nooks_ioremap_ctr == 2)
		nooks_ioremap_ctr = 0;
	nooks_ioremap_addr[nooks_ioremap_ctr] = result;
	nooks_ioremap_ctr++;
	NOOKS_WRAPPER_EXIT();
	return (result);
}

void NOOKS_KWRAPPER(__netif_schedule) (struct net_device * dev) {
	NOOKS_SWRAPPER(__netif_schedule) (dev);
}

void NOOKS_WRAPPER(__netif_schedule) (struct net_device * dev) {
	NOOKS_KERNEL_CALL1(nooks_k___netif_schedule, dev);
	return;
}

/* Adds the driver structure to the list of registered drivers. 
 * How useful is the new owner field - 
 */

int NOOKS_KWRAPPER(__pci_register_driver) (struct pci_driver *
					   dev, struct module * owner) {
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	int result = 0;
	char *kernel_name = NULL;
	int shadow_driver = 0;
	NOOKS_KWRAPPER_ENTER();

	printk ("__In pci_register_driver.\n");

	if (nook == NULL)	{
		result = -ENOMEM;
		printk ("***NOOKS Fatal(pci_register_driver): Nook is NULL.\n");
		goto Cleanup;
	}
	
	//nook_print_drivers(nook);

	driver = shadow_find_driver(nook, pciops_e, 0); // Was NULL and working earlier :((
	if (driver != NULL) {
		shadow_driver = 1;
		dev = driver->data;
		nook_print_drivers(nook);
 		printk ("Driver found %s.\n", dev->name);
		printk ("Trying to free %p %d %d.\n", __resp, __iobase, __lenarg);
		__release_region(__resp, __iobase, __lenarg); //only hack for dex45
	} else {
		printk ("No shadow found.\n");
		/* Create the driver */
		result =
		    nooks_get_driver_string(dev->name, &kernel_name);
		if (result < 0)
			goto Cleanup;

		strcpy (dev->name, kernel_name);
		printk(" Registering pci driver %s, %s\n",
		       kernel_name, dev->name);

		driver =
		    nooks_create_driver(nook, pciops_e,
					NOOK_DRIVER_NO_MAJOR, kernel_name);
		if (driver->nook == NULL)
			driver->nook = nook;
		if (driver == NULL) {
			PERROR
			    (" Failed to create driver for register pci driver\n");
			result = -ENOMEM;
			goto err3;
		}
		driver->data = dev;
	

	nooks_iwrapper_copy_pciops(dev,
				   &driver->operations.pciops,
				   driver->iwrapper_operations.
				   pciops, dev);
	} //end of else
	
	result =
	    nooks_add_resource_elem(nook, pci_driver_e, dev, driver);

	if (result < 0)
		goto err4;
  	
		
		
	if (!shadow_driver && (result >= 0))
		result = shadow_add_spciops(driver, dev);
	
	/* Call the actual register function. */
	result = NOOKS_SWRAPPER(__pci_register_driver) (dev, owner);
	printk("pci_register_driver returns %d.\n", result);

	if (result != 0)
		goto err4;

	printk ("pci_register_driver initialized correctly.\n");
	nooks_pci_driver = dev;
	goto Cleanup;


      err4:
	if (driver != NULL)
	  nooks_free_driver(driver);
      err3:
	nooks_free_driver_string(kernel_name);
      Cleanup:
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(__pci_register_driver) (struct pci_driver *
					  arg0, struct module * arg1) {
	int result;
	result = (int)
	    NOOKS_KERNEL_CALL2(nooks_k___pci_register_driver, arg0, arg1);
	return (result);
}

/* Linux has moved from task queues to work queues. Each work queue now has
 * a separate calling thread. We treat the schedule_work as schedule_task was 
 * treated.
 */
int NOOKS_WRAPPER(schedule_work) (struct work_struct * work) {
	int result;
	result = schedule_work(work);
	return (result);
}

int NOOKS_WRAPPER(pci_set_consistent_dma_mask) (struct pci_dev * dev, u64 mask) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = pci_set_consistent_dma_mask(dev, mask);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

	/* param_get_int is a helper function for kernel parameter handling. */
int NOOKS_WRAPPER(param_get_int) (char *buffer, struct kernel_param * kp) {
	int result;
	result = (int)NOOKS_KERNEL_CALL2(param_get_int, buffer, kp);
	return (result);
}

void *NOOKS_KWRAPPER(vmalloc) (unsigned long size) {
	pnook_t nook = GET_CURRENT_NOOK();
	void *result;
	int rc = 0;
	gfp_t gfp_mask = GFP_KERNEL | __GFP_HIGHMEM;
	NOOKS_KWRAPPER_ENTER();
	if (nook == NULL)
		BUG();
	
	result = vmalloc(size);
 	rc = nooks_add_resource_elem(nook, vmalloc_e, result, size);			
	if (rc < 0)
		printk ("Error tracking vmalloced memory.\n");

	NOOKS_KWRAPPER_EXIT();
	return (result);
}

void *NOOKS_WRAPPER(vmalloc) (unsigned long size) {
	void *result;
	NOOKS_WRAPPER_ENTER();
	result = (void *)NOOKS_KERNEL_CALL1(nooks_k_vmalloc, size);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(pci_bus_write_config_word) (struct pci_bus *
					      bus,
					      unsigned int
					      devfn, int where, u16 val) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = (int)pci_bus_write_config_word(bus, devfn, where, val);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(param_set_int) (const char *val, struct kernel_param * kp) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = (int)NOOKS_KERNEL_CALL2(param_set_int, val, kp);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

	/* RELOOK */
void NOOKS_WRAPPER(raise_softirq_irqoff) (unsigned int nr) {
	NOOKS_KERNEL_CALL1(raise_softirq_irqoff, nr);
	return;
}

	/* Should be similar to  pci_free_consistent. */
void NOOKS_KWRAPPER(dma_free_coherent) (struct device * dev,
					size_t size,
					void *vaddr, dma_addr_t dma_handle) {
	dma_free_coherent(dev, size, vaddr, size);
	return;
}

void NOOKS_WRAPPER(dma_free_coherent) (struct device * dev,
				       size_t size, void *vaddr,
				       dma_addr_t dma_handle) {
	NOOKS_KERNEL_CALL4(nooks_k_dma_free_coherent, dev, size,
			   vaddr, dma_handle);
	return;
}

	/* Implementing from the 2.4 version. */
int NOOKS_KWRAPPER(pci_enable_device) (struct pci_dev * dev) {
	int result = 0;
	NOOKS_KWRAPPER_ENTER();

	result = NOOKS_SWRAPPER(pci_enable_device) (dev);
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(pci_enable_device) (struct pci_dev * dev) {
	int result = 0;
	NOOKS_WRAPPER_ENTER();
	nooks_pci_device = dev;
	result = (int)NOOKS_KERNEL_CALL1(nooks_k_pci_enable_device, dev);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* In Linux 2.4, when a packet is received, netif_rx is called which
 * enqueues the received packet in the interrupted CPU's backlog queue 
 * schedules a softirq. On the other hand, 2.6 calls the function 
 * below that puts a reference to the device in a queue attached to 
 * the interrupted CPU and a softirq is scheduled again.
 * The difference in scheduling the softirq is obvious - instead of 
 * pulling packets in backlog, the CPU polls the device to get the
 * packets from the rx_ring or from the backlog. Hence, we don't 
 * need to perform deferred packet processing at this stage, neither
 * we can do it given the interface of the below function.
 */
void NOOKS_KWRAPPER(__netif_rx_schedule) (struct net_device * dev) {

	NOOKS_SWRAPPER(__netif_rx_schedule) (dev);
}

void NOOKS_WRAPPER(__netif_rx_schedule) (struct net_device * dev) {
	NOOKS_KERNEL_CALL1(nooks_k___netif_rx_schedule, dev);
	return;
}

void NOOKS_WRAPPER(_spin_lock) (spinlock_t * lock) {
	NOOKS_KERNEL_CALL1(_spin_lock, lock);
	return;
}

unsigned long NOOKS_WRAPPER(_spin_lock_irqsave) (spinlock_t * lock) {
	unsigned long result;
	result = NOOKS_KERNEL_CALL1(_spin_lock_irqsave, lock);
	return result;
}

void NOOKS_WRAPPER(synchronize_irq) (unsigned int irq) {
	NOOKS_KERNEL_CALL1(synchronize_irq, irq);
	return;
}

int NOOKS_WRAPPER(pci_bus_read_config_dword) (struct pci_bus *
					      bus,
					      unsigned int
					      devfn, int where, u32 * val) {
	int result;
	result =
	    (int)NOOKS_KERNEL_CALL4(pci_bus_read_config_dword,
				    bus, devfn, where, val);
	return (result);
}

/* The minimalist add_timer */
void NOOKS_KWRAPPER(add_timer) (struct timer_list * nook_timer,
				unsigned long expires) {
	pnook_t nook = GET_CURRENT_NOOK();
	int result = 0;
	pnook_driver_t driver = NULL;
	NOOKS_KWRAPPER_ENTER();

	driver =
	    nooks_create_driver(nook, timerops_e, NOOK_DRIVER_NO_MAJOR, NULL);
	if (driver == NULL) {
		result = -ENOMEM;
		goto Cleanup;
	}

	driver->data = nook_timer;
	result = nooks_add_resource_elem(nook, timer_e, nook_timer, driver);
	NOOKS_SWRAPPER(add_timer) (nook_timer);
      Cleanup:
	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(add_timer) (struct timer_list * nook_timer) {
	NOOKS_KERNEL_CALL2(nooks_k_add_timer, nook_timer, 0);
}

/* mod_timer - Modifies a timer's timeout. */
int NOOKS_KWRAPPER(mod_timer) (struct timer_list * nook_timer,
			       unsigned long expires) {
	int result = 0;
	pnook_t nook;
	pnook_driver_t driver = NULL;
	char *kernel_name = NULL;
	char * name = "timer";
	NOOKS_KWRAPPER_ENTER ();

	nook = GET_CURRENT_NOOK();
	if (nook == NULL)
		BUG();
	result = nooks_del_resource_elem(nook, timer_e, nook_timer, &driver);

		//nooks_find_resource_elem(nook, timer_e, nook_timer,
		//				 	&driver);
		
	if (result < 0)	{
		nooks_get_driver_string("timer", &kernel_name);
		strcpy(name, kernel_name);
		driver = nooks_create_driver(nook, timerops_e, NOOK_DRIVER_NO_MAJOR, kernel_name);
/*
		if (driver != NULL)	{
			driver->need_delete = 0;
			driver->data = nook_timer;
			nooks_iwrapper_copy_timerops(nook_timer->function,
                                     &driver->operations.timerops,
                                     driver->iwrapper_operations.timerops,
                                     &driver->wrapper_operations.timerops);
		}
*/
		nooks_add_resource_elem(nook, timer_e, nook_timer, driver);
                printk ("Added timerops while calling mod_timer object tracking at mod_timer.\n");
	}
	
	result = NOOKS_SWRAPPER(mod_timer) (nook_timer, expires);
	
	NOOKS_KWRAPPER_EXIT ();
	return (result);
}

int NOOKS_WRAPPER(mod_timer) (struct timer_list * timer, unsigned long expires) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = (int)NOOKS_KERNEL_CALL2(nooks_k_mod_timer, timer, expires);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* Implemeneted similar to Nooks 2.4 version taking into care appropriate
 * interface changes.
 */

struct sk_buff *NOOKS_WRAPPER(__alloc_skb) (unsigned int size,
                                            gfp_t priority, int fclone) {
        struct sk_buff *skb;
	int result = -1;

	pnook_t nook = GET_CURRENT_NOOK();
	
	skb = __alloc_skb(size, priority, fclone);
	skb->nook_info = nook;
	skb_shinfo(skb)->nook_info.nook = nook;

	if (nook != NULL)	
		result = nooks_mm_move_skb_to_nook(nook, skb);
	
	if (result < 0)
		PDEBUG("Fatal: Error adding skb as a nook resource in __alloc_skb.");
	
	return (skb);
}

void NOOKS_WRAPPER(init_timer) (struct timer_list * timer) {
#if 0	
        pnook_t nook = GET_CURRENT_NOOK();
        int result = 0;
        struct timer_list * timer = NULL;
        pnook_driver_t driver = NULL;
        NOOKS_ENTER();

        /* create a new timerops driver */

        driver = nooks_create_driver(nook, timerops_e, NOOK_DRIVER_NO_MAJOR, NULL);
        if (driver == NULL) {
                result = -ENOMEM;
                goto err2;
        }

                // Copy the operations, preserving NULL entries
        nooks_iwrapper_copy_timerops(timer->function,
                                     &driver->operations.timerops,
                                     driver->iwrapper_operations.timerops,
                                     &driver->wrapper_operations.timerops);

        driver->data = timer;

#endif
	init_timer (timer);
	return;
}

void nooks_w_kfree(const void *ptr)
{
	int size = 0;
	int result = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_WRAPPER_ENTER();
	kfree(ptr);
	result =  nooks_del_from_hash(&nook->resources.area_map,
                                area_map_e,
                                ptr,
                                (void **) &size );
	if (result < 0)
		printk ("Object tracker in nooks_w_kfree is having problems.\n");
/*
	if (ptr != NULL) {
#ifdef NOOKS_KEEP_STATS
		u8 *ptr_s = (u8 *) ptr - NOOKS_MM_KMALLOC_OVERHEAD;
		NOOKS_STATS_ADD(NOOKS_OPS_W_KMALLOC_BYTES,
				-(*(size_t *) ptr_s));
		nooks_mm_kfree(GET_CURRENT_NOOK(), ptr_s);
#else				// NOOKS_KEEP_STATS
		nooks_mm_kfree(GET_CURRENT_NOOK(), ptr);
#endif				// NOOKS_KEEP_STATS
	}
*/
	NOOKS_WRAPPER_EXIT();
}

void NOOKS_WRAPPER(pci_clear_mwi) (struct pci_dev * dev) {
	NOOKS_KERNEL_CALL1(pci_clear_mwi, dev);
	return;
}

/* We let go the buffers as is for now. */
int NOOKS_KWRAPPER(netif_receive_skb) (struct sk_buff * skb) {
	int result;
	result = NOOKS_SWRAPPER(netif_receive_skb) (skb);
	return (result);
}

int NOOKS_WRAPPER(netif_receive_skb) (struct sk_buff * skb) {
	int result;
	pnook_t nook = GET_CURRENT_NOOK();
        if (nook == NULL)       {
                PDEBUG ("WARN(netif_receive_skb): Nook not found - tracking for nook1.\n");
                nook = nooks_find_nook_internal("nook1");
                SET_CURRENT_NOOK(nook);
        }

	result = (int)NOOKS_KERNEL_CALL1(nooks_k_netif_receive_skb, skb);
	if (nook !=NULL)
		result = nooks_mm_move_skb_to_nook(nook, skb);
	else
		printk("Fatal: Nooks is null(netif_receive_skb)");
	return (result);
}

int NOOKS_WRAPPER(pci_bus_read_config_byte) (struct pci_bus *
					     bus,
					     unsigned int devfn,
					     int where, u8 * val) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = pci_bus_read_config_byte(bus, devfn, where, val);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

void *nooks_w_memcpy(void *dest, const void *src, size_t n)
{
	void *result;
	result = memcpy(dest, src, n);
	return result;
}

extern void nooks_w_put_page(struct page *page)
{
	NOOKS_KERNEL_CALL1(put_page, page);
}

/* Allocate nooks local pages instead of kernel pages. */
struct page *NOOKS_KWRAPPER(__alloc_pages) (gfp_t gfp_mask,
					    unsigned int order,
					    struct zonelist * zonelist) {
	struct page *result;
	unsigned long temp;
	pnook_t nook = GET_CURRENT_NOOK();

	NOOKS_KWRAPPER_ENTER();

       temp = nooks_mm_get_free_pages(nook, gfp_mask, order);
        if (temp == 0)
                result = NULL;
        else
                result = (struct page *)virt_to_page(temp);


	NOOKS_KWRAPPER_EXIT();
	return (result);
}

struct page *NOOKS_WRAPPER(__alloc_pages) (gfp_t gfp_mask,
					   unsigned int order,
					   struct zonelist * zonelist) {
	struct page *result;
	result =
	    NOOKS_KERNEL_CALL3(nooks_k___alloc_pages, gfp_mask,
			       order, zonelist);
	return result;
}

dma_addr_t NOOKS_KWRAPPER(dma_map_single) (struct device * dev,
					   void *ptr,
					   size_t size,
					   enum dma_data_direction direction) {
	dma_addr_t result;
	result = NOOKS_SWRAPPER(dma_map_single) (dev, ptr, size, direction);
	nooks_store_dma_info(GET_CURRENT_NOOK(), dev, ptr, size, direction);
	return (result);
}

dma_addr_t NOOKS_WRAPPER(dma_map_single) (struct device * dev,
					  void *ptr,
					  size_t size,
					  enum dma_data_direction direction) {
	dma_addr_t result;
	result =
	    NOOKS_KERNEL_CALL4(nooks_k_dma_map_single, dev, ptr,
			       size, direction);
	return (result);
}

/* Dummy System Calls */

int NOOKS_KWRAPPER(dummy_call0) (void) {
	return (0);
}

int NOOKS_KWRAPPER(dummy_call1) (int arg1) {
	return (0);
}

int NOOKS_KWRAPPER(dummy_call2) (int arg1, int arg2) {
	return (0);
}

int NOOKS_KWRAPPER(dummy_call3) (int arg1, int arg2, int arg3) {
	return (0);
}

int NOOKS_KWRAPPER(dummy_call4) (int arg1, int arg2, int arg3, int arg4) {
	return (0);
}

int nooks_w_dummy_call0(void)
{
	return (NOOKS_KERNEL_CALL0(nooks_k_dummy_call0));
}

int nooks_w_dummy_call1(int arg1)
{
	return (NOOKS_KERNEL_CALL1(nooks_k_dummy_call1, arg1));
}

int nooks_w_dummy_call2(int arg1, int arg2)
{
	return (NOOKS_KERNEL_CALL2(nooks_k_dummy_call2, arg1, arg2));

}

int nooks_w_dummy_call3(int arg1, int arg2, int arg3)
{
	return (NOOKS_KERNEL_CALL3(nooks_k_dummy_call3, arg1, arg2, arg3));
}

int nooks_w_dummy_call4(int arg1, int arg2, int arg3, int arg4)
{
	return (NOOKS_KERNEL_CALL4
		(nooks_k_dummy_call4, arg1, arg2, arg3, arg4));

}

/* Initialize pool information -- used in nooks-data.c */

int nooks_w_init_pools(void)
{
	int result = 0;

#if 0
	pnook_t nook = GET_CURRENT_NOOK();
	pnooks_lookaside_list_t list = &nook->locals->elem_pool;
	void *elem;

	//
	// Do the list elem pool
	//

	elem = nooks_mm_lookaside_alloc(list);
	if (elem != NULL) {
		nooks_mm_lookaside_free(list, elem);
	} else {
		result = -ENOMEM;
	}

	//
	// Now do the work item pool
	//

	list = &nook->locals->work_pool;
	elem = nooks_mm_lookaside_alloc(list);
	if (elem != NULL) {
		nooks_mm_lookaside_free(list, elem);
	} else {
		result = -ENOMEM;
	}
#endif
	return (result);
}

void nooks_free_ctl_table(pnook_t nook, ctl_table * table)
{
	int i = 0;
	pnook_driver_t driver = NULL;
	if (table->ctl_name != 0) {
		if (table->procname != NULL) {
			nooks_free_driver_string((char *)table->procname);
		}
		if (table->child != NULL) {
			while (table->child[i].ctl_name != 0) {
				nooks_free_ctl_table(nook, &table->child[i]);
				i++;
			}
			nooks_kfree(table->child);
		}
		nooks_del_resource_elem(nook, ctl_table_e, table,
					(void **)&driver);
		if (driver != NULL) {
			nooks_free_driver(driver);
		}
		table->ctl_name = 0;
	}
}

#define NOOKS_MAX_CTL_TABLE 20

/**********************************************************
 *
 * Function: free_pages
 *
 * Type: cache kernel
 *
 * Notes: Needs to update per-nook state in kernel, so call
 *        nooks version.
 *
 *********************************************************/

void nooks_mm_w_free_pages(unsigned long addr, unsigned int order)
{
	NOOKS_KERNEL_CALL2(nooks_free_pages, addr, order);
}

void nooks_w_free_pages(unsigned long addr, unsigned int order)
{
	NOOKS_WRAPPER_ENTER();

	if (addr != 0) {
		if (order == 0) {
			pnook_t nook = GET_CURRENT_NOOK();
			pnooks_lookaside_list_t list = &nook->locals->page_pool;
			nooks_mm_lookaside_free(list, (void *)addr);
		} else {
			nooks_mm_w_free_pages(addr, order);
		}
	}

	NOOKS_WRAPPER_EXIT();

}

int NOOKS_KWRAPPER(printk) (char *buf) {
	char *kern_buf;
	int result = 0;
	//NOOKS_KWRAPPER_ENTER(); FIXME TBD May need to remove later

	if (nooks_get_driver_string(buf, &kern_buf) >= 0) {
		result = printk(kern_buf);
		nooks_kfree(kern_buf);
	}
	//NOOKS_KWRAPPER_EXIT(); FIXME TBD May need to remove later
	return (result);

}

int nooks_w_printk(const char *fmt, ...)
{
	int result;
	va_list args;
	char nooks_w_printk_buf[200];

	//  NOOKS_WRAPPER_ENTER();

	//
	// Build the string to print locally
	//

	va_start(args, fmt);
	vsnprintf(nooks_w_printk_buf, sizeof(nooks_w_printk_buf), fmt, args);
	va_end(args);

	//
	// Call into the kernel to perform the call.
	//

	result = NOOKS_KERNEL_CALL1(nooks_k_printk, nooks_w_printk_buf);

	//  NOOKS_WRAPPER_EXIT();
	return (result);

}

int nooks_w_sprintf(char *buf, const char *fmt, ...)
{
	va_list args;
	int i;

	NOOKS_WRAPPER_ENTER();
	va_start(args, fmt);
	i = vsprintf(buf, fmt, args);
	va_end(args);
	NOOKS_WRAPPER_EXIT();
	return i;
}

void nooks_w_ether_setup(struct net_device *dev)
{
	NOOKS_WRAPPER_ENTER();
	pnook_t nook = GET_CURRENT_NOOK();
	if (nook == NULL)
		printk("Nook not initialized correctly.\n");
	else
		printk("Nook values: %d.\n",
		       get_current()->nook_state.curr.flags);
	NOOKS_KERNEL_CALL1(ether_setup, dev);
	NOOKS_WRAPPER_EXIT();
}

void *nooks_w_kmem_cache_alloc(struct kmem_cache *cachep, gfp_t flags)
{
	void *result;
	//NOOKS_WRAPPER_ENTER();
	result = kmem_cache_alloc(cachep, flags);
	//NOOKS_WRAPPER_EXIT();
	return (result);
}

void print_nooks_hash_tables(void)
{
	pnook_t nook = GET_CURRENT_NOOK();
	printk ("********Hash tables are ********\n");
	printk("skb table is %p.\n", &nook->resources.active_skbs);
	printk("Area map hash table is at %p.\n", &nook->resources.area_map);
	printk("***************************************");
}

int nooks_store_dma_info(pnook_t nook, struct device *dev,
                            dma_addr_t dma_handle,
                            size_t size,
                            enum dma_data_direction direction)	{
	if (0 == is_dma_init)	{
	 	INIT_LIST_HEAD(&nooks_dma_info.list);
		is_dma_init ++;
	}

	struct nooks_dma_info_list *tmp;
	tmp = (struct nooks_dma_info_list *) kzalloc(sizeof(struct nooks_dma_info_list), GFP_KERNEL);	
	tmp->nook = nook;
	tmp->dev = dev;
	tmp->dma_handle = dma_handle;
	tmp->size = size;
	tmp->direction = direction;
	//tmp->type = type; // 0 = single , 1 = page
	list_add((&tmp->list), (&nooks_dma_info.list));

	return 0;
}


int list_dma_info(void)	{
	struct list_head * pos;
	struct nooks_dma_info_list *tmp;
	
	list_for_each(pos, &nooks_dma_info.list){
		tmp = list_entry(pos, struct nooks_dma_info_list, list);
		printk ("DMA Entry: dev: %p, handle %p, size %d, direction %d.\n", tmp->dev,
								tmp->dma_handle, tmp->size, tmp->direction);

	}
	return 0;
}

int del_dma_info (struct device * dev,
                   dma_addr_t dma_addr, size_t size,
                   enum dma_data_direction direction)	{
        struct list_head * pos, *q;
        struct nooks_dma_info_list *tmp;
	
	if (0 == is_dma_init)   {
        	return -ENOMEM;
	}

	
        list_for_each_safe(pos, q, &nooks_dma_info.list)	{
                tmp = list_entry(pos, struct nooks_dma_info_list, list);
		if ((tmp->dev == dev) && (tmp->dma_handle == dma_addr) && 
		 	(tmp->direction == direction))	{
				PDEBUG("Deleting stuff.\n");
				list_del(pos);
				kfree(tmp);
        	}	
	}
	return 0;
}

#ifdef NOOKS_MIGRATE
void nooks_w_shadow_arp()	{
	shadow_ioctl("nook1", 13000, 0);
}
#endif

void nooks_w_trigger_recovery()	{
	shadow_ioctl("nook1", 10000, 0);
}
#ifdef NOOKS_MIGRATE
int nooks_transfer_dma_to_cpu(void)	{

	struct list_head * pos;
        struct nooks_dma_info_list *tmp;
	pnook_t nook = GET_CURRENT_NOOK();
	/*
	if (0 == is_dma_init)   {
		printk("No PCI steps to perform pre migration.\n");
                return 0;
        }
	*/
        list_for_each(pos, &nooks_dma_info.list){
                tmp= list_entry(pos, struct nooks_dma_info_list, list);
		//printk ("Transferring DMA to CPU.\n");
                dma_unmap_single(tmp->dev,
                                        tmp->dma_handle, tmp->size, tmp->direction);

        }
	//printk ("ioremap records are %p and %p.\n",nooks_ioremap_addr[0],nooks_ioremap_addr[1]);

	nooks_net_device->stop(nooks_net_device);
	dma_release_declared_memory(nooks_pci_device);
	pci_release_regions(nooks_pci_device);
	iounmap(nooks_ioremap_addr[0]);
	//iounmap(nooks_ioremap_addr[1]);
	//unregister_netdev(nooks_net_device); // Should only clear nooks related informatio
	return 0;
}

int nooks_fix_pci(void)		{

	//printk ("Setting shadow migration flag.\n");
	shadow_migrated = 1;	

}

#endif
