#ifndef _LINUX_NOOKS_WRAPPERS_GEN_H
#define _LINUX_NOOKS_WRAPPERS_GEN_H

#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/blkdev.h>
#include <linux/blkpg.h>
#include <linux/interrupt.h>
#include <asm/io.h>
#include <asm/semaphore.h>
#include <linux/ioport.h>
#include <linux/netdevice.h>
#include <linux/genhd.h>
#include <linux/etherdevice.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/module.h>
#include <asm/delay.h>
#include <linux/proc_fs.h>
#include <linux/pci.h>
#include <linux/isapnp.h>
#include <net/pkt_sched.h>
#include <net/sock.h>
#include <net/tcp.h>
#include <../drivers/scsi/scsi.h>
#include <scsi/scsi_host.h>
#include <linux/nls.h>
#include <linux/mii.h>
/* doug added */
#include <asm/mtrr.h>
#include <linux/mm.h>
#include <asm/processor.h>
#include <linux/miscdevice.h>
#include <linux/pm.h>
#include <linux/pci.h>
#include <asm/page.h>
#include <linux/rwsem.h>
/* /doug */
#include "nooks-switch.h"
#include "nooks-local.h"
#include "nooks-swrappers.h"
#include <asm/dma.h>
#include <linux/irq.h>
#include <linux/file.h>
#include <linux/kmod.h>

// index of where filp arguments are stashed
#define NOOKS_FILP_ARG 0
#define NOOKS_NOOK_FILP_ARG 1
#define NOOKS_POLL_ARG 1
#define NOOKS_DENTRY_ARG 2
#define NOOKS_NOOK_DENTRY_ARG 3
#define NOOKS_FILLDIR_ARG 1
#define NOOKS_BUF_ARG 2
#define NOOKS_NOOK_PAGE_ARG 2
#define NOOKS_PAGE_ARG 3
#define NOOKS_INODE_ARG 0
#define NOOKS_NOOK_INODE_ARG 1



/*********************************************************************
 *
 * Categories of wrapper functions
 *
 * 1. PURELOCAL: Functions that can be handled purely locally, and don't 
 *    access shared data. (e.g. Helper functions like sprintf())
 * 2. CACHELOCAL: Functions that can be handled locally with caching, but
 *    may need to synchronously call the kernel (e.g. kmalloc())
 * 3. DEFFERRED: Functions that can be deferred safely and executed later
 *    used (e.g. functions that free resources like kfree())
 * 4. KERNEL: Functions that must be executed in the kernel, because 
 *    they involve kernel state or scheduling (e.g. registering devices, 
 *    waiting)
 * 5. LOW-IMP: Low-importance functions that are executed in the kernel 
 *    because that is easiest. These may change categories later
 *********************************************************************/

#define NOOKS_WRAPPER(_x_) nooks_w_ ## _x_
#define NOOKS_KWRAPPER(_x_) nooks_k_ ## _x_
#define NOOKS_WRAPPER_ENTER()  \
do { \
ENTER_NOOK(); \
} while (0)

#define NOOKS_WRAPPER_EXIT() \
do { \
EXIT_NOOK();	\
} while (0)

void NOOKS_WRAPPER(add_timer)(struct timer_list * timer);

#define NOOKS_IRQ_TO_MAJOR(_irq_) ((_irq_) << 16)
#define NOOKS_MAJOR_TO_IRQ(_maj_) ((_maj_) >> 16)

extern struct softnet_data *nooks_w_softnet_data;

struct nooks_dma_info_list  {
  pnook_t nook;
  struct list_head list;
  struct device *dev;
  dma_addr_t dma_handle;
  size_t size;
  enum dma_data_direction direction;
};


int nooks_store_dma_info(pnook_t nook, struct device *dev,
                            dma_addr_t dma_handle,
                            size_t size,
                            enum dma_data_direction direction);

int list_dma_info(void);

int nooks_transfer_dma_to_cpu(void);
int nooks_fix_pci(void);
//int shadow_migrated = 0;

int del_dma_info (struct device * dev,
                   dma_addr_t dma_addr, size_t size,
                   enum dma_data_direction direction);

int nooks_transfer_dma_to_cpu(void);

extern void NOOKS_WRAPPER(netif_device_attach)(struct net_device *dev);

extern struct sk_buff * NOOKS_WRAPPER(__netdev_alloc_skb)(struct net_device *dev, unsigned int length, gfp_t gfp_mask);

extern void NOOKS_WRAPPER(free_netdev)(struct net_device *dev);

extern void NOOKS_WRAPPER(netif_device_detach)(struct net_device *dev);

extern dma_addr_t NOOKS_WRAPPER(dma_map_page)(struct device *dev, struct page *page, unsigned long offset, size_t size, enum dma_data_direction direction);

extern void * NOOKS_WRAPPER(dma_alloc_coherent)(struct device *dev, size_t size, dma_addr_t *dma_handle, gfp_t flag);

//extern void FASTCALL(NOOKS_WRAPPER(_spin_unlock_irqrestore)(spinlock_t *lock, unsigned long flags));

extern void (NOOKS_WRAPPER(_spin_unlock_irqrestore)(spinlock_t *lock, unsigned long flags));

extern void NOOKS_WRAPPER(pci_release_regions)(struct pci_dev * arg0);

extern void NOOKS_WRAPPER(dump_stack)(void);

extern void NOOKS_WRAPPER(__const_udelay)(unsigned long usecs);

extern void NOOKS_WRAPPER(pci_unregister_driver)(struct pci_driver * arg0);

extern void NOOKS_WRAPPER(netif_carrier_off)(struct net_device *dev);

extern void NOOKS_WRAPPER(kunmap_atomic)(void *kvaddr, enum km_type type);

extern struct net_device * NOOKS_WRAPPER(alloc_etherdev)(int sizeof_priv);

extern pci_power_t NOOKS_WRAPPER(pci_choose_state)(struct pci_dev *dev, pm_message_t state);

extern int NOOKS_WRAPPER(pci_set_power_state)(struct pci_dev *dev, pci_power_t state);

extern void NOOKS_WRAPPER(iounmap)(volatile void *addr);

extern void NOOKS_WRAPPER(netif_carrier_on)(struct net_device *dev);

extern signed long NOOKS_WRAPPER(schedule_timeout_interruptible)(signed long timeout);

extern int NOOKS_WRAPPER(dev_close)(struct net_device *dev);

extern int NOOKS_WRAPPER(pci_request_regions)(struct pci_dev * arg0, const char * arg1);

extern int NOOKS_WRAPPER(pci_bus_write_config_dword)(struct pci_bus *bus, unsigned int devfn, int where, u32 val);

extern __be16 NOOKS_WRAPPER(eth_type_trans)(struct sk_buff *skb, struct net_device *dev);

extern int NOOKS_WRAPPER(dev_open)(struct net_device *dev);

extern int NOOKS_WRAPPER(pci_set_dma_mask)(struct pci_dev *dev, u64 mask);

extern int NOOKS_WRAPPER(pci_bus_read_config_word)(struct pci_bus *bus, unsigned int devfn, int where, u16 *val);

extern void * NOOKS_WRAPPER(kmap_atomic)(struct page *page, enum km_type type);

extern void NOOKS_WRAPPER(__udelay)(unsigned long usecs);

extern void NOOKS_WRAPPER(free_irq)(unsigned int arg0, void * arg1);

extern void * NOOKS_WRAPPER(__kmalloc)(size_t arg0, gfp_t arg1);

extern void NOOKS_WRAPPER(dma_sync_single_for_device)(struct device *dev, dma_addr_t dma_handle, size_t size, enum dma_data_direction direction);

extern void NOOKS_WRAPPER(pci_set_master)(struct pci_dev *dev);

extern int NOOKS_WRAPPER(netpoll_trap)(void);

//extern int FASTCALL(NOOKS_WRAPPER(_spin_trylock)(spinlock_t *lock));
extern int (NOOKS_WRAPPER(_spin_trylock)(spinlock_t *lock));

extern void NOOKS_WRAPPER(unregister_netdev)(struct net_device *dev);

extern void NOOKS_WRAPPER(disable_irq)(unsigned int irq);

extern int NOOKS_WRAPPER(request_irq)(unsigned int arg0, irqreturn_t (*handler)(int, void *, struct pt_regs *), unsigned long arg1, const char * arg2, void * arg3);

extern int NOOKS_WRAPPER(pci_enable_wake)(struct pci_dev *dev, pci_power_t state, int enable);

extern void NOOKS_WRAPPER(dev_kfree_skb_any)(struct sk_buff *skb);

#ifdef NOOKS_MIGRATE
extern void NOOKS_WRAPPER(shadow_arp)(void);
#endif
extern void NOOKS_WRAPPER(trigger_recovery)(void);

extern int NOOKS_WRAPPER(pci_set_mwi)(struct pci_dev *dev);

extern int NOOKS_WRAPPER(___pskb_trim)(struct sk_buff *skb, unsigned int len);

extern void NOOKS_WRAPPER(dma_unmap_single)(struct device *dev, dma_addr_t dma_addr, size_t size, enum dma_data_direction direction);

extern void NOOKS_WRAPPER(force_evtchn_callback)(void);

extern void NOOKS_WRAPPER(dma_sync_single_for_cpu)(struct device *dev, dma_addr_t dma_handle, size_t size, enum dma_data_direction direction);

extern int NOOKS_WRAPPER(param_array_set)(const char *val, struct kernel_param *kp);

extern void NOOKS_WRAPPER(flush_scheduled_work)(void);

extern void NOOKS_WRAPPER(put_page)(struct page *page);

extern int NOOKS_WRAPPER(param_array_get)(char *buffer, struct kernel_param *kp);

extern int NOOKS_WRAPPER(capable)(int cap);

extern unsigned char * NOOKS_WRAPPER(__pskb_pull_tail)(struct sk_buff *skb, int delta);

extern int NOOKS_WRAPPER(register_netdev)(struct net_device *dev);

extern void NOOKS_WRAPPER(vfree)(void *addr);

extern int NOOKS_WRAPPER(pskb_expand_head)(struct sk_buff *skb, int nhead, int ntail, gfp_t gfp_mask);

extern void NOOKS_WRAPPER(enable_irq)(unsigned int irq);

extern void NOOKS_WRAPPER(pci_disable_device)(struct pci_dev *dev);

extern int NOOKS_WRAPPER(del_timer_sync)(struct timer_list *timer);

extern void NOOKS_WRAPPER(kfree_skb)(struct sk_buff *skb);

extern void NOOKS_WRAPPER(skb_over_panic)(struct sk_buff *skb, int len, void *here);

extern void NOOKS_WRAPPER(dma_unmap_page)(struct device *dev, dma_addr_t dma_address, size_t size, enum dma_data_direction direction);

extern void * NOOKS_WRAPPER(__ioremap)(unsigned long offset, unsigned long size, unsigned long flags);

extern void NOOKS_WRAPPER(__netif_schedule)(struct net_device *dev);

extern int NOOKS_WRAPPER(__pci_register_driver)(struct pci_driver * arg0, struct module * arg1);

//extern int FASTCALL(NOOKS_WRAPPER(schedule_work)(struct work_struct *work));
extern int (NOOKS_WRAPPER(schedule_work)(struct work_struct *work));

extern int NOOKS_WRAPPER(pci_set_consistent_dma_mask)(struct pci_dev *dev, u64 mask);

extern int NOOKS_WRAPPER(param_get_int)(char *buffer, struct kernel_param *kp);

//extern unsigned long FASTCALL(NOOKS_WRAPPER(_spin_lock_irqsave)(spinlock_t *lock));
extern unsigned long NOOKS_WRAPPER(_spin_lock_irqsave)(spinlock_t *lock);

extern void * NOOKS_WRAPPER(vmalloc)(unsigned long size);

extern int NOOKS_WRAPPER(pci_bus_write_config_word)(struct pci_bus *bus, unsigned int devfn, int where, u16 val);

extern int NOOKS_WRAPPER(param_set_int)(const char *val, struct kernel_param *kp);

//extern void FASTCALL(NOOKS_WRAPPER(raise_softirq_irqoff)(unsigned int nr));
extern void (NOOKS_WRAPPER(raise_softirq_irqoff)(unsigned int nr));

extern void NOOKS_WRAPPER(dma_free_coherent)(struct device *dev, size_t size, void *vaddr, dma_addr_t dma_handle);

extern int NOOKS_WRAPPER(pci_enable_device)(struct pci_dev *dev);

extern void NOOKS_WRAPPER(__netif_rx_schedule)(struct net_device *dev);

//extern void FASTCALL(NOOKS_WRAPPER(_spin_lock)(spinlock_t *lock));
extern void (NOOKS_WRAPPER(_spin_lock)(spinlock_t *lock));

extern void NOOKS_WRAPPER(synchronize_irq)(unsigned int irq);

extern int NOOKS_WRAPPER(pci_bus_read_config_dword)(struct pci_bus *bus, unsigned int devfn, int where, u32 *val);

extern int NOOKS_WRAPPER(mod_timer)(struct timer_list *timer, unsigned long expires);

extern struct sk_buff * NOOKS_WRAPPER(__alloc_skb)(unsigned int size, gfp_t priority, int fclone);

//extern void FASTCALL(NOOKS_WRAPPER(init_timer)(struct timer_list * timer));
extern void (NOOKS_WRAPPER(init_timer)(struct timer_list * timer));

extern void NOOKS_WRAPPER(kfree)(const void * arg0);

extern void NOOKS_WRAPPER(pci_clear_mwi)(struct pci_dev *dev);

extern int NOOKS_WRAPPER(netif_receive_skb)(struct sk_buff *skb);

extern int NOOKS_WRAPPER(pci_bus_read_config_byte)(struct pci_bus *bus, unsigned int devfn, int where, u8 *val);

//extern struct page * FASTCALL(NOOKS_WRAPPER(__alloc_pages)(gfp_t arg0, unsigned int arg1, struct zonelist * arg2));
extern struct page * (NOOKS_WRAPPER(__alloc_pages)(gfp_t arg0, unsigned int arg1, struct zonelist * arg2));

extern dma_addr_t NOOKS_WRAPPER(dma_map_single)(struct device *dev, void *ptr, size_t size, enum dma_data_direction direction);

extern void *nooks_w_kmalloc(size_t size, int flags);

extern unsigned long nooks_mm_w_get_free_pages(unsigned int gfp_mask,
                                        unsigned int order);

extern void nooks_mm_w_free_pages(unsigned long addr, unsigned int order);

extern u32 NOOKS_WRAPPER(ethtool_op_get_link) (struct net_device *dev);

extern struct ethtool_perm_addr;

extern int NOOKS_WRAPPER(ethtool_op_get_perm_addr) (struct net_device *dev, struct ethtool_perm_addr *addr, u8 *data);

extern u32 NOOKS_WRAPPER(ethtool_op_get_sg) (struct net_device *dev);

extern u32 NOOKS_WRAPPER(ethtool_op_get_tso) (struct net_device *dev);

extern int NOOKS_WRAPPER(ethtool_op_set_sg) (struct net_device *dev, u32 data);

extern void NOOKS_WRAPPER(msleep) (unsigned int msecs);

extern unsigned long NOOKS_WRAPPER(msleep_interruptible) (unsigned int msecs);

extern unsigned long volatile nooks_w_jiffies;

extern int nooks_w_sprintf(char *buf, const char *fmt, ...);

//extern struct shared_info_t;

//extern shared_info_t * nooks_w_HYPERVISOR_shared_info;

//extern struct pglist_data; 

extern struct pglist_data nooks_w_contig_page_data;

extern void * nooks_w_memcpy(void * dest, const void *src, size_t n);

extern void nooks_w_put_page(struct page *page);

extern unsigned long nooks_w___per_cpu_offset[NR_CPUS];

void nooks_w___shadow_recover(void);

int
nooks_w_dummy_call0(void);

int
nooks_w_dummy_call1(int arg1);

int
nooks_w_dummy_call2(int arg1, int arg2);

int
nooks_w_dummy_call3(int arg1, int arg2, int arg3);

int
nooks_w_dummy_call4(int arg1, int arg2, int arg3, int arg4);


extern int nooks_w_init_pools(void);

#ifdef __i386__
#define NET_CALLER(arg) (*(((void**)&arg)-1))
#else
#define NET_CALLER(arg) __builtin_return_address(0)
#endif

void
nooks_free_ctl_table(pnook_t nook, ctl_table * table);

typedef struct nooks_tasklet_struct {
  struct tasklet_struct t;
  unsigned long data;
  struct tasklet_struct *nt;
} nooks_tasklet_t, *pnooks_tasklet_t;

void *nooks_w_kmem_cache_alloc(struct kmem_cache * cachep, gfp_t flags);

void nooks_w_ether_setup(struct net_device *dev);

struct net_device *NOOKS_WRAPPER(alloc_netdev) (int sizeof_priv, const char *name,
                                                void (*setup)(struct net_device *)); 

int NOOKS_WRAPPER(netif_rx) (struct sk_buff *skb);

void print_nooks_hash_tables(void);

struct resource * NOOKS_WRAPPER(__request_region) (struct resource * rp, resource_size_t arg0, resource_size_t arg1, const char *arg2);
#endif
