#ifndef _LINUX_NOOKS_DEBUG_H
#define _LINUX_NOOKS_DEBUG_H


/*
 * Macros to help debugging
 */

int nooks_w_printk(const char * fmt, ...);
//
// turn on to enable debug output
//

// #define NOOKS_DEBUG 1
#undef NOOKS_DEBUG

//
// More advanced debugging - not just printouts
//
#define NOOKS_DEBUG2 1
//#undef NOOKS_DEBUG2



#undef PDEBUG             /* undef it, just in case */
#undef PERROR             /* undef it, just in case */
#undef PDEBUG2             /* undef it, just in case */


#ifdef NOOKS_DEBUG2

#define NOOKS_ASSERT(_x_) \
do { \
  if (!(_x_)) { \
    PDEBUG(" Assertion failed in function %s, line %d of file %s\n",  \
	   __FUNCTION__, \
           __LINE__,  \
	   __FILE__); \
    BUG(); \
  } } while (0) 

#define NASSERT(_x_) \
do { \
  if (!(_x_)) { \
    asm ("int $0x3"); \
  } } while (0) 

#define nooks_kmalloc(_x_, _y_) nooks_dbg_kmalloc((_x_), (_y_))
#define nooks_kfree(_x_) nooks_dbg_kfree((_x_))

#else // NOOKS_DEBUG2

#define NOOKS_ASSERT(_x_)
#define NASSERT(_x_)
#define nooks_kmalloc(_x_, _y_) kmalloc((_x_), (_y_))
#define nooks_kfree(_x_) kfree((_x_))
#endif // NOOKS_DEBUG2


#ifdef NOOKS_DEBUG
extern unsigned int nooks_debug_level;
#  ifdef __KERNEL__

     /* This one if debugging is on, and kernel space */

#define PDEBUG(fmt, args...) \
do { \
  if (nooks_debug_level > 1) { \
    if (IN_NOOK()) {\
      nooks_w_printk(KERN_ALERT "NOOKS: " fmt, ## args);  \
    } else { \
      printk( KERN_ALERT "NOOKS: " fmt, ## args); \
    } \
  } \
} while (0)

#define PDEBUG2(_s_) \
do { \
  if (nooks_debug_level > 1) {\
    if (IN_NOOK()) { \
      nooks_w_printk(KERN_ALERT "NOOKS: %s", _s_);  \
    } else { \
      printk( KERN_ALERT "NOOKS:  %s", _s_); \
    } \
  } \
} while (0)

#define PERROR(fmt, args...) \
do { \
  if (nooks_debug_level > 0) { \
    if (IN_NOOK()) { \
      nooks_w_printk(KERN_ALERT "NOOKS-err: " fmt, ## args);  \
    } else { \
      printk( KERN_ALERT "NOOKS-err: " fmt, ## args); \
    } \
  } \
} while (0)



#  else
     /* This one for user space */
#    define PDEBUG(fmt, args...) fprintf(stderr, fmt, ## args)
#  endif
#else
#define PDEBUG(fmt, args...) /* not debugging: nothing */
#define PDEBUG2(_s_) /* not debugging: nothing */
#define PERROR(fmt, args...) /* not debugging: nothing */
#endif



#endif /* _LINUX_NOOKS_DEBUG_H */
