#ifndef _LINUX_NOOKS_DWRAPPERS_H
#define _LINUX_NOOKS_DWRAPPERS_H

#include "nooks-switch.h"
#include "nooks-work.h"
#include "nooks-wrappers.h"



#define NOOKS_DWRAPPER(_x_) nooks_d_ ## _x_



#define NOOKS_DWRAPPER_ENTER() \
do { \
NOOKS_STATS_ADD_EVENT(NOOKS_EVENT_DWRAPPER_ENTER); \
NOOKS_STATS_INC(NOOKS_OPS_DWRAPPER_ENTER); \
PDEBUG(" Entering driver function %s\n", __FUNCTION__); \
} while (0)



#define NOOKS_DWRAPPER_EXIT() \
do { \
PDEBUG(" Exiting driver function %s\n", __FUNCTION__); \
NOOKS_STATS_ADD_EVENT(NOOKS_EVENT_DWRAPPER_EXIT); \
} while (0)

#define NOOKS_DWRAPPER_EXIT2(_n_) \
do { \
PDEBUG(" Exiting driver function %s\n", __FUNCTION__); \
} while (0)

#if 0
/**********************************************************
 *
 * Module operations wrappers
 *
 *********************************************************/

int NOOKS_DWRAPPER(mops_init)(void);
void NOOKS_DWRAPPER(mops_cleanup)(void);
int NOOKS_DWRAPPER(mops_can_unload)(void);

extern struct module_operations NOOKS_DWRAPPER(mops);

/**********************************************************
 *
 * File operations wrappers
 *
 *********************************************************/

loff_t NOOKS_DWRAPPER(fops_llseek) (struct file * file, 
				    loff_t offset, 
				    int orig);

ssize_t NOOKS_DWRAPPER(fops_read) (struct file * file, 
				   char * buf, 
				   size_t size, 
				   loff_t *ppos);

int NOOKS_DWRAPPER(fops_release) (struct inode * inode, struct file * file);

ssize_t NOOKS_DWRAPPER(fops_write) (struct file * file, 
				    const char * buf, size_t size, 
				    loff_t * ppos);

int NOOKS_DWRAPPER(fops_readdir) (struct file * file, void * v, filldir_t f);

unsigned int NOOKS_DWRAPPER(fops_poll) (struct file * file, 
					struct poll_table_struct * poll);

int NOOKS_DWRAPPER(fops_ioctl) (struct inode * inode, 
				struct file * file, 
				unsigned int cmd, 
				unsigned long arg);


int NOOKS_DWRAPPER(fops_mmap) (struct file * file, 
			       struct vm_area_struct * area);

int NOOKS_DWRAPPER(fops_open) (struct inode * inode, struct file * file);

int NOOKS_DWRAPPER(fops_flush) (struct file * file);

int NOOKS_DWRAPPER(fops_release) (struct inode * inode, struct file * file);

int NOOKS_DWRAPPER(fops_fsync) (struct file * filp, 
				struct dentry * entry, 
				int datasync);

int NOOKS_DWRAPPER(fops_fasync) (int fd, struct file * filp, int on);

int NOOKS_DWRAPPER(fops_lock) (struct file * filp, 
			       int flags, 
			       struct file_lock * fl);

ssize_t NOOKS_DWRAPPER(fops_readv) (struct file * filp, 
				    const struct iovec * iov, 
				    unsigned long count, 
				    loff_t * pos);

ssize_t NOOKS_DWRAPPER(fops_writev) (struct file * filp, 
				     const struct iovec * iov, 
				     unsigned long count, 
				     loff_t * pos);

ssize_t NOOKS_DWRAPPER(fops_sendpage) (struct file * filp, 
				       struct page * page, 
				       int offset, 
				       size_t count, 
				       loff_t * pos, 
				       int more);

unsigned long NOOKS_DWRAPPER(fops_get_unmapped_area)(struct file * filp, 
						     unsigned long addr, 
						     unsigned long len, 
						     unsigned long pgoff, 
						     unsigned long flags);


extern struct file_operations NOOKS_DWRAPPER(fops);

/**********************************************************
 *
 * Block device operations wrappers
 *
 *********************************************************/


int NOOKS_DWRAPPER(bdops_open) (struct inode * inode, struct file * filp);

int NOOKS_DWRAPPER(bdops_release) (struct inode * inode, struct file * filp);

int NOOKS_DWRAPPER(bdops_ioctl) (struct inode * inode, 
				 struct file * filp, 
				 unsigned int cmd, 
				 unsigned long arg);

int NOOKS_DWRAPPER(bdops_check_media_change) (kdev_t device);

int NOOKS_DWRAPPER(bdops_revalidate) (kdev_t device);

extern struct block_device_operations NOOKS_DWRAPPER(bdops);

/**********************************************************
 *
 * VM  operations wrappers
 *
 *********************************************************/


void NOOKS_DWRAPPER(vmops_open)(struct vm_area_struct * area);

void NOOKS_DWRAPPER(vmops_close)(struct vm_area_struct * area);

struct page * NOOKS_DWRAPPER(vmops_nopage)(struct vm_area_struct * area, 
					   unsigned long address, 
					   int write_access);

extern struct vm_operations_struct NOOKS_DWRAPPER(vmops);

/**********************************************************
 *
 * Block operations wrappers
 *
 *********************************************************/

void NOOKS_DWRAPPER(blkops_request) (request_queue_t *q);

int NOOKS_DWRAPPER(blkops_back_merge) (request_queue_t *q, 
				       struct request  *req,
				       struct buffer_head *bh,
				       int max_segments);

int NOOKS_DWRAPPER(blkops_front_merge) (request_queue_t *q, 
					struct request  *req,
					struct buffer_head *bh,
					int max_segments);

int NOOKS_DWRAPPER(blkops_merge_requests) (request_queue_t *q, 
					   struct request  *req,
					   struct request  *next,
					   int max_segments);

	 

int NOOKS_DWRAPPER(blkops_make_request) (request_queue_t *q, 
					 int rw, 
					 struct buffer_head *bh);

void NOOKS_DWRAPPER(blkops_plug_device) (request_queue_t *q, kdev_t device);

extern struct blk_operations NOOKS_DWRAPPER(blkops);

/**********************************************************
 *
 * ProcFs operations wrappers
 *
 *********************************************************/

int NOOKS_DWRAPPER(procops_read_proc)(char *page, 
				      char **start, 
				      off_t off,
				      int count, 
				      int *eof, 
				      void *data);

int NOOKS_DWRAPPER(procops_write_proc)(struct file *file, 
				       const char *buffer,
				       unsigned long count, 
				       void *data);

int NOOKS_DWRAPPER(procops_get_info)(char * bug, 
				     char ** start, 
				     off_t off, 
				     int len);

extern struct procfs_operations NOOKS_DWRAPPER(procops);

/**********************************************************
 *
 * net device operations wrappers
 *
 *********************************************************/


int NOOKS_DWRAPPER(netops_init)(struct net_device *dev);
struct net_device_stats * NOOKS_DWRAPPER(netops_get_stats)(struct net_device *dev);
struct iw_statistics * NOOKS_DWRAPPER(netops_get_wireless_stats)(struct net_device *dev);

void NOOKS_DWRAPPER(netops_uninit)(struct net_device *dev);
void NOOKS_DWRAPPER(netops_destructor)(struct net_device *dev);
int NOOKS_DWRAPPER(netops_open)(struct net_device *dev);
int NOOKS_DWRAPPER(netops_stop)(struct net_device *dev);
int NOOKS_DWRAPPER(netops_hard_start_xmit) (struct sk_buff *skb,
					    struct net_device *dev);
int NOOKS_DWRAPPER(netops_hard_header) (struct sk_buff *skb,
					struct net_device *dev,
					unsigned short type,
					void *daddr,
					void *saddr,
					unsigned len);
int NOOKS_DWRAPPER(netops_rebuild_header)(struct sk_buff *skb);
void NOOKS_DWRAPPER(netops_set_multicast_list)(struct net_device *dev);
int NOOKS_DWRAPPER(netops_set_mac_address)(struct net_device *dev,
					   void *addr);
int NOOKS_DWRAPPER(netops_do_ioctl)(struct net_device *dev,
				    struct ifreq *ifr, int cmd);
int NOOKS_DWRAPPER(netops_set_config)(struct net_device *dev,
				      struct ifmap *map);
int NOOKS_DWRAPPER(netops_hard_header_cache)(struct neighbour *neigh,
					     struct hh_cache *hh);
void NOOKS_DWRAPPER(netops_header_cache_update)(struct hh_cache *hh,
						struct net_device *dev,
						unsigned char *  haddr);
int NOOKS_DWRAPPER(netops_change_mtu)(struct net_device *dev, int new_mtu);
void NOOKS_DWRAPPER(netops_tx_timeout) (struct net_device *dev);
int NOOKS_DWRAPPER(netops_hard_header_parse)(struct sk_buff *skb,
					     unsigned char *haddr);
int NOOKS_DWRAPPER(netops_neigh_setup)(struct net_device *dev, 
				       struct neigh_parms *);
int NOOKS_DWRAPPER(netops_accept_fastpath)(struct net_device * dev, 
					   struct dst_entry * entry);

extern struct netdev_operations NOOKS_DWRAPPER(netdev_ops);



void NOOKS_DWRAPPER(irqops_handler)(int irq, 
				    void * dev_id, 
				    struct pt_regs * regs);

extern struct irq_operations NOOKS_DWRAPPER(irq_ops);




int NOOKS_DWRAPPER(pciops_prepare)(struct pci_dev *dev);	
int NOOKS_DWRAPPER(pciops_activate)(struct pci_dev *dev);
int NOOKS_DWRAPPER(pciops_deactivate)(struct pci_dev *dev);

int  NOOKS_DWRAPPER(pciops_probe)(struct pci_dev *dev, 
				  const struct pci_device_id *id);
void NOOKS_DWRAPPER(pciops_remove) (struct pci_dev *dev);	
//int  NOOKS_DWRAPPER(pciops_save_state) (struct pci_dev *dev, u32 state);
void NOOKS_DWRAPPER(pciops_shutdown) (struct pci_dev *dev);
int  NOOKS_DWRAPPER(pciops_suspend)(struct pci_dev *dev, u32 state);	
int  NOOKS_DWRAPPER(pciops_resume) (struct pci_dev *dev);	                
int  NOOKS_DWRAPPER(pciops_enable_wake) (struct pci_dev *dev, u32 state, int enable);

extern  struct pci_operations NOOKS_DWRAPPER(pci_ops);

int NOOKS_DWRAPPER(callbackops_handler)(unsigned long arg);
int NOOKS_DWRAPPER(callbackops_pm_handler)(struct pm_dev *dev, pm_request_t rqst, void *data);

extern struct callback_operations NOOKS_DWRAPPER(callback_ops);

void NOOKS_DWRAPPER(timerops_handler)(unsigned long arg);

extern struct timer_operations NOOKS_DWRAPPER(timer_ops);

//
// inode operations
//

int NOOKS_DWRAPPER(inodeops_create) (struct inode * in,struct dentry * d,int i);
struct dentry * NOOKS_DWRAPPER(inodeops_lookup) (struct inode * in,struct dentry * d);
int NOOKS_DWRAPPER(inodeops_link) (struct dentry * d,struct inode * in,struct dentry * d2);
int NOOKS_DWRAPPER(inodeops_unlink) (struct inode * in,struct dentry * d);
int NOOKS_DWRAPPER(inodeops_symlink) (struct inode * in,struct dentry * d,const char * s);
int NOOKS_DWRAPPER(inodeops_mkdir) (struct inode * in,struct dentry * d,int i);
int NOOKS_DWRAPPER(inodeops_rmdir) (struct inode * in,struct dentry * d);
int NOOKS_DWRAPPER(inodeops_mknod) (struct inode * in,struct dentry * d,int i,int j);
int NOOKS_DWRAPPER(inodeops_rename) (struct inode * in, struct dentry * d,
			    struct inode * in2, struct dentry * d2);
int NOOKS_DWRAPPER(inodeops_readlink) (struct dentry * d, char * s,int i);
int NOOKS_DWRAPPER(inodeops_follow_link) (struct dentry * d, struct nameidata * n);
void NOOKS_DWRAPPER(inodeops_truncate) (struct inode * in);
int NOOKS_DWRAPPER(inodeops_permission) (struct inode * in, int i);
int NOOKS_DWRAPPER(inodeops_revalidate) (struct dentry * d);
int NOOKS_DWRAPPER(inodeops_setattr) (struct dentry * d, struct iattr * ia);
int NOOKS_DWRAPPER(inodeops_getattr) (struct dentry * d, struct iattr * ia);

extern struct inode_operations NOOKS_DWRAPPER(inode_ops);


int NOOKS_DWRAPPER(sysctlops_strategy)(ctl_table *table, int *name, 
				  int nlen, void *oldval, 
				  size_t *oldlenp, void *newval, 
				  size_t newlen, void **context);

int NOOKS_DWRAPPER(sysctlops_proc_handler)(ctl_table *ctl, int write, 
				   struct file * filp,
				   void *buffer, size_t *lenp);


extern struct sysctl_operations NOOKS_DWRAPPER(sysctl_ops);


/**********************************************************
 *
 * actor operations wrappers
 *
 *********************************************************/

int 
NOOKS_DWRAPPER(actorops_read)(read_descriptor_t * desc, 
			      struct page * page, 
			      unsigned long offset, 
			      unsigned long size);

/**********************************************************
 *
 * address_space_operations
 *
 *********************************************************/


int 
NOOKS_DWRAPPER(asops_writepage)(struct page * page);

int 
NOOKS_DWRAPPER(asops_readpage)(struct file * filp, struct page * page);

int 
NOOKS_DWRAPPER(asops_sync_page)(struct page * page);

int 
NOOKS_DWRAPPER(asops_prepare_write)(struct file * filp, 
					   struct page * page, 
					   unsigned from, 
					   unsigned to);

int 
NOOKS_DWRAPPER(asops_commit_write)(struct file * filp, 
					  struct page * page, 
					  unsigned from, 
					  unsigned to); 

int 
NOOKS_DWRAPPER(asops_bmap)(struct address_space * mapping, long block); 

int 
NOOKS_DWRAPPER(asops_flushpage) (struct page * page, 
					unsigned long offset); 

int 
NOOKS_DWRAPPER(asops_releasepage) (struct page * page, int gfp_mask); 

int 
NOOKS_DWRAPPER(asops_direct_IO)(int rw, 
				       struct inode * inode, 
				       struct kiobuf * iobuf, 
				       unsigned long blocknr, 
				       int blocksize); 


/**********************************************************
 *
 * dentry_operations
 *
 *********************************************************/


int 
NOOKS_DWRAPPER(dentryops_d_revalidate)(struct dentry * dentry, 
					      int flags); 

int 
NOOKS_DWRAPPER(dentryops_d_hash)(struct dentry * parent, 
					struct qstr * qstr); 

int 
NOOKS_DWRAPPER(dentryops_d_compare)(struct dentry * dentry, 
					   struct qstr * a, 
					   struct qstr * b); 

int 
NOOKS_DWRAPPER(dentryops_d_delete)(struct dentry * dentry); 

void 
NOOKS_DWRAPPER(dentryops_d_release)(struct dentry * dentry); 

void 
NOOKS_DWRAPPER(dentryops_d_iput)(struct dentry * dentry, 
					struct inode * inode); 


/**********************************************************
 *
 * super_operations
 *
 *********************************************************/


void 
NOOKS_DWRAPPER(superops_read_inode)(struct inode * inode); 

void 
NOOKS_DWRAPPER(superops_read_inode2)(struct inode * inode, 
					    void * opaque);  

void 
NOOKS_DWRAPPER(superops_dirty_inode)(struct inode * inode); 

void 
NOOKS_DWRAPPER(superops_write_inode)(struct inode * inode, 
					    int sync); 

void 
NOOKS_DWRAPPER(superops_put_inode)(struct inode * inode); 

void 
NOOKS_DWRAPPER(superops_delete_inode)(struct inode * inode); 

void 
NOOKS_DWRAPPER(superops_put_super)(struct super_block * sb); 

void 
NOOKS_DWRAPPER(superops_write_super)(struct super_block * sb); 

void 
NOOKS_DWRAPPER(superops_write_super_lockfs)(struct super_block * sb); 

void 
NOOKS_DWRAPPER(superops_unlockfs)(struct super_block * sb); 

int 
NOOKS_DWRAPPER(superops_statfs)(struct super_block * sb, 
				       struct statfs * buf); 

int 
NOOKS_DWRAPPER(superops_remount_fs)(struct super_block * sb, 
					   int * flags, 
					   char * data); 

void 
NOOKS_DWRAPPER(superops_clear_inode)(struct inode * inode); 

void 
NOOKS_DWRAPPER(superops_umount_begin)(struct super_block * sb); 

struct dentry * 
NOOKS_DWRAPPER(superops_fh_to_dentry)(struct super_block *sb, 
					     __u32 *fh, 
					     int len, 
					     int fhtype, 
				      int parent);

int 
NOOKS_DWRAPPER(superops_dentry_to_fh)(struct dentry * dentry, 
				      __u32 *fh, 
				      int *lenp, 
				      int need_parent);
int 
NOOKS_DWRAPPER(superops_show_options)(struct seq_file * seq, 
				       struct vfsmount * mnt);

/**********************************************************
 *
 * get_block_operations
 *
 *********************************************************/


int 
NOOKS_DWRAPPER(getblockops_get_block)(struct inode* inode,
					     long iblock,
					     struct buffer_head * bh_result,
					     int create);

int 
NOOKS_DWRAPPER(getblockops_find_inode)(struct inode * inode, 
					      unsigned long ino, 
				       void * opaque);

/**********************************************************
 *
 * sb_operations
 *
 *********************************************************/

struct super_block *
NOOKS_DWRAPPER(fsops_read_super)(struct super_block * sb, 
				 void * data, 
				 int silent);
#endif
#include "nooks-dwrap-h-gen.h"

//#include "nooks-dwrappers-muthu.h"

#endif // NOOKS_DWRAPPERS_H
