#ifndef _LINUX_NOOKS_I_H
#define _LINUX_NOOKS_I_H
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/spinlock.h>
#include <linux/highmem.h>
#include <linux/smp_lock.h>
#include <linux/blkdev.h>
#include <linux/proc_fs.h>
#include <linux/nooks.h>
#include <linux/nooks-trap.h>
#include <linux/netdevice.h>
#include <linux/pci.h>
#include <linux/isapnp.h>
#include <linux/sysctl.h>
#include <linux/rwsem.h>
#include <linux/pm.h>
#include <linux/reboot.h>


#include <asm/uaccess.h>
#include <asm/pgalloc.h>

#include "nooks-debug.h"
#include "nooks-stats.h"
#include "nooks-work.h"

#define USE_GENERATED_FILES 1
//
// Global definitions
//
#ifndef STATS
#define STATS 0
#endif

#ifndef DEBUG
#define DEBUG 1
#endif

extern unsigned long nooks_global_flag;
#define NOOKS_FLAG_DEBUG_1 0x01
#define NOOKS_FLAG_DEBUG_2 0x02
#define NOOKS_FLAG_DEBUG_3 0x04
#define NOOKS_FLAG_DEBUG_4 0x08
#define NOOKS_FLAG_DEBUG_5 0x10
#define NOOKS_FLAG_DEBUG_6 0x20
#define NOOKS_FLAG_DEBUG_7 0x40
#define NOOKS_FLAG_DEBUG_8 0x80

#define TEST_FLAG(_x_) (nooks_global_flag & (1 << ((_x_) - 1)))

#define NOOK_DRIVER_NO_MAJOR -1
#define NOOKS_MEM_FLAGS (in_interrupt() ? GFP_ATOMIC : GFP_KERNEL)
//Added by Muthu
#define NOOKS_FOPS_BLKDEV ((void *) 2)

typedef struct nook nook_t, *pnook_t;
typedef struct nook_driver nook_driver_t, *pnook_driver_t;

/*
struct nooks_dma_info_list  {
  pnook_t nook; 
  struct list_head list;
  struct device *dev;
  dma_addr_t dma_handle;
  size_t size;
  enum dma_data_direction direction;
};
*/
//extern struct nooks_dma_info_list nooks_dma_info;


typedef enum nooks_environment_type {
  syscall_e = 0,
  softirq_e = 1,
  irq_e = 2
} nooks_environment_t;

typedef enum nooks_resource_type {
  fasync_e = 1,
  waitqueue_e,
  waitqueue_sync_e,
  devfs_handle_e,
  netdev_e,
  netdev_stats_e, //6
  netdev_nook_e,
  timer_e,
  pci_driver_e,
  pci_memory_e,
  worker_e, // 11
  proc_entry_e,
  tasklet_e,
  skbuff_k_e,
  skbuff_n_e, //15
  area_map_e, // 16 
  pcidev_e,	
  irq_dev_e,
  region_e,
  ctl_table_e,
  ctl_head_e,	//21
  socket_e,
  waitqueuehead_e,
  sock_e,
  locked_sock_e,
  tcp_open_req_e,	//26
  file_e,
  actor_driver_e,
  fs_e,
  superblock_e,
  inode_e,	//31
  inode_ops_e,
  dentry_e,
  dentry_ops_e,
  buffer_head_e,
  super_ops_e,
  file_ops_e,
  iomap_e,
  as_ops_e,
  address_space_e,
  getblock_driver_e, //41
  buffer_data_e,
  dma_e,
  vm_ops_e,
  misc_device_e,
  mmap_e,
  module_data_e,  //49
  irqaction_e,
  ioctl_e,
  pm_dev_e,
  pciregion_e,
  notifier_e,
#ifdef CONFIG_SWIFI
  module_code_e,
#endif // CONFIG_SWIFI  

//added by Muthu
  gendisk_e,
  sector_info_e,	
  bd_ops_e,
  major_e,	
//end 
  vmalloc_e,
  bss_start_e,
  bss_end_e,
  kmap_e,
  dmamap_e,
  dma_device_e,
  dma_direction_e,
  dma_alloc_e,
  dma_map_page_e,
  dma_map_single_e,
  last_restype_e,
} nooks_resource_type_t, *pnooks_resource_type_t;

//
// data structure for handling i-wrappers
// Added by dmg
//
struct nook_driver;
typedef struct _nook_iwrapper {
	unsigned char  iwrapper[16]; // pointer to the specific wrapper
	int            nb_args;      // number of argument of the wrapper
	int            is_void;      // the function is void ?
	void          *dwrapper;     // pointer to the dwrapper
	void          *wrapped;      // pointer to the wrapped function, used for shadow operations
	struct nook_driver *driver;  // the nook driver
	int            errnum;       // the return error number
} nook_iwrapper_t;
typedef nook_iwrapper_t *pnook_iwrapper_t;


//
// Data structures for memory management within a nook.
//
// There are four key data structures:
//
// nooks_page_map: a linked list of pages that are writeable by the
//                 a nook
//
// nooks_area_map: a linked list of areas of allocation that are
//                 writeable by the nook
//
// nooks_cache_map: an array of kmem_caches for various sizes of allocation
//
// nooks_localcache_map: an array of kmem_caches for various sizes of 
//                       allocation accessible within the nook
//

#define CACHE_NAMELEN	20	/* max name length for a slab cache */


/*
 * The slab lists for all objects.
 */
struct kmem_list3 {
        struct list_head slabs_partial; /* partial list first, better asm code */
        struct list_head slabs_full;
        struct list_head slabs_free;
        unsigned long free_objects;
        unsigned int free_limit;
        unsigned int colour_next;       /* Per-node cache coloring */
        spinlock_t list_lock;
        struct array_cache *shared;     /* shared per node */
        struct array_cache **alien;     /* on other nodes */
        unsigned long next_reap;        /* updated without locking */
        int free_touched;               /* updated without locking */
};


/*
 * struct kmem_cache
 *
 * manages a cache.
 */

struct kmem_cache {
/* 1) per-cpu data, touched during every alloc/free */
        struct array_cache *array[NR_CPUS];
/* 2) Cache tunables. Protected by cache_chain_mutex */
        unsigned int batchcount;
        unsigned int limit;
        unsigned int shared;

        unsigned int buffer_size;
/* 3) touched by every alloc & free from the backend */
        struct kmem_list3 *nodelists[MAX_NUMNODES];

        unsigned int flags;             /* constant flags */
        unsigned int num;               /* # of objs per slab */

/* 4) cache_grow/shrink */
        /* order of pgs per slab (2^n) */
        unsigned int gfporder;

        /* force GFP flags, e.g. GFP_DMA */
        gfp_t gfpflags;

        size_t colour;                  /* cache colouring range */
        unsigned int colour_off;        /* colour offset */
        struct kmem_cache *slabp_cache;
        unsigned int slab_size;
        unsigned int dflags;            /* dynamic flags */

        /* constructor func */
        void (*ctor) (void *, struct kmem_cache *, unsigned long);

        /* de-constructor func */
        void (*dtor) (void *, struct kmem_cache *, unsigned long);

/* 5) cache creation/removal */
        const char *name;
        struct list_head next;

/* 6) statistics */
#if STATS
        unsigned long num_active;
        unsigned long num_allocations;
        unsigned long high_mark;
        unsigned long grown;
        unsigned long reaped;
        unsigned long errors;
        unsigned long max_freeable;
        unsigned long node_allocs;
        unsigned long node_frees;
        unsigned long node_overflow;
        atomic_t allochit;
        atomic_t allocmiss;
        atomic_t freehit;
        atomic_t freemiss;
#endif
#if DEBUG
        /*
         * If debugging is enabled, then the allocator can add additional
         * fields and/or padding to every object. buffer_size contains the total
         * object size including these internal fields, the following two
         * variables contain the offset to the user object and its size.
         */
        int obj_offset;
        int obj_size;
#endif
};



typedef struct _nooks_mm_map {
  struct list_head next;
  unsigned long address;
  unsigned long size;
  int prot;
} nooks_mm_map_t, *pnooks_mm_map_t;

typedef struct nooks_cache_sizes {
  size_t cs_size;
  kmem_cache_t * cs_cachep;
  kmem_cache_t * cs_dmacachep;
} nooks_cache_sizes_t;

typedef struct nooks_local_cache_sizes {
  size_t cs_size;
  struct kmem_cache cs_cache;
  struct kmem_cache cs_dmacache;
} nooks_local_cache_sizes_t;

#define NOOKS_NUM_CACHE_SIZES 1
#define NOOKS_CACHE_FIRST_SIZE 32

typedef struct _nooks_cache_map {
  nooks_cache_sizes_t caches[NOOKS_NUM_CACHE_SIZES];
} nooks_cache_map_t, *pnooks_cache_map_t;

typedef struct _nooks_local_cache_map {
  nooks_local_cache_sizes_t caches[NOOKS_NUM_CACHE_SIZES];
} nooks_local_cache_map_t, *pnooks_local_cache_map_t;



#define NOOKS_ENTER() PDEBUG(" Entering function %s\n", __FUNCTION__)
#define NOOKS_EXIT() PDEBUG(" Leaving function %s\n", __FUNCTION__)
int
nooks_mm_init(void);

void
nooks_mm_cleanup(void);

int
nooks_wrappers_init(void);

void 
nooks_cleanup(void);

/*
typedef struct blk_operations {
  request_fn_proc	* request_fn;
  merge_request_fn	* back_merge_fn;
  merge_request_fn	* front_merge_fn;
  merge_requests_fn	* merge_requests_fn;
  make_request_fn	* make_request_fn;
  plug_device_fn	* plug_device_fn;
} blk_operations_t;
i*/
//
// This type must be kept in sync with the module type in
// linux/include/module.h
//

#define NOOKS_MOPS_MAJOR -1

/*
typedef struct module_operations {
  int (*init)(void);
  void (*cleanup)(void);
  int (*can_unload)(void);
} module_operations_t;
*/

typedef struct module_operations {
  int (*init)(void);
  void (*exit)(void);
  //int (unsafe);
} module_operations_t;



typedef struct procfs_operations {
  get_info_t * get_info;
  read_proc_t * read_proc;
  write_proc_t * write_proc;
} procfs_operations_t;


typedef struct netdev_operations {

  int (*init)(struct net_device *dev);
  struct net_device_stats* (*get_stats)(struct net_device *dev);
  struct iw_statistics*	(*get_wireless_stats)(struct net_device *dev);
  void (*uninit)(struct net_device *dev);
  void (*destructor)(struct net_device *dev);
  int (*open)(struct net_device *dev);
  int (*stop)(struct net_device *dev);
  int (*hard_start_xmit) (struct sk_buff *skb,
			  struct net_device *dev);
  int (*hard_header) (struct sk_buff *skb,
		      struct net_device *dev,
		      unsigned short type,
		      void *daddr,
		      void *saddr,
		      unsigned len);
  int (*rebuild_header)(struct sk_buff *skb);
  void (*set_multicast_list)(struct net_device *dev);
  int (*set_mac_address)(struct net_device *dev,
			 void *addr);
  int (*do_ioctl)(struct net_device *dev,
		  struct ifreq *ifr, 
		  int cmd);
  int (*set_config)(struct net_device *dev,
		    struct ifmap *map);
  int (*hard_header_cache)(struct neighbour *neigh,
			   struct hh_cache *hh);
  void (*header_cache_update)(struct hh_cache *hh,
			      struct net_device *dev,
			      unsigned char *  haddr);
  int (*change_mtu)(struct net_device *dev, int new_mtu);
  void (*tx_timeout) (struct net_device *dev);
  int (*hard_header_parse)(struct sk_buff *skb,
			   unsigned char *haddr);
  int (*neigh_setup)(struct net_device *dev, 
		     struct neigh_parms * parms);
} netdev_operations_t;

typedef struct irq_operations {
  void (*handler)(int irq, 
		  void * dev_id, 
		  struct pt_regs * regs);
} irq_operations_t;

//Added by Muthu

typedef struct rqueue_operations {

	void (*request_fn)(request_queue_t *q);	

} rqueue_operations_t;


//



typedef int (*isapnp_probe_fn)(struct pci_dev *dev,
			       const struct isapnp_device_id *id);
typedef	int (*pci_probe_fn)  (struct pci_dev *dev, 
			       const struct pci_device_id *id);

typedef struct pci_operations {
  /* pci device functions */
	int (*prepare)(struct pci_dev *dev);	
	int (*activate)(struct pci_dev *dev);
	int (*deactivate)(struct pci_dev *dev);
  /* pci driver functions */
        pci_probe_fn probe;
	void (*remove) (struct pci_dev *dev);
//	int  (*save_state) (struct pci_dev *dev, u32 state);
	int  (*suspend)(struct pci_dev *dev, u32 state);
	int  (*resume) (struct pci_dev *dev);
	int  (*enable_wake) (struct pci_dev *dev, u32 state, int enable);
	void (*shutdown) (struct pci_dev *dev);
} pci_operations_t;


typedef int callback_handler_t(unsigned long arg);
typedef void tasklet_handler_t(unsigned long arg);
typedef int thread_handler_t(void * arg);

typedef struct callback_operations {
  callback_handler_t *handler;
  pm_callback pm_handler;
} callback_operations_t;

typedef struct sysctl_operations {
  ctl_handler * strategy;
  proc_handler * proc_handler;
} sysctl_operations_t;
//
// This is set in the data field if the callback is a thread, meaning
// that the driver must be deleted after the call completes
//

#define NOOKS_CALLBACK_THREAD ((void *) 1)
#define NOOKS_CALLBACK_TASKLET ((void *) 2)
#define NOOKS_CALLBACK_PM ((void *) 3)
#define NOOKS_FOPS_CHRDEV ((void *) 1)

#define NOOKS_PCIOPS_ISAPNP ((void *) 1)

typedef struct timer_operations {
  void (*handler)(unsigned long arg);
} timer_operations_t;


typedef struct actor_operations {
  read_actor_t read;
} actor_operations_t;

/*
typedef struct getblock_operations {
  get_block_t * get_block;
  find_inode_t find_inode;
} getblock_operations_t;
*/
typedef struct fs_operations {
  struct super_block *(*read_super) (struct super_block *, void *, int);
} fs_operations_t;

typedef struct notifier_operations {
  int (*notifier_call)(struct notifier_block *self, unsigned long, void *);
} notifier_operations_t;

#include "nooks-iwrapper-gen.h"
#ifdef USE_GENERATED_FILES
#include "nooks-i-h-gen.h"
#else 
#define NOOKS_IWRAPPER_TYPE_COUNT 21
typedef enum nooks_iwrapper_type {
  mops_e = 1,
  fops_e, 
  bdops_e, 
  vmops_e, 
  blkops_e,
  procops_e,
  netops_e,
  irqops_e,
  pciops_e,
  callbackops_e,
  timerops_e,
  inodeops_e,
  sysctlops_e,
  actorops_e,
  asops_e,
  dentryops_e,
  superops_e,
  getblockops_e,
  fsops_e,
  notifierops_e,
  last_ops_e
} nooks_iwrapper_type_t, *pnooks_iwrapper_type_t;

typedef union nook_operations {
  struct module_operations mops;
  struct file_operations fops;
  struct block_device_operations bdops;
  struct vm_operations_struct vmops;
  struct blk_operations blkops;
  struct procfs_operations procops;
  struct netdev_operations netops;
  struct irq_operations irqops;
  struct pci_operations pciops;
  struct callback_operations callbackops;
  struct timer_operations timerops;
  struct inode_operations inodeops;
  struct sysctl_operations sysctlops;
  struct actor_operations actorops;
  struct address_space_operations asops;
  struct dentry_operations dentryops;
  struct super_operations superops;
  struct getblock_operations getblockops;
  struct fs_operations fsops;
  struct notifier_operations notifierops;

  //ADDED BY MUTHU
  struct block_device_operations blkdev_ops;
  struct rqueue_operations rqueueops;
  //END		

} nook_operations_t;

typedef union nook_iwrapper_operations {
  void * dummy_ops;
  struct iwrapper_netdev_operations * netops;
  struct iwrapper_dentry_operations * dentryops;
  struct iwrapper_module_operations * mops;
  struct iwrapper_file_operations * fops;
  struct iwrapper_actor_operations * actorops;
  struct iwrapper_procfs_operations * procops;
  struct iwrapper_address_space_operations * asops;
  struct iwrapper_inode_operations * inodeops;
  struct iwrapper_pci_operations * pciops;
  struct iwrapper_notifier_operations * notifierops;
  struct iwrapper_fs_operations * fsops;
  struct iwrapper_blk_operations * blkops;
  struct iwrapper_getblock_operations * getblockops;
  struct iwrapper_callback_operations * callbackops;
  struct iwrapper_super_operations * superops;
  struct iwrapper_block_device_operations * bdops;
  struct iwrapper_vm_operations_struct * vmops;
  struct iwrapper_timer_operations * timerops;
  struct iwrapper_sysctl_operations * sysctlops;
  struct iwrapper_irq_operations * irqops;
} nook_iwrapper_operations_t;

#endif

#if 0
typedef struct nooks_resource_elem {
  struct list_head next;
  nooks_resource_type_t restype;
  void * resource;
  void * value;
} nooks_resource_elem_t, *pnooks_resource_elem_t;
#endif

// for request_region structures
typedef struct nooks_resource_struct {
  char * name;
  struct resource * parent;
  struct resource * res;
  unsigned long start;
  unsigned long n;
} nooks_resource_t, *pnooks_resource_t;

#define NOOKS_STACK_POOL_SIZE 8
typedef struct nook_stack_pool {
  unsigned long free_mask; // zero means stack is available
  spinlock_t lock;
  void * stacks[NOOKS_STACK_POOL_SIZE];
} nook_stack_pool_t, *pnook_stack_pool_t;

typedef struct nooks_lookaside_elem {
  struct nooks_lookaside_elem * next;
} nooks_lookaside_elem_t, *pnooks_lookaside_elem_t;

#define NOOKS_LOOKASIDE_KERNEL 1
#define NOOKS_LOOKASIDE_NOOK   2
#define NOOKS_LOOKASIDE_PAGE   4
#define NOOKS_LOOKASIDE_NOOK_PAGE 8

typedef struct nooks_lookaside_list {
  pnooks_lookaside_elem_t list;
  int qlen;
  spinlock_t lock;
  int highwater;
  size_t size;
  int flags;
#ifdef NOOKS_KEEP_STATS
  unsigned long alloc_heap;
  unsigned long alloc_pool;
  unsigned long free_heap;
  unsigned long free_pool;
#endif // NOOKS_KEEP_STATS
} nooks_lookaside_list_t, *pnooks_lookaside_list_t;

#include "nooks-hash.h"
#include "nooks-recover.h"

typedef struct nooks_resource_list {
  nooks_hash_table_t hash;
} nooks_resource_list_t, *pnooks_resource_list_t;
 

typedef struct nook_resources {
  nooks_cache_map_t cache_map;
  nooks_hash_table_t area_map;
  nooks_hash_table_t active_skbs;
  struct list_head page_map;
  struct list_head cache_list;
  nooks_resource_list_t resource_list;
  nooks_work_queue_t kernel_work_list;
  int res_qlen;
  int maps_qlen;
  struct sk_buff_head skb_pool;
  spinlock_t lock;
  //struct mm_struct * mm; No shadow pagetable
  nooks_lookaside_list_t stack_pool;
  nooks_lookaside_list_t hash_pool;
  nooks_lookaside_list_t map_pool;
  nooks_lookaside_list_t work_pool;

} nook_resources_t;


#define NOOKS_LOCK_RESOURCES(_n_,_f_) spin_lock_irqsave(&(_n_)->resources.lock, _f_)
#define NOOKS_UNLOCK_RESOURCES(_n_,_f_) spin_unlock_irqrestore(&(_n_)->resources.lock, _f_)


typedef struct nooks_list_elem {
  struct list_head next;
  void * data;
} nooks_list_elem_t, *pnooks_list_elem_t;

typedef struct nook_locals {
  nooks_local_cache_map_t cache_map;
  struct list_head cache_list;
  struct sk_buff_head skb_pool;
  spinlock_t lock;
    struct list_head work_list;
  struct list_head free_work_list;           // entries to be freed
  nooks_lookaside_list_t work_pool;
  nooks_lookaside_list_t elem_pool;
  nooks_lookaside_list_t page_pool;
  struct sk_buff_head skb_rx_list;
  struct list_head skb_free_list;
} nook_locals_t, *pnook_locals_t;


typedef enum nook_state {
  nook_state_init = 0,
  nook_state_running,
  nook_state_faulted,
  nook_state_cleanup,
  nook_state_zombie
} nook_state_t;



typedef struct nooks_recovery_state {
  struct work_struct recovery_task;
  const char * file;
  int line;
  int type;
  struct pt_regs regs;
  nooks_recovery_info_t info;
} nooks_recovery_state_t, *pnooks_recovery_state_t;

#define NOOKS_DEADLOCK_CALL_LIMIT 10000

struct shadow_driver;
typedef struct nooks_livelock_state {
  unsigned long last_clock;
  unsigned long calls_since_last_clock;
} nooks_livelock_state_t, *pnooks_livelock_state_t;

#define NOOK_MAX_NAME_LEN 30
struct nook {
  struct list_head next;
  struct list_head drivers;
  unsigned int refs;
  nook_state_t state;
  char * name;
  nook_resources_t resources;
  pnook_locals_t locals;
  spinlock_t lock;
  nooks_recovery_state_t recovery_state;
  nooks_livelock_state_t livelock_state;  
  struct shadow_driver * shadow;
#ifdef CONFIG_NOOKS_PROC
  char *shadow_driver_name;
  struct proc_dir_entry *proc_ent;
  struct timeval last_fault_time;
  char *last_fault_reason;
  char *last_fault_mod;
  unsigned last_fault_eip;
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
  struct timeval last_replug_time;
#endif
#endif
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
  long waiting_time;
#endif
};

union nooks_shadow_operations;
union nooks_shadow_kernel_operations;

struct nook_driver {
  struct list_head next;
  struct list_head next_in_nook;
  struct list_head next_in_shadow;
  char * name;
  void * data;
  unsigned int refs;
  unsigned int major;
  unsigned int need_delete;
  nooks_iwrapper_type_t driver_type;
  pnook_t nook;
  unsigned int shadow_state;
  struct nook_driver * parent;

  //
  // Pointers to real functions in drivers
  //
  nook_operations_t operations;
  //
  // Pointer to nooks wrappers for real functions
  //
  nook_iwrapper_operations_t iwrapper_operations;
  
  //
  // Function tables for passing to kernel, which
  // may need NULL pointers for non-existant functions
  //

  nook_operations_t wrapper_operations;
  union nooks_shadow_operations * shadow_operations;
  union nooks_shadow_kernel_operations * shadow_kernel_operations;
};



extern pnook_t nooks_global_nook;
extern spinlock_t nook_list_lock;

//  List of all current nooks
extern struct list_head nook_list;

// list of all current drivers
extern struct list_head nook_driver_list;

//
// Nooks page table
//

extern struct mm_struct nooks_mm_struct;


// array of possible drivers, for an array of function tables
#define MAX_NOOK_DRIVERS 100

// BUGBUG: temporary routines to lookup current nook

#define ENTER_NOOK() \
do { \
get_current()->nook_state.curr.flags |= NOOK_STATE_IN_NOOK; } while (0)

#define EXIT_NOOK() \
do { \
get_current()->nook_state.curr.flags &= ~NOOK_STATE_IN_NOOK; } while (0)

#define CLEAR_NOOK() \
do { \
get_current()->nook_state.curr.current_nook = NULL;  \
get_current()->nook_state.curr.flags &= ~NOOK_STATE_IN_NOOK; } while (0)

#define SET_CURRENT_NOOK(_n_) \
do { \
get_current()->nook_state.curr.current_nook = (_n_); } while (0)

#define GET_NOOK_STATE() (get_current()->nook_state)

#define GET_TASK_CURRENT_NOOK(_t_) (_t_)->nook_state.curr.current_nook
#define GET_CURRENT_NOOK() GET_TASK_CURRENT_NOOK(get_current())

#define GET_NOOK_ARG(_i_) get_current()->nook_state.curr.args[(_i_)]

#define SET_NOOK_ARG(_i_,_v_) do { \
get_current()->nook_state.curr.args[(_i_)] = (unsigned long) (_v_); } while (0)

#define SWAP_NOOK_ARG(_i_, _v_, _o_) do { \
(_o_) = get_current()->nook_state.curr.args[(_i_)]; \
get_current()->nook_state.curr.args[(_i_)] = (unsigned long) (_v_); } while (0)

 
//
// Routines for referencing and dereferencing nooks for multithreaded
// safety
//

#define NOOKS_REFERENCE(_n_) (_n_)->refs++
#define NOOKS_DEREFERENCE(_n_) (--(_n_)->refs)

#define FIELD_OFFSET(_t_, _f_) ((unsigned long) & ((( _t_ *) 0)->_f_))
#define struct_entry(ptr, type, member) \
	((type *)((char *)(ptr)-(unsigned long)(&((type *)0)->member)))

#define ROUND_UP_COUNT(_i_, _r_) (((_i_) + (_r_) - 1) & ~((_r_) - 1))

#include "nooks-util.h"
#include "nooks-mm.h"

#endif /* _LINUX_NOOKS_I_H */
