/*
 * nooks-init.c -- initialization code for nooks
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  
 * No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

#include <linux/percpu.h> 
#include <asm/processor.h>
 
#include "nooks-i.h"
#include "nooks-iwrappers.h"
#include "nooks-dwrappers.h"
#include "nooks-recover.h"
#ifdef CONFIG_NOOKS_PROC
#include <../shadow/shadow-proc.h>
#endif

//
// Global variables for nooks
//

LIST_HEAD(nook_list);
LIST_HEAD(nook_driver_list);
nook_driver_t * nook_driver_array[NOOKS_IWRAPPER_TYPE_COUNT][MAX_NOOK_DRIVERS];

pnook_t nooks_global_nook;

unsigned long nooks_global_flag = 0;
#ifdef NOOKS_DEBUG
unsigned int nooks_debug_level = 0;
#endif


/**********************************************************
 *
 * nooks_init_drivers
 *
 * Initiatlizes data structures required by nooks code
 *
 *********************************************************/
int
nooks_init_drivers(void)
{
  int i,j;
  INIT_LIST_HEAD(&nook_driver_list);

  for (i = 0; i < NOOKS_IWRAPPER_TYPE_COUNT; i++) {
    for (j = 0; j < MAX_NOOK_DRIVERS; j++) {
      nook_driver_array[i][j] = NULL;
    }
  }


  return(0);
  
}

/**********************************************************
 *
 * nooks_init_nooks
 *
 * Initiatlizes data structures required by nooks code
 *
 *********************************************************/
int
nooks_init_nooks(void)
{
  int result = 0;

  //
  // Initialize global lists and arrays
  //


  INIT_LIST_HEAD(&nook_list);
  spin_lock_init(&nook_list_lock);
  //
  // BUGBUG: for now, create a nook and set it as the current nook on all
  // CPUs.
  //
  
  result = nooks_i_create_nook("global-nook", &nooks_global_nook);
  if (result < 0) {
    goto Cleanup;
  }

 Cleanup:
  if (result < 0) {
    if (nooks_global_nook != NULL) {
      nooks_free_nook(nooks_global_nook);
      printk ("Global nooks creation failed.\n");
      nooks_global_nook = NULL;
    }
  }
  
  return(result);
}

/**********************************************************
 *
 * nooks_init
 *
 * Initiatlizes data structures required by nooks code
 *
 *********************************************************/
extern unsigned long doublefault_stack[];
extern struct tss_struct doublefault_tss;
extern struct desc_struct gdt_table[];

int nooks_init(void)
{
  int result = 0;
#ifndef CONFIG_X86_NO_TSS
  int cpu = smp_processor_id();
  struct tss_struct * t = &per_cpu(init_tss, cpu);
#endif
  // result = nooks_mm_init(); We do not copy shadow page tables now. FIXME
  if (result < 0) {
    goto Cleanup;
  }
  result = nooks_init_nooks();
  if (result < 0) {
    goto Cleanup;
  }
#ifdef NOOKS_KEEP_STATS
  nooks_init_nooks_stats();
#endif // NOOKS_KEEP_STATS
  
#ifdef CONFIG_NOOKS_PROC
  result = shadow_proc_init();
  if (result < 0)
	  goto Cleanup;
#endif
  
  result = nooks_wrappers_init();
  if (result < 0) {
    goto Cleanup;
  }
  result = nooks_init_drivers();
  if (result < 0) {
    goto Cleanup;
  }
  result = nooks_init_iwrappers();
  if (result < 0) {
    goto Cleanup;
  }
  result = nooks_init_recovery();
  if (result < 0) {
    goto Cleanup;
  }
 
  printk ("All nooks initialization successful.\n"); 
  Cleanup:
  if (result < 0) {
    printk("<1> Failed to initialize nooks: %d\n",result);
    nooks_cleanup();
  }
  return(result);  
}


/**********************************************************
 *
 * nooks_cleanup
 *
 * Cleans up and frees data structures required by nooks code
 *
 *********************************************************/
void 
nooks_cleanup(void)
{
  nooks_mm_cleanup();
}



