#ifndef _LINUX_NOOKS_IWRAPPERS_H
#define _LINUX_NOOKS_IWRAPPERS_H

#include <../shadow/netshadow.h>
#include "nooks-i.h"
#include "nooks-dwrappers.h"
/**********************************************************
 *
 * This file declares macros that define sets of interfaces
 * so that each instance of the interface is used by a different
 * driver. The linux kernel code calls one of these functions,
 * which sets the current nook and then calls the 
 *
 *********************************************************/

#define NOOKS_IWRAPPER_REGISTERED_ITF     0x1
#define NOOKS_IWRAPPER_NAMED              0x10
#define NOOKS_IWRAPPER_NUMBERED           0x20

void nooks_iwrapper_debug(void);

/*
 * Bunch of assembly codes for wrappers call from the
 * generic iwrapper
 */
// push the argument
#define NOOKS_IWRAPPER_PUSH(p) \
  asm ( "pushl %0\n" : : "r" (p) )

// reduce the stack of s word (s must be a string)
#define NOOKS_IWRAPPER_REDUCE_STACK(s) \
  asm ( "addl $" s ", %esp\n")

// call the function p with n arguments from the stack
// r is for getting the result
// "r" => Store in a register
// :Output:Input:List of Clobbered Registers
#define NOOKS_IWRAPPER_CALL(p, n, r) \
  asm ( \
	"movl %0, %%eax\n" \
	"subl %%eax, %%esp\n" \
	"pushl %%esi\n" \
	"pushl %%edi\n" \
	"movl %%esp, %%edi\n" \
	"movl %%ebp, %%esi\n" \
	"addl $8, %%edi\n" \
	"addl $8, %%esi\n" \
	"shr  $2, %%eax\n" \
	"movl %%eax, %%ecx\n" \
	"repz movsd\n" \
	"popl %%edi\n" \
	"popl %%esi\n" \
	: : "r" (n*4) : "esi","edi","eax","ecx","esp" ); \
  r ( (int (*)(void)) p)(); \
  asm ("addl %0, %%esp\n" : : "r" (n*4))

// call function p with n arguments from the stack putting the result in r
#define NOOKS_IWRAPPER_CALL_RES(p, n, r)\
	NOOKS_IWRAPPER_CALL(p, n, r = )
// call function p with n arguments from the stack + d as last arg, result in r
#define NOOKS_IWRAPPER_CALL_RES_DRIV(p, n, r, d)\
	NOOKS_IWRAPPER_PUSH(d); \
	NOOKS_IWRAPPER_PUSH(r); \
	NOOKS_IWRAPPER_CALL(p, n, r = ); \
	NOOKS_IWRAPPER_REDUCE_STACK("8")
// call function p with n args from the stack + d
#define NOOKS_IWRAPPER_CALL_DRIV(p, n, d) \
	NOOKS_IWRAPPER_PUSH(d); \
	NOOKS_IWRAPPER_CALL(p, n, ); \
	NOOKS_IWRAPPER_REDUCE_STACK("4")

#define NOOKS_IWRAPPER_DEBUG() nooks_iwrapper_debug()

#define NOOKS_ICALL static
#define NOOKS_IWRAPPER(_v_, _x_) nooks_i_ ## _v_ ## _ ##  _x_
#define NOOKS_IWRAPPER_NOVER(_x_) nooks_i_ ##  _x_
// dmg - the assembly line is to get the nook_iwrapper structure pointer
#define NOOKS_IWRAPPER_ENTER \
pnook_iwrapper_t _iwrap; \
pnook_t _old_nook, __nook;\
pnook_driver_t _old_driver,__driver;\
asm volatile("movl %%ecx, %0\n" : "=r" (_iwrap) : ); \
NOOKS_IWRAPPER_DEBUG(); \
NOOKS_STATS_ADD_EVENT(NOOKS_EVENT_IWRAPPER_ENTER); \
_old_nook = GET_CURRENT_NOOK(); \
_old_driver = GET_CURRENT_NOOK_DRIVER(); \
PDEBUG(" Entering driver function %s\n", __FUNCTION__); \
__driver = _iwrap->driver; \
SET_CURRENT_NOOK_DRIVER3(__driver);\
__nook = GET_CURRENT_NOOK();


#define NOOKS_IWRAPPER_ENTER_CHECK(_v_, _t_, _ty_ ) \
NOOKS_IWRAPPER_ENTER(_v_,_t_); \
NOOKS_IWRAPPER_CHECK(_ty_)

#define NOOKS_IWRAPPER_CHECK( _t_ ) \
if ((__nook == NULL) || (__nook->state != nook_state_running)) { \
   NOOKS_IWRAPPER_FAIL( _t_ ); \
} \

#define NOOKS_IWRAPPER_FAIL( _t_ ) \
NOOKS_IWRAPPER_EXIT(); \
return _t_

#define NOOKS_IWRAPPER_EXIT() \
do { \
  SET_CURRENT_NOOK_DRIVER2(_old_driver); \
  SET_CURRENT_NOOK(_old_nook); \
  PDEBUG(" Exiting driver function %s\n", __FUNCTION__); \
  NOOKS_STATS_ADD_EVENT(NOOKS_EVENT_IWRAPPER_EXIT); \
} while (0)


int
nooks_init_iwrappers(void);


typedef void * nooks_iwrapper_itf;

// Array of drivers using each itf instance
extern unsigned long nooks_iwrapper_flags[];
extern char * nooks_iwrapper_names[];

#define GET_CURRENT_NOOK_DRIVER() get_current()->nook_state.curr.current_driver
#define SET_CURRENT_NOOK_DRIVER(_n_,_t_) \
do { \
pnook_driver_t driver = nooks_iwrapper_driver_array[_t_][_n_]; \
if (driver != NULL) { \
   get_current()->nook_state.curr.current_driver  = driver; \
   SET_CURRENT_NOOK(driver->nook); \
  } else {\
    SET_CURRENT_NOOK(NULL); \
  }  \
} while(0)

#define SET_CURRENT_NOOK_DRIVER2(_p_) \
do { get_current()->nook_state.curr.current_driver  = _p_; \
} while(0)

#define SET_CURRENT_NOOK_DRIVER3(_p_) \
do { \
if ((_p_) != NULL) { \
   get_current()->nook_state.curr.current_driver  = (_p_);  \
   SET_CURRENT_NOOK((_p_)->nook); \
  } else {\
    SET_CURRENT_NOOK(NULL); \
  }  \
} while(0)

nooks_iwrapper_itf nooks_iwrapper_get_itf(nooks_iwrapper_type_t itf_type,
					  pnook_driver_t driver);

void nooks_iwrapper_release_itf(nooks_iwrapper_type_t itf_type,
				nooks_iwrapper_itf itf);

// dmg - now it's iwrapper_*_operations instead of *_operations that
// contains the iwrappers informations
void
nooks_iwrapper_copy_mops(struct module * module_in,
			 struct module_operations * wrapper_operations, 
			 struct iwrapper_module_operations * iwrapper_operations,
			 struct module * module_out);

void
nooks_iwrapper_copy_fops(struct file_operations * driver_operations, 
			 struct file_operations * wrapper_operations,
			 struct iwrapper_file_operations * iwrapper_operations,
			 struct file_operations * system_operations);

void
nooks_iwrapper_copy_vmops(struct vm_operations_struct * driver_operations, 
			  struct vm_operations_struct * wrapper_operations,
			  struct iwrapper_vm_operations_struct * iwrapper_operations,
			  struct vm_operations_struct * system_operations);


void
nooks_iwrapper_copy_netops(struct net_device * netdev_in,
			   struct netdev_operations * wrapper_operations, 
			   struct iwrapper_netdev_operations * iwrapper_operations,
			   struct net_device * netdev_out);

void
nooks_iwrapper_copy_irqops(void (*handler)(int, void *, struct pt_regs *),
			   struct irq_operations * wrapper_operations,
			   struct iwrapper_irq_operations * iwrapper_operations,
			   struct irq_operations * system_operations);

void
nooks_iwrapper_copy_callbackops(callback_handler_t * handler,
				struct callback_operations * wrapper_operations,
				struct iwrapper_callback_operations * iwrapper_operations,
				struct callback_operations * system_operations);
void
nooks_iwrapper_copy_pm_callbackops(pm_callback pm_handler,
				   struct callback_operations * wrapper_operations,
				   struct iwrapper_callback_operations * iwrapper_operations,
				   struct callback_operations * system_operations);

void
nooks_iwrapper_copy_timerops(void (*handler)(unsigned long),
				struct timer_operations * wrapper_operations,
				struct iwrapper_timer_operations * iwrapper_operations,
				struct timer_operations * system_operations);

void
nooks_iwrapper_copy_pciops(struct pci_driver * pcidrv_in,
			   struct pci_operations * wrapper_operations, 
			   struct iwrapper_pci_operations * iwrapper_operations,
			   struct pci_driver * pcidrv_out);

void
nooks_iwrapper_copy_isapnp_pciops(isapnp_probe_fn probe,
				  struct pci_operations * wrapper_operations, 
				  struct iwrapper_pci_operations * iwrapper_operations,
				  struct pci_operations * system_operations);

void
nooks_iwrapper_copy_procops(struct proc_dir_entry * pentry_in,
			    struct procfs_operations * wrapper_operations, 
			    struct iwrapper_procfs_operations * iwrapper_operations,
			    struct proc_dir_entry * pentry_out);

void
nooks_iwrapper_copy_inodeops(struct inode_operations * driver_operations, 
			     struct inode_operations * wrapper_operations,
			     struct iwrapper_inode_operations * iwrapper_operations,
			     struct inode_operations * system_operations);


void
nooks_iwrapper_copy_sysctlops(ctl_table * ctl_in,
			   struct sysctl_operations * wrapper_operations, 
			   struct iwrapper_sysctl_operations * iwrapper_operations,
			   ctl_table * ctl_out);

void
nooks_iwrapper_copy_actorops(read_actor_t actor,
			     struct actor_operations * wrapper_operations,
			     struct iwrapper_actor_operations * iwrapper_operations,
			     struct actor_operations * system_operations);

void
nooks_iwrapper_copy_fsops(struct file_system_type * fs_in,
			  struct fs_operations * wrapper_operations, 
			  struct iwrapper_fs_operations * iwrapper_operations,
			  struct file_system_type * fs_out);

void
nooks_iwrapper_copy_dentryops(struct dentry_operations * driver_operations, 
			      struct dentry_operations * wrapper_operations,
			      struct iwrapper_dentry_operations * iwrapper_operations,
			      struct dentry_operations * system_operations);

void
nooks_iwrapper_copy_superops(struct super_operations * driver_operations, 
			     struct super_operations * wrapper_operations,
			     struct iwrapper_super_operations * iwrapper_operations,
			     struct super_operations * system_operations);

void
nooks_iwrapper_copy_asops(struct address_space_operations * driver_operations, 
			  struct address_space_operations * wrapper_operations,
			  struct iwrapper_address_space_operations * iwrapper_operations,
			  struct address_space_operations * system_operations);

void
nooks_iwrapper_copy_notifierops(struct notifier_block * notifier_in,
				struct notifier_operations * wrapper_operations, 
				struct iwrapper_notifier_operations * iwrapper_operations,
				struct notifier_block * notifier_out);

void
nooks_iwrapper_copy_bdops( 
		struct block_device_operations * driver_operations, 
		struct block_device_operations * wrapper_operations,
		struct iwrapper_block_device_operations * iwrapper_operations,
		struct block_device_operations * system_operations);

#endif // NOOKS_IWRAPPERS_H
