#ifndef _LINUX_NOOKS_RECOVER_H
#define _LINUX_NOOKS_RECOVER_H

#include <linux/module.h>

#define NOOKS_TASK_ABORT_RETRY_COUNT 3


#define NOOKS_FAULT(X) nooks_ ## X ## _fault(__FILE__, __LINE__)
#define FAULT_PARAMETER 1
#define FAULT_ENOMEM 	2
#define FAULT_RESOURCE 	3
#define FAULT_STACK 	4
#define FAULT_DEADLOCK 	5

extern int nooks_parameter_fault(const char * file, const int line);
extern int nooks_enomem_fault(const char * file, const int line);
extern int nooks_resource_fault(const char * file, const int line);
extern int nooks_deadlock_fault(const char * file, const int line);
	
extern void * nooks_driver_call_fault;
extern void * nooks_driver_fault;

void
nooks_fault_handler(pnook_t nook);

void 
nooks_fault_alert_usermode(void * u_nook);

int 
nooks_init_recovery(void);

long
nooks_do_unload_nook(pnook_t nook);

int
nooks_clean_nook(pnook_t nook);

void nook_print_drivers(pnook_t);

void 
nooks_poll_device(unsigned int irq);

void 
nooks_unpoll_device(unsigned int irq, int from_cleanup);

int
nooks_reset_nook_state(pnook_t nook);

void
nooks_free_resource_hash(pnook_t nook, pnooks_hash_table_t table, int pass);

//
// return values:
// 0 means the driver can be deleted
// < 0 means the driver should not be deleted yet
//

#define NOOKS_RECOVERY_STOP_PASS 0
#define NOOKS_RECOVERY_UNLOAD_PASS1 1
#define NOOKS_RECOVERY_UNLOAD_PASS2 2
#define NOOKS_RECOVERY_CLEANUP_PASS 3

typedef int (nooks_driver_recovery_fn)(pnook_t nook, 
				    pnook_driver_t driver, 
				    int pass);

typedef int (nooks_resource_recovery_fn)(pnook_t nook, 
					 void * key, 
					 void * data, 
					 int pass);


typedef long (nooks_recovery_fn)(pnook_t nook);

typedef struct nooks_recovery_info {
  nooks_recovery_fn * recover;
  nooks_driver_recovery_fn ** driver_recovery;
  nooks_resource_recovery_fn ** resource_recovery;
} nooks_recovery_info_t, *pnooks_recovery_info_t;


extern nooks_resource_recovery_fn * nooks_resource_recovery[];
extern nooks_driver_recovery_fn * nooks_driver_recovery[];


void
nooks_set_recovering_interrupt(unsigned int irq);

void
nooks_clear_recovering_interrupt(unsigned int irq);


nooks_driver_recovery_fn nooks_recover_generic;
nooks_driver_recovery_fn nooks_recover_mops;
nooks_driver_recovery_fn nooks_recover_fops;
//nooks_driver_recovery_fn nooks_recover_bdops;
//nooks_driver_recovery_fn nooks_recover_vmops;
//nooks_driver_recovery_fn nooks_recover_blkops;
nooks_driver_recovery_fn nooks_recover_procops;
nooks_driver_recovery_fn nooks_recover_netops;
nooks_driver_recovery_fn nooks_recover_irqops;
nooks_driver_recovery_fn nooks_recover_pciops;
nooks_driver_recovery_fn nooks_recover_callbackops;
nooks_driver_recovery_fn nooks_recover_timerops;
//nooks_driver_recovery_fn nooks_recover_inodeops;
nooks_driver_recovery_fn nooks_recover_sysctlops;
nooks_driver_recovery_fn nooks_recover_actorops;
//nooks_driver_recovery_fn nooks_recover_asops;
//nooks_driver_recovery_fn nooks_recover_dentryops;
//nooks_driver_recovery_fn nooks_recover_superops;
//nooks_driver_recovery_fn nooks_recover_getblockops;
//nooks_driver_recovery_fn nooks_recover_fsops;
nooks_driver_recovery_fn nooks_recover_notifierops;
nooks_driver_recovery_fn nooks_test_recover_bdops;


nooks_resource_recovery_fn nooks_res_recover_generic;
nooks_resource_recovery_fn nooks_res_recover_fasync;
nooks_resource_recovery_fn nooks_res_recover_waitqueue;
nooks_resource_recovery_fn nooks_res_recover_waitqueue_sync;
nooks_resource_recovery_fn nooks_res_recover_devfs_handle;
nooks_resource_recovery_fn nooks_res_recover_netdev;
nooks_resource_recovery_fn nooks_res_recover_netdev_stats;
nooks_resource_recovery_fn nooks_res_recover_netdev_nook;
nooks_resource_recovery_fn nooks_res_recover_timer;
nooks_resource_recovery_fn nooks_res_recover_pci_driver;
nooks_resource_recovery_fn nooks_res_recover_pci_memory;
nooks_resource_recovery_fn nooks_res_recover_worker;
nooks_resource_recovery_fn nooks_res_recover_proc_entry;
nooks_resource_recovery_fn nooks_res_recover_tasklet;
nooks_resource_recovery_fn nooks_res_recover_skbuff_k;
nooks_resource_recovery_fn nooks_res_recover_skbuff_n;
nooks_resource_recovery_fn nooks_res_recover_area_map;
nooks_resource_recovery_fn nooks_res_recover_pcidev;
nooks_resource_recovery_fn nooks_res_recover_irq_dev;
nooks_resource_recovery_fn nooks_res_recover_region;
nooks_resource_recovery_fn nooks_res_recover_ctl_table;
nooks_resource_recovery_fn nooks_res_recover_ctl_head;
nooks_resource_recovery_fn nooks_res_recover_ctl_head;
nooks_resource_recovery_fn nooks_res_recover_socket;
nooks_resource_recovery_fn nooks_res_recover_waitqueuehead;
nooks_resource_recovery_fn nooks_res_recover_sock;
nooks_resource_recovery_fn nooks_res_recover_locked_sock;
nooks_resource_recovery_fn nooks_res_recover_tcp_open_req;
nooks_resource_recovery_fn nooks_res_recover_file;
nooks_resource_recovery_fn nooks_res_recover_actor_driver;
nooks_resource_recovery_fn nooks_res_recover_fs;
nooks_resource_recovery_fn nooks_res_recover_superblock;
nooks_resource_recovery_fn nooks_res_recover_inode;
nooks_resource_recovery_fn nooks_res_recover_inode_ops;
nooks_resource_recovery_fn nooks_res_recover_dentry;
nooks_resource_recovery_fn nooks_res_recover_dentry_ops;
nooks_resource_recovery_fn nooks_res_recover_buffer_head;
nooks_resource_recovery_fn nooks_res_recover_super_ops;
nooks_resource_recovery_fn nooks_res_recover_file_ops;
nooks_resource_recovery_fn nooks_res_recover_iomap;
nooks_resource_recovery_fn nooks_res_recover_as_ops;
nooks_resource_recovery_fn nooks_res_recover_address_space;
nooks_resource_recovery_fn nooks_res_recover_getblock_driver;
nooks_resource_recovery_fn nooks_res_recover_buffer_data;
nooks_resource_recovery_fn nooks_res_recover_dma;
nooks_resource_recovery_fn nooks_res_recover_vm_ops;
nooks_resource_recovery_fn nooks_res_recover_misc_device;
nooks_resource_recovery_fn nooks_res_recover_mmap;
nooks_resource_recovery_fn nooks_res_recover_module_data;
nooks_resource_recovery_fn nooks_res_recover_irqaction;
nooks_resource_recovery_fn nooks_res_recover_ioctl;
nooks_resource_recovery_fn nooks_res_recover_pm_dev;
nooks_resource_recovery_fn nooks_res_recover_pciregion;
nooks_resource_recovery_fn nooks_res_recover_notifier;
nooks_resource_recovery_fn nooks_res_recover_dummy1;
nooks_resource_recovery_fn nooks_res_recover_dummy2;
nooks_resource_recovery_fn nooks_res_recover_gendisk;
nooks_resource_recovery_fn nooks_res_recover_sector_info;
nooks_resource_recovery_fn nooks_res_recover_bd_ops;
nooks_resource_recovery_fn nooks_res_recover_major;


nooks_resource_recovery_fn  nooks_res_recover_vmalloc;
nooks_resource_recovery_fn  nooks_res_recover_bssstart;
nooks_resource_recovery_fn  nooks_res_recover_bssend;
nooks_resource_recovery_fn  nooks_res_recover_kmap;
nooks_resource_recovery_fn  nooks_res_recover_dmamap;
nooks_resource_recovery_fn  nooks_res_recover_dmadevice;
nooks_resource_recovery_fn  nooks_res_recover_dmadir;
nooks_resource_recovery_fn  nooks_res_recover_dmaalloc;
nooks_resource_recovery_fn  nooks_res_recover_dmamappage;
nooks_resource_recovery_fn  nooks_res_recover_dmamapsingle;





#define NOOKS_RECOVER_FREED 1

#endif // _LINUX_NOOKS_RECOVER_H
