/*
 * nooks-resource.c -- resource lists for nooks
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  
 * No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

 
#include "nooks-i.h"


/**********************************************************
 *
 * nooks_add_resource_elem
 *
 * add an elem to the nook lists of objects to remember
 *
 *********************************************************/

int
nooks_add_resource_elem(pnook_t nook, 
			nooks_resource_type_t restype,
			void * resource,
			void * value)
{
  int result = 0;
 
  result = nooks_add_to_hash(&nook->resources.resource_list.hash,
			     restype,
			     resource,
			     value);
  return(result);
}

/**********************************************************
 *
 * nooks_del_resource_elem
 *
 * add a fasync struct to the nook lists of fasync structs
 *
 *********************************************************/

int
nooks_del_resource_elem(pnook_t nook, 
			nooks_resource_type_t restype,
			void * resource,
			void ** value)
{
  int result;
  if (nook == NULL)
	return -ENOMEM;

  if (resource == NULL) 
  {
	printk("Resource is null in nooks_del_resource_elem.\n");
	return -EFAULT;
  }
  result = nooks_del_from_hash(&nook->resources.resource_list.hash,
			       restype,
			       resource,
			       value);
  return(result);
}

/**********************************************************
 *
 * nooks_find_resource_elem
 *
 * add a fasync struct to the nook lists of fasync structs
 *
 *********************************************************/

int
nooks_find_resource_elem(pnook_t nook, 
			 nooks_resource_type_t restype,
			 void * resource,
			 void ** value)
{
  int result;

  if (resource == NULL)
	return -ENOMEM;
  result = nooks_find_in_hash(&nook->resources.resource_list.hash,
			      restype,
			      resource,
			      value);

  return(result);
}


/**********************************************************
 *
 * nooks_del_resource_elem_reverse
 *
 * add a fasync struct to the nook lists of fasync structs
 *
 *********************************************************/

int
nooks_del_resource_elem_reverse(pnook_t nook, 
				nooks_resource_type_t restype,
				void * value,
				void ** resource)
{
  int result;
  result = nooks_del_from_hash_reverse(&nook->resources.resource_list.hash,
				       restype,
				       value,
				       resource);

  return(result);
}

/**********************************************************
 *
 * nooks_find_resource_elem_reverse
 *
 * add a fasync struct to the nook lists of fasync structs
 *
 *********************************************************/

int
nooks_find_resource_elem_reverse(pnook_t nook, 
				 nooks_resource_type_t restype,
				 void * value,
				 void ** resource)
{
  int result;
  result = nooks_find_in_hash_reverse(&nook->resources.resource_list.hash,
				      restype,
				      value,
				      resource);


  return(result);
}

