/*
 * nooks-stats.c -- statistics code for nooks
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  
 * No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

#include "nooks-i.h"
#include <linux/kernel_stat.h>


#define LIMIT (PAGE_SIZE-128) /* don't print any more after this size */

u_int64_t nooks_call_time;

#ifdef NOOKS_KEEP_STATS
int nooks_ops_num = NOOKS_OPS_NUM;
int nooks_timer_num = NOOKS_TIMER_NUM;

char * nooks_ops_name[NOOKS_OPS_NUM] = {
"NOOKS_OPS_DWRAPPER_ENTER",
"NOOKS_OPS_ALLOC_SKB",
"NOOKS_OPS_ALLOC_SKB_POOL",
"NOOKS_OPS_ALLOC_SKB_NOOK_POOL",
"NOOKS_OPS_KFREE_SKB_POOL",
"NOOKS_OPS_CONTEXT_IN",
"NOOKS_OPS_STACK_POOL_HITS",
"NOOKS_OPS_STACK_PAGE_HITS",
"NOOKS_OPS_TRAP1",
"NOOKS_OPS_ILLEGAL_INST",
"NOOKS_OPS_NETIF_RX",
"NOOKS_OPS_SKB_MOVED",
"NOOKS_OPS_KMALLOC_BYTES",
"NOOKS_OPS_SKB_MOVED_TO_NOOK",
"NOOKS_OPS_SKB_MOVED_TO_KERNEL",
"NOOKS_OPS_W_KMALLOC_BYTES",
"NOOKS_OPS_STACK_POOL_FREE",
"NOOKS_OPS_STACK_PAGE_FREE",
"NOOKS_OPS_STACK_NOT_FREE",
"NOOKS_OPS_STACK_NOT_ALLOC",
"NOOKS_OPS_KERNEL_NETIF_RX",
"NOOKS_OPS_PAGES_ALLOCED",
"NOOKS_OPS_PAGES_FREED",
"NOOKS_OPS_DRIVER_CALLS",
"NOOKS_OPS_KERNEL_CALLS"
};

char * nooks_timers_name[NOOKS_OPS_NUM] = {
"NOOKS_TIMER_CONTEXT_IN",
"NOOKS_TIMER_CONTEXT_OUT",
"NOOKS_TIMER_DWRAPPER_IN",
"NOOKS_TIMER_DWRAPPER_OUT"
};

unsigned long nooks_ops_count[NOOKS_OPS_NUM] = {0};
u_int64_t nooks_ops_timers[NOOKS_TIMER_NUM] = {0};

u_int64_t nooks_history_timers[NOOKS_TIMER_HISTORY] = {0};
int nooks_history_events[NOOKS_TIMER_HISTORY] = {0};
int nooks_history_entry = 0;

#endif // NOOKS_KEEP_STATS

#ifdef CONFIG_NOOKS_STATS    

/* MS-addition */
int ms_ops_num = MS_OPS_NUM;
char * ms_ops_name[MS_OPS_NUM] = {
  "MS_OPS_IN",
  "MS_OPS_OUT",
  "MS_OPS_DMA",
  "MS_OPS_BLK_OPEN",
  "MS_OPS_BLK_RELEASE",
  "MS_OPS_BLK_IOCTL",
  "MS_OPS_BLK_CHECK",
  "MS_OPS_BLK_REVALIDATE",
  "MS_OPS_AS_WRITE",
  "MS_OPS_AS_READ",
  "MS_OPS_AS_SYNC",
  "MS_OPS_AS_PREPARE",
  "MS_OPS_AS_COMMIT",
  "MS_OPS_AS_BMAP",
  "MS_OPS_AS_DIRECT",
  "MS_OPS_FS_LLSEEK",
  "MS_OPS_FS_READ",
  "MS_OPS_FS_WRITE",
  "MS_OPS_FS_READDIR",
  "MS_OPS_FS_POLL",
  "MS_OPS_FS_IOCTL",
  "MS_OPS_FS_MMAP",
  "MS_OPS_FS_OPEN",
  "MS_OPS_FS_FLUSH",
  "MS_OPS_FS_RELEASE",
  "MS_OPS_FS_FSYNC",
  "MS_OPS_FS_FASYNC",
  "MS_OPS_FS_LOCK",
  "MS_OPS_FS_READV",
  "MS_OPS_FS_WRITEV",
  "MS_OPS_FS_SENDPAGE",
  "MS_OPS_FS_GET_UNMAPPED" ,
  "MS_OPS_CLI",
  "MS_OPS_STI",
  "MS_OPS_SAVE_FLAGS",
  "MS_OPS_RESTORE_FLAGS",
  "MS_OPS_IRQ_EVENT"
};

unsigned long ms_ops_count[MS_OPS_NUM] = {0};
/* end-MS-addition */

void ms_access_int(void) {}
void ms_access_dma(void) {}
void ms_access_io(void) {}

#endif // CONFIG_NOOKS_STATS    


#ifdef NOOKS_KEEP_STATS

#define PRINT_NOOK_LOOKASIDE_STAT(_n_,_v_) { \
    len += sprintf(buf+len, "%s: %ld %ld %ld %ld\n", (_n_), \
                   (_v_)->alloc_heap, \
                   (_v_)->alloc_pool, \
                   (_v_)->free_heap, \
                   (_v_)->free_pool); \
    if (len > LIMIT) goto Cleanup; \
    }


int nooks_read_driver_time(char *buf, char **start, off_t offset,
                   int len, int *eof, void *data)
{
    len=0;

    // the time
    len += sprintf(buf+len,"%lli",  nooks_call_time);
    
    *eof = 1;
    return len;
}

int nooks_read_per_nook_stats(char *buf, int len)
{
  pnook_t nook = NULL;
  struct list_head * next = NULL;

  //
  // Enumerate the nooks, counting the size of each nook
  //

  list_for_each(next, &nook_list) {
    nook = list_entry(next,nook_t, next);
    len += sprintf(buf+len, "Nook: %s ah ap fh fp\n", nook->name);
    if (len > LIMIT) 
      goto Cleanup;
    PRINT_NOOK_LOOKASIDE_STAT("hash_pool" ,&nook->resources.hash_pool);
    PRINT_NOOK_LOOKASIDE_STAT("map_pool" ,&nook->resources.map_pool);
    PRINT_NOOK_LOOKASIDE_STAT("work_pool" ,&nook->resources.work_pool);
    PRINT_NOOK_LOOKASIDE_STAT("stack_pool" ,&nook->resources.stack_pool);

    PRINT_NOOK_LOOKASIDE_STAT("local_work_pool" ,&nook->locals->work_pool);
    PRINT_NOOK_LOOKASIDE_STAT("local_elem_pool" ,&nook->locals->elem_pool);
    PRINT_NOOK_LOOKASIDE_STAT("local_page_pool" ,&nook->locals->page_pool);
    
    
  }
 Cleanup:
  return(len);
}

#define CLEAR_NOOK_LOOKASIDE_STAT(_v_) { \
                   (_v_)->alloc_heap = \
                   (_v_)->alloc_pool = \
                   (_v_)->free_heap =  \
                   (_v_)->free_pool = 0; } \

int nooks_clear_per_nook_stats(void)
{
  pnook_t nook = NULL;
  struct list_head * next = NULL;

  //
  // Enumerate the nooks, counting the size of each nook
  //

  list_for_each(next, &nook_list) {
    nook = list_entry(next,nook_t, next);
    CLEAR_NOOK_LOOKASIDE_STAT(&nook->resources.hash_pool);
    CLEAR_NOOK_LOOKASIDE_STAT(&nook->resources.map_pool);
    CLEAR_NOOK_LOOKASIDE_STAT(&nook->resources.work_pool);
    CLEAR_NOOK_LOOKASIDE_STAT(&nook->resources.stack_pool);

    CLEAR_NOOK_LOOKASIDE_STAT(&nook->locals->work_pool);
    CLEAR_NOOK_LOOKASIDE_STAT(&nook->locals->elem_pool);       
  }
  return(0);
}

/*
  Steve's additions

  This code inits a proc directory to output the results of
  various nooks benchmarks.

  This proc entry is different than Mike's because it can
  also be written to in order to clear stats.
*/
// This is the read function
int nooks_read_nooks_stats(char *buf, char **start, off_t offset,
                   int len, int *eof, void *data)
{
  int i;
  len=0;
	

  len += sprintf(buf+len, "Counters:\n");
  for (i = 0; i < nooks_ops_num; i++ ) {
    len += sprintf(buf+len, "%s: %ld\n", 
		   nooks_ops_name[i], 
		   nooks_ops_count[i]);
    if (len > LIMIT) {
      goto Cleanup;
    }
  }


  len += sprintf(buf+len, "Timers:\n");
  for (i = 0; i < nooks_timer_num; i++ ) {
    len += sprintf(buf+len, "%s: %lli\n", 
		   nooks_timers_name[i], 
		   nooks_ops_timers[i]);
    if (len > LIMIT) {
      goto Cleanup;
    }
  }

  len = nooks_read_per_nook_stats(buf, len);
 Cleanup:

  *eof = 1;
  return len;
}


int nooks_clear_nooks_stats(struct file* file, const char* buffer, unsigned long count, void* data)
{
    // dont really care WHAT was written, if you're writing into the
    // proc entry, stats should be clered.

    // cast the value into an int pointer
    int value = 0; // int * value = (int *) data
    int i = 0;
    
    int len = sizeof(int); // we want to copy an int
    
    // do the copy from userspace into kernel space
    if(copy_from_user(&value, buffer, len)){ 
	printk("error in proc write!\n");
	return -EFAULT;
    }

    for (i = 0; i < nooks_ops_num; i++) {
      nooks_ops_count[i] = 0;
    }

    for (i = 0; i< nooks_timer_num; i++) {
      nooks_ops_timers[i] = 0;
    }

    for (i = 0; i< NOOKS_TIMER_HISTORY; i++) {
      nooks_history_events[i] = 0;
      nooks_history_timers[i] = 0;
    }

    nooks_clear_per_nook_stats();
    return len;

}


int nooks_read_nooks_timers(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
  int i;
  int len=0;
  int total_len = 0;
  int page_offset = 0;
  int passed_offset = 0;

  len = sprintf(buf+len, "Last entry:  %d\n", nooks_history_entry);
  total_len += len;
  page_offset += len;
  for (i = (nooks_history_entry+(NOOKS_TIMER_HISTORY/2)) 
	 & (NOOKS_TIMER_HISTORY - 1); 
       i != nooks_history_entry; 
       i = (i + 1) & (NOOKS_TIMER_HISTORY - 1)) {
    len = sprintf(buf+page_offset, "[%d]: %d : %lli\n", 
		  i,
		  nooks_history_events[i], 
		  nooks_history_timers[i]);
    total_len += len;
    page_offset += len;
    if (page_offset > LIMIT) {
      goto Cleanup;
    }
  }
	
#if 0
  *start = buf;
  if (offset == 0) {
    passed_offset = 1;
  }
  len = sprintf(buf+len, "Last entry:  %d\n", nooks_history_entry);
  total_len += len;
  page_offset += len;
  for (i = (nooks_history_entry+1) % NOOKS_TIMER_HISTORY; 
       i != nooks_history_entry; 
       i = (i + 1) % NOOKS_TIMER_HISTORY) {
    len = sprintf(buf+page_offset, "[%d]: %d : %lli\n", 
		  i,
		  nooks_history_events[i], 
		  nooks_history_timers[i]);
    total_len += len;
    page_offset += len;
    if ((total_len > offset) && (!passed_offset)) {
      memmove(buf, 
	      buf + page_offset - (total_len - offset), 
	      (total_len - offset));
      passed_offset = 1;
    } else if (page_offset > LIMIT) {
      //
      // We are copying good data here
      //
      if (passed_offset) {
	goto Cleanup;
      }
      page_offset = 0;
    }
    else if (total_len > offset + count) {
      goto Cleanup;
    }
  }

  if (total_len < offset) {
    page_offset = 0;
  }
#endif
  *eof = 1;
 Cleanup:
  return (page_offset);
}


// the init function
int nooks_init_nooks_stats(void)
{
  struct proc_dir_entry * ptr;

  // create the proc entry for the nooks context stuff
  ptr = create_proc_entry("nooks_stats", 0755, NULL);

  // set the read function
  ptr->read_proc = nooks_read_nooks_stats;
  ptr->write_proc = nooks_clear_nooks_stats;

  // create proc entry for driver
  ptr = create_proc_entry("inside_driver", 0755, NULL);

  // set the read function
  ptr->read_proc = nooks_read_driver_time;


  ptr = create_proc_entry("nooks_timers", 0755, NULL);
  ptr->read_proc = nooks_read_nooks_timers;

  //
  // Make the locals writeable
  //

  /* nooks_mm_update_pagetable(nooks_global_nook,
			    NOOKS_WRITEABLE, 
			    (unsigned long) &nooks_ops_count,
			    sizeof(int) +
			    (unsigned long) ((char *) &nooks_history_entry -
					     (char *) &nooks_ops_count));

  */
  return 0; /* succeed */
}


#endif // NOOKS_KEEP_STATS

unsigned long ms_trace_buffer;
unsigned long ms_trace_buffer_size;

int ms_init_trace_buffer(void)
{
  ms_trace_buffer_size = 2 * PAGE_SIZE;
  ms_trace_buffer = __get_free_pages(GFP_KERNEL, 1);
  if (ms_trace_buffer == 0) {
    return(-ENOMEM);
  }
  memset((void *) ms_trace_buffer, 0xef, ms_trace_buffer_size);
  return(0);
}
