#ifndef _NOOKS_STATS_H
#define _NOOKS_STATS_H

// #define NOOKS_KEEP_STATS 1
#undef NOOKS_KEEP_STATS


#define NOOKS_OPS_DWRAPPER_ENTER                  0
#define NOOKS_OPS_ALLOC_SKB                       1
#define NOOKS_OPS_ALLOC_SKB_POOL                  2 
#define NOOKS_OPS_ALLOC_SKB_NOOK_POOL             3
#define NOOKS_OPS_KFREE_SKB_POOL                  4
#define NOOKS_OPS_CONTEXT_IN                      5
#define NOOKS_OPS_STACK_POOL_HITS                 6
#define NOOKS_OPS_STACK_PAGE_HITS                 7
#define NOOKS_OPS_TRAP1                           8
#define NOOKS_OPS_ILLEGAL_INST                    9
#define NOOKS_OPS_NETIF_RX                        10
#define NOOKS_OPS_SKB_MOVED                       11
#define NOOKS_OPS_KMALLOC_BYTES                   12
#define NOOKS_OPS_SKB_MOVED_TO_NOOK               13
#define NOOKS_OPS_SKB_MOVED_TO_KERNEL             14
#define NOOKS_OPS_W_KMALLOC_BYTES                 15
#define NOOKS_OPS_STACK_POOL_FREE                 16
#define NOOKS_OPS_STACK_PAGE_FREE                 17
#define NOOKS_OPS_STACK_NOT_FREE                  18
#define NOOKS_OPS_STACK_NOT_ALLOC                 19
#define NOOKS_OPS_KERNEL_NETIF_RX                 20
#define NOOKS_OPS_PAGES_ALLOCED                   21
#define NOOKS_OPS_PAGES_FREED                     22
#define NOOKS_OPS_DRIVER_CALLS                    23
#define NOOKS_OPS_KERNEL_CALLS                    24

#define NOOKS_OPS_NUM                             25


#define NOOKS_TIMER_CONTEXT_IN                    0
#define NOOKS_TIMER_CONTEXT_OUT                   1
#define NOOKS_TIMER_DWRAPPER_OUT                 2 
#define NOOKS_TIMER_DWRAPPER_IN                   3
#define NOOKS_TIMER_PT_UPDATE                     4

#define NOOKS_TIMER_NUM                           5

//
// Timer events
//

#define NOOKS_EVENT_IWRAPPER_ENTER 1
#define NOOKS_EVENT_IWRAPPER_EXIT 2

#define NOOKS_EVENT_DWRAPPER_ENTER 3
#define NOOKS_EVENT_DWRAPPER_EXIT 4

#define NOOKS_EVENT_DRIVER_CALL 5
#define NOOKS_EVENT_DRIVER_RETURN 6

#define NOOKS_EVENT_DRIVER_CALLOUT 7
#define NOOKS_EVENT_DRIVER_CALLRETURN 8

#define NOOKS_EVENT_KERNEL_CALL 9
#define NOOKS_EVENT_KERNEL_RETURN 10

#define NOOKS_EVENT_KERNEL_CALLOUT 11
#define NOOKS_EVENT_KERNEL_CALLRETURN 12

#define NOOKS_EVENT_DRIVER_COPIED_ARGS      101
#define NOOKS_EVENT_DRIVER_PROCESSED_LIST1  102
#define NOOKS_EVENT_DRIVER_CS_IN            103
#define NOOKS_EVENT_DRIVER_SWAP_REGS        104
#define NOOKS_EVENT_DRIVER_SWAP_PT          105
#define NOOKS_EVENT_DRIVER_PROCESSED_LIST2  106
#define NOOKS_EVENT_DRIVER_PUSH_ARGS        107
#define NOOKS_EVENT_DRIVER_PROCESS_LIST3    108
#define NOOKS_EVENT_DRIVER_SWAP_PT2         109
#define NOOKS_EVENT_DRIVER_SWAP_REGS2       110
#define NOOKS_EVENT_DRIVER_CS_OUT           111
#define NOOKS_EVENT_DRIVER_PROCESSED_LIST4  112
#define NOOKS_EVENT_KERNEL_COPIED_ARGS      113
#define NOOKS_EVENT_KERNEL_PROCESSED_LIST1  114
#define NOOKS_EVENT_KERNEL_SWAP_PT          115
#define NOOKS_EVENT_KERNEL_SWAP_REGS        116
#define NOOKS_EVENT_KERNEL_CS_OUT           117
#define NOOKS_EVENT_KERNEL_PROCESSED_LIST2  118
#define NOOKS_EVENT_KERNEL_PUSH_ARGS        119
#define NOOKS_EVENT_KERNEL_RETURN_GOOP      120
#define NOOKS_EVENT_KERNEL_CS_IN            121
#define NOOKS_EVENT_KERNEL_PROCESSED_LIST3  122

// macros

#define NOOK_TASK_TIME() get_current()->nook_state.curr.time

extern u_int64_t nooks_call_time;
#ifdef NOOKS_KEEP_STATS
extern int nooks_ops_num;
extern int nooks_timer_num;
extern char * nooks_ops_name[];
extern char * nooks_timers_name[];
extern unsigned long nooks_ops_count[];
extern u_int64_t nooks_ops_timers[];



#define NOOKS_STATS_INC(_x_) do { nooks_ops_count[_x_]++;} while (0)
#define NOOKS_STATS_ADD(_x_, _y_) do { nooks_ops_count[_x_] += (_y_); } while (0)
#define NOOKS_STATS_START_TIME() \
{ u_int64_t __start_time, __end_time; \
  rdtscll(__start_time);

#define NOOKS_STATS_TIME(_x_) \
  rdtscll(__end_time); \
  nooks_ops_timers[_x_] += (__end_time - __start_time); \
 }

#define NOOKS_RDTSCLL(_x_) rdtscll(_x_)

#define NOOKS_TIMER_HISTORY 256
#define NOOKS_TIMER_HISTORY_STRING "256"

extern u_int64_t nooks_history_timers[NOOKS_TIMER_HISTORY];
extern int nooks_history_events[NOOKS_TIMER_HISTORY];
extern int nooks_history_entry;
#if 0
#define NOOKS_STATS_ADD_EVENT(_x_) \
do { \
  nooks_history_events[nooks_history_entry] = (_x_); \
  rdtscll(nooks_history_timers[nooks_history_entry]); \
  nooks_history_entry = (nooks_history_entry + 1) & ~(NOOKS_TIMER_HISTORY-1); \
} while (0)
#endif


#define NOOKS_STATS_ADD_EVENT(_x_) \
  __asm__ __volatile__ \
    (   "movl	nooks_history_entry, %%ebx \n\t" \
        "movl	%0, %%esi \n\t" \
        "movl	%%esi, nooks_history_events(,%%ebx,4) \n\t" \
        "rdtsc  \n\t" \
	"movl	%%eax, %%esi \n\t" \
	"leal	1(%%ebx), %%eax \n\t" \
	"movl	%%edx, %%edi \n\t" \
	"movl	%%esi, nooks_history_timers(,%%ebx,8) \n\t" \
	"andl	$" NOOKS_TIMER_HISTORY_STRING " - 1, %%eax \n\t" \
	"movl	%%edi, nooks_history_timers+4(,%%ebx,8) \n\t" \
	"movl	%%eax, nooks_history_entry \n\t" \
        : : "i" (_x_) : "eax", "ebx", "edx", "esi", "edi", "memory" );


// #define NOOKS_KEEP_DETAILED_STATS 1
#undef NOOKS_KEEP_DETAILED_STATS

#ifdef NOOKS_KEEP_DETAILED_STATS
#define NOOKS_STATS_ADD_EVENT2(_x_) NOOKS_STATS_ADD_EVENT(_x_)
#else // NOOKS_KEP_DETAILED_STATS
#define NOOKS_STATS_ADD_EVENT2(_x_) 
#endif // NOOKS_KEP_DETAILED_STATS

#else //  NOOKS_KEEP_STATS

#define NOOKS_STATS_INC(_x_)
#define NOOKS_STATS_ADD(_x_, _y_) 
#define NOOKS_STATS_START_TIME()
#define NOOKS_STATS_TIME(_x_) 
#define NOOKS_STATS_ADD_EVENT(_x_)
#define NOOKS_STATS_ADD_EVENT2(_x_)
#define NOOKS_RDTSCLL(_x_)

#endif // NOOKS_KEEP_STATS

int nooks_init_nooks_stats(void);

#endif // _NOOKS_STATS_H




