#ifndef _LINUX_NOOKS_SWITCH_H
#define _LINUX_NOOKS_SWITCH_H

// How many pages are on a nook stack?
//#include <current.h>
#define NOOKS_STACK_SIZE PAGE_SIZE


//extern struct tlb_state cpu_tlbstate[NR_CPUS];


//
// Stack layout
//

typedef struct nooks_stack_frame {
  struct list_head work_list;
} nooks_stack_frame_t , *pnooks_stack_frame_t;

typedef struct nooks_alloc_stack_frame {
  pnooks_stack_frame_t old_frame;
  void * old_stack;
} nooks_alloc_stack_frame_t, * pnooks_alloc_stack_frame_t;

typedef struct nooks_kernel_stack_frame {
  struct nook_thread_state thread_state;
  void * frame_register;
} nooks_kernel_stack_frame_t , *pnooks_kernel_stack_frame_t;


void
nooks_context_switch_set_param(unsigned long param);

#ifdef CONFIG_NOOKS_NOPROT

#define nooks_dwrapper_stack_alloc(_size_) (alloca(_size_))
#define nooks_dwrapper_stack_release()

#else

void * 
nooks_dwrapper_stack_alloc(size_t size);

void
nooks_dwrapper_stack_release(void);

#endif

#define SWITCH_CALLOUT       0
#define SWITCH_OUT           1

#define SWITCH_IN            2
#define SWITCH_CALLRETURN    3


//
// Context switch macros (x86 only, sorry)
//

#define NOOKS_SWAP_GLOBAL_PGD(_pgd_) \
  do {								\
		__asm__ __volatile__(					\
			"movl %1, %%cr4;  # turn off PGE     \n"	\
			"movl %0, %%cr3;                     \n"	\
			"movl %2, %%cr4;  # turn PGE back on \n"	\
			: : "r" (_pgd_),				\
			    "r" (mmu_cr4_features & ~X86_CR4_PGE),	\
			    "r" (mmu_cr4_features)			\
			: "memory");					\
	} while (0)

#ifdef CONFIG_X86_PGE
#define NOOKS_SWAP_PGD(_pgd_)  NOOKS_SWAP_GLOBAL_PGD(_pgd_)
#else // CONFIG_X86_PGE
#define NOOKS_SWAP_PGD(_pgd_) \
  do { \
    if (cpu_has_pge) { \
      NOOKS_SWAP_GLOBAL_PGD(_pgd_); \
    } else { \
      asm volatile("movl %0,%%cr3" : :"r" (_pgd_)); \
    } \
} while (0)
#endif // CONFIG_X86_PGE

static inline void nooks_switch_in_mm(struct mm_struct *prev, 
				      struct mm_struct *next, 
				      unsigned cpu)
{
  NASSERT(prev->pgd);
  NASSERT(next->pgd);
  /* stop flush ipis for the previous mm */
  clear_bit(cpu, (void *)&prev->cpu_vm_mask);
  
#ifdef CONFIG_SMP
  per_cpu(cpu_tlbstate, cpu).state = TLBSTATE_OK;
  per_cpu(cpu_tlbstate, cpu).active_mm = next; 
  //cpu_tlbstate[cpu].state = TLBSTATE_OK;
  //cpu_tlbstate[cpu].active_mm = next;
#endif
  
  set_bit(cpu, (void *)&next->cpu_vm_mask);
  // FIXME set_bit(cpu, (void *)&next->context.cpuvalid);
  
  /* Re-load page tables, with flush of global pages */
  NOOKS_SWAP_PGD(__pa(next->pgd));  
}

static inline void nooks_switch_out_mm(struct mm_struct *prev, 
				       struct mm_struct *next, 
				       unsigned cpu)
{
  NASSERT(prev->pgd);
  NASSERT(next->pgd);

  /* Re-load page tables. there should be no global pages to flush */  

  asm volatile("movl %0,%%cr3": :"r" (__pa(next->pgd)));
  
  //
  // BUGBUG: do we have a race condition if we do this
  // after changing the PGD pointer?
  //
  
  /* stop flush ipis for the previous mm */
  
  clear_bit(cpu, (void *) &prev->cpu_vm_mask);
  
#ifdef CONFIG_SMP
  per_cpu(cpu_tlbstate, cpu).state = TLBSTATE_OK;
  per_cpu(cpu_tlbstate, cpu).active_mm = next; 
  //cpu_tlbstate[cpu].state = TLBSTATE_OK;
  //cpu_tlbstate[cpu].active_mm = next;
#endif
  
  set_bit(cpu, (void *) &next->cpu_vm_mask);
// FIXME  set_bit(cpu, &next->context.cpuvalid);
}


//
//  Macros for calling from the kernel into nooks
//

unsigned long nooks_driver_call(void * func, int argc, ...);

#define NOOKS_DRIVER_CALL_MAX_PARAM 10

//#ifdef CONFIG_NOOKS_NOPROT

#if 1

#define NOOKS_PLAIN_DRIVER_CALL0( _f_ ) \
    _f_( ) 

#define NOOKS_PLAIN_DRIVER_CALL1( _f_, _a1_ ) \
    _f_( (_a1_)) 

#define NOOKS_PLAIN_DRIVER_CALL2( _f_, _a1_, _a2_ )  \
    _f_( (_a1_), (_a2_)) 

#define NOOKS_DRIVER_CALL0( _f_ ) \
    GET_CURRENT_NOOK_DRIVER()->operations._f_( ) 

#define NOOKS_DRIVER_CALL1( _f_, _a1_ ) \
    GET_CURRENT_NOOK_DRIVER()->operations._f_( (_a1_)) 

#define NOOKS_DRIVER_CALL2( _f_, _a1_, _a2_ )  \
    GET_CURRENT_NOOK_DRIVER()->operations._f_( \
		      (_a1_), (_a2_)) 

#define NOOKS_DRIVER_CALL3( _f_, _a1_, _a2_, _a3_ ) \
    GET_CURRENT_NOOK_DRIVER()->operations._f_( \
		      (_a1_), (_a2_), (_a3_)) 

#define NOOKS_DRIVER_CALL4( _f_, _a1_, _a2_, _a3_, _a4_ ) \
    GET_CURRENT_NOOK_DRIVER()->operations._f_( \
		      (_a1_), (_a2_), (_a3_), (_a4_)) 

#define NOOKS_DRIVER_CALL5( _f_, _a1_, _a2_, _a3_, _a4_, _a5_ ) \
    GET_CURRENT_NOOK_DRIVER()->operations._f_( \
		      (_a1_), (_a2_), (_a3_), (_a4_), (_a5_)) 

#define NOOKS_DRIVER_CALL6( _f_, _a1_, _a2_, _a3_, _a4_, _a5_, _a6_ ) \
    GET_CURRENT_NOOK_DRIVER()->operations._f_( \
		      (_a1_), (_a2_), (_a3_), (_a4_), (_a5_), (_a6_)) 

#define NOOKS_DRIVER_CALL7( _f_, _a1_, _a2_, _a3_, _a4_, _a5_, _a6_, _a7_ ) \
    GET_CURRENT_NOOK_DRIVER()->operations._f_( \
		      (_a1_), (_a2_), (_a3_), (_a4_), (_a5_), (_a6_), (_a7_)) 

#define NOOKS_DRIVER_CALL8( _f_, _a1_, _a2_, _a3_, _a4_, _a5_, _a6_, _a7_, _a8_ ) \
    GET_CURRENT_NOOK_DRIVER()->operations._f_( \
		      (_a1_), (_a2_), (_a3_), (_a4_), (_a5_), (_a6_), (_a7_), (_a8_)) 

#else
#define NOOKS_PLAIN_DRIVER_CALL0( _f_ ) \
    nooks_driver_call(_f_, 0)

#define NOOKS_PLAIN_DRIVER_CALL1( _f_, _a1_ ) \
    nooks_driver_call(_f_, 1, (_a1_))

#define NOOKS_PLAIN_DRIVER_CALL2( _f_, _a1_, _a2_ )  \
    nooks_driver_call(_f_, 2, \
		      (_a1_), (_a2_))


#define NOOKS_DRIVER_CALL0( _f_ ) \
    nooks_driver_call(GET_CURRENT_NOOK_DRIVER()->operations._f_, 0) 222

#define NOOKS_DRIVER_CALL1( _f_, _a1_ ) \
    nooks_driver_call(GET_CURRENT_NOOK_DRIVER()->operations._f_, 1, (_a1_))

#define NOOKS_DRIVER_CALL2( _f_, _a1_, _a2_ )  \
    nooks_driver_call(GET_CURRENT_NOOK_DRIVER()->operations._f_, 2, \
		      (_a1_), (_a2_))

#define NOOKS_DRIVER_CALL3( _f_, _a1_, _a2_, _a3_ ) \
    nooks_driver_call(GET_CURRENT_NOOK_DRIVER()->operations._f_, 3, \
		      (_a1_), (_a2_), (_a3_))

#define NOOKS_DRIVER_CALL4( _f_, _a1_, _a2_, _a3_, _a4_ ) \
    nooks_driver_call(GET_CURRENT_NOOK_DRIVER()->operations._f_, 4, \
		      (_a1_), (_a2_), (_a3_), (_a4_))

#define NOOKS_DRIVER_CALL5( _f_, _a1_, _a2_, _a3_, _a4_, _a5_ ) \
    nooks_driver_call(GET_CURRENT_NOOK_DRIVER()->operations._f_, 5, \
		      (_a1_), (_a2_), (_a3_), (_a4_), (_a5_))

#define NOOKS_DRIVER_CALL6( _f_, _a1_, _a2_, _a3_, _a4_, _a5_, _a6_ ) \
    nooks_driver_call(GET_CURRENT_NOOK_DRIVER()->operations._f_, 6, \
		      (_a1_), (_a2_), (_a3_), (_a4_), (_a5_), (_a6_))

#define NOOKS_DRIVER_CALL7( _f_, _a1_, _a2_, _a3_, _a4_, _a5_, _a6_, _a7_ ) \
    nooks_driver_call(GET_CURRENT_NOOK_DRIVER()->operations._f_, 7, \
		      (_a1_), (_a2_), (_a3_), (_a4_), (_a5_), (_a6_), (_a7_))

#define NOOKS_DRIVER_CALL8( _f_, _a1_, _a2_, _a3_, _a4_, _a5_, _a6_, _a7_, _a8_ ) \
    nooks_driver_call(GET_CURRENT_NOOK_DRIVER()->operations._f_, 8, \
		      (_a1_), (_a2_), (_a3_), (_a4_), (_a5_), (_a6_), (_a7_), (_a8_))
#endif // CONFIG_NOOKS_NOPROT



//
//  Macros for calling from nooks into the kernel
//

unsigned long 
nooks_kernel_call(void * func, int argc, ...);

//#ifdef CONFIG_NOOKS_NOPROT

#if 1
#define NOOKS_KERNEL_CALL_PRE()
#define NOOKS_KERNREL_CALL_POST()
#define NOOKS_KERNEL_CALL0( _f_ ) \
    (_f_)( ) \

#define NOOKS_KERNEL_CALL1( _f_, _a1_ ) \
    (_f_)(  (_a1_))

#define NOOKS_KERNEL_CALL2( _f_, _a1_, _a2_ )  \
    (_f_)(       (_a1_), (_a2_))

#define NOOKS_KERNEL_CALL3( _f_, _a1_, _a2_, _a3_ ) \
    (_f_)(       (_a1_), (_a2_), (_a3_))

#define NOOKS_KERNEL_CALL4( _f_, _a1_, _a2_, _a3_, _a4_ ) \
    (_f_)(       (_a1_), (_a2_), (_a3_), (_a4_))

#define NOOKS_KERNEL_CALL5( _f_, _a1_, _a2_, _a3_, _a4_, _a5_ ) \
    (_f_)(       (_a1_), (_a2_), (_a3_), (_a4_), (_a5_))

#define NOOKS_KERNEL_CALL6( _f_, _a1_, _a2_, _a3_, _a4_, _a5_, _a6_ ) \
    (_f_)(       (_a1_), (_a2_), (_a3_), (_a4_), (_a5_), (_a6_))

#define NOOKS_KERNEL_CALL7( _f_, _a1_, _a2_, _a3_, _a4_, _a5_, _a6_, _a7_ ) \
    (_f_)(    (_a1_), (_a2_), (_a3_), (_a4_), (_a5_), (_a6_), (_a7_))

#define NOOKS_KERNEL_CALL8( _f_, _a1_, _a2_, _a3_, _a4_, _a5_, _a6_, _a7_, _a8_ ) \
    (_f_)( 8, \
      (_a1_), (_a2_), (_a3_), (_a4_), (_a5_), (_a6_), (_a7_), (_a8_))
#else // CONFIG_NOOKS_NOPROT
#define NOOKS_KERNEL_CALL0( _f_ ) \
    nooks_kernel_call(_f_, 0)

#define NOOKS_KERNEL_CALL1( _f_, _a1_ ) \
    nooks_kernel_call(_f_, 1, (_a1_))

#define NOOKS_KERNEL_CALL2( _f_, _a1_, _a2_ )  \
    nooks_kernel_call(_f_, 2, \
		      (_a1_), (_a2_))

#define NOOKS_KERNEL_CALL3( _f_, _a1_, _a2_, _a3_ ) \
    nooks_kernel_call(_f_, 3, \
		      (_a1_), (_a2_), (_a3_))

#define NOOKS_KERNEL_CALL4( _f_, _a1_, _a2_, _a3_, _a4_ ) \
    nooks_kernel_call(_f_, 4, \
		      (_a1_), (_a2_), (_a3_), (_a4_))

#define NOOKS_KERNEL_CALL5( _f_, _a1_, _a2_, _a3_, _a4_, _a5_ ) \
    nooks_kernel_call(_f_, 5, \
		      (_a1_), (_a2_), (_a3_), (_a4_), (_a5_))

#define NOOKS_KERNEL_CALL6( _f_, _a1_, _a2_, _a3_, _a4_, _a5_, _a6_ ) \
    nooks_kernel_call(_f_, 6, \
		      (_a1_), (_a2_), (_a3_), (_a4_), (_a5_), (_a6_))

#define NOOKS_KERNEL_CALL7( _f_, _a1_, _a2_, _a3_, _a4_, _a5_, _a6_, _a7_ ) \
    nooks_kernel_call(_f_, 7, \
	      (_a1_), (_a2_), (_a3_), (_a4_), (_a5_), (_a6_), (_a7_))

#define NOOKS_KERNEL_CALL8( _f_, _a1_, _a2_, _a3_, _a4_, _a5_, _a6_, _a7_, _a8_ ) \
    nooks_kernel_call(_f_, 8, \
      (_a1_), (_a2_), (_a3_), (_a4_), (_a5_), (_a6_), (_a7_), (_a8_))
#endif // CONFIG_NOOKS_NOPROT

static inline int NOOK_OK(pnook_t _n_) 
{
  return(((_n_)->state == nook_state_running || 
	  (IN_SHADOW_RECOVERY() && (_n_)->state == nook_state_cleanup)));
}

#ifdef CONFIG_NOOKS_NOPROT

#define NOOKS_ENTER_DRIVER() \
{ \
  pnook_t curr_nook = GET_CURRENT_NOOK(); \
  pnook_thread_state_t nook_thread_state = &(GET_NOOK_STATE().thread_state); \
  nooks_kernel_stack_frame_t frame; \
  nooks_stack_frame_t nook_frame; \
  struct_cpy(&frame.thread_state, nook_thread_state); \
  struct_cpy(&nooks_w_irq_stat[0], &irq_stat[0]); \
  if (!NOOK_OK(curr_nook)) { \
    printk("Nook not ok!\n"); \
  }  \
  memset(&nook_frame, 0, sizeof(nooks_stack_frame_t)); \
  nook_thread_state->nook_frame = &nook_frame; \
  nook_thread_state->kernel_frame = &frame; \
  NOOKS_PROCESS_LIST(curr_nook, NOOKS_WORK_FLAG_ENTRY | NOOKS_WORK_FLAG_KERNEL); \
  ENTER_NOOK(); \
  NOOKS_PROCESS_LIST(curr_nook, NOOKS_WORK_FLAG_ENTRY | NOOKS_WORK_FLAG_NOOK); \
  nooks_free_work_list(curr_nook); 

#define NOOKS_LEAVE_DRIVER() \
  if (!NOOK_OK(curr_nook)) { \
    if (signal_pending(current) && sigismember(&current->pending.signal, SIGCRASH)) { \
      struct k_sigaction *ka; \
      ka = &current->sig->action[SIGCRASH-1]; \
      ka->sa.sa_handler = SIG_IGN; \
 printk(" nook died while calling driver - should return ERESTARTSYS\n"); \
    } \
  } \
  NOOKS_PROCESS_LIST(curr_nook, NOOKS_WORK_FLAG_EXIT | NOOKS_WORK_FLAG_NOOK); \
  EXIT_NOOK(); \
  nooks_process_work_list(curr_nook, SWITCH_OUT); \
  NOOKS_PROCESS_LIST(curr_nook, NOOKS_WORK_FLAG_EXIT | NOOKS_WORK_FLAG_KERNEL); \
  struct_cpy(nook_thread_state, &frame.thread_state); \
}


#define NOOKS_KWRAPPER_ENTER() \
{ \
  pnook_t curr_nook = GET_CURRENT_NOOK(); \
  NOOKS_PROCESS_LIST(curr_nook, NOOKS_WORK_FLAG_EXIT | NOOKS_WORK_FLAG_NOOK); \
//  printk ("NOOKS_KWRAPPER_ENTER:Entered function  function %s\n", __FUNCTION__); \
  EXIT_NOOK(); \
  nooks_process_work_list(curr_nook, SWITCH_CALLOUT); \
  NOOKS_PROCESS_LIST(curr_nook, NOOKS_WORK_FLAG_EXIT | NOOKS_WORK_FLAG_KERNEL); 

#define NOOKS_KWRAPPER_EXIT() \
  if (!NOOK_OK(curr_nook)) { \
	printk("Nook faulted while in kernel!\n"); \
  } \
  NOOKS_PROCESS_LIST(GET_CURRENT_NOOK(), NOOKS_WORK_FLAG_ENTRY | NOOKS_WORK_FLAG_KERNEL); \
//  printk ("NOOKS_KWRAPPER_EXIT:Exiting function  function %s\n", __FUNCTION__); \
  ENTER_NOOK(); \
  NOOKS_PROCESS_LIST(curr_nook, NOOKS_WORK_FLAG_ENTRY | NOOKS_WORK_FLAG_NOOK);\
  nooks_free_work_list(curr_nook); \
}

#else  // CONFIG_NOOKS_NOPROT
#define NOOKS_ENTER_DRIVER()
#define NOOKS_LEAVE_DRIVER()
#define NOOKS_KWRAPPER_ENTER()  \
do { \
PDEBUG ("NOOKS_KWRAPPER_ENTER:Entered function  function %s\n", __FUNCTION__); \ 
} while (0)

#define NOOKS_KWRAPPER_EXIT() \
do { \
PDEBUG ("NOOKS_KWRAPPER_EXIT:Exiting function  function %s\n", __FUNCTION__); \ 
} while (0)
#endif // CONFIG_NOOKS_NOPROT



#endif // _LINUX_NOOKS_SWITCH_H
