#ifndef _LINUX_NOOKS_UTIL_H
#define _LINUX_NOOKS_UTIL_H

extern spinlock_t nooks_list_lock;

int
nooks_get_user_string(const char *user_name, char **buf, int * len);

void
nooks_free_user_string(char * kernel_name);

#define NOOKS_MAX_DRIVER_STRING 128
int
nooks_get_driver_string(const char *driver_name, char **buf);

void
nooks_free_driver_string(char * kernel_name);

int
nooks_memcpy(void * out, void * in, size_t len);

int
nooks_memcpy_out(void * out, void * in, size_t len);

char *
nooks_strdup(const char * kernel_name, int flags);

int
nooks_i_create_nook(char *name, pnook_t * ret_nook);

void
nooks_free_nook(pnook_t nook);

/* defined in nooks.h 
pnook_t 
nooks_find_nook(char * name);
*/

void
nooks_reference_nook(pnook_t nook);

void
nooks_dereference_nook(pnook_t nook);


pnook_driver_t
nooks_create_driver(pnook_t nook, 
		    nooks_iwrapper_type_t driver_type, 
		    int major, 
		    char * name);

void
nooks_free_driver(pnook_driver_t driver);

void
nooks_delayed_free_driver(pnook_driver_t driver);

pnook_driver_t 
nooks_find_driver_in_nook(int major, pnook_t nook);

pnook_driver_t 
nooks_find_driver_by_name_in_nook(char * name, 
				  nooks_iwrapper_type_t driver_type,
				  pnook_t nook);

pnook_driver_t 
nooks_find_driver_by_type_in_nook(nooks_iwrapper_type_t driver_type,
				  void * data,
				  pnook_t nook);

pnook_driver_t 
nooks_find_driver_by_data_in_nook(void * data, 
				  nooks_iwrapper_type_t driver_type,
				  pnook_t nook);

pnook_driver_t 
nooks_find_driver(int major);

void
nooks_validate_nook(pnook_t nook);


void
nooks_dump_nook_mm(pnook_t nook);

void 
nooks_dump_nook_drivers(pnook_t nook);


int 
nooks_copy_softnet_in(void);

void
nooks_copy_softnet_out(void);

unsigned int
nooks_log2pages(unsigned int num);


int 
nooks_list_del(struct list_head * entry);

int
nooks_list_add(struct list_head * new,
	       struct list_head * head);

#endif /* _LINUX_NOOKS_UTIL_H */
