#ifndef _LINUX_NOOKS_WORK_H
#define _LINUX_NOOKS_WORK_H

/**********************************************************
 *
 * This file declares types and macros used for deferred
 * work processing, as in when a request made by a driver
 * is deferred until it completes execution.
 *
 *********************************************************/

#define NOOKS_CURRENT_ENV() (in_irq() ? irq_e : (in_interrupt() ? softirq_e : syscall_e))


typedef int (nooks_work_handler_t) (void * data1,
				    void * data2,
				    void * data3);

typedef struct nooks_work_item {
	struct list_head next;
	void * data1;
	void * data2;
	void * data3;
	nooks_work_handler_t * handler;	
        int flags;
  int result;
} nooks_work_item_t, *pnooks_work_item_t;


typedef struct nooks_work_queue {
  struct list_head list;
  int flags;
  int count;
} nooks_work_queue_t, *pnooks_work_queue_t;

//
// Flags indicating whether the work item should be deleted after
// executing or should be executed at every opportunity
//
#define NOOKS_WORK_FLAG_ONE_SHOT 1
#define NOOKS_WORK_FLAG_RECUR 2

//
// Flags indicating whether the job should be executed in the nook
// environment or the kernel environment
//

#define NOOKS_WORK_FLAG_NOOK 4
#define NOOKS_WORK_FLAG_KERNEL 8

//
// Flags indicating whether the job should be exected at entry to a nook
// or at exit, or both
//

#define NOOKS_WORK_FLAG_ENTRY 16
#define NOOKS_WORK_FLAG_EXIT 32

//
// Flag indicating that job should only be handled on a return to the kernel
//

#define NOOKS_WORK_FLAG_RETURN 64

struct nook;

int 
nooks_process_work_list(struct nook * nook, int callout);

pnooks_work_item_t 
nooks_add_work_item(struct nook * nook, 
	   	    void * data1, 
	   	    void * data2, 
	   	    void * data3, 
	            nooks_work_handler_t * handler);

pnooks_work_item_t 
nooks_add_kernel_work_item(struct nook * nook, 
			   void * data1, 
			   void * data2, 
			   void * data3, 
			   nooks_work_handler_t * handler,
			   int mem_flags);

int 
nooks_process_entry_exit_list(struct nook * nook, int flags);

#define NOOKS_PROCESS_LIST(_nook_, _flags_) { \
  if ((((_nook_)->resources.kernel_work_list.flags & (_flags_)) == (_flags_)) && \
      ((_nook_)->resources.kernel_work_list.count >= 0)) { \
	nooks_process_entry_exit_list((_nook_), (_flags_)); \
      } \
}


void
nooks_free_work_list(struct nook * nook);

void 
nooks_delete_kernel_work_item(struct nook * nook, pnooks_work_item_t item);

void
nooks_delete_work_list(struct nook * nook);

#endif // NOOKS_WORK_H






