/*
 * nooks-workers.c -- Routines for executing deferred work
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  The citation
 * should list that the code comes from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.   No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

#if 0 

#include "nooks-i.h"
#include "nooks-iwrappers.h"
#include "nooks-wrappers.h"
#include <net/sock.h>
#include <linux/rtnetlink.h> // rtnl_lock()
#include "nooks-recover.h"

int nooks_handler_kfree(void * data1, void * data2, void * data3) 
{
  if (data1 != NULL) {
    nooks_w_kfree(data1);
  }
  return(0);
}

int nooks_handler_add_timer(void * data1, void * data2, void * data3) 
{
	pnook_t nook = GET_CURRENT_NOOK();
	int result = 0;
	struct timer_list * nook_timer = (struct timer_list *) data1;
	struct timer_list * timer = NULL;
	pnook_driver_t driver = NULL;

	NOOKS_ENTER();

	timer = nooks_kmalloc(sizeof(struct timer_list), NOOKS_MEM_FLAGS);
	if (timer == NULL) {
		result = -ENOMEM;
		goto out;
	}

	//
	// Fill in the fields
	//

	result = nooks_memcpy(timer, nook_timer, sizeof(struct timer_list));
	if (result < 0) {
		PERROR("Failed to copy timer from nook\n");
		goto err1;
	}

	driver = nooks_create_driver(nook, timerops_e, NOOK_DRIVER_NO_MAJOR, NULL);
	if (driver == NULL) {
		result = -ENOMEM;
		goto err1;
	}

	// Copy the operations, preserving NULL entries
	//

	nooks_iwrapper_copy_timerops(timer->function,
			&driver->operations.timerops,
			driver->iwrapper_operations.timerops,
			&driver->wrapper_operations.timerops);

	driver->data = timer;

	// reset the pointers
	//
	init_timer(timer);

	// BUGBUG: what about the function pointer?
	//
	timer->function = driver->wrapper_operations.timerops.handler;

	// add the mapping
	//
	result = nooks_add_resource_elem(nook, timer_e, nook_timer, driver);
	if (result < 0) {
		goto err2;
	}

	shadow_add_stimerops(driver, timer);
	add_timer((struct timer_list *) timer);
	
	NOOKS_EXIT();
	return(0);





err2:
	nooks_free_driver(driver);
err1:
	nooks_kfree(timer);
out:
	NOOKS_EXIT();
	return(0);
}

int nooks_handler_del_timer(void * data1, void * data2, void * data3) 
{
  int result;
  pnook_t nook;
  pnook_driver_t driver = NULL;
  NOOKS_ENTER();
  nook = GET_CURRENT_NOOK();

  result = nooks_del_resource_elem(nook, timer_e, data1, (void **) &driver);
  if (result >= 0) {
    del_timer((struct timer_list *) driver->data);
    nooks_kfree(driver->data);
    nooks_free_driver(driver);
  } else {
    PERROR("Failed to find timer %p for nook %p for del_timer\n",
	   data1, nook);
  }
  NOOKS_EXIT();
  return(0);
}

int nooks_handler_mod_timer(void * data1, void * data2, void * data3) 
{
  int result;
  pnook_t nook;
  pnook_driver_t driver = NULL;
  NOOKS_ENTER();
  nook = GET_CURRENT_NOOK();

  result = nooks_find_resource_elem(nook, timer_e, data1, (void **) &driver);
  if (result >= 0) {
    mod_timer((struct timer_list *) driver->data, (unsigned long) data2);

  } else {
    PERROR("Failed to find timer %p for nook %p for mod_timer\n",
	   data1, nook);
  }
  NOOKS_EXIT();
  return(0);
}

int nooks_handler_netif_rx(void * data1, void * data2, void * data3) 
{
	int result;
	pnook_t nook;
	pnook_driver_t driver = NULL;
	struct net_device * kernel_dev = NULL;
	struct net_device * nook_dev = NULL;
	struct sk_buff_head * queue;
	struct sk_buff * skb;

	NOOKS_ENTER();

	nook = GET_CURRENT_NOOK();
	queue = &(nook->locals->skb_rx_list);

	result = nooks_copy_softnet_in();
	if (result < 0) {
		PERROR("Failed to copy softnet data to kernel\n");
		goto Cleanup;
	}

	while ((skb = skb_dequeue(queue)) != NULL) {

		result = nooks_mm_move_skb_to_kernel(nook, skb);
		if (result < 0) {
			PERROR("Unable to move skb to kernel: %p, %d\n", skb, result);
			// BUG();
			goto Cleanup;
		}

		if (nook_dev != skb->dev) {
			// BUGBUG: illegal dereference of skb here.
			//
			result = nooks_find_resource_elem(nook, netdev_nook_e, skb->dev, (void **) &driver);
			if (result < 0) {
				NOOKS_FAULT(resource);
				// BUG();
				PERROR("Failed to find netdevice on softnet queue\n");
				goto Cleanup;
			}

			kernel_dev = (struct net_device *) driver->data;
			nook_dev = skb->dev;

			// BUGBUG: illegal dereference of skb here
			//
			nooks_memcpy(&kernel_dev->state, &nook_dev->state, sizeof(unsigned long));

		}

		skb->dev = kernel_dev;

		// Give up on this skb
		//
		NOOKS_STATS_INC(NOOKS_OPS_KERNEL_NETIF_RX);
		result = netif_rx(skb);
	}

Cleanup:
	nooks_copy_softnet_out();
	NOOKS_EXIT();
	return(result);
}

int nooks_handler_enter_netdev(void * data1, void * data2, void * data3) 
{
  //  pnook_t nook = (pnook_t) data1;
  struct net_device * kernel_dev = (struct net_device *) data2;
  struct net_device * nook_dev = (struct net_device *) data3;
#ifdef NOOKS_DEBUG2
  static int diffs = 0;
  if (nook_dev->state != kernel_dev->state) {
    diffs++;
  }
#endif
  

  nooks_memcpy_out(&nook_dev->state, &kernel_dev->state, sizeof(unsigned long));
  return 0;
}

int nooks_handler_exit_netdev(void * data1, void * data2, void * data3) 
{
  //  pnook_t nook = (pnook_t) data1;
  struct net_device * kernel_dev = (struct net_device *) data2;
  struct net_device * nook_dev = (struct net_device *) data3;
#ifdef NOOKS_DEBUG2
  static int diffs = 0;

  if (nook_dev->state != kernel_dev->state) {
    diffs++;
  }
#endif
  nooks_memcpy(&kernel_dev->state, &nook_dev->state, sizeof(unsigned long));
  nooks_copy_softnet_in();
  return 0;
}

/**********************************************************
 *
 * nooks_handler_proc_entry
 *
 * Do the work of copying function pointers from
 * the driver into the kernel
 *
 *********************************************************/

int nooks_handler_proc_entry(void * data1, void * data2, void * data3) 
{
	int result = 0;
	pnook_t nook = (pnook_t) data1;
	struct proc_dir_entry * kern_entry = (struct proc_dir_entry *) data2;
	struct proc_dir_entry * nook_entry = (struct proc_dir_entry *) data3;
	struct proc_dir_entry temp_entry;
	struct proc_dir_entry * temp_kern_entry = NULL;
	pnook_driver_t driver = NULL;
	char * kern_name = NULL;

	result = nooks_memcpy(&temp_entry, nook_entry, sizeof(struct proc_dir_entry));
	if (result < 0) {
		goto out;
	}

	// Make sure the entry is still valid
	//
	result = nooks_find_resource_elem(nook, proc_entry_e, nook_entry, (void **) &temp_kern_entry);
	if (result < 0) {
		PERROR("Failed to find resource for proc_entry handler: %p, %p\n",
				nook, nook_entry);
		NOOKS_FAULT(resource);
		goto out;
	}

	if (temp_kern_entry != kern_entry) {
		PERROR("Stored kernel entry and supplied kernel entry don't match: %p vs %p (%p)\n",
				kern_entry, temp_kern_entry, nook);
		result = -ENOENT;
		goto out;
	}

	// Allocate a driver for the functions
	//
	kern_name = nooks_strdup(kern_entry->name, GFP_KERNEL);
	if (kern_name == NULL) {
		result = -ENOMEM;
		goto out;
	}

	driver = nooks_create_driver(nook, procops_e, NOOK_DRIVER_NO_MAJOR, kern_name);
	if (driver == NULL) {
		PERROR(" Failed to create driver %s for create_proc_entry\n", kern_entry->name);
		result = -ENOMEM;
		goto err1;
	}

	// Set the name to NULL so we the driver can free it.
	//
	kern_name = NULL;

	driver->data = kern_entry;

	// Copy the operations, preserving NULL entries
	//
	nooks_iwrapper_copy_procops(&temp_entry,
			&driver->operations.procops,
			driver->iwrapper_operations.procops,
			kern_entry);

	kern_entry->data = temp_entry.data;
	shadow_add_sprocops(driver, kern_entry);

	return result;


/*  not used
err2:
	nooks_free_driver(driver);
*/
err1:
	nooks_kfree(kern_name);
out:
	return result;
}


int nooks_handler_tasklet_schedule(void * data1, void * data2, void * data3)
{
	struct tasklet_struct * t = (struct tasklet_struct *) data1;
	struct tasklet_struct nt;
	int lo_hi = (int) data2;
	pnooks_tasklet_t kt = NULL;
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	int result = 0;

	NOOKS_ENTER();

	// Copy the state of the tasklet
	//
	result = nooks_memcpy(&nt, t, sizeof(struct tasklet_struct));
	if (result < 0) {
		goto out;
	}

	result = nooks_find_resource_elem(nook, tasklet_e, t, (void **)  &kt);
	if (result < 0) {

		kt = nooks_kmalloc(sizeof(nooks_tasklet_t), NOOKS_MEM_FLAGS) ;
		if (kt == NULL) {
			result = -ENOMEM;
			goto out;
		}

		driver = nooks_create_driver(nook, callbackops_e, NOOK_DRIVER_NO_MAJOR, NULL);
		if (driver == NULL) {
			PERROR(" Failed to create driver for tasklet_schedule\n");
			nooks_kfree(kt);
			result = -ENOMEM;
			goto err1;
		}

		result = nooks_add_resource_elem(nook, tasklet_e, kt, driver);
		if (result < 0) {
			goto err2;
		}

		driver->data = NOOKS_CALLBACK_TASKLET;

		// Copy the operations, preserving NULL entries
		//
		nooks_iwrapper_copy_callbackops((callback_handler_t *) nt.func,
				&driver->operations.callbackops,
				driver->iwrapper_operations.callbackops,
				&driver->wrapper_operations.callbackops);

		// We store a pointer to the tasklet itself in the data field, and the
		// tasklet has an extra field for the real data
		//
		tasklet_init(&kt->t, (tasklet_handler_t *) 
				driver->wrapper_operations.callbackops.handler, 
				(unsigned long) kt);
		kt->data = nt.data;
		//
		// Store a pointer to the nook tasklet here
		//
		kt->nt = t;
		nooks_add_resource_elem(nook, tasklet_e, t, kt);
		shadow_add_scallbackops(driver, kt); // BUGUBG why kt
	}

	//	clear_bit(TASKLET_STATE_SCHED, &(t)->state);

	if (lo_hi) {
	  tasklet_hi_schedule(&kt->t);
	} else {
	  tasklet_schedule(&kt->t);
	}

	NOOKS_EXIT();
	return 0;





err2:
	nooks_del_resource_elem(nook, tasklet_e, kt, NULL);	
	nooks_free_driver(driver);
err1:
	nooks_kfree(kt);
out:
	NOOKS_EXIT();
	return 0;
}


int nooks_handler_tasklet_enable(void * data1, void * data2, void * data3) 
{
  int result;
  pnook_t nook;
  struct tasklet_struct * kt = NULL;
  NOOKS_ENTER();
  nook = GET_CURRENT_NOOK();

  result = nooks_find_resource_elem(nook, tasklet_e, data1, (void **) &kt);
  if (result >= 0) {
    tasklet_enable(kt);

  } else {
    PERROR("Failed to find tasklet %p for nook %p for tasklet_enable\n",
	   data1, nook);
  }
  NOOKS_EXIT();
  return(0);
}

int nooks_handler_tasklet_disable(void * data1, void * data2, void * data3) 
{
  int result;
  pnook_t nook;
  struct tasklet_struct * kt = NULL;
  NOOKS_ENTER();
  nook = GET_CURRENT_NOOK();

  result = nooks_find_resource_elem(nook, tasklet_e, data1, (void **) &kt);
  if (result >= 0) {
    tasklet_enable(kt);

  } else {
    PERROR("Failed to find tasklet %p for nook %p for tasklet_enable\n",
	   data1, nook);
  }
  NOOKS_EXIT();
  return(0);
}



int nooks_handler_dev_kfree_skb_irq(void * data1, void * data2, void * data3) 
{
  int result = 0;
  pnook_t nook = GET_CURRENT_NOOK();
  struct sk_buff * skb;
  struct list_head * queue;
  struct list_head * elem;
  NOOKS_ENTER();
  queue = &(nook->locals->skb_free_list);
  
  while ((elem = list_pop(queue)) != NULL) {
    skb = (struct sk_buff *) list_entry(elem, nooks_list_elem_t, next)->data;
    nooks_mm_lookaside_free(&nook->locals->elem_pool, elem);
    
    result = nooks_mm_move_skb_to_kernel(nook, skb);
    if (result >= 0) {
      //
      // BUGBUG: Only free it if it is not on a list. 
      // Better to leak than to crash...
      //
      if (skb->list == NULL) {
	dev_kfree_skb_irq(skb);
      } else {
	PERROR("Couldn't free skb %p for nook %p - still on list\n",
	       skb, nook);
      }
      
    } else {
      PERROR("Failed to find skb %p for nook %p for dev_kfree_skb_irq\n",
	     skb, nook);
    }
  }
  NOOKS_EXIT();
  return(0);
}



int nooks_handler_init_etherdev(void * data1, void * data2, void * data3)
{
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	int result = 0;
	char * kernel_name = NULL;
	struct net_device * kernel_dev = NULL;
	struct net_device * nook_dev = (struct net_device *) data1;
	struct net_device dev;
	pnooks_work_item_t enter_work_item = NULL;
	pnooks_work_item_t exit_work_item = NULL;
	int shadow_driver = 0;

	result = nooks_find_resource_elem_reverse(nook, 
						  netdev_e, 
						  nook_dev, 
						  (void **) &kernel_dev);
	if (result < 0) {
	  PERROR("Failed to find netdev for %p\n",dev);
	  goto out;
	}
	if (kernel_dev != (struct net_device *) data2) {
	  PERROR("Kernel dev mismatch: %p vs %p\n",
		 kernel_dev, data1);
	  goto out;
	}
	result = nooks_memcpy(&dev, nook_dev, sizeof(struct net_device));
	if (result < 0) {
	  PERROR(" failed to copy netdevice from %p\n",dev);
	  goto out;
	}

	driver = shadow_find_driver(nook, netops_e, NULL); // BUGBUG: null key
	if (driver != NULL) {
	  shadow_driver = 1;
	  kernel_dev = (struct net_device *) driver->data;
	} else {
	
	  
	  result = nooks_get_driver_string(kernel_dev->name, &kernel_name);
	  if (result < 0) {
	    goto out;
	  }
	  PDEBUG(" Registering netdevice %s, %s\n", kernel_name, kernel_dev->name);
	  
	  driver = nooks_create_driver(nook, netops_e, NOOK_DRIVER_NO_MAJOR, kernel_name);
	  if (driver == NULL) {
	    PERROR(" Failed to create driver for register netdev\n");
	    result = -ENOMEM;
	    goto err1;
	  }
	  
	}
	// Delete this so only the driver has a reference
	//
	//  commented cause of err2 then err1
	// kernel_name = NULL;
	
	// Copy the operations, preserving NULL entries
	//
	PDEBUG("nooksdev->init = %p\n", dev->init);

	
	//
	// SHADOW: register with shadow
	//
	
	if (!shadow_driver) {
	  shadow_add_snetops(driver, kernel_dev);
	}
	
	nooks_iwrapper_copy_netops(&dev, &driver->operations.netops, driver->iwrapper_operations.netops, kernel_dev); 
	
	driver->data =  kernel_dev; 
	result = nooks_add_resource_elem(nook, netdev_nook_e, nook_dev, driver);
	if (result < 0) {
	  goto err2;
	}
	
	enter_work_item = nooks_add_kernel_work_item(nook, nook, kernel_dev,
						     nook_dev, nooks_handler_enter_netdev,
						     NOOKS_WORK_FLAG_RECUR | NOOKS_WORK_FLAG_ENTRY |
						     NOOKS_WORK_FLAG_KERNEL);
	
	exit_work_item = nooks_add_kernel_work_item(nook, nook, kernel_dev, nook_dev,
						    nooks_handler_exit_netdev,
						    NOOKS_WORK_FLAG_RECUR | NOOKS_WORK_FLAG_EXIT |
						    NOOKS_WORK_FLAG_KERNEL);
	
	nooks_add_resource_elem(nook, worker_e, driver, enter_work_item);
	nooks_add_resource_elem(nook, worker_e, driver, exit_work_item);


out:
	return result;




err2:
	nooks_free_driver(driver);
err1:
	nooks_free_driver_string(kernel_name);
	nooks_kfree(kernel_dev);
	return result;
}

int nooks_handler_enter_listen_sock(void * data1, void * data2, void * data3) 
{
  struct sock * sock = (struct sock *) data1;
  struct sock * nook_sock = (struct sock *) data2;
  
  memcpy(&nook_sock->tp_pinfo.af_tcp, &sock->tp_pinfo.af_tcp, sizeof(struct tcp_opt));
  nook_sock->state = sock->state;

  // BUGBUG: is this right?
  //
  nook_sock->receive_queue = sock->receive_queue;
  return(0);
}

int nooks_handler_add_wait(void * data1, void * data2, void * data3) 
{
  pnook_t nook = GET_CURRENT_NOOK();
  wait_queue_head_t * nook_head = (wait_queue_head_t *) data1;
  wait_queue_t * nook_wait = (wait_queue_t *) data2;
  wait_queue_head_t * kernel_head = NULL;
  wait_queue_t * kernel_wait = NULL;

  wait_queue_head_t * stored_kernel_head = NULL;
  if (nooks_find_resource_elem(nook,
			       waitqueuehead_e,
			       nook_head,
			       (void **) &stored_kernel_head) < 0) {
    goto Cleanup;
  }
  if (stored_kernel_head != nook_head->u.kernel_q) {
    goto Cleanup;
  }


  kernel_wait = (wait_queue_t *) nooks_kmalloc(sizeof(wait_queue_t), 
					       GFP_KERNEL);
  if (kernel_wait == NULL) {
    goto Cleanup;
  }

  // BUGBUG: unsafe at the extreme

  kernel_head = nook_head->u.kernel_q;

  init_waitqueue_entry(kernel_wait, get_current());
  nooks_add_resource_elem(nook, waitqueue_sync_e, nook_wait, kernel_wait);
  
  if ((nook_wait->flags & WQ_FLAG_EXCLUSIVE) == 0) {
    add_wait_queue(kernel_head, kernel_wait);
  } else {
    add_wait_queue_exclusive(kernel_head, kernel_wait);
  }

 Cleanup:
  return(0);
}

int nooks_handler_remove_wait(void * data1, void * data2, void * data3) 
{
  pnook_t nook = GET_CURRENT_NOOK();
  wait_queue_head_t * nook_head = (wait_queue_head_t *) data1;
  wait_queue_t * nook_wait = (wait_queue_t *) data2;
  wait_queue_head_t * kernel_head = NULL;
  wait_queue_t * kernel_wait = NULL;
  
  wait_queue_head_t * stored_kernel_head = NULL;
  if (nooks_find_resource_elem(nook, waitqueuehead_e, nook_head,
			       (void **) &stored_kernel_head) < 0) {
	  NOOKS_FAULT(resource);
    goto Cleanup;
  }
  if (stored_kernel_head != nook_head->u.kernel_q) {
    goto Cleanup;
  }

  if (nooks_del_resource_elem(nook, waitqueue_sync_e, nook_wait, 
			      (void **) &kernel_wait) < 0) {
    goto Cleanup;
  }

  // BUGBUG: unsafe at the extreme

  if (nooks_memcpy(&kernel_head, &nook_head->u.kernel_q, sizeof(wait_queue_head_t *)) < 0) {
    goto Cleanup;
  }

  remove_wait_queue(kernel_head, kernel_wait);
  nooks_kfree(kernel_wait);

 Cleanup:
  return(0);
}


int nooks_handler_pm_register(void * data1, void * data2, void * data3) 
{
  struct pm_dev * dev = NULL;
  struct pm_dev * nook_dev = data1;
  NOOKS_ENTER();
  
  if (nooks_find_resource_elem(GET_CURRENT_NOOK(),
			       pm_dev_e,
			       nook_dev,
			       (void **) &dev) >= 0) {
    nooks_memcpy(&dev->data, &nook_dev->data, sizeof(void *));
  }
 
  NOOKS_EXIT();
  return(0);
}

#endif
