/*
 * nooks-wrappers.c -- Wrappers for system functions escpecially
 * for the e1000 driver. This to implement direct device access
 * VM migration support.
 * 2008 - Asim Kadav
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.
 * No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

#include "nooks-i.h"
#include "nooks-wrappers.h"
#include "nooks-iwrappers.h"
#include "nooks-workers.h"
#include "nooks-recover.h"
#include <../shadow/netshadow.h>
#include <net/tcp.h>
#include <linux/ctype.h>
#include <linux/rtnetlink.h>	// rtnl_lock()
#include <linux/ethtool.h>
#include <xen/interface/xen.h>

struct tasklet_struct nooks_w_bh_task_vec[32];
struct mm_struct nooks_w_init_mm;
struct resource nooks_w_iomem_resource;
struct softnet_data *nooks_w_softnet_data;
struct resource nooks_w_ioport_resource;
struct list_head nooks_w_pci_devices;
struct page *nooks_w_mem_map;
unsigned long nooks_last_wrapper_local;
unsigned long volatile nooks_w_jiffies;
shared_info_t *nooks_w_HYPERVISOR_shared_info;
struct pglist_data nooks_w_contig_page_data;
unsigned long nooks_w___per_cpu_offset[NR_CPUS];

/* Initialize the wrappers.
 */

int nooks_wrappers_init(void)
{
	int cpu_id;
	nooks_w_init_mm = init_mm;
	nooks_w_iomem_resource = iomem_resource;
	nooks_w_ioport_resource = ioport_resource;
	nooks_w_mem_map = mem_map;
#ifdef CONFIG_NOOKS_STATS
	memcpy(nooks_w_ms_ops_count, ms_ops_count,
	       MS_OPS_NUM * sizeof(unsigned long));
#endif				// CONFIG_NOOKS_STATS
	cpu_id = smp_processor_id();
	nooks_w_softnet_data = &per_cpu(softnet_data, cpu_id);
	for (cpu_id = 0; cpuid < NR_CPUS; cpu_id++)
		nooks_w___per_cpu_offset[cpu_id] = __per_cpu_offset[cpu_id];

	nooks_w_pci_devices = pci_devices;

	nooks_w_HYPERVISOR_shared_info = (shared_info_t *) empty_zero_page;
	return (0);

}

/* This is for hot-plugging of network devices. */
void NOOKS_KWRAPPER(netif_device_attach) (struct net_device * dev) {
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	struct net_device *kernel_dev = NULL;
	int result = 0;
	NOOKS_KWRAPPER_ENTER();

	result =
	    nooks_find_resource_elem(nook, netdev_nook_e, dev,
				     (void **)&driver);

	if (result < 0) {
		PERROR("Failed to find netdev for netif_device_attach.\n", dev);
		goto Cleanup;
	}

	kernel_dev = (struct net_device *)driver->data;
	NOOKS_SWRAPPER(netif_device_attach) (kernel_dev);

      Cleanup:
	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(netif_device_attach) (struct net_device * dev) {
	NOOKS_WRAPPER_ENTER();

	NOOKS_KERNEL_CALL1(nooks_k_netif_device_attach, dev);
	NOOKS_WRAPPER_EXIT();
	return;
}

/*FIXME */
struct sk_buff *NOOKS_WRAPPER(__netdev_alloc_skb) (struct net_device * dev,
						   unsigned int length,
						   gfp_t gfp_mask) {
	struct sk_buff *skb;

	NOOKS_WRAPPER_ENTER();
	skb = nooks_w___alloc_skb(length + NET_SKB_PAD, gfp_mask, 0);
	if (skb == NULL)
		PDEBUG("Error allocating skb in __netdev_alloc_skb.");
	goto Cleanup;

	if (likely(skb)) {
		skb_reserve(skb, NET_SKB_PAD);
		skb->dev = dev;
	}

      Cleanup:
	NOOKS_WRAPPER_EXIT();
	return (skb);
}

/* This can't  be caught by the kfree semantics. 
void NOOKS_KWRAPPER (free_netdev) (struct net_device * dev)
{
  NOOKS_KWRAPPER_ENTER ();
  NOOKS_SWRAPPER (free_netdev) (dev);
  NOOKS_KWRAPPER_EXIT ();
  return;
}
*/

void nooks_w_free_netdev(struct net_device *dev)
{
#ifdef CONFIG_SYSFS
	/*  Compatibility with error handling in drivers */
	if (dev->reg_state == NETREG_UNINITIALIZED) {
		nooks_w_kfree((char *)dev - dev->padded);
		return;
	}

	BUG_ON(dev->reg_state != NETREG_UNREGISTERED);
	dev->reg_state = NETREG_RELEASED;

	/* will free via class release BigTBD FIXME 
	   if (&dev->class_dev != NULL)
	   if (&dev->class_dev->kobj) {
	   struct kobj = dev->class_dev->kobj;

	   if (atomic_dec_and_test(&(kobj->kref)->refcount))
	   kobject_release(kobj->kref);

	   if (dev ! = NULL)
	   nooks_w_kfree((char *)dev - dev->padded);
	   } */
	class_device_put(&dev->class_dev);
#else
	/* We malloced the dev using nooks_w_kmalloc. */
	nooks_w_kfree((char *)dev - dev->padded);
#endif
}

void NOOKS_KWRAPPER(netif_device_detach) (struct net_device * dev) {
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	struct net_device *kernel_dev = NULL;
	int result = 0;
	NOOKS_KWRAPPER_ENTER();

	result =
	    nooks_find_resource_elem(nook, netdev_nook_e, dev,
				     (void **)&driver);

	if (result < 0) {
		PERROR("Failed to find netdev for netif_device_detach: %p\n",
		       dev);
		goto Cleanup;
	}

	kernel_dev = (struct net_device *)driver->data;
	NOOKS_SWRAPPER(netif_device_detach) (kernel_dev);

      Cleanup:
	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(netif_device_detach) (struct net_device * dev) {
	NOOKS_WRAPPER_ENTER();

	NOOKS_KERNEL_CALL1(nooks_k_netif_device_detach, dev);
	NOOKS_WRAPPER_EXIT();
	return;
}

/* BigTBD  */
dma_addr_t NOOKS_KWRAPPER(dma_map_page) (struct device * dev,
					 struct page * page,
					 unsigned long offset, size_t size,
					 enum dma_data_direction direction) {
	pnook_t nook = GET_CURRENT_NOOK();
	dma_addr_t result;
	NOOKS_KWRAPPER_ENTER();

	result =
	    NOOKS_SWRAPPER(dma_map_page) (dev, page, offset, size, direction);

	if (nooks_add_to_hash
	    (&nook->resources.area_map, dmamap_e, (void *)result,
	     (void *)page) < 0) {
		PDEBUG
		    ("Failed to track dma_map pointer by the object tracker in dma_map_page.\n");
	}

	NOOKS_KWRAPPER_EXIT();
	return (result);
}

dma_addr_t NOOKS_WRAPPER(dma_map_page) (struct device * dev, struct page * page,
					unsigned long offset, size_t size,
					enum dma_data_direction direction) {
	dma_addr_t result = NULL;
	result =
	    NOOKS_KERNEL_CALL5(nooks_k_dma_map_page, dev, page, offset, size,
			       direction);
	return (result);
}

/* This needs to be similar to pci_alloc consistent in 2.4. There have 
 * been changes in the kernel since then (additional search for pages
 * from a bitmap. Also, there are additional xen modifications done 
 * over 2.6 that needs to be taken care of.
 * 
 */
void *NOOKS_KWRAPPER(dma_alloc_coherent) (struct device * dev, size_t size,
					  dma_addr_t * dma_handle, gfp_t flag) {
	void *result, *ret;
	u64 mask;
	size_t order;
	unsigned long vstart = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	struct dma_coherent_mem *mem = dev ? dev->dma_mem : NULL;
	NOOKS_KWRAPPER_ENTER();

	order = get_order(size);

	if (mem) ;
	/* Get continuous aligned free memory. How to do this FIXME MIKS  */

	ret = (void *)nooks_mm_get_free_pages(nook, flag, order);

	if (ret != NULL) {
		if (nooks_add_resource_elem
		    (nook, pci_memory_e, ret, (void *)size) < 0) {
			nooks_mm_free_pages(nook, (unsigned long)ret, order);
			ret = NULL;
		} else
			goto Cleanup;
		vstart = (unsigned long)(void *)ret;	//FIXME
	}

	if (dev != NULL && dev->coherent_dma_mask)
		mask = dev->coherent_dma_mask;
	else
		mask = 0xffffffff;

	if (ret != NULL) {
		/* Ensures multi-page extents are contiguous in machine memory. */
		if (xen_create_contiguous_region(vstart, order, fls64(mask)) !=
		    0) {
			nooks_mm_free_pages(nook, (unsigned long)ret, order);
			result = NULL;
		}
	}
      Cleanup:
	NOOKS_KWRAPPER_EXIT();
	return (ret);
}

void *NOOKS_WRAPPER(dma_alloc_coherent) (struct device * hwdev,
					 size_t size,
					 dma_addr_t * dma_handle, gfp_t gfp) {
	void *ret;
	gfp &= ~(__GFP_DMA | __GFP_HIGHMEM);
	if (hwdev == NULL || *(hwdev->dma_mask) != 0xffffffff) {
		gfp |= GFP_DMA;
	}
	ret =
	    (void *)NOOKS_KERNEL_CALL4(nooks_k_dma_alloc_coherent,
				       hwdev, size, dma_handle, gfp);

	if (ret != NULL) {
		memset(ret, 0, size);
		*dma_handle = virt_to_bus(ret);
	}

	NOOKS_WRAPPER_EXIT();
	return (ret);
}

/*TBD  -- We don't do anything for now, later we may need to do 
 * fix this if irqs cannt be restored safely.
 */
void NOOKS_KWRAPPER(_spin_unlock_irqrestore) (spinlock_t * lock,
					      unsigned long flags) {
	NOOKS_KWRAPPER_ENTER();
	NOOKS_SWRAPPER(_spin_unlock_irqrestore) (lock, flags);
	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(_spin_unlock_irqrestore) (spinlock_t * lock,
					     unsigned long flags) {
	NOOKS_KERNEL_CALL2(nooks_k__spin_unlock_irqrestore, lock, flags);
	return;
}

/* Implemented as is from the 2.4 Nooks version. */
void NOOKS_KWRAPPER(pci_release_regions) (struct pci_dev * dev) {
	int result = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	struct pci_dev *kernel_dev = NULL;
	NOOKS_KWRAPPER_ENTER();

	result =
	    nooks_find_resource_elem(nook, pcidev_e, dev, (void **)&kernel_dev);

	/* Clear associated pci_region. */
	if (result >= 0) {
		result =
		    nooks_del_resource_elem(nook, pciregion_e,
					    kernel_dev, NULL);
		if (result >= 0)
			NOOKS_SWRAPPER(pci_release_regions)
			    (kernel_dev);
	}
	NOOKS_KWRAPPER_EXIT();
}

void NOOKS_WRAPPER(pci_release_regions) (struct pci_dev * dev) {
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL1(nooks_k_pci_release_regions, dev);
	NOOKS_WRAPPER_EXIT();
}

void NOOKS_WRAPPER(dump_stack) (void) {
	dump_stack();
}

/* No wrappers entry/exit calls here - probably to avoid upsetting the
 * purpose of the function.
*/

void NOOKS_WRAPPER(__const_udelay) (unsigned long usecs) {
	__const_udelay(usecs);
	return;
}

void NOOKS_KWRAPPER(pci_unregister_driver) (struct pci_driver * pci_driver) {
	int result = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	struct pci_driver *kernel_drv = NULL;

	NOOKS_KWRAPPER_ENTER();

	result =
	    nooks_del_resource_elem(nook, pci_driver_e, pci_driver,
				    (void **)&driver);

	if (result < 0) {
		PERROR(" Failed to find pci driver for cleanup: %p\n",
		       pci_driver);
		goto Cleanup;
	}
	kernel_drv = (struct pci_driver *)driver->data;

	NOOKS_SWRAPPER(pci_unregister_driver) (kernel_drv);

	//Freeing the driver will release any interfaces it is using.

	nooks_free_driver(driver);

	if (kernel_drv->id_table != NULL) {
		nooks_kfree(kernel_drv->id_table);
	}

	nooks_kfree(kernel_drv);

      Cleanup:
	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(pci_unregister_driver) (struct pci_driver * arg0) {
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL1(nooks_k_pci_unregister_driver, arg0);
	NOOKS_WRAPPER_EXIT();
	return;
}

/* The network interface drivers use the netif_carrier_on() and 
 * netif_carrier_off() functions to update their interface card 
 * status.
 */

void NOOKS_KWRAPPER(netif_carrier_off) (struct net_device * dev) {
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	struct net_device *kernel_dev = NULL;
	int result = 0;

	NOOKS_KWRAPPER_ENTER();

	result =
	    nooks_find_resource_elem(nook, netdev_nook_e, dev,
				     (void **)&driver);

	if (result < 0) {
		PERROR
		    (" Failed to find netdev for netif_carrier_off: %p\n", dev);
		goto Cleanup;
	}
	kernel_dev = (struct net_device *)driver->data;
	NOOKS_SWRAPPER(netif_carrier_off) (kernel_dev);
      Cleanup:
	NOOKS_KWRAPPER_EXIT();
}

void NOOKS_WRAPPER(netif_carrier_off) (struct net_device * dev) {
	NOOKS_KERNEL_CALL1(nooks_k_netif_carrier_off, dev);
	return;
}

/* Are there any changes required here? TBD  - relook. */
void NOOKS_KWRAPPER(kunmap_atomic) (void *kvaddr, enum km_type type) {
	unsigned long pgP;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_KWRAPPER_ENTER();
	NOOKS_SWRAPPER(kunmap_atomic) (kvaddr, type);
	if (nooks_del_from_hash
	    (&nook->resources.area_map, kmap_e, (void *)kvaddr,
	     (void **)&pgP) < 0)
		PDEBUG("Object tracking failed in kunmap_atomic.\n");

	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(kunmap_atomic) (void *kvaddr, enum km_type type) {
	NOOKS_KERNEL_CALL2(nooks_k_kunmap_atomic, kvaddr, type);
	return;
}

void *nooks_w_kmalloc(size_t size, int flags)
{
	void *result;
	printk("In nooks_w_kmalloc.");
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_WRAPPER_ENTER();

	//result = nooks_mm_kmalloc(nook, size, flags);
	result = kmalloc(size, flags);
	if (result == NULL)
		printk ("FATAL:kmalloc failed for size %d.\n", size);
	NOOKS_WRAPPER_EXIT();

	return (result);
}

/* This is similar to the Nooks 2.4 version. We write wrapper for alloc_netdev
 * the function which is called by alloc_etherdev.
 */

/*Forward Declerations.*/
void nooks_w_ether_setup(struct net_device *dev);
struct net_device *NOOKS_WRAPPER(alloc_netdev) (int sizeof_priv,
						const char *name,
						void (*setup) (struct net_device
							       *));

struct net_device *NOOKS_WRAPPER(alloc_etherdev) (int sizeof_priv) {

	return nooks_w_alloc_netdev(sizeof_priv, "eth%d", ether_setup);
}

struct net_device *NOOKS_WRAPPER(alloc_netdev) (int sizeof_priv,
						const char *name,
						void (*setup) (struct net_device
							       *)) {

	//return alloc_netdev(sizeof_priv, name, setup);

	printk ("In alloc_netdev.\n");

	struct net_device *dev = NULL;
	void *p;
	int alloc_size;

	NOOKS_WRAPPER_ENTER();

	/* Ensure 32-byte alignment of both the device and private area. */
	alloc_size = (sizeof(*dev) + NETDEV_ALIGN_CONST) & ~NETDEV_ALIGN_CONST;
	alloc_size += sizeof_priv + NETDEV_ALIGN_CONST;

	p = nooks_w_kmalloc(alloc_size, GFP_KERNEL);
	//p = kzalloc(alloc_size, GFP_KERNEL);
	if (!p) {
		PERROR(KERN_ERR "alloc_dev: Unable to allocate device.\n");
		goto Cleanup;
	}

	dev = (struct net_device *)
	    (((long)p + NETDEV_ALIGN_CONST) & ~NETDEV_ALIGN_CONST);
	dev->padded = (char *)dev - (char *)p;

	memset(dev, 0, alloc_size);

	if (sizeof_priv != 0) {
		dev->priv = (char *)dev + ((sizeof(struct net_device) +
					    NETDEV_ALIGN_CONST) &
					   ~NETDEV_ALIGN_CONST);
	}

  	printk ("Calling setup.\n");
	setup(dev);
	strcpy(dev->name, name);

      Cleanup:
	NOOKS_WRAPPER_EXIT();
	return (dev);

}

pci_power_t NOOKS_KWRAPPER(pci_choose_state) (struct pci_dev * dev,
					      pm_message_t state) {
	int rc = 0;
	pci_power_t result = NULL;
	pnook_t nook = GET_CURRENT_NOOK();
	struct pci_dev *kernel_dev = NULL;
	NOOKS_KWRAPPER_ENTER();
	rc = nooks_find_resource_elem(nook, pcidev_e, dev,
				      (void **)&kernel_dev);

	if (rc == 0)
		result = NOOKS_SWRAPPER(pci_choose_state) (kernel_dev, state);
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

pci_power_t NOOKS_WRAPPER(pci_choose_state) (struct pci_dev * dev,
					     pm_message_t state) {
	pci_power_t result;
	result =
	    (pci_power_t) NOOKS_KERNEL_CALL2(nooks_k_pci_choose_state,
					     dev, state);
	return (result);
}

/* This wrapper is not required. Is this because this function is never
 * called in the kernel domain (or called by the kernel?

int NOOKS_KWRAPPER(pci_set_power_state)(struct pci_dev *dev, pci_power_t state)
{
  int result;
  NOOKS_KWRAPPER_ENTER();
  result = NOOKS_SWRAPPER(pci_set_power_state)(dev, state);
  NOOKS_KWRAPPER_EXIT();
  return(result);
}
*/

/* The wrapper does nothing here TBD */
int NOOKS_WRAPPER(pci_set_power_state) (struct pci_dev * dev, pci_power_t state) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = pci_set_power_state(dev, state);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* iounmap() - This function unmaps a I/O address range that was 
 * mapped into the kernel address range by the ioremap function.
 * iomap_e is a resource type. 
 *
 */
void NOOKS_KWRAPPER(iounmap) (volatile void *addr) {
	pnook_t nook = GET_CURRENT_NOOK();
	unsigned long size = 0;
	if (nooks_del_resource_elem(nook, iomap_e, addr, (void **)&size) >= 0) {
		nooks_mm_release_pages(nook, (unsigned long)addr, size);
		NOOKS_SWRAPPER(iounmap) (addr);
	}
}

/* Just call the kernel iounmap function. */
void NOOKS_WRAPPER(iounmap) (volatile void *addr) {
	NOOKS_KERNEL_CALL1(nooks_k_iounmap, addr);
	return;
}

void NOOKS_KWRAPPER(netif_carrier_on) (struct net_device * dev) {
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	struct net_device *kernel_dev = NULL;
	int result = 0;

	NOOKS_KWRAPPER_ENTER();

	result =
	    nooks_find_resource_elem(nook, netdev_nook_e, dev,
				     (void **)&driver);

	if (result < 0) {
		PERROR
		    (" Failed to find netdev for netif_carrier_on: %p\n", dev);
		goto Cleanup;
	}
	kernel_dev = (struct net_device *)driver->data;
	NOOKS_SWRAPPER(netif_carrier_on) (kernel_dev);
      Cleanup:
	NOOKS_KWRAPPER_EXIT();
}

void NOOKS_WRAPPER(netif_carrier_on) (struct net_device * dev) {
	NOOKS_KERNEL_CALL1(nooks_k_netif_carrier_on, dev);
	return;
}

unsigned long NOOKS_KWRAPPER(schedule_generic) (long state, long timeout) {
	unsigned long ret = 0;
	NOOKS_KWRAPPER_ENTER();
	if (state != -1) {
		current->state = state;
	}
	if (timeout != -1) {
		ret = NOOKS_SWRAPPER(schedule_timeout) (timeout);
	} else {
		NOOKS_SWRAPPER(schedule) ();
	}
	NOOKS_KWRAPPER_EXIT();
	return ret;
}

signed long NOOKS_WRAPPER(schedule_timeout_interruptible) (signed long timeout) {
	signed long result;
	result =
	    (signed long)NOOKS_KERNEL_CALL2(nooks_k_schedule_generic,
					    TASK_INTERRUPTIBLE, timeout);
	return (result);
}

void NOOKS_WRAPPER(msleep) (unsigned int msecs) {
	NOOKS_SWRAPPER(msleep) (msecs);
}

unsigned long NOOKS_WRAPPER(msleep_interruptible) (unsigned int msecs) {
	unsigned long result;
	result = (unsigned long)NOOKS_SWRAPPER(msleep_interruptible) (msecs);
	return result;
}

/*TBD Will need a relook */
int NOOKS_KWRAPPER(dev_close) (struct net_device * dev) {
	int result;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_KWRAPPER_ENTER();
	result = NOOKS_SWRAPPER(dev_close) (dev);

	if (result >= 0)
		if (nooks_del_resource_elem(nook, netdev_e, dev, NULL)
		    <= 0)
			PDEBUG("Object tracking failed at dev_close.\n");

	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(dev_close) (struct net_device * dev) {
	int result;
	result = (int)NOOKS_KERNEL_CALL1(nooks_k_dev_close, dev);
	return (result);
}

/* Implemented as is from the 2.4 verison with very minor changes.
 */
int NOOKS_KWRAPPER(pci_request_regions) (struct pci_dev * dev, const char *name) {

	int result = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	struct pci_dev *kernel_dev = NULL;
	char *kernel_name = NULL;
	NOOKS_KWRAPPER_ENTER();

	if (nooks_get_driver_string(name, &kernel_name) < 0) {
		PERROR("Illegal name to request region: %p\n", name);
		goto Cleanup;
	}

	result =
	    nooks_find_resource_elem(nook, pcidev_e, dev, (void **)&kernel_dev);
	if (result >= 0) {
		result =
		    nooks_add_resource_elem(nook, pciregion_e,
					    kernel_dev, NULL);
		if (result >= 0) {
			result = NOOKS_SWRAPPER(pci_request_regions)
			    (kernel_dev, kernel_name);
			if (result < 0)
				nooks_del_resource_elem(nook,
							pciregion_e,
							kernel_dev, NULL);
		}
	}

      Cleanup:
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(pci_request_regions) (struct pci_dev * arg0, const char *arg1) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result =
	    (int)NOOKS_KERNEL_CALL2(nooks_k_pci_request_regions, arg0, arg1);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* Implemented similar to Nooks 2.4 */
int NOOKS_WRAPPER(pci_bus_write_config_dword) (struct pci_bus * bus,
					       unsigned int devfn,
					       int where, u32 val) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = (int)pci_bus_write_config_dword(bus, devfn, where, val);

	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* The function eth_type_trans is used to determine the packet's 
 * protocol ID. No WRAPPER required here.
 */

__be16 NOOKS_WRAPPER(eth_type_trans) (struct sk_buff * skb,
				      struct net_device * dev) {
	__be16 result;
	NOOKS_WRAPPER_ENTER();
	result = eth_type_trans(skb, dev);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

struct net_device *(dev_from_nooks_dev) (struct net_device * dev) {
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	struct net_device *kernel_dev = NULL;
	int result = 0;
	result =
	    nooks_find_resource_elem(nook, netdev_nook_e, dev,
				     (void **)&driver);

	if (result < 0) {
		PERROR("Failed to find netdev for ethtool operations.\n", dev);
		goto Cleanup;
	}

	kernel_dev = (struct net_device *)driver->data;
	return (kernel_dev);

      Cleanup:
	return NULL;
}

u32 NOOKS_WRAPPER(ethtool_op_get_link) (struct net_device * dev) {
	u32 result;
	NOOKS_WRAPPER_ENTER();
	result = ethtool_op_get_link(dev_from_nooks_dev(dev));
	NOOKS_WRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(ethtool_op_get_perm_addr) (struct net_device * dev,
					     struct ethtool_perm_addr * addr,
					     u8 * data) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = ethtool_op_get_perm_addr(dev_from_nooks_dev(dev), addr, data);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

u32 NOOKS_WRAPPER(ethtool_op_get_sg) (struct net_device * dev) {
	u32 result;
	NOOKS_WRAPPER_ENTER();
	result = ethtool_op_get_sg(dev_from_nooks_dev(dev));
	NOOKS_WRAPPER_EXIT();
	return (result);
}

u32 NOOKS_WRAPPER(ethtool_op_get_tso) (struct net_device * dev) {
	u32 result;
	NOOKS_WRAPPER_ENTER();
	result = ethtool_op_get_tso(dev_from_nooks_dev(dev));
	NOOKS_WRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(ethtool_op_set_sg) (struct net_device * dev, u32 data) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = ethtool_op_set_sg(dev_from_nooks_dev(dev), data);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* TBD Will need a relook. */
int NOOKS_KWRAPPER(dev_open) (struct net_device * dev) {
	int result;
	pnook_t nook = GET_CURRENT_NOOK();
	//struct net_device *kernel_dev = NULL;
	NOOKS_KWRAPPER_ENTER();
	result = NOOKS_SWRAPPER(dev_open) (dev);

	if (result >= 0)
		if ((nooks_add_resource_elem(nook, netdev_e, dev, NULL)
		     < 0))
			PDEBUG
			    ("Object tracker failed to track net_device at dev_open.");

	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(dev_open) (struct net_device * dev) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = (int)NOOKS_KERNEL_CALL1(nooks_k_dev_open, dev);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* Written from the 2.4 version. */
int NOOKS_WRAPPER(pci_set_dma_mask) (struct pci_dev * dev, u64 mask) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = pci_set_dma_mask(dev, mask);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* TBD */
int NOOKS_WRAPPER(pci_bus_read_config_word) (struct pci_bus * bus,
					     unsigned int devfn,
					     int where, u16 * val) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = (int)pci_bus_read_config_word(bus, devfn, where, val);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* TBD -- May need a relook */
void *NOOKS_KWRAPPER(kmap_atomic) (struct page * page, enum km_type type) {
	pnook_t nook = GET_CURRENT_NOOK();
	void *result;
	NOOKS_KWRAPPER_ENTER();

	result = NOOKS_SWRAPPER(kmap_atomic) (page, type);

	if (nooks_add_to_hash
	    (&nook->resources.area_map, kmap_e, (void *)result,
	     (void *)page) < 0) {
		PDEBUG
		    ("Failed to track kmap_atomic pointer by the object tracker.\n");
	}
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

void *NOOKS_WRAPPER(kmap_atomic) (struct page * page, enum km_type type) {
	void *result;
	result = (void *)NOOKS_KERNEL_CALL2(nooks_k_kmap_atomic, page, type);
	return (result);
}

void NOOKS_WRAPPER(__udelay) (unsigned long usecs) {
	udelay(usecs);
	return;
}

/* Free an interrupt. */
void NOOKS_KWRAPPER(free_irq) (unsigned int irq, void *dev_id) {
	pnook_driver_t driver = NULL;
	void *kernel_dev_id = NULL;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_KWRAPPER_ENTER();

	//Find the driver to cleanup.

	driver = nooks_find_driver_in_nook(NOOKS_IRQ_TO_MAJOR(irq), nook);
	if (driver == NULL)
		PERROR(" ERROR: no driver %d in nook %s, function %s\n",
		       NOOKS_IRQ_TO_MAJOR(irq), nook->name, __FUNCTION__);
	else {
		int result;
		result =
		    nooks_find_resource_elem(nook, irq_dev_e,
					     driver->data, &kernel_dev_id);
		if ((result >= 0) && (kernel_dev_id == dev_id)) {
			nooks_del_resource_elem(nook, irq_dev_e,
						driver->data, &kernel_dev_id);
			NOOKS_SWRAPPER(free_irq) (irq, dev_id);
			//Freeing the driver will release any interface it is using.
			nooks_free_driver(driver);
		}
	}
	NOOKS_KWRAPPER_EXIT();
}

void NOOKS_WRAPPER(free_irq) (unsigned int arg0, void *arg1) {
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL2(nooks_k_free_irq, arg0, arg1);
	NOOKS_WRAPPER_EXIT();
	return;
}

/*
 *  We allocate local memory for the kmalloc call. 
 */

void *NOOKS_WRAPPER(__kmalloc) (size_t size, gfp_t flags) {
	void *result;
	NOOKS_WRAPPER_ENTER();

	result = __kmalloc(size, flags);
/*	result = nooks_mm_kmalloc(GET_CURRENT_NOOK(),
#ifdef NOOKS_KEEP_STATS
				  size + NOOKS_MM_KMALLOC_OVERHEAD,
#else				// NOOKS_KEEP_STATS
				  size,
#endif				// NOOKS_KEEP_STATS
				  flags);
#ifdef NOOKS_KEEP_STATS
	if (result != NULL) {
		NOOKS_STATS_ADD(NOOKS_OPS_W_KMALLOC_BYTES, size);
		*(size_t *) result = size;
		result = (u8 *) result + NOOKS_MM_KMALLOC_OVERHEAD;
	}
#endif
*/
	NOOKS_WRAPPER_EXIT();
	return (result);
}

	/* Nothing to be done. */
void NOOKS_WRAPPER(dma_sync_single_for_device) (struct device * dev,
						dma_addr_t dma_handle,
						size_t size,
						enum dma_data_direction
						direction) {
	NOOKS_KERNEL_CALL4(dma_sync_single_for_device, dev, dma_handle,
			   size, direction);
	return;
}

/* Implementing from the 2.4 version.
 */

void NOOKS_KWRAPPER(pci_set_master) (struct pci_dev * nook_dev) {
	pnook_t nook = GET_CURRENT_NOOK();
	struct pci_dev *kernel_dev = NULL;

	//Copy device changes back to kernel.

	if (nooks_find_resource_elem
	    (nook, pcidev_e, nook_dev, (void **)&kernel_dev) >= 0) {
		//kernel_dev->driver_data = nook_dev->driver_data;
		kernel_dev->dma_mask = nook_dev->dma_mask;
		NOOKS_SWRAPPER(pci_set_master) (kernel_dev);
	}

}

void NOOKS_WRAPPER(pci_set_master) (struct pci_dev * dev) {
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL1(nooks_k_pci_set_master, dev);
	NOOKS_WRAPPER_EXIT();
}

int NOOKS_WRAPPER(netpoll_trap) (void) {
	int result;
	result = (int)NOOKS_KERNEL_CALL0(netpoll_trap);
	return (result);
}

int NOOKS_WRAPPER(_spin_trylock) (spinlock_t * lock) {
	int result;
	result = (int)NOOKS_KERNEL_CALL1(_spin_trylock, lock);
	return (result);
}

/* unregister_netdev() shuts down a device interface and removes it
 * from the kernel tables. This code is coming from the 2.4 version 
 * as-is.
 */
/* We do not create any extra drivers/devices so no need to delete them.
 * We just free the driver and netdev_stats.
 */

void NOOKS_KWRAPPER(unregister_netdev) (struct net_device * dev) {
	int result = 0;
	pnook_driver_t driver = NULL;
	pnook_t nook = GET_CURRENT_NOOK();
	struct net_device *kernel_dev = NULL;
	struct net_device_stats *kernel_stats = NULL;
	pnooks_work_item_t item;
	short in_nook = 0;

	printk("Entered unregister_netdev.\n");
	if (nook == NULL) {
		return -ENOMEM;
		printk("Nook is null.\n");
	} else
		printk("In nook %s", nook->name);
	if (IN_NOOK()) {
		in_nook = 1;
		EXIT_NOOK();
	}

	NOOKS_KWRAPPER_ENTER();

	printk("In unregister_netdev.\n");
	result =
	    nooks_del_resource_elem(nook, netdev_stats_e, dev,
				    (void **)&kernel_stats);

	if (result < 0) {
		PERROR(" failed to find netdev_stats for kernel dev %p\n", dev);
		//NOOKS_FAULT(resource);
		result = -EFAULT;
		goto Cleanup;
	}

	nooks_kfree(kernel_stats);

	result =
	    nooks_del_resource_elem(nook, netdev_nook_e, dev, (void **)&driver);
	if (result < 0) {
		PERROR(" Failed to find netdev for cleanup: %p\n", dev);
		//NOOKS_FAULT(resource);
		goto Cleanup;
	}

	if (driver)
		if (driver->data)
			kernel_dev = (struct net_device *)driver->data;

	if (kernel_dev)
		NOOKS_SWRAPPER(unregister_netdev) (kernel_dev);
	else {
		NOOKS_FAULT(resource);
		goto Cleanup;
	}

//      NOOKS_SWRAPPER(unregister_netdev) (kernel_dev);

	if (driver)
		nooks_free_driver(driver);
/*
	if (kernel_dev)
	nooks_kfree(kernel_dev);
*/

	printk("Unregister_netdev completes fine.\n");
      Cleanup:
	NOOKS_KWRAPPER_EXIT();
	if (in_nook)
		ENTER_NOOK();

	return;
}

void NOOKS_WRAPPER(unregister_netdev) (struct net_device * dev) {
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL1(nooks_k_unregister_netdev, dev);
	NOOKS_WRAPPER_EXIT();
	return;
}

/* Just directly call the associated kernel function. */
void NOOKS_WRAPPER(disable_irq) (unsigned int irq) {
	NOOKS_KERNEL_CALL1(disable_irq, irq);
	return;
}

/* Request for an irq - from the 2.4v version as-is.
 */

extern unsigned long nooks_recovering_interrupts[NR_IRQS];

int NOOKS_KWRAPPER(request_irq) (unsigned int irq,
				 irqreturn_t(*handler) (int, void *,
							struct pt_regs
							*),
				 unsigned long irqflags,
				 const char *devname, void *dev_id) {
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	int result = 0, shadow_driver = 0;
	char *kernel_name = NULL;

	NOOKS_KWRAPPER_ENTER();

	result = nooks_get_driver_string(devname, &kernel_name);

	if (result < 0)
		goto Cleanup;

	if (IN_SHADOW_RECOVERY())
		driver = shadow_find_driver2(nook, irqops_e, irq, kernel_name);
	else
		driver = NULL;

	if (driver != NULL)
		shadow_driver = 1;
	else {
		//Allocate a new driver object.
		driver =
		    nooks_create_driver(nook, irqops_e,
					NOOKS_IRQ_TO_MAJOR(irq), kernel_name);
		if (driver == NULL) {
			PERROR
			    (" Failed to reate driver %d for request irq\n",
			     NOOKS_IRQ_TO_MAJOR(irq));
			result = -ENOMEM;
			goto err1;
		}

		driver->data = (void *)irq;
		kernel_name = NULL;
	}

	nooks_iwrapper_copy_irqops(handler, &driver->operations.irqops,
				   driver->iwrapper_operations.irqops,
				   &driver->wrapper_operations.irqops);
	if (!shadow_driver)
		shadow_add_sirqops(driver, (void *)irq);

	result =
	    NOOKS_SWRAPPER(request_irq) (irq,
					 driver->wrapper_operations.irqops.
					 handler, irqflags, driver->name,
					 dev_id);
	printk("THIS IS BROKEN NEED TO IMPLEMENT request_nooks_irq instead");

	if (result < 0)
		goto err2;

	// Re-enable the irq (disabled by disable drivers)
	//
	result = nooks_add_resource_elem(nook, irq_dev_e, (void *)irq, dev_id);

	if (nooks_recovering_interrupts[irq] == 0) {
		nooks_unpoll_device(irq, 0);

	} else {
		nooks_poll_device(irq);
	}
	enable_irq(irq);
	goto Cleanup;
      err2:
	nooks_free_driver(driver);
      err1:
	nooks_free_driver_string(kernel_name);

      Cleanup:
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(request_irq) (unsigned int arg0,
				irqreturn_t(*handler) (int, void *,
						       struct pt_regs
						       *),
				unsigned long arg1, const char *arg2,
				void *arg3) {
	int result;
	result =
	    (int)NOOKS_KERNEL_CALL5(nooks_k_request_irq, arg0, handler,
				    arg1, arg2, arg3);
	return (result);
}

int NOOKS_WRAPPER(pci_enable_wake) (struct pci_dev * dev,
				    pci_power_t state, int enable) {
	int result;
	result = (int)pci_enable_wake(dev, state, enable);
	return (result);
}

/* Use the nooks_w_dev_kfree_skb_irq from 2.4 implementation
 * to implement this function. This function can be called
 * within or out of interrupt context. This has been  
 * implemented without  deferred function using work queues.
 */
void NOOKS_WRAPPER(dev_kfree_skb_any) (struct sk_buff * skb) {
	int result = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	//int empty = 0;
	NOOKS_WRAPPER_ENTER();

	result = nooks_mm_move_skb_to_kernel(nook, skb);
	if (result < 0) {
		PDEBUG("Error moving skb to kernel in dev_kfree_skb_any.\n");
		goto Cleanup;
	}
	dev_kfree_skb_any(skb);
      Cleanup:
	NOOKS_WRAPPER_EXIT();
	return;
}

int NOOKS_KWRAPPER(pci_set_mwi) (struct pci_dev * dev) {
	int result;
	NOOKS_KWRAPPER_ENTER();
	result = NOOKS_SWRAPPER(pci_set_mwi) (dev);
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(pci_set_mwi) (struct pci_dev * dev) {
	int result;
	result = (int)NOOKS_KERNEL_CALL1(nooks_k_pci_set_mwi, dev);
	return (result);
}

/* May need a relook - TBD */
int NOOKS_WRAPPER(___pskb_trim) (struct sk_buff * skb, unsigned int len) {
	int result;
	int err = 0;
	pnook_t nook = GET_CURRENT_NOOK();

	NOOKS_WRAPPER_ENTER();
	// Cloned buffers are shared data so must not be written. 
	if (skb_cloned(skb)
	    &&
	    unlikely((err = nooks_w_pskb_expand_head(skb, 0, 0, GFP_ATOMIC)))) {
		result = -EACCES;
		goto Cleanup;
	}

	nooks_mm_move_skb_to_kernel(nook, skb);
	result = (int)NOOKS_KERNEL_CALL2(___pskb_trim, skb, len);

	if (skb != NULL)
		nooks_mm_move_skb_to_nook(nook, skb);

      Cleanup:
	NOOKS_WRAPPER_EXIT();
	return (result);
}

/* BigTBD  */
void NOOKS_KWRAPPER(dma_unmap_single) (struct device * dev,
				       dma_addr_t dma_addr, size_t size,
				       enum dma_data_direction direction) {
	pnook_t nook = GET_CURRENT_NOOK();
	unsigned long pgP;
	NOOKS_KWRAPPER_ENTER();
	NOOKS_SWRAPPER(dma_unmap_single) (dev, dma_addr, size, direction);

	if (nooks_del_from_hash
	    (&nook->resources.area_map, dmamap_e, (void *)dma_addr,
	     (void **)&pgP) < 0)
		PDEBUG("Object tracking failed in dma_unmap_single.\n");

	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(dma_unmap_single) (struct device * dev,
				      dma_addr_t dma_addr, size_t size,
				      enum dma_data_direction direction) {
	NOOKS_KERNEL_CALL4(nooks_k_dma_unmap_single, dev, dma_addr,
			   size, direction);
	return;
}

/* Force an event channel callback. */
void NOOKS_WRAPPER(force_evtchn_callback) (void) {
	NOOKS_KERNEL_CALL0(force_evtchn_callback);
	return;
}

void NOOKS_KWRAPPER(dma_sync_single_for_cpu) (struct device * dev,
					      dma_addr_t dma_handle,
					      size_t size,
					      enum dma_data_direction
					      direction) {
	NOOKS_KWRAPPER_ENTER();
	NOOKS_SWRAPPER(dma_sync_single_for_cpu) (dev, dma_handle, size,
						 direction);
	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(dma_sync_single_for_cpu) (struct device * dev,
					     dma_addr_t dma_handle,
					     size_t size,
					     enum dma_data_direction
					     direction) {
	NOOKS_KERNEL_CALL4(nooks_k_dma_sync_single_for_cpu, dev,
			   dma_handle, size, direction);
	return;
}

/* This is helper function for kernel parameters. */

int NOOKS_WRAPPER(param_array_set) (const char *val, struct kernel_param * kp) {
	int result;
	result = (int)NOOKS_KERNEL_CALL2(param_array_set, val, kp);
	return (result);
}

/* Flush the work-queue. */
void NOOKS_WRAPPER(flush_scheduled_work) (void) {
	NOOKS_KERNEL_CALL0(flush_scheduled_work);
	return;
}

int NOOKS_WRAPPER(param_array_get) (char *buffer, struct kernel_param * kp) {
	int result;
	result = (int)NOOKS_KERNEL_CALL2(param_array_get, buffer, kp);
	return (result);
}

/* Process security function. */
int NOOKS_WRAPPER(capable) (int cap) {
	int result;
	result = (int)NOOKS_KERNEL_CALL1(capable, cap);
	return (result);
}

/* TBD -- This may need a relook as we do not know if such trading
 * of skb works correctly.  
 */

unsigned char *NOOKS_WRAPPER(__pskb_pull_tail) (struct sk_buff * skb, int delta) {
	unsigned char *result = NULL;
	int rc = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_WRAPPER_ENTER();
	rc = nooks_mm_move_skb_to_kernel(nook, skb);

	if (rc < 0) {
		PDEBUG("Error moving skb to kernel in __pskb_pull_tail.\n");
		goto Cleanup;
	}

	result =
	    (unsigned char *)NOOKS_KERNEL_CALL2(__pskb_pull_tail, skb, delta);
	if (result == NULL)
		PDEBUG("__pskb_pull_tail wrapper call failed.\n");

	/* Play safe */
	if (skb != NULL) {
		rc = nooks_mm_move_skb_to_nook(nook, skb);
		if (rc < 0)
			PDEBUG
			    ("Error moving skb to nook in __pskb_pull_tail.\n");
	} else
		PDEBUG
		    ("__pskb_pull_tail wrapper call failed while moving skb to nook. \n");
      Cleanup:
	NOOKS_WRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(netif_rx) (struct sk_buff * skb) {
	int rc = 0;
	int result = -1;
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_WRAPPER_ENTER();

	rc = nooks_mm_move_skb_to_kernel(nook, skb);

	if (rc < 0) {
		PDEBUG("Error moving skb to kernel in nooks_w_netif_rx.\n");
		goto Cleanup;
	}

	result = NOOKS_KERNEL_CALL1(netif_rx, skb);

	if (result != NET_RX_DROP)
		rc = nooks_mm_move_skb_to_nook(nook, skb);

	if (rc < 0) {
		PDEBUG("Error moving skb to nook in nooks_w_netif_rx.\n");
		goto Cleanup;
	}

      Cleanup:
	NOOKS_WRAPPER_EXIT();
	return result;
}

/* Implementing from the 2.4 version. Needs to be re-implemented to take 
 * care of the migration of task queues to work queues. Code cleanup
 * pending. 
 */

int NOOKS_KWRAPPER(register_netdev) (struct net_device * dev) {
	int result = 0, shadow_driver = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	char *kernel_name = NULL;
	struct net_device *kernel_dev = NULL, *kernel_dev2 = NULL;
	struct net_device_stats *kernel_stats = NULL;

	short in_nook = 0;

	if (nook == NULL)
		printk("Fatal nook is null.\n");
	if (IN_NOOK()) {
		in_nook = 1;
		EXIT_NOOK();
	}

	NOOKS_KWRAPPER_ENTER();

	printk("In register_netdev.\n");

	/* Allocate kernel version of the net device structure. */
	kernel_dev2 = nooks_kmalloc(sizeof(struct net_device), GFP_KERNEL);
	if (kernel_dev2 == NULL)
		goto Cleanup;

	/* Copy the nooks version to the kernel version. This may not be required? BigTBD */
	result = copy_from_user(kernel_dev2, dev, sizeof(struct net_device));
	if (result < 0) {
		printk
		    ("Nooks(register_netdev): Failed to copy netdevice from %p.\n",
		     dev);
		goto err2;
	}

	kernel_dev = kernel_dev2;
#if 0
	/* now create the nooks driver. */

	/* Get the driver name from the kernel space. */
	result =
	    // nooks_get_driver_string(dev->name, &kernel_name);
	    nooks_get_driver_string(dev->name, &kernel_name);
	if (result < 0)
		goto err2;

#endif
	kernel_dev = kernel_dev2;

	result = nooks_get_driver_string(dev->name, &kernel_name);
	if (result < 0) {
		printk("nooks_get_driver_string failed.");
		goto err3;
	}

	/* Now create the driver. */
	driver =
	    nooks_create_driver(nook, netops_e,
				NOOK_DRIVER_NO_MAJOR, kernel_name);
	if (driver == NULL) {

		printk
		    ("Nooks(register_netdev): Failed to create driver for register_netdev.\n");
		result = -ENOMEM;
		goto err3;
	}

	if (driver->nook == NULL)
		driver->nook = nook;

	strcpy(kernel_dev->name, kernel_name);
	printk(" NOOKS(register_netdev): Registering netdevice %s, %s.\n",
	       kernel_name, kernel_dev->name);

	/* Copy the operations. 
	   nooks_iwrapper_copy_netops(kernel_dev2,
	   &driver->operations.netops,
	   driver->iwrapper_operations.
	   netops, kernel_dev);

	   driver->data = kernel_dev;
	 */

	nooks_iwrapper_copy_netops(dev,
				   &driver->operations.netops,
				   driver->iwrapper_operations.netops, dev);

	driver->data = dev;

	printk("Copy netops complete.\n");
	result = nooks_add_resource_elem(nook, netdev_nook_e, dev, driver);	// Wrong FIXME
	if (result < 0)
		goto err5;
	kernel_dev->priv = dev;

	kernel_stats =
	    nooks_kmalloc(sizeof(struct net_device_stats), GFP_KERNEL);
	if (kernel_stats == NULL)
		goto err1;

	result = nooks_add_resource_elem(nook, netdev_stats_e, dev, kernel_stats);	//WRONG FIXME

	if (result < 0)
		goto err4;

	dev->priv = dev;
	/* No shadow work for now.!
	   if (!shadow_driver)
	   result = shadow_add_snetops(driver, dev);

	 */

	/*Execute the original code */
	//result = register_netdev (kernel_dev);
	printk ("Calling register_netdev with dev %p.\n",dev);
	result = register_netdev(dev);
	printk(" register_netdev returns %d\n", result);

#if 0
	// Copy the name back
	//
	if (result >= 0) {
		kernel_dev2 = NULL;
		nooks_memcpy_out(dev->name, kernel_dev->name, IFNAMSIZ);
		// BUGBUG: is this safe?
		//
		//NASSERT(strlen(driver->name) >= strlen(kernel_dev->name));
		strcpy(driver->name, kernel_dev->name);
	} else {
		goto err5;
	}
#endif

	printk("**NOOKS**:Register netdev completed successfully.\n");
	goto Cleanup;

	/* 6. Cleanup */
      err5:
	nooks_del_resource_elem(nook, netdev_e, dev, (void **)&dev);
	nooks_del_resource_elem(nook, netdev_stats_e,
				dev, (void **)&kernel_stats);
	nooks_del_resource_elem(nook, netdev_nook_e, dev, (void **)&driver);
      err4:
	nooks_free_driver(driver);
      err3:
	nooks_free_driver_string(kernel_name);
      err2:
	nooks_kfree(kernel_stats);
      err1:

      Cleanup:

	NOOKS_KWRAPPER_EXIT();
	if (in_nook)
		ENTER_NOOK();

	driver->nook = GET_CURRENT_NOOK();

	return result;
}

int NOOKS_WRAPPER(register_netdev) (struct net_device * dev) {
	int result;
	result = (int)NOOKS_KERNEL_CALL1(nooks_k_register_netdev, dev);
	return (result);
}

void NOOKS_KWRAPPER(vfree) (void *addr) {
	pnook_t nook = GET_CURRENT_NOOK();
	NOOKS_KWRAPPER_ENTER();
	nooks_mm_vfree(nook, addr);
	NOOKS_KWRAPPER_EXIT();
}

void nooks_w_vfree(void *addr)
{
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL1(nooks_k_vfree, addr);
	NOOKS_WRAPPER_EXIT();
	return;
}

/* TBD */
int NOOKS_WRAPPER(pskb_expand_head) (struct sk_buff * skb,
				     int nhead, int ntail, gfp_t gfp_mask) {
	int result = 0;
	int i = 0;
	struct sk_buff *skb2;
	pnook_t nook = GET_CURRENT_NOOK();
	int new_size = 0;
	NOOKS_WRAPPER_ENTER();
	new_size = nhead + (skb->end - skb->head) + ntail;
	new_size = SKB_DATA_ALIGN(new_size);

	/* Getting a new buffer is cleaner than getting dirty because 
	 * alloc_skb allocates memory for skb as a whole.
	 */
	skb2 = nooks_w___alloc_skb(new_size, gfp_mask, 0);
	if (skb2 == NULL) {
		result = -ENOMEM;
		PDEBUG
		    ("pskb_expand_head wrapper failed.\n Error allocating new skb");
		goto Cleanup;
	}

	/* Copy the header + data and sharedinfo respectively. */
	memcpy(skb2->head + nhead, skb->head, skb->tail - skb->head);
	memcpy(skb_shinfo(skb2), skb_shinfo(skb),
	       sizeof(struct skb_shared_info));

	/* Any frag related changes. TBD This may need a relook */
	for (i = 0; i < skb_shinfo(skb2)->nr_frags; i++)
		get_page(skb_shinfo(skb2)->frags[i].page);

	if (skb_shinfo(skb2)->frag_list)
		skb_clone_fraglist(skb2);

	skb2->csum = skb->csum;
	skb2->ip_summed = skb->ip_summed;
	skb2->data_len = skb->data_len;
	copy_skb_header(skb2, skb);
	skb2->nook_info = &skb_shinfo(skb)->nook_info;

	nooks_mm_move_skb_to_kernel(nook, skb);
	if (result < 0) {
		PDEBUG("Error moving skb to kernel in pskb_expand_head.\n");
		goto Cleanup;
	}

	NOOKS_KERNEL_CALL1(skb_release_data, skb);
	skb = skb2;
      Cleanup:
	NOOKS_WRAPPER_EXIT();
	return (result);
}

void NOOKS_WRAPPER(enable_irq) (unsigned int irq) {
	NOOKS_KERNEL_CALL1(enable_irq, irq);
	return;
}

void NOOKS_KWRAPPER(pci_disable_device) (struct pci_dev * dev) {
	int result = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	struct pci_dev *kernel_dev = NULL;
	NOOKS_KWRAPPER_ENTER();

	result =
	    nooks_find_resource_elem(nook, pcidev_e, dev, (void **)&kernel_dev);

	if (result >= 0)
		NOOKS_SWRAPPER(pci_disable_device) (kernel_dev);
	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(pci_disable_device) (struct pci_dev * dev) {
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL1(nooks_k_pci_disable_device, dev);
	NOOKS_WRAPPER_EXIT();
	return;
}

	/* The following function behaves like del_timer except on SMPs. 
	 * besides deactivating the timer it also makes sure the handler 
	 * has finished executing on other CPUs.
	 */
int NOOKS_KWRAPPER(del_timer_sync) (struct timer_list * nook_timer) {
	int result;
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;

	NOOKS_KWRAPPER_ENTER();

	result =
	    nooks_del_resource_elem(nook, timer_e, nook_timer,
				    (void **)&driver);
	if (result >= 0) {
		NOOKS_SWRAPPER(del_timer) ((struct timer_list *)
					   driver->data);
		/* Free Nooks resources. */
		//nooks_kfree(driver->data); -- In 2.6 version we do not maintain copy of things.
		nooks_free_driver(driver);
	} else {
		PERROR
		    ("Failed to find timer %p for nook %p for del_timer\n",
		     data1, nook);
		result = -1;
	}

	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(del_timer_sync) (struct timer_list * timer) {
	int result;
	result = NOOKS_KERNEL_CALL1(nooks_k_del_timer_sync, timer);
	return (result);
}

	/* Remove access by the nook to an SKB so it can be safely
	 * used in the kernel and appropriately free it.
	 */
void NOOKS_WRAPPER(kfree_skb) (struct sk_buff * skb) {
	int result = 0;
	NOOKS_WRAPPER_ENTER();
	nooks_mm_move_skb_to_kernel(GET_CURRENT_NOOK(), skb);
	if (result < 0) {
		PDEBUG("Error moving skb to kernel in kfree_skb.\n");
		goto Cleanup;
	}

	kfree_skb(skb);

      Cleanup:
	NOOKS_WRAPPER_EXIT();
	return;
}

/*Implemented from the 2.4 version. */
void nooks_w_skb_over_panic(struct sk_buff *skb, int len, void *here)
{
	NOOKS_WRAPPER_ENTER();
	skb_over_panic(skb, len, here);
	NOOKS_WRAPPER_EXIT();
}

/* BigTBD  */
void NOOKS_KWRAPPER(dma_unmap_page) (struct device * dev,
				     dma_addr_t dma_address,
				     size_t size,
				     enum dma_data_direction direction) {
	pnook_t nook = GET_CURRENT_NOOK();
	unsigned long pgP;
	NOOKS_KWRAPPER_ENTER();
	NOOKS_SWRAPPER(dma_unmap_page) (dev, dma_address, size, direction);
	if (nooks_del_from_hash
	    (&nook->resources.area_map, dmamap_e,
	     (void *)dma_address, (void **)&pgP) < 0)
		PDEBUG("Object tracking failed in dma_unmap_page.\n");

	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(dma_unmap_page) (struct device * dev,
				    dma_addr_t dma_address,
				    size_t size,
				    enum dma_data_direction direction) {
	NOOKS_KERNEL_CALL4(nooks_k_dma_unmap_page, dev,
			   dma_address, size, direction);
	return;
}

/* ioremap - Map a range of device address to the kernel address. */
void *NOOKS_KWRAPPER(__ioremap) (unsigned long offset,
				 unsigned long size, unsigned long flags) {
	pnook_t nook = GET_CURRENT_NOOK();
	void *result;
	NOOKS_KWRAPPER_ENTER();
	result = NOOKS_SWRAPPER(__ioremap) (offset, size, flags);
	if (result != NULL) {
		unsigned long area_size = ROUND_UP_COUNT((offset + size),
							 PAGE_SIZE) -
		    (offset & PAGE_MASK);
		if (nooks_add_resource_elem
		    (nook, iomap_e, result, (void *)area_size) < 0) {
			NOOKS_SWRAPPER(iounmap) (result);
			result = NULL;
		} else {
			if (nooks_mm_add_pages
			    (nook, (unsigned long)result,
			     GFP_KERNEL, area_size) < 0) {
				nooks_del_resource_elem(nook,
							iomap_e, result, NULL);
				NOOKS_SWRAPPER(iounmap)
				    (result);
				result = NULL;
			}
		}
	}
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

void *NOOKS_WRAPPER(__ioremap) (unsigned long offset,
				unsigned long size, unsigned long flags) {
	void *result;
	NOOKS_WRAPPER_ENTER();
	result =
	    (void *)NOOKS_KERNEL_CALL3(nooks_k___ioremap, offset, size, flags);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

void NOOKS_KWRAPPER(__netif_schedule) (struct net_device * dev) {
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	struct net_device *kernel_dev = NULL;
	int result = 0;

	NOOKS_KWRAPPER_ENTER();

	result =
	    nooks_find_resource_elem(nook, netdev_nook_e, dev,
				     (void **)&driver);

	if (result < 0) {
		PERROR(" Failed to find netdev for netif_schedule: %p\n", dev);
		goto Cleanup;
	}
	kernel_dev = (struct net_device *)driver->data;
	NOOKS_SWRAPPER(__netif_schedule) (kernel_dev);
      Cleanup:
	NOOKS_KWRAPPER_EXIT();
}

void NOOKS_WRAPPER(__netif_schedule) (struct net_device * dev) {
	NOOKS_WRAPPER_ENTER();
	NOOKS_KERNEL_CALL1(nooks_k___netif_schedule, dev);
	NOOKS_WRAPPER_EXIT();
	return;
}

/* Adds the driver structure to the list of registered drivers. 
 * Implementing from the 2.4 version.
 * How useful is the new owner field - 
 */

int NOOKS_KWRAPPER(__pci_register_driver) (struct pci_driver *
					   pci_driver, struct module * owner) {
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	int result = 0;

	char *kernel_name = NULL;
	struct pci_driver *kernel_drv = NULL;
	struct pci_driver *kernel_drv2 = NULL;
	int shadow_driver = 0;
	NOOKS_KWRAPPER_ENTER();
	kernel_drv = nooks_kmalloc(sizeof(struct pci_driver), GFP_KERNEL);
	if (kernel_drv == NULL)
		goto Cleanup;

	result =
	    nooks_memcpy(kernel_drv, pci_driver, sizeof(struct pci_driver));
	if (result < 0) {
		PERROR(" failed to copy pci driver from %p\n", pci_driver);
		goto err1;
	}

	/* ID Table contain vendor, device information in a table form. */
	if (kernel_drv->id_table != NULL) {
		struct pci_device_id *ids =
		    (struct pci_device_id *)kernel_drv->id_table;
		int count = 0;
		unsigned int vendor;
		do {
			result =
			    nooks_memcpy(&vendor,
					 &(ids + count)->vendor,
					 sizeof(unsigned int));
			if (result < 0)
				goto err1;
			count++;
		}
		while (vendor != 0);	/* Count the size of the id-table */

		/* Allocate space and copy the table. This copy may not be possible. FIXME */
		kernel_drv->id_table = (const struct pci_device_id *)
		    nooks_kmalloc(sizeof(struct pci_device_id) *
				  count, GFP_KERNEL);

		if (kernel_drv->id_table == NULL) {
			PERROR(" failed to copy pci driver from %p\n", driver);
			goto err2;
		}
	}

	driver = shadow_find_driver(nook, pciops_e, NULL);
	if (driver != NULL) {
		shadow_driver = 1;
		kernel_drv2 = (struct pci_driver *)driver->data;
	} else {
		kernel_drv2 = kernel_drv;
		result =
		    nooks_get_driver_string(kernel_drv->name, &kernel_name);
		if (result < 0)
			goto err2;

		PDEBUG(" Registering pci driver %s, %s\n",
		       kernel_name, kernel_drv->name);

		driver =
		    nooks_create_driver(nook, pciops_e,
					NOOK_DRIVER_NO_MAJOR, kernel_name);
		if (driver == NULL) {
			PERROR
			    (" Failed to create driver for register pci driver\n");
			result = -ENOMEM;
			goto err3;
		}

		kernel_drv->name = kernel_name;
	}

	nooks_iwrapper_copy_pciops(kernel_drv,
				   &driver->operations.pciops,
				   driver->iwrapper_operations.
				   pciops, kernel_drv2);
	driver->data = kernel_drv2;
	result =
	    nooks_add_resource_elem(nook, pci_driver_e, pci_driver, driver);

	if (result < 0)
		goto err4;

	if (!shadow_driver && (result >= 0))
		result = shadow_add_spciops(driver, kernel_drv2);

	/* Call the actual register function. */
	result = NOOKS_SWRAPPER(pci_register_driver) (kernel_drv2);

	PDEBUG(" pci_register_driver returns %d\n", result);

	if (result == 0)
		goto err4;

	goto Cleanup;

      err4:
	nooks_free_driver(driver);
      err3:
	nooks_free_driver_string(kernel_name);
      err2:
	if (kernel_drv2->id_table)
		nooks_kfree(kernel_drv2->id_table);
      err1:
	nooks_kfree(kernel_drv2);
      Cleanup:
	if (shadow_driver && kernel_drv) {
		if (kernel_drv->id_table)
			nooks_kfree(kernel_drv);
		nooks_kfree(kernel_drv);
	}
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(__pci_register_driver) (struct pci_driver *
					  arg0, struct module * arg1) {
	int result;
	result = (int)
	    NOOKS_KERNEL_CALL2(nooks_k___pci_register_driver, arg0, arg1);
	return (result);
}

/* Linux has moved from task queues to work queues. Each work queue now has
 * a separate calling thread. We treat the schedule_work as schedule_task was 
 * treated.
 */
int NOOKS_WRAPPER(schedule_work) (struct work_struct * work) {
	int result;
	result = schedule_work(work);
	return (result);
}

int NOOKS_WRAPPER(pci_set_consistent_dma_mask) (struct pci_dev * dev, u64 mask) {
	int result;
	NOOKS_WRAPPER_ENTER();
	result = pci_set_consistent_dma_mask(dev, mask);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

	/* param_get_int is a helper function for kernel parameter handling. */
int NOOKS_WRAPPER(param_get_int) (char *buffer, struct kernel_param * kp) {
	int result;
	result = (int)NOOKS_KERNEL_CALL2(param_get_int, buffer, kp);
	return (result);
}

void *NOOKS_KWRAPPER(vmalloc) (unsigned long size) {
	pnook_t nook = GET_CURRENT_NOOK();
	void *result;
	gfp_t gfp_mask = GFP_KERNEL | __GFP_HIGHMEM;
	NOOKS_KWRAPPER_ENTER();
	result = nooks_mm_vmalloc(nook, size, gfp_mask, PAGE_KERNEL);
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

void *NOOKS_WRAPPER(vmalloc) (unsigned long size) {
	void *result;
	NOOKS_WRAPPER_ENTER();
	result = (void *)NOOKS_KERNEL_CALL1(nooks_k_vmalloc, size);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(pci_bus_write_config_word) (struct pci_bus *
					      bus,
					      unsigned int
					      devfn, int where, u16 val) {
	int result;
	result = (int)pci_bus_write_config_word(bus, devfn, where, val);
	return (result);
}

int NOOKS_WRAPPER(param_set_int) (const char *val, struct kernel_param * kp) {
	int result;
	result = (int)NOOKS_KERNEL_CALL2(param_set_int, val, kp);
	return (result);
}

	/* RELOOK */
void NOOKS_WRAPPER(raise_softirq_irqoff) (unsigned int nr) {
	NOOKS_KERNEL_CALL1(raise_softirq_irqoff, nr);
	return;
}

	/* Should be similar to  pci_free_consistent. */
void NOOKS_KWRAPPER(dma_free_coherent) (struct device * dev,
					size_t size,
					void *vaddr, dma_addr_t dma_handle) {
	pnook_t nook = GET_CURRENT_NOOK();
	size_t stored_size;
	struct dma_coherent_mem *mem = dev ? dev->dma_mem : NULL;
	int order = get_order(size);
	NOOKS_KWRAPPER_ENTER();
	if (mem) ;
	/* BigTBD Need to do the continguous memory stuff here. */

	xen_destroy_contiguous_region((unsigned long)vaddr, order);
	if (nooks_del_resource_elem
	    (nook, pci_memory_e, vaddr, (void **)&stored_size) >= 0) {
		if (stored_size == size) {
			nooks_mm_free_pages(nook, (unsigned long)
					    vaddr, get_order(size));
		} else {
			PERROR
			    ("Stored pci memory size (%x) does not equal passed (%x)\n",
			     stored_size, size);
		}
	}

	NOOKS_KWRAPPER_EXIT();
	return;
}

void NOOKS_WRAPPER(dma_free_coherent) (struct device * dev,
				       size_t size, void *vaddr,
				       dma_addr_t dma_handle) {
	NOOKS_KERNEL_CALL4(nooks_k_dma_free_coherent, dev, size,
			   vaddr, dma_handle);
	return;
}

	/* Implementing from the 2.4 version. */
int NOOKS_KWRAPPER(pci_enable_device) (struct pci_dev * dev) {
	int result = 0;
	pnook_t nook = GET_CURRENT_NOOK();
	struct pci_dev *kernel_dev = NULL;
	NOOKS_KWRAPPER_ENTER();
	result =
	    nooks_find_resource_elem(nook, pcidev_e, dev, (void **)&kernel_dev);

	if (result >= 0)
		result = NOOKS_SWRAPPER(pci_enable_device) (dev);
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(pci_enable_device) (struct pci_dev * dev) {
	int result = 0;
	NOOKS_WRAPPER_ENTER();
	result = (int)NOOKS_KERNEL_CALL1(nooks_k_pci_enable_device, dev);
	NOOKS_WRAPPER_EXIT();
	return (result);
}

	/* In Linux 2.4, when a packet is received, netif_rx is called which
	 * enqueues the received packet in the interrupted CPU's backlog queue 
	 * schedules a softirq. On the other hand, 2.6 calls the function 
	 * beloew that puts a reference to the device in a queue attached to 
	 * the interrupted CPU and a softirq is scheduled again.
	 * The difference in scheduling the softirq is obvious - instead of 
	 * pulling packets in backlog, the CPU polls the device to get the
	 * packets from the rx_ring or from the backlog. Hence, we don't 
	 * need to perform deferred packet processing at this stage, neither
	 * we can do it given the interface of the below function.
	 */
void NOOKS_KWRAPPER(__netif_rx_schedule) (struct net_device * dev) {
	pnook_t nook = GET_CURRENT_NOOK();
	pnook_driver_t driver = NULL;
	struct net_device *kernel_dev = NULL;
	int result = 0;

	NOOKS_KWRAPPER_ENTER();

	result =
	    nooks_find_resource_elem(nook, netdev_nook_e, dev,
				     (void **)&driver);

	if (result < 0) {
		PERROR
		    (" Failed to find netdev for netif_carrier_on: %p\n", dev);
		goto Cleanup;
	}
	kernel_dev = (struct net_device *)driver->data;
	NOOKS_SWRAPPER(__netif_rx_schedule) (kernel_dev);
      Cleanup:
	NOOKS_KWRAPPER_EXIT();
}

void NOOKS_WRAPPER(__netif_rx_schedule) (struct net_device * dev) {
	NOOKS_KERNEL_CALL1(nooks_k___netif_rx_schedule, dev);
	return;
}

void NOOKS_WRAPPER(_spin_lock) (spinlock_t * lock) {
	NOOKS_KERNEL_CALL1(_spin_lock, lock);
	return;
}

unsigned long NOOKS_WRAPPER(_spin_lock_irqsave) (spinlock_t * lock) {
	unsigned long result;
	result = NOOKS_KERNEL_CALL1(_spin_lock_irqsave, lock);
	return result;
}

void NOOKS_WRAPPER(synchronize_irq) (unsigned int irq) {
	NOOKS_KERNEL_CALL1(synchronize_irq, irq);
	return;
}

int NOOKS_WRAPPER(pci_bus_read_config_dword) (struct pci_bus *
					      bus,
					      unsigned int
					      devfn, int where, u32 * val) {
	int result;
	result =
	    (int)NOOKS_KERNEL_CALL4(pci_bus_read_config_dword,
				    bus, devfn, where, val);
	return (result);
}

/* The minimalist add_timer */
void NOOKS_KWRAPPER(add_timer) (struct timer_list * nook_timer,
				unsigned long expires) {
	pnook_t nook = GET_CURRENT_NOOK();
	int result = 0;
	struct timer_list *timer = NULL;
	pnook_driver_t driver = NULL;
	NOOKS_ENTER();

	driver =
	    nooks_create_driver(nook, timerops_e, NOOK_DRIVER_NO_MAJOR, NULL);
	if (driver == NULL) {
		result = -ENOMEM;
		goto Cleanup;
	}

	driver->data = nook_timer;
	result = nooks_add_resource_elem(nook, timer_e, nook_timer, driver);
	NOOKS_SWRAPPER(add_timer) ((struct timer_list *)timer);
      Cleanup:
	NOOKS_EXIT();
	return;
}

void NOOKS_WRAPPER(add_timer) (struct timer_list * nook_timer) {
	NOOKS_KERNEL_CALL2(nooks_k_add_timer, nook_timer, 0);
}

/* mod_timer - Modifies a timer's timeout. */
int NOOKS_KWRAPPER(mod_timer) (struct timer_list * nook_timer,
			       unsigned long expires) {
	int result;
	pnook_t nook;
	pnook_driver_t driver = NULL;
	NOOKS_ENTER();		//Debug printk calls nothing more. 
	//NOOKS_KWRAPPER_ENTER ();

	nook = GET_CURRENT_NOOK();
	result =
	    nooks_find_resource_elem(nook, timer_e, nook_timer,
				     (void **)&driver);

	if (result >= 0) {
		driver->need_delete = 0;
		NOOKS_SWRAPPER(mod_timer) ((struct timer_list *)
					   driver->data, expires);
	} else {
		//printk("THIS IS BROKEN.FIX ME");
		nooks_k_add_timer(nook_timer, expires);
		result = 1;
	}

	NOOKS_EXIT();
	//NOOKS_KWRAPPER_EXIT ();
	return (result);
}

int NOOKS_WRAPPER(mod_timer) (struct timer_list * timer, unsigned long expires) {
	int result;
	result = (int)NOOKS_KERNEL_CALL2(nooks_k_mod_timer, timer, expires);
	return (result);
}

	/* Implemeneted similar to Nooks 2.4 version taking into care appropriate
	 * interface changes.
	 */
struct sk_buff *NOOKS_WRAPPER(__alloc_skb) (unsigned int size,
					    gfp_t priority, int fclone) {
	//struct skb_shared_info *shinfo;
	struct sk_buff *skb;
	u8 *data;
	gfp_t gfp_mask = priority;
	unsigned int total_size, order;
	pnook_t nook = GET_CURRENT_NOOK();

	size = SKB_DATA_ALIGN(size);
	total_size = size	// body
	    + SKB_DATA_ALIGN(sizeof(struct sk_buff))	// header
	    + sizeof(struct skb_shared_info);	// shared information
	if (in_interrupt() && (gfp_mask & __GFP_WAIT)) {
		static int count = 0;
		if (++count < 5)
			printk(KERN_ERR
			       "alloc_skb called nonatomically from interrupt %p\n",
			       NET_CALLER(size));
		gfp_mask &= ~__GFP_WAIT;
	}

	/* Get the head. */
	order = nooks_log2pages(total_size);
	skb = nooks_alloc_skb(nook, order);

	if (skb == NULL) {
		skb = (struct sk_buff *)
		    NOOKS_KERNEL_CALL2(nooks_mm_alloc_skb, order, gfp_mask);
		if (skb == NULL) {
			PERROR("Failed to get pages %d for alloc_skb\n", order);
			goto Cleanup;
		}
	}

	PDEBUG("Allocate order %d pages at %p\n", order, skb);

	/* Skip the header to go to the data. */
	data = (u8 *) skb + SKB_DATA_ALIGN(sizeof(struct sk_buff));
	//skb_headerinit(skb, NULL, 0);
	fill_in_skb(skb, data, size);

	/* Fill in the information about the nook. This information lives in the data,
	 * but the header has a pointer to it. We have two references - one for the
	 * header reference to the page of data, and one for the  reference to the
	 * page. 
	 */

	// BUGBUG: storing this in memory writeable by a driver is unsafe.
	// it should be stored somewhere else, or copied in later.
	skb_shinfo(skb)->nook_info.nook = nook;
	skb_shinfo(skb)->nook_info.order = order;
	atomic_set(&(skb_shinfo(skb)->nook_info.refs), 2);
	skb->nook_info = &skb_shinfo(skb)->nook_info;

	if (fclone) {
		struct sk_buff *child = skb + 1;
		atomic_t *fclone_ref = (atomic_t *) (child + 1);

		skb->fclone = SKB_FCLONE_ORIG;
		atomic_set(fclone_ref, 1);

		child->fclone = SKB_FCLONE_UNAVAILABLE;
	}

	NOOKS_STATS_INC(NOOKS_OPS_ALLOC_SKB);
      Cleanup:
	NOOKS_WRAPPER_EXIT();
	return skb;
}

void NOOKS_WRAPPER(init_timer) (struct timer_list * timer) {
	NOOKS_KERNEL_CALL1(init_timer, timer);
	return;
}

void nooks_w_kfree(const void *ptr)
{
	NOOKS_WRAPPER_ENTER();
	if (ptr != NULL) {
#ifdef NOOKS_KEEP_STATS
		u8 *ptr_s = (u8 *) ptr - NOOKS_MM_KMALLOC_OVERHEAD;
		NOOKS_STATS_ADD(NOOKS_OPS_W_KMALLOC_BYTES,
				-(*(size_t *) ptr_s));
		nooks_mm_kfree(GET_CURRENT_NOOK(), ptr_s);
#else				// NOOKS_KEEP_STATS
		nooks_mm_kfree(GET_CURRENT_NOOK(), ptr);
#endif				// NOOKS_KEEP_STATS
	}
	NOOKS_WRAPPER_EXIT();
}

void NOOKS_WRAPPER(pci_clear_mwi) (struct pci_dev * dev) {
	NOOKS_KERNEL_CALL1(pci_clear_mwi, dev);
	return;
}

	/* We let go the buffers as is for now. */
int NOOKS_KWRAPPER(netif_receive_skb) (struct sk_buff * skb) {
	int result;
	NOOKS_KWRAPPER_ENTER();
	result = NOOKS_SWRAPPER(netif_receive_skb) (skb);
	NOOKS_KWRAPPER_EXIT();
	return (result);
}

int NOOKS_WRAPPER(netif_receive_skb) (struct sk_buff * skb) {
	int result;
	result = (int)NOOKS_KERNEL_CALL1(nooks_k_netif_receive_skb, skb);
	return (result);
}

int NOOKS_WRAPPER(pci_bus_read_config_byte) (struct pci_bus *
					     bus,
					     unsigned int devfn,
					     int where, u8 * val) {
	int result;
	result = pci_bus_read_config_byte(bus, devfn, where, val);
	return (result);
}

void *nooks_w_memcpy(void *dest, const void *src, size_t n)
{
	void *result;
	result = memcpy(dest, src, n);
	return result;
}

extern void nooks_w_put_page(struct page *page)
{
	NOOKS_KERNEL_CALL1(put_page, page);
}

/* Allocate nooks local pages instead of kernel pages. */
struct page *NOOKS_KWRAPPER(__alloc_pages) (gfp_t gfp_mask,
					    unsigned int order,
					    struct zonelist * zonelist) {
	struct page *result;
	unsigned long temp;
	pnook_t nook = GET_CURRENT_NOOK();

	NOOKS_KWRAPPER_ENTER();

	temp = nooks_mm_get_free_pages(nook, gfp_mask, order);
	if (temp == 0)
		result = NULL;
	else
		result = (struct page *)virt_to_page(temp);

	NOOKS_KWRAPPER_EXIT();
	return (result);
}

struct page *NOOKS_WRAPPER(__alloc_pages) (gfp_t gfp_mask,
					   unsigned int order,
					   struct zonelist * zonelist) {
	struct page *result;
	result =
	    NOOKS_KERNEL_CALL3(nooks_k___alloc_pages, gfp_mask,
			       order, zonelist);
	return result;
}

	/* Big TBD */
dma_addr_t NOOKS_KWRAPPER(dma_map_single) (struct device * dev,
					   void *ptr,
					   size_t size,
					   enum dma_data_direction direction) {
	pnook_t nook = GET_CURRENT_NOOK();
	dma_addr_t result;
	NOOKS_KWRAPPER_ENTER();
	result = NOOKS_SWRAPPER(dma_map_single) (dev, ptr, size, direction);

	if (nooks_add_to_hash
	    (&nook->resources.area_map, dmamap_e,
	     (void *)result, (void *)ptr) < 0) {
		PDEBUG
		    ("Failed to track dma_map pointer by the object tracker in dma_map_single.\n");
	}

	NOOKS_KWRAPPER_EXIT();
	return (result);
}

dma_addr_t NOOKS_WRAPPER(dma_map_single) (struct device * dev,
					  void *ptr,
					  size_t size,
					  enum dma_data_direction direction) {
	dma_addr_t result;
	result =
	    NOOKS_KERNEL_CALL4(nooks_k_dma_map_single, dev, ptr,
			       size, direction);
	return (result);
}

/* Dummy System Calls */

int NOOKS_KWRAPPER(dummy_call0) (void) {
	return (0);
}

int NOOKS_KWRAPPER(dummy_call1) (int arg1) {
	return (0);
}

int NOOKS_KWRAPPER(dummy_call2) (int arg1, int arg2) {
	return (0);
}

int NOOKS_KWRAPPER(dummy_call3) (int arg1, int arg2, int arg3) {
	return (0);
}

int NOOKS_KWRAPPER(dummy_call4) (int arg1, int arg2, int arg3, int arg4) {
	return (0);
}

int nooks_w_dummy_call0(void)
{
	return (NOOKS_KERNEL_CALL0(nooks_k_dummy_call0));
}

int nooks_w_dummy_call1(int arg1)
{
	return (NOOKS_KERNEL_CALL1(nooks_k_dummy_call1, arg1));
}

int nooks_w_dummy_call2(int arg1, int arg2)
{
	return (NOOKS_KERNEL_CALL2(nooks_k_dummy_call2, arg1, arg2));

}

int nooks_w_dummy_call3(int arg1, int arg2, int arg3)
{
	return (NOOKS_KERNEL_CALL3(nooks_k_dummy_call3, arg1, arg2, arg3));
}

int nooks_w_dummy_call4(int arg1, int arg2, int arg3, int arg4)
{
	return (NOOKS_KERNEL_CALL4
		(nooks_k_dummy_call4, arg1, arg2, arg3, arg4));

}

/* Initialize pool information -- used in nooks-data.c */

int nooks_w_init_pools(void)
{
	int result = 0;

#if 0
	pnook_t nook = GET_CURRENT_NOOK();
	pnooks_lookaside_list_t list = &nook->locals->elem_pool;
	void *elem;

	//
	// Do the list elem pool
	//

	elem = nooks_mm_lookaside_alloc(list);
	if (elem != NULL) {
		nooks_mm_lookaside_free(list, elem);
	} else {
		result = -ENOMEM;
	}

	//
	// Now do the work item pool
	//

	list = &nook->locals->work_pool;
	elem = nooks_mm_lookaside_alloc(list);
	if (elem != NULL) {
		nooks_mm_lookaside_free(list, elem);
	} else {
		result = -ENOMEM;
	}
#endif
	return (result);
}

void nooks_free_ctl_table(pnook_t nook, ctl_table * table)
{
	int i = 0;
	pnook_driver_t driver = NULL;
	if (table->ctl_name != 0) {
		if (table->procname != NULL) {
			nooks_free_driver_string((char *)table->procname);
		}
		if (table->child != NULL) {
			while (table->child[i].ctl_name != 0) {
				nooks_free_ctl_table(nook, &table->child[i]);
				i++;
			}
			nooks_kfree(table->child);
		}
		nooks_del_resource_elem(nook, ctl_table_e, table,
					(void **)&driver);
		if (driver != NULL) {
			nooks_free_driver(driver);
		}
		table->ctl_name = 0;
	}
}

#define NOOKS_MAX_CTL_TABLE 20

/**********************************************************
 *
 * Function: free_pages
 *
 * Type: cache kernel
 *
 * Notes: Needs to update per-nook state in kernel, so call
 *        nooks version.
 *
 *********************************************************/

void nooks_mm_w_free_pages(unsigned long addr, unsigned int order)
{
	NOOKS_KERNEL_CALL2(nooks_free_pages, addr, order);
}

void nooks_w_free_pages(unsigned long addr, unsigned int order)
{
	NOOKS_WRAPPER_ENTER();

	if (addr != 0) {
		if (order == 0) {
			pnook_t nook = GET_CURRENT_NOOK();
			pnooks_lookaside_list_t list = &nook->locals->page_pool;
			nooks_mm_lookaside_free(list, (void *)addr);
		} else {
			nooks_mm_w_free_pages(addr, order);
		}
	}

	NOOKS_WRAPPER_EXIT();

}

int NOOKS_KWRAPPER(printk) (char *buf) {
	char *kern_buf;
	int result = 0;
	NOOKS_KWRAPPER_ENTER();

	if (nooks_get_driver_string(buf, &kern_buf) >= 0) {
		result = printk(kern_buf);
		nooks_kfree(kern_buf);
	}
	NOOKS_KWRAPPER_EXIT();
	return (result);

}

int nooks_w_printk(const char *fmt, ...)
{
	int result;
	va_list args;
	char nooks_w_printk_buf[200];

	//  NOOKS_WRAPPER_ENTER();

	//
	// Build the string to print locally
	//

	va_start(args, fmt);
	vsnprintf(nooks_w_printk_buf, sizeof(nooks_w_printk_buf), fmt, args);
	va_end(args);

	//
	// Call into the kernel to perform the call.
	//

	result = NOOKS_KERNEL_CALL1(nooks_k_printk, nooks_w_printk_buf);

	//  NOOKS_WRAPPER_EXIT();
	return (result);

}

int nooks_w_sprintf(char *buf, const char *fmt, ...)
{
	va_list args;
	int i;

	NOOKS_WRAPPER_ENTER();
	va_start(args, fmt);
	i = vsprintf(buf, fmt, args);
	va_end(args);
	NOOKS_WRAPPER_EXIT();
	return i;
}

void nooks_w_ether_setup(struct net_device *dev)
{
	NOOKS_WRAPPER_ENTER();
	pnook_t nook = GET_CURRENT_NOOK();
	if (nook == NULL)
		printk("Nook not initialized correctly.\n");
	else
		printk("Nook values: %d.\n",
		       get_current()->nook_state.curr.flags);
	NOOKS_KERNEL_CALL1(ether_setup, dev);
	NOOKS_WRAPPER_EXIT();
}

void *nooks_w_kmem_cache_alloc(struct kmem_cache *cachep, gfp_t flags)
{
	void *result;
	NOOKS_WRAPPER_ENTER();
	result = kmem_cache_alloc(cachep, flags);
	NOOKS_WRAPPER_EXIT();
	return (result);
}
