/*
 * Utility to generate asm-i386/offsets.h.
 *
 * Copyright (C) 1999-2001 Hewlett-Packard Co
 * Copyright (C) 1999-2001 David Mosberger-Tang <davidm@hpl.hp.com>
 *
 * Note that this file has dual use: when building the kernel
 * natively, the file is translated into a binary and executed.  When
 * building the kernel in a cross-development environment, this file
 * gets translated into an assembly file which, in turn, is processed
 * by awk to generate offsets.h.  So if you make any changes to this
 * file, be sure to verify that the awk procedure still works (see
 * prin_offsets.awk).
 */
#include <linux/config.h>

#include <linux/sched.h>

#include <asm-i386/processor.h>
#include <asm-i386/ptrace.h>
#include <asm-i386/siginfo.h>
#include <asm-i386/sigcontext.h>
#include "../nooks-i.h"
#include "../nooks-switch.h"

#ifdef offsetof
# undef offsetof
#endif

/*
 * We _can't_ include the host's standard header file, as those are in
 *  potential conflict with the what the Linux kernel declares for the
 *  target system.
 */
extern int printf (const char *, ...);

#define offsetof(type,field)	((char *) &((type *) 0)->field - (char *) 0)

struct
  {
    const char name[256];
    unsigned long value;
  }
tab[] =
  {
    { "ASM_TASK_SIZE",			sizeof (struct task_struct) },
    { "ASM_PT_REGS_SIZE",		sizeof (struct pt_regs) },
    { "ASM_SIGINFO_SIZE",		sizeof (struct siginfo) },
    { "", 0 },			        /* spacer */
    { "ASM_TASK_PTRACE_OFFSET",	offsetof (struct task_struct, ptrace) },
    { "ASM_TASK_SIGPENDING_OFFSET",	offsetof (struct task_struct, sigpending) },
    { "ASM_TASK_NEED_RESCHED_OFFSET",	offsetof (struct task_struct, need_resched) },
    { "ASM_TASK_PROCESSOR_OFFSET",	offsetof (struct task_struct, processor) },
    { "ASM_TASK_THREAD_OFFSET",	offsetof (struct task_struct, thread) },
#ifdef CONFIG_NOOKS
    { "ASM_TASK_KERNEL_STACK_OFFSET",	offsetof (struct task_struct, nook_state.thread_state.trap_state.kernel_stack) },
    { "ASM_TASK_NOOK_STACK_OFFSET",	offsetof (struct task_struct, nook_state.thread_state.trap_state.nook_stack) },
    { "ASM_TASK_NOOK_PGD_OFFSET",	offsetof (struct task_struct, nook_state.thread_state.trap_state.pgd) },
    { "ASM_TASK_KERNEL_FRAME_OFFSET",	offsetof (struct task_struct, nook_state.thread_state.kernel_frame) },
    { "ASM_TASK_NOOK_OFFSET",	        offsetof (struct task_struct, nook_state.thread_state.nook) },
    { "ASM_TASK_CURRENT_DRIVER_OFFSET",offsetof (struct task_struct, nook_state.curr.current_driver) },
    {" ASM_TASK_NOOK_FRAME_OFFSET", offsetof(struct task_struct, nook_state.thread_state.nook_frame) },
    { "ASM_TASK_KERNEL_FRAME_OFFSET", offsetof(struct task_struct, nook_state.thread_state.kernel_frame) },
    { "ASM_NOOK_KERNEL_FRAME_OFFSET", offsetof(struct nooks_kernel_stack_frame, frame_register) },
    { "ASM_NOOK_KERNEL_FRAME_FRAME_OFFSET", offsetof(struct nooks_kernel_stack_frame, thread_state.kernel_frame) },
#endif // CONFIG_NOOKS
    { "ASM_TASK_PID_OFFSET",		offsetof (struct task_struct, pid) },
    { "ASM_TASK_MM_OFFSET",		offsetof (struct task_struct, mm) },
    { "ASM_TASK_ACTIVE_MM_OFFSET",	offsetof (struct task_struct, active_mm) },
    { "ASM_PT_REGS_EBX_OFFSET",	offsetof (struct pt_regs, ebx) },
    { "ASM_PT_REGS_ECX_OFFSET",	offsetof (struct pt_regs, ecx) },
    { "ASM_PT_REGS_EDX_OFFSET",	offsetof (struct pt_regs, edx) },
    { "ASM_PT_REGS_ESI_OFFSET",	offsetof (struct pt_regs, esi) },
    { "ASM_PT_REGS_EDI_OFFSET",	offsetof (struct pt_regs, edi) },
    { "ASM_PT_REGS_EBP_OFFSET",	offsetof (struct pt_regs, ebp) },
    { "ASM_PT_REGS_EAX_OFFSET",	offsetof (struct pt_regs, eax) },
#if defined(CONFIG_NOOKS) && !defined(CONFIG_NOOKS_NOPROT)
    { "ASM_PT_REGS_NOOK1_OFFSET",	offsetof (struct pt_regs, nook1) },
    { "ASM_PT_REGS_NOOK2_OFFSET",	offsetof (struct pt_regs, nook2) },
    { "ASM_PT_REGS_NOOK3_OFFSET",	offsetof (struct pt_regs, nook3) },
#endif // CONFIG_NOOKS
    { "ASM_PT_REGS_XDS_OFFSET",	offsetof (struct pt_regs, xds) },
    { "ASM_PT_REGS_XES_OFFSET",	offsetof (struct pt_regs, xes) },
    { "ASM_PT_REGS_ORIG_EAX_OFFSET",	offsetof (struct pt_regs, orig_eax) },
    { "ASM_PT_REGS_EIP_OFFSET",	offsetof (struct pt_regs, eip) },
    { "ASM_PT_REGS_XCS_OFFSET",	offsetof (struct pt_regs, xcs) },
    { "ASM_PT_REGS_EFLAGS_OFFSET",	offsetof (struct pt_regs, eflags) },
    { "ASM_PT_REGS_ESP_OFFSET",	offsetof (struct pt_regs, esp) },
    { "ASM_PT_REGS_XSS_OFFSET",	offsetof (struct pt_regs, xss) }

};

static const char *tabs = "\t\t\t\t\t\t\t\t\t\t";

int
main (int argc, char **argv)
{
  const char *space;
  int i, num_tabs;
  size_t len;

  printf ("#ifndef _ASM_OFFSETS_H\n");
  printf ("#define _ASM_OFFSETS_H\n\n");

  printf ("/*\n * DO NOT MODIFY\n *\n * This file was generated by "
	  "nooks/tools/print_offsets.\n *\n */\n\n");


  for (i = 0; i < sizeof (tab) / sizeof (tab[0]); ++i)
    {
      if (tab[i].name[0] == '\0')
	printf ("\n");
      else
	{
	  len = strlen (tab[i].name) + 1;

	  num_tabs = (48 - len) / 8;
	  if (num_tabs <= 0)
	    space = " ";
	  else
	    space = strchr(tabs, '\0') - (48 - len) / 8;

	  printf ("#define %s%s%lu\t/* 0x%lx */\n",
		  tab[i].name, space, tab[i].value, tab[i].value);
	  len += 4;
	  num_tabs = (48 - len) / 8;
	  if (num_tabs <= 0)
	    space = " ";
	  else
	    space = strchr(tabs, '\0') - (48 - len) / 8;

	  printf ("#define %s_STR%s\"%lu\"\t\n",
		  tab[i].name, space, tab[i].value);
	}
    }

  printf ("\n#endif /* _ASM_OFFSETS_H */\n");
  return 0;
}
