/*
 * netshadow-ops.c --  driver interface operations shadow net devices
 *
 * Copyright (C) 2004 Mike Swift
 */

#include "netshadow-i.h"

#ifdef CONFIG_NOOKS_SHADOW
#include "netshadow-c-gen.h"
#include <../nooks/nooks-iwrappers.h>
#include <linux/random.h>
#include <linux/poll.h>
#include <linux/soundcard.h>

#define struct_cpy(x,y)                         \
({                                              \
        if (sizeof(*(x)) != sizeof(*(y)))       \
                printk ("Fatal:Error.\n");               \
        memcpy(x, y, sizeof(*(x)));             \
})



/**********************************************************
 * spciops_probe
 * 
 * need to do special logging here
 * 
 *
 *********************************************************/

int spciops_probe(struct pci_dev *dev, const struct pci_device_id *id, int __driver_result, pnook_driver_t __driver)
{
  int result = __driver_result;
  pnook_driver_t driver;
  pshadow_driver_t shadow = __driver->nook->shadow;


  shadow_log(__driver, __FUNCTION__, (int) __driver_result);

  if (DRIVER_RUNNING(__driver)) {
    //
    // On normal state, we recover what pci device is being used and 
    // what device id
    //
    
    // 
    // Get the driver name
    //
    driver = GET_CURRENT_NOOK_DRIVER();
    if (driver->driver_type == pciops_e) {
      struct pci_driver * pci_driver = driver->data;
      pshadow_pciops_info_t pci_info = NULL;

      pci_info = kmalloc(sizeof(shadow_pciops_info_t), GFP_KERNEL);
      if (pci_info == NULL) {
	result = -ENOMEM;
	goto Cleanup;
      }
      pci_info->dev = dev;
      if (driver->data == NOOKS_PCIOPS_ISAPNP) {
	struct_cpy(&pci_info->u.isapnp_id, (struct isapnp_device_id *) id); 
      } else {
	struct_cpy(&pci_info->u.pci_id,  id); 
      }
      pci_info->driver = pci_driver;
      
      result = nooks_add_to_hash(&shadow->kernel_objects,
				 pcidev_e,
				 pci_driver,
				 pci_info);
      printk ("Added probe information to hash %d.\n", result);
      if (result < 0) {
	kfree(pci_info);
	pci_info = NULL;
	goto Cleanup;
      }
    }
  }
 Cleanup:
  return(result);
}

struct spci_operations shadow_pciops = {
  //prepare: spciops_prepare,
  resume: spciops_resume,
  remove: spciops_remove,
  suspend: spciops_suspend,
  shutdown: spciops_shutdown,
  enable_wake: spciops_enable_wake,
  //deactivate: spciops_deactivate,
  probe: spciops_probe,
  //save_state: spciops_save_state,
  //activate: spciops_activate
};

int smops_init(int __driver_result, pnook_driver_t __driver)
{
  int result = __driver_result;
  printk ("smops_init:I should be called. \n");
  shadow_log(__driver, __FUNCTION__, (int) __driver_result);
  return(result);
}



int
shadow_init_shadow_module(struct module * module, pnook_driver_t __driver)
{
  int result = 0;
  pshadow_driver_t shadow = __driver->nook->shadow;
  char * module_data = NULL;


  //
  // Only do this if shadow exists for this nook
  //

  if (shadow == NULL) {
    goto Cleanup;
  }
  printk ("Initalizing shadow module. writeable start %p, end %p, differnece %d.\n",module->writeable_start, module->writeable_end, module->writeable_end - module->writeable_start);
  module_data = vmalloc(module->writeable_end - module->writeable_start);
  if (module_data == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }
  memcpy(module_data, module->writeable_start, module->writeable_end - module->writeable_start);
  
  result = nooks_add_to_hash(&shadow->kernel_objects,
			     module_data_e,
			     module,
			     module_data );

  if (result < 0) {
    goto Cleanup;
  }

  result = nooks_add_to_hash(&shadow->kernel_objects,
				bss_start_e,
				module,
				module->bss_start);
  result = nooks_add_to_hash(&shadow->kernel_objects,
				bss_end_e,
				module,
			 	module->bss_end);
  //printk ("Added bss start %p and end %p information.\n", module->bss_start, module->bss_end);
  

#ifdef CONFIG_SWIFI
  
  module_data = vmalloc(module->size - sizeof(struct module));
  if (module_data == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }
  memcpy(module_data,
	 (char *) module + sizeof(struct module),
	 module->size - sizeof(struct module));

  result = nooks_add_to_hash(&shadow->kernel_objects,
			      module_code_e,
			      module,
			      module_data);
	   
#endif // CONFIG_SWIFI
 Cleanup:
  if (result < 0) {
    if (module_data != NULL) {
      vfree(module_data);
    }
  }
		    
  return(result);
}
struct smodule_operations shadow_mops = {
  cleanup: smops_cleanup,
//  can_unload: smops_can_unload,
  init: smops_init
};

/**********************************************************
 *
 * snetops_get_stats
 *
 * handle returning real stats during cleanup
 *
 *********************************************************/


struct net_device_stats* snetops_get_stats(struct net_device *dev, struct net_device_stats* __driver_result, pnook_driver_t __driver)
{
  struct net_device_stats* result = NULL;
  int res;
  pshadow_driver_t shadow = __driver->nook->shadow;
  struct net_device_stats * kernel_stats = NULL;
  shadow_log(__driver, __FUNCTION__, (int) __driver_result);


  if (DRIVER_RUNNING(__driver)) {
    return(__driver_result);
  } else {
    res = nooks_find_in_hash(&shadow->kernel_objects, 
			     netdev_stats_e, 
			     dev,
			     (void **) &kernel_stats);
    if (res < 0) {
      return(result);
    } else {
      return(kernel_stats);
    }
  }
}

struct snetdev_operations shadow_netops = {
  hard_header_cache: snetops_hard_header_cache,
  open: snetops_open,
  hard_header: snetops_hard_header,
  destructor: snetops_destructor,
  init: snetops_init,
  do_ioctl: snetops_do_ioctl,
  hard_header_parse: snetops_hard_header_parse,
  stop: snetops_stop,
  get_wireless_stats: snetops_get_wireless_stats,
  header_cache_update: snetops_header_cache_update,
  hard_start_xmit: snetops_hard_start_xmit,
  uninit: snetops_uninit,
  tx_timeout: snetops_tx_timeout,
  set_mac_address: snetops_set_mac_address,
  neigh_setup: snetops_neigh_setup,
  set_multicast_list: snetops_set_multicast_list,
  accept_fastpath: snetops_accept_fastpath,
  rebuild_header: snetops_rebuild_header,
  get_stats: snetops_get_stats,
  set_config: snetops_set_config,
  change_mtu: snetops_change_mtu
};



/**********************************************************
 *
 * shadow_update_ioctl
 *
 * add a new ioctl to the existing list of ioctls
 * we have a list per separately-opened files (e.g. must implement
 * open/release call, so we can scavenge these)
 *
 * For each file, we maintain a list, in time-order, of ioctls
 * one for each ioctl commmand.
 *
 *********************************************************/

int shadow_update_ioctl(pshadow_driver_t shadow,
			pnook_driver_t driver,
			struct file * filp, 
			pshadow_ioctl_data_t data)
{
  int result = 0;
  pshadow_ioctl_head_t head = NULL;
  pshadow_ioctl_data_t old_data = NULL;
  struct list_head * next, *spare;
  int new_head = 0;

  down(&shadow->recovery_lock);

  result = nooks_find_in_hash(&shadow->kernel_objects,
			      ioctl_e,
			      filp,
			      (void **) &head);

  if (result < 0) {
    new_head = 1;
    head = (pshadow_ioctl_head_t) kmalloc(sizeof(shadow_ioctl_head_t), GFP_KERNEL);
    if (head == NULL) {
      result = -ENOMEM;
      goto Cleanup;
    }
    memset(head, 0, sizeof(shadow_ioctl_head_t));
    INIT_LIST_HEAD(&head->next);
    INIT_LIST_HEAD(&head->list);
    head->opener = driver;
    head->inode = filp->f_dentry->d_inode;
    head->channels = 0;
    head->samples_per_second = 44100;
    head->sample_size = 2;
  }

  if (data != NULL) {

    //
    // BUGBUG: grab lock here
    //

    //
    // Find an old entry
    //
    
    list_for_each_safe(next, spare, &head->list) {
      old_data = list_entry(next, shadow_ioctl_data_t, next);
      if (old_data->command == data->command) {
	list_del_init(&old_data->next);
	break;
      }
      old_data = NULL;
    }
    
    //
    // Add new entry
    //
    
    list_add_tail(&data->next,&head->list);
    
    //
    // Update header with new sample information
    //

    switch(data->command) {
    case SNDCTL_DSP_SETFMT :
      {
	int format = * (int *) data->data;
	switch(format) {
	case AFMT_S16_BE:
	case AFMT_U16_BE:
	case AFMT_S16_LE:
	case AFMT_U16_LE:
	  head->sample_size = 2;
	  break;
	case AFMT_U8:
	case AFMT_S8:
	case AFMT_MU_LAW:
	case AFMT_A_LAW:
	  
	  head->sample_size = 1;
	  break;
	default:
	  head->sample_size = 1;
	  break;
	}
	break;
      }

    case SNDCTL_DSP_STEREO :
      {
	int channels = * (int *) data->data;
	head->channels = channels + 1;
	break;
      }
    case SNDCTL_DSP_SPEED :
      {
	int rate = * (int *) data->data;
	head->samples_per_second = rate;
	break;
      }
    case SNDCTL_DSP_GETIPTR :
      {
	count_info * info = (count_info *) data->data;
	info->bytes += head->read_offset;
	head->bytes_read = info->bytes;
      }
    default:
      break;
    }
  }
  if (new_head) {
    
    //
    // Update hash
    //
    result = nooks_add_to_hash(&shadow->kernel_objects,
			       ioctl_e,
			       filp,
			       head );
    if (result < 0) {
      goto Cleanup;
    }
  }
  
  head = NULL;
    
 Cleanup:

  up(&shadow->recovery_lock);

  if (result < 0) {
    if (head != NULL) {
      list_del_init(&data->next);
      kfree(head);
    }
  }
  if (old_data != NULL) {
    kfree(old_data);
  }
  return(result);
}

int shadow_find_ioctl(pshadow_driver_t shadow,
		      struct file * filp,
		      unsigned int command,
		      pshadow_ioctl_data_t * data)
{
  int result = 0;
  struct list_head * next;
  pshadow_ioctl_head_t head = NULL;
  pshadow_ioctl_data_t old_data = NULL;

  down(&shadow->recovery_lock);

  result = nooks_find_in_hash(&shadow->kernel_objects,
			      ioctl_e,
			      filp,
			      (void **) &head);

  if (result < 0) {
    goto Cleanup;
  }


  //
  // Find an old entry
  //

  result = -ENOENT;
  list_for_each(next, &head->list) {
    old_data = list_entry(next, shadow_ioctl_data_t, next);
    if (old_data->command == command) {
      *data = old_data;
      result = 0;
      break;
    }
  }

 Cleanup:
  up(&shadow->recovery_lock);

  return(result);
}


int shadow_find_sound_rate(pshadow_driver_t shadow,
			   struct file * filp,
			   unsigned int * bytes_per_second)
{
  int result = 0;
  pshadow_ioctl_head_t head = NULL;
  unsigned int samples, size, channels;

  down(&shadow->recovery_lock);

  result = nooks_find_in_hash(&shadow->kernel_objects,
			      ioctl_e,
			      filp,
			      (void **) &head);

  if (result < 0) {
    goto Cleanup;
  }

  if (head->samples_per_second != 0 ) {
    samples = head->samples_per_second;
  } else {
    samples = 44100;
  }

  if (head->sample_size != 0) {
    size = head->sample_size;
  } else {
    size = 2;
  }

  if (head->channels != 0) {
    channels = head->channels;
  } else {
    channels = 2;
  }

  *bytes_per_second = samples * size * channels;

 Cleanup:
  up(&shadow->recovery_lock);
  return(result);
}
		      
int shadow_del_ioctls(pshadow_driver_t shadow,
		     struct file * filp) 
{
  int result = 0;
  struct list_head * next, *safe;
  pshadow_ioctl_head_t head = NULL;
  pshadow_ioctl_data_t old_data = NULL;

  down(&shadow->recovery_lock);

  result = nooks_del_from_hash(&shadow->kernel_objects,
			       ioctl_e,
			       filp,
			       (void **) &head);

  if (result < 0) {
    goto Cleanup;
  }

  //
  // Find an old entry
  //

  list_for_each_safe(next, safe, &head->list) {
    old_data = list_entry(next, shadow_ioctl_data_t, next);
    list_del(&old_data->next);
    kfree(old_data);
  }
  kfree(head);

 Cleanup:
  up(&shadow->recovery_lock);
  return(result);
}


int shadow_verbose_ioctl = 0;


#endif // CONFIG_NOOKS_SHADOW
