typedef struct snetdev_operations {
  int (*hard_header_cache)(struct neighbour *neigh, struct hh_cache *hh, int driver_result, pnook_driver_t driver);
  int (*open)(struct net_device *dev, int driver_result, pnook_driver_t driver);
  int (*hard_header)(struct sk_buff *skb, struct net_device *dev, unsigned short type, void *daddr, void *saddr, unsigned len, int driver_result, pnook_driver_t driver);
  int (*init)(struct net_device *dev, int driver_result, pnook_driver_t driver);
  int (*do_ioctl)(struct net_device *dev, struct ifreq *ifr, int cmd, int driver_result, pnook_driver_t driver);
  int (*hard_header_parse)(struct sk_buff *skb, unsigned char *haddr, int driver_result, pnook_driver_t driver);
  int (*stop)(struct net_device *dev, int driver_result, pnook_driver_t driver);
  struct iw_statistics* (*get_wireless_stats)(struct net_device *dev, struct iw_statistics* driver_result, pnook_driver_t driver);
  void (*header_cache_update)(struct hh_cache *hh, struct net_device *dev, unsigned char * haddr, pnook_driver_t driver);
  void (*uninit)(struct net_device *dev, pnook_driver_t driver);
  void (*tx_timeout)(struct net_device *dev, pnook_driver_t driver);
  void (*set_multicast_list)(struct net_device *dev, pnook_driver_t driver);
  void (*destructor)(struct net_device *dev, pnook_driver_t driver);
  int (*hard_start_xmit)(struct sk_buff *skb, struct net_device *dev, int driver_result, pnook_driver_t driver);
  int (*set_mac_address)(struct net_device *dev, void *addr, int driver_result, pnook_driver_t driver);
  int (*neigh_setup)(struct net_device *dev, struct neigh_parms * parms, int driver_result, pnook_driver_t driver);
  int (*accept_fastpath)(struct net_device *dev, struct dst_entry *entry, int driver_result, pnook_driver_t driver);
  int (*rebuild_header)(struct sk_buff *skb, int driver_result, pnook_driver_t driver);
  struct net_device_stats* (*get_stats)(struct net_device *dev, struct net_device_stats* driver_result, pnook_driver_t driver);
  int (*set_config)(struct net_device *dev, struct ifmap *map, int driver_result, pnook_driver_t driver);
  int (*change_mtu)(struct net_device *dev, int new_mtu, int driver_result, pnook_driver_t driver);
} snetdev_operations_t, *psnetdev_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_snetops(pnook_driver_t driver, void * arg);
void shadow_del_snetops(pnook_driver_t driver);
extern struct snetdev_operations shadow_netops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_snetops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_snetops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct smodule_operations {
  int (*init)(int driver_result, pnook_driver_t driver);
  void (*cleanup)(pnook_driver_t driver);
  int (*can_unload)(int driver_result, pnook_driver_t driver);
} smodule_operations_t, *psmodule_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_smops(pnook_driver_t driver, void * arg);
void shadow_del_smops(pnook_driver_t driver);
extern struct smodule_operations shadow_mops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_smops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_smops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct sdentry_operations {
  int (*d_compare)(struct dentry * arg0, struct qstr * arg1, struct qstr * arg2, int driver_result, pnook_driver_t driver);
  int (*d_delete)(struct dentry * arg0, int driver_result, pnook_driver_t driver);
  int (*d_hash)(struct dentry * arg0, struct qstr * arg1, int driver_result, pnook_driver_t driver);
  void (*d_release)(struct dentry * arg0, pnook_driver_t driver);
  int (*d_revalidate)(struct dentry * arg0, struct nameidata * arg1, int driver_result, pnook_driver_t driver);
  void (*d_iput)(struct dentry * arg0, struct inode * arg1, pnook_driver_t driver);
} sdentry_operations_t, *psdentry_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_sdentryops(pnook_driver_t driver, void * arg);
void shadow_del_sdentryops(pnook_driver_t driver);
extern struct sdentry_operations shadow_dentryops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_sdentryops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_sdentryops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct sfile_operations {
  unsigned int (*poll)(struct file * arg0, struct poll_table_struct * arg1, unsigned int driver_result, pnook_driver_t driver);
  int (*release)(struct inode * arg0, struct file * arg1, int driver_result, pnook_driver_t driver);
  int (*mmap)(struct file * arg0, struct vm_area_struct * arg1, int driver_result, pnook_driver_t driver);
  loff_t (*llseek)(struct file * arg0, loff_t arg1, int arg2, loff_t driver_result, pnook_driver_t driver);
  ssize_t (*sendfile)(struct file * arg0, loff_t * arg1, size_t arg2, read_actor_t arg3, void * arg4, ssize_t driver_result, pnook_driver_t driver);
  int (*check_flags)(int arg0, int driver_result, pnook_driver_t driver);
  ssize_t (*write)(struct file * arg0, const char * arg1, size_t arg2, loff_t * arg3, ssize_t driver_result, pnook_driver_t driver);
  ssize_t (*splice_write)(struct pipe_inode_info * arg0, struct file * arg1, loff_t * arg2, size_t arg3, unsigned int arg4, ssize_t driver_result, pnook_driver_t driver);
  int (*readdir)(struct file * arg0, void * arg1, filldir_t arg2, int driver_result, pnook_driver_t driver);
  ssize_t (*splice_read)(struct file * arg0, loff_t * arg1, struct pipe_inode_info * arg2, size_t arg3, unsigned int arg4, ssize_t driver_result, pnook_driver_t driver);
  ssize_t (*writev)(struct file * arg0, const struct iovec * arg1, unsigned long arg2, loff_t * arg3, ssize_t driver_result, pnook_driver_t driver);
  long (*unlocked_ioctl)(struct file * arg0, unsigned int arg1, unsigned long arg2, long driver_result, pnook_driver_t driver);
  ssize_t (*sendpage)(struct file * arg0, struct page * arg1, int arg2, size_t arg3, loff_t * arg4, int arg5, ssize_t driver_result, pnook_driver_t driver);
  int (*fasync)(int arg0, struct file * arg1, int arg2, int driver_result, pnook_driver_t driver);
  int (*flush)(struct file * arg0, fl_owner_t id, int driver_result, pnook_driver_t driver);
  ssize_t (*aio_write)(struct kiocb * arg0, const char * arg1, size_t arg2, loff_t arg3, ssize_t driver_result, pnook_driver_t driver);
  int (*fsync)(struct file * arg0, struct dentry * arg1, int datasync, int driver_result, pnook_driver_t driver);
  int (*ioctl)(struct inode * arg0, struct file * arg1, unsigned int arg2, unsigned long arg3, int driver_result, pnook_driver_t driver);
  int (*dir_notify)(struct file *filp, unsigned long arg, int driver_result, pnook_driver_t driver);
  ssize_t (*read)(struct file * arg0, char * arg1, size_t arg2, loff_t * arg3, ssize_t driver_result, pnook_driver_t driver);
  ssize_t (*readv)(struct file * arg0, const struct iovec * arg1, unsigned long arg2, loff_t * arg3, ssize_t driver_result, pnook_driver_t driver);
  int (*aio_fsync)(struct kiocb * arg0, int datasync, int driver_result, pnook_driver_t driver);
  long (*compat_ioctl)(struct file * arg0, unsigned int arg1, unsigned long arg2, long driver_result, pnook_driver_t driver);
  int (*lock)(struct file * arg0, int arg1, struct file_lock * arg2, int driver_result, pnook_driver_t driver);
  ssize_t (*aio_read)(struct kiocb * arg0, char * arg1, size_t arg2, loff_t arg3, ssize_t driver_result, pnook_driver_t driver);
  int (*open)(struct inode * arg0, struct file * arg1, int driver_result, pnook_driver_t driver);
  unsigned long (*get_unmapped_area)(struct file * arg0, unsigned long arg1, unsigned long arg2, unsigned long arg3, unsigned long arg4, unsigned long driver_result, pnook_driver_t driver);
  int (*flock)(struct file * arg0, int arg1, struct file_lock * arg2, int driver_result, pnook_driver_t driver);
} sfile_operations_t, *psfile_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_sfops(pnook_driver_t driver, void * arg);
void shadow_del_sfops(pnook_driver_t driver);
extern struct sfile_operations shadow_fops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_sfops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_sfops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct sactor_operations {
  int (*read)(read_descriptor_t * arg0, struct page * arg1, unsigned long arg2, unsigned long arg3, int driver_result, pnook_driver_t driver);
} sactor_operations_t, *psactor_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_sactorops(pnook_driver_t driver, void * arg);
void shadow_del_sactorops(pnook_driver_t driver);
extern struct sactor_operations shadow_actorops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_sactorops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_sactorops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct sprocfs_operations {
  int (*write_proc)(struct file *file, const char *buffer, unsigned long count, void *data, int driver_result, pnook_driver_t driver);
  int (*read_proc)(char *page, char **start, off_t off, int count, int *eof, void *data, int driver_result, pnook_driver_t driver);
  int (*get_info)(char * arg0, char ** arg1, off_t arg2, int arg3, int driver_result, pnook_driver_t driver);
} sprocfs_operations_t, *psprocfs_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_sprocops(pnook_driver_t driver, void * arg);
void shadow_del_sprocops(pnook_driver_t driver);
extern struct sprocfs_operations shadow_procops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_sprocops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_sprocops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct saddress_space_operations {
  int (*releasepage)(struct page * arg0, gfp_t arg1, int driver_result, pnook_driver_t driver);
  int (*writepages)(struct address_space * arg0, struct writeback_control * arg1, int driver_result, pnook_driver_t driver);
  void (*sync_page)(struct page * arg0, pnook_driver_t driver);
  int (*commit_write)(struct file * arg0, struct page * arg1, unsigned arg2, unsigned arg3, int driver_result, pnook_driver_t driver);
  int (*writepage)(struct page *page, struct writeback_control *wbc, int driver_result, pnook_driver_t driver);
  void (*invalidatepage)(struct page * arg0, unsigned long arg1, pnook_driver_t driver);
  ssize_t (*direct_IO)(int arg0, struct kiocb * arg1, const struct iovec *iov, loff_t offset, unsigned long nr_segs, ssize_t driver_result, pnook_driver_t driver);
  struct page* (*get_xip_page)(struct address_space * arg0, sector_t arg1, int arg2, struct page* driver_result, pnook_driver_t driver);
  int (*readpage)(struct file * arg0, struct page * arg1, int driver_result, pnook_driver_t driver);
  int (*set_page_dirty)(struct page *page, int driver_result, pnook_driver_t driver);
  int (*prepare_write)(struct file * arg0, struct page * arg1, unsigned arg2, unsigned arg3, int driver_result, pnook_driver_t driver);
  int (*migratepage)(struct address_space * arg0, struct page * arg1, struct page * arg2, int driver_result, pnook_driver_t driver);
  sector_t (*bmap)(struct address_space * arg0, sector_t arg1, sector_t driver_result, pnook_driver_t driver);
  int (*readpages)(struct file *filp, struct address_space *mapping, struct list_head *pages, unsigned nr_pages, int driver_result, pnook_driver_t driver);
} saddress_space_operations_t, *psaddress_space_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_sasops(pnook_driver_t driver, void * arg);
void shadow_del_sasops(pnook_driver_t driver);
extern struct saddress_space_operations shadow_asops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_sasops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_sasops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct sinode_operations {
  int (*removexattr)(struct dentry * arg0, const char * arg1, int driver_result, pnook_driver_t driver);
  ssize_t (*listxattr)(struct dentry * arg0, char * arg1, size_t arg2, ssize_t driver_result, pnook_driver_t driver);
  int (*link)(struct dentry * arg0, struct inode * arg1, struct dentry * arg2, int driver_result, pnook_driver_t driver);
  struct dentry * (*lookup)(struct inode * arg0, struct dentry * arg1, struct nameidata * arg2, struct dentry * driver_result, pnook_driver_t driver);
  int (*getattr)(struct vfsmount *mnt, struct dentry * arg0, struct kstat * arg1, int driver_result, pnook_driver_t driver);
  int (*setxattr)(struct dentry * arg0, const char * arg1, const void * arg2, size_t arg3, int arg4, int driver_result, pnook_driver_t driver);
  void (*truncate)(struct inode * arg0, pnook_driver_t driver);
  int (*readlink)(struct dentry * arg0, char * arg1, int arg2, int driver_result, pnook_driver_t driver);
  void (*truncate_range)(struct inode * arg0, loff_t arg1, loff_t arg2, pnook_driver_t driver);
  void (*put_link)(struct dentry * arg0, struct nameidata * arg1, void * arg2, pnook_driver_t driver);
  int (*unlink)(struct inode * arg0, struct dentry * arg1, int driver_result, pnook_driver_t driver);
  int (*setattr)(struct dentry * arg0, struct iattr * arg1, int driver_result, pnook_driver_t driver);
  ssize_t (*getxattr)(struct dentry * arg0, const char * arg1, void * arg2, size_t arg3, ssize_t driver_result, pnook_driver_t driver);
  int (*create)(struct inode * arg0, struct dentry * arg1, int arg2, struct nameidata * arg3, int driver_result, pnook_driver_t driver);
  int (*permission)(struct inode * arg0, int arg1, struct nameidata * arg2, int driver_result, pnook_driver_t driver);
  int (*symlink)(struct inode * arg0, struct dentry * arg1, const char * arg2, int driver_result, pnook_driver_t driver);
  int (*mknod)(struct inode * arg0, struct dentry * arg1, int arg2, dev_t arg3, int driver_result, pnook_driver_t driver);
  int (*rmdir)(struct inode * arg0, struct dentry * arg1, int driver_result, pnook_driver_t driver);
  int (*mkdir)(struct inode * arg0, struct dentry * arg1, int arg2, int driver_result, pnook_driver_t driver);
  void * (*follow_link)(struct dentry * arg0, struct nameidata * arg1, void * driver_result, pnook_driver_t driver);
  int (*rename)(struct inode * arg0, struct dentry * arg1, struct inode * arg2, struct dentry * arg3, int driver_result, pnook_driver_t driver);
} sinode_operations_t, *psinode_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_sinodeops(pnook_driver_t driver, void * arg);
void shadow_del_sinodeops(pnook_driver_t driver);
extern struct sinode_operations shadow_inodeops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_sinodeops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_sinodeops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct spci_operations {
 // int (*prepare)(struct pci_dev *dev, int driver_result, pnook_driver_t driver);
  void (*remove)(struct pci_dev *dev, pnook_driver_t driver);
  void (*shutdown)(struct pci_dev *dev, pnook_driver_t driver);
  int (*enable_wake)(struct pci_dev *dev, u32 state, int enable, int driver_result, pnook_driver_t driver);
  int (*save_state)(struct pci_dev *dev, u32 state, int driver_result, pnook_driver_t driver);
//  int (*activate)(struct pci_dev *dev, int driver_result, pnook_driver_t driver);
  int (*resume)(struct pci_dev *dev, int driver_result, pnook_driver_t driver);
  int (*suspend)(struct pci_dev *dev, u32 state, int driver_result, pnook_driver_t driver);
//  int (*deactivate)(struct pci_dev *dev, int driver_result, pnook_driver_t driver);
  int (*probe)(struct pci_dev *dev, const struct pci_device_id *id, int driver_result, pnook_driver_t driver);
} spci_operations_t, *pspci_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_spciops(pnook_driver_t driver, void * arg);
void shadow_del_spciops(pnook_driver_t driver);
extern struct spci_operations shadow_pciops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_spciops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_spciops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct snotifier_operations {
  int (*notifier_call)(struct notifier_block *self, unsigned long arg0, void * arg1, int driver_result, pnook_driver_t driver);
} snotifier_operations_t, *psnotifier_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_snotifierops(pnook_driver_t driver, void * arg);
void shadow_del_snotifierops(pnook_driver_t driver);
extern struct snotifier_operations shadow_notifierops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_snotifierops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_snotifierops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct sfs_operations {
  struct super_block * (*read_super)(struct super_block * arg0, void * arg1, int arg2, struct super_block * driver_result, pnook_driver_t driver);
} sfs_operations_t, *psfs_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_sfsops(pnook_driver_t driver, void * arg);
void shadow_del_sfsops(pnook_driver_t driver);
extern struct sfs_operations shadow_fsops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_sfsops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_sfsops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct sblk_operations {
  void (*request_fn)(request_queue_t *q, pnook_driver_t driver);
  int (*front_merge_fn)(request_queue_t * arg0, struct request * arg1, struct bio * arg2, int driver_result, pnook_driver_t driver);
  int (*back_merge_fn)(request_queue_t * arg0, struct request * arg1, struct bio * arg2, int driver_result, pnook_driver_t driver);
  int (*make_request_fn)(request_queue_t *q, struct bio *bio, int driver_result, pnook_driver_t driver);
  int (*merge_requests_fn)(request_queue_t * arg0, struct request * arg1, struct request * arg2, int driver_result, pnook_driver_t driver);
} sblk_operations_t, *psblk_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_sblkops(pnook_driver_t driver, void * arg);
void shadow_del_sblkops(pnook_driver_t driver);
extern struct sblk_operations shadow_blkops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_sblkops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_sblkops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct sgetblock_operations {
  int (*get_block)(struct inode *inode, sector_t iblock, struct buffer_head *bh_result, int create, int driver_result, pnook_driver_t driver);
} sgetblock_operations_t, *psgetblock_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_sgetblockops(pnook_driver_t driver, void * arg);
void shadow_del_sgetblockops(pnook_driver_t driver);
extern struct sgetblock_operations shadow_getblockops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_sgetblockops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_sgetblockops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct scallback_operations {
  int (*handler)(unsigned long arg, int driver_result, pnook_driver_t driver);
  int (*pm_handler)(struct pm_dev *dev, pm_request_t rqst, void *data, int driver_result, pnook_driver_t driver);
} scallback_operations_t, *pscallback_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_scallbackops(pnook_driver_t driver, void * arg);
void shadow_del_scallbackops(pnook_driver_t driver);
extern struct scallback_operations shadow_callbackops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_scallbackops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_scallbackops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct ssuper_operations {
  int (*sync_fs)(struct super_block *sb, int wait, int driver_result, pnook_driver_t driver);
  void (*write_super_lockfs)(struct super_block * arg0, pnook_driver_t driver);
  void (*clear_inode)(struct inode * arg0, pnook_driver_t driver);
  void (*delete_inode)(struct inode * arg0, pnook_driver_t driver);
  ssize_t (*quota_read)(struct super_block * arg0, int arg1, char * arg2, size_t arg3, loff_t arg4, ssize_t driver_result, pnook_driver_t driver);
  void (*unlockfs)(struct super_block * arg0, pnook_driver_t driver);
  int (*show_options)(struct seq_file * arg0, struct vfsmount * arg1, int driver_result, pnook_driver_t driver);
  void (*umount_begin)(struct vfsmount * arg0, int arg1, pnook_driver_t driver);
  void (*write_super)(struct super_block * arg0, pnook_driver_t driver);
  void (*put_inode)(struct inode * arg0, pnook_driver_t driver);
  void (*put_super)(struct super_block * arg0, pnook_driver_t driver);
  void (*drop_inode)(struct inode * arg0, pnook_driver_t driver);
  int (*write_inode)(struct inode * arg0, int arg1, int driver_result, pnook_driver_t driver);
  void (*dirty_inode)(struct inode * arg0, pnook_driver_t driver);
  int (*remount_fs)(struct super_block * arg0, int * arg1, char * arg2, int driver_result, pnook_driver_t driver);
  void (*read_inode)(struct inode * arg0, pnook_driver_t driver);
  void (*destroy_inode)(struct inode * arg0, pnook_driver_t driver);
  struct inode * (*alloc_inode)(struct super_block *sb, struct inode * driver_result, pnook_driver_t driver);
  int (*statfs)(struct dentry * arg0, struct kstatfs * arg1, int driver_result, pnook_driver_t driver);
  ssize_t (*quota_write)(struct super_block * arg0, int arg1, const char * arg2, size_t arg3, loff_t arg4, ssize_t driver_result, pnook_driver_t driver);
  int (*show_stats)(struct seq_file * arg0, struct vfsmount * arg1, int driver_result, pnook_driver_t driver);
} ssuper_operations_t, *pssuper_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_ssuperops(pnook_driver_t driver, void * arg);
void shadow_del_ssuperops(pnook_driver_t driver);
extern struct ssuper_operations shadow_superops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_ssuperops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_ssuperops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct sblock_device_operations {
  long (*unlocked_ioctl)(struct file * arg0, unsigned arg1, unsigned long arg2, long driver_result, pnook_driver_t driver);
  int (*media_changed)(struct gendisk * arg0, int driver_result, pnook_driver_t driver);
  int (*getgeo)(struct block_device * arg0, struct hd_geometry * arg1, int driver_result, pnook_driver_t driver);
  int (*revalidate_disk)(struct gendisk * arg0, int driver_result, pnook_driver_t driver);
  long (*compat_ioctl)(struct file * arg0, unsigned arg1, unsigned long arg2, long driver_result, pnook_driver_t driver);
  int (*ioctl)(struct inode * arg0, struct file * arg1, unsigned arg2, unsigned long arg3, int driver_result, pnook_driver_t driver);
  int (*open)(struct inode * arg0, struct file * arg1, int driver_result, pnook_driver_t driver);
  int (*release)(struct inode * arg0, struct file * arg1, int driver_result, pnook_driver_t driver);
  int (*direct_access)(struct block_device * arg0, sector_t arg1, unsigned long * arg2, int driver_result, pnook_driver_t driver);
} sblock_device_operations_t, *psblock_device_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_sbdops(pnook_driver_t driver, void * arg);
void shadow_del_sbdops(pnook_driver_t driver);
extern struct sblock_device_operations shadow_bdops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_sbdops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_sbdops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct svm_operations_struct {
  int (*page_mkwrite)(struct vm_area_struct *vma, struct page *page, int driver_result, pnook_driver_t driver);
  void (*close)(struct vm_area_struct * area, pnook_driver_t driver);
  pte_t (*zap_pte)(struct vm_area_struct *vma, unsigned long addr, pte_t *ptep, int is_fullmm, pte_t driver_result, pnook_driver_t driver);
  int (*populate)(struct vm_area_struct * area, unsigned long address, unsigned long len, pgprot_t prot, unsigned long pgoff, int nonblock, int driver_result, pnook_driver_t driver);
  void (*open)(struct vm_area_struct * area, pnook_driver_t driver);
  struct page * (*nopage)(struct vm_area_struct * area, unsigned long address, int *type, struct page * driver_result, pnook_driver_t driver);
} svm_operations_struct_t, *psvm_operations_struct_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_svmops(pnook_driver_t driver, void * arg);
void shadow_del_svmops(pnook_driver_t driver);
extern struct svm_operations_struct shadow_vmops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_svmops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_svmops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct stimer_operations {
  void (*handler)(unsigned long arg, pnook_driver_t driver);
} stimer_operations_t, *pstimer_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_stimerops(pnook_driver_t driver, void * arg);
void shadow_del_stimerops(pnook_driver_t driver);
extern struct stimer_operations shadow_timerops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_stimerops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_stimerops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct ssysctl_operations {
  int (*strategy)(ctl_table *table, int *name, int nlen, void *oldval, size_t *oldlenp, void *newval, size_t newlen, void **context, int driver_result, pnook_driver_t driver);
  int (*proc_handler)(ctl_table *ctl, int write, struct file * filp, void *buffer, size_t *lenp, loff_t *ppos, int driver_result, pnook_driver_t driver);
} ssysctl_operations_t, *pssysctl_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_ssysctlops(pnook_driver_t driver, void * arg);
void shadow_del_ssysctlops(pnook_driver_t driver);
extern struct ssysctl_operations shadow_sysctlops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_ssysctlops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_ssysctlops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
typedef struct sirq_operations {
  void (*handler)(int irq, void * dev_id, struct pt_regs * regs, pnook_driver_t driver);
} sirq_operations_t, *psirq_operations_t;

#ifdef CONFIG_NOOKS_SHADOW
int shadow_add_sirqops(pnook_driver_t driver, void * arg);
void shadow_del_sirqops(pnook_driver_t driver);
extern struct sirq_operations shadow_irqops;

#else // CONFIG_NOOKS_SHADOW
static inline int shadow_add_sirqops(pnook_driver_t driver, void * arg) { return 0;} 
static inline void shadow_del_sirqops(pnook_driver_t driver) {return;  };
#endif // CONFIG_NOOKS_SHADOW
