/*
 * netshadow-kops.c --  Shadow driver for network devices
 *
 * Copyright (C) 2003 Mike Swift
 */

#include "netshadow-i.h"
//#include "netshadowk-c-gen.h"
#include <../nooks/nooks-wrappers.h>
#include <linux/pci.h>

/**********************************************************
 * sknetops_pci_register_driver
 * 
 * normally: do nothing
 * fault mode: use stored pci information to make callback to driver
 *   instead of having PCI do it. 
 * Who stores
 * performs callback to register driver without asking pci first
 *
 *********************************************************/

int sknetops_pci_register_driver(struct pci_driver * arg0, int __kernel_result, pnook_driver_t __driver)
{
  pnook_t nook = __driver->nook;
  int result = __kernel_result;
  pshadow_driver_t shadow = __driver->nook->shadow;
  printk ("I CANNOT BE HERE - EVER FOR e1000.\n");
if (nook->state != nook_state_running) {
    pnook_driver_t driver;
    pshadow_pciops_info_t pci_info = NULL;

    driver = shadow_find_driver(nook, pciops_e, NULL);
    if (driver != NULL) {
      driver->shadow_state = shadow_state_running;
      result = nooks_find_in_hash(&shadow->kernel_objects,
				  pcidev_e,
				  arg0,
				  (void **) &pci_info );
      if (result >= 0) {
	result = arg0->probe(pci_info->dev,
			     &pci_info->u.pci_id);
	if (result >= 0) {
	  return(1);
	} else {
	  return(0);
	}
      }
    }
    result = pci_register_driver(arg0);

  }
    
  return(result);
}



/**********************************************************
 *
 * sknetops_request_nooks_irq
 *
 * requets irq without asking  first
 *
 *********************************************************/

int sknetops_request_nooks_irq(unsigned int irq, void (*handler) (int, void *, struct pt_regs *), unsigned long irqflags, const char *devname, void *dev_id, int __kernel_result, pnook_driver_t __driver)
{
  int result = __kernel_result;
  pnook_t nook = __driver->nook;
  pshadow_driver_t shadow = nook->shadow;
  struct irqaction * action = NULL;

  if (nook->state != nook_state_running) {
    pnook_driver_t driver;
    driver = shadow_find_driver(__driver->nook, irqops_e, (void *) irq);

    
    if (driver != NULL) {
      driver->shadow_state = shadow_state_running;
      result = nooks_find_in_hash(&shadow->kernel_objects,
			     irqaction_e,
			     (void *) irq,
			     (void **) &action );
      if (result >= 0) {
      
	action->dev_id = dev_id;
	return(result);
      }
    }
    //
    // Fall back to doing the routine itself
    //
    
    result = request_nooks_irq(irq, handler, irqflags, devname, dev_id);
     
  } 
  action = get_irq_action(irq,
			  dev_id);
  
  if (action != NULL) {
    //
    // Save the irq action so we can get it back
    //
    result = nooks_add_to_hash(&shadow->kernel_objects,
			       irqaction_e,
			       (void *) irq,
			       action );
  }
  

  return(result);
}


/**********************************************************
 *
 * sknetops_request_irq
 *
 * requests irq without asking kernel first
 *
 *********************************************************/

int sknetops_request_irq(unsigned int irq, void (*handler)(int, void *, struct pt_regs *), unsigned long irqflags, const char * devname, void * dev_id, int __kernel_result, pnook_driver_t __driver)
{
  int result = __kernel_result;
  pnook_t nook = __driver->nook;
  pshadow_driver_t shadow = nook->shadow;
  struct irqaction * action = NULL;

  if (nook->state != nook_state_running) {
    pnook_driver_t driver;

    driver = shadow_find_driver(nook, irqops_e, (void *) irq);

    if (driver != NULL) {
      driver->shadow_state = shadow_state_running;
      result = nooks_find_in_hash(&shadow->kernel_objects,
			     irqaction_e,
			     (void *) irq,
			     (void **) &action );
      if (result >= 0) {
      
	action->dev_id = dev_id;
	return(result);
      }
    }
    result = request_irq(irq, handler, irqflags, devname, dev_id);
  }
  
  action = get_irq_action(irq,
			  dev_id);
  
  if (action != NULL) {
    //
    // Save the irq action so we can get it back
    //
    result = nooks_add_to_hash(&shadow->kernel_objects,
			       irqaction_e,
			       (void *) irq,
			       action );
  }
  return(result);
}

/**********************************************************
 *
 * sknetops_free_irq
 *
 * 
 *
 *********************************************************/

void sknetops_free_irq(unsigned int irq, void * devid, pnook_driver_t __driver)
{
  struct irqaction * action = NULL;
  int result = 0;
  result = nooks_del_from_hash(&__driver->nook->shadow->kernel_objects,
			       irqaction_e,
			       (void *) irq,
			       (void *) &action);
  return;
}

/**********************************************************
 *
 * sknetops_request_region
 *
 * remember requested region and possibly return it
 *
 *********************************************************/

struct resource * sknetops___request_region(struct resource * parent, 
					    unsigned long start, 
					    unsigned long n, 
					    const char *name, 
					    struct resource * __kernel_result, 
					    pnook_driver_t __driver)
{
  struct resource * result = __kernel_result;
  pnook_t nook = __driver->nook;
  pshadow_driver_t shadow = nook->shadow;

  if (nook->state != nook_state_running) {
    int res;
    pnooks_resource_t resource = NULL;

    res = nooks_find_in_hash(&shadow->kernel_objects,
			     region_e,
			     (void *) start,
			     (void **) &resource);
    if (res >= 0) {
      if ((resource->parent == parent) &&
	  (resource->n == n)) {
	res = nooks_del_from_hash(&__driver->nook->shadow->kernel_objects,
			     region_e,
			     (void *) start,
			     (void **) &resource);
	return(resource->res);
      }
    }
    result = __request_region(parent, start, n, name);
  }
  return(result);
}

/**********************************************************
 *
 * sknetops___release_region
 *
 * release a region from memory
 *
 *********************************************************/

void sknetops___release_region(struct resource * parent, 
			       unsigned long start, 
			       unsigned long n, 
			       pnook_driver_t __driver)
{
  return;
}


int sknetops___check_region(struct resource * parent, unsigned long start, unsigned long n, int __kernel_result, pnook_driver_t __driver)
{
  int result = __kernel_result;
  pnook_t nook = __driver->nook;
  pshadow_driver_t shadow = nook->shadow;

  if (nook->state != nook_state_running) {
    int res;
    pnooks_resource_t resource = NULL;

    res = nooks_find_in_hash(&shadow->kernel_objects,
			     region_e,
			     (void *) start,
			     (void **) &resource);
    if (res >= 0) {
      if ((resource->parent == parent) &&
	  (resource->n == n)) {
	return(0);
      }
    }
    result = __check_region(parent, start, n);
  }

  return(result);

}

/**********************************************************
 *
 * sknetops_register_netdevice 
 *
 * 
 *
 *********************************************************/


int sknetops_register_netdevice(struct net_device *dev, int __kernel_result, pnook_driver_t __driver)
{
  int result = __kernel_result;
  pnook_t nook = __driver->nook;

  //printk ("***In sknetops_register_netdevice. ***\n");
  if (nook->state != nook_state_running) {

    pnook_driver_t driver;

    driver = shadow_find_driver(__driver->nook, netops_e, NULL); // BUGBUG

    if (driver != NULL) {
      driver->shadow_state = shadow_state_running;
      return(0);      
    }
    result = register_netdev(dev);
  } 
  return(result);
}

/**********************************************************
 *
 * sknetops_init_etherdev 
 *
 * 
 *
 *********************************************************/

struct net_device * sknetops_alloc_etherdev(int sizeof_priv, struct net_device * __kernel_result, pnook_driver_t __driver)
{
  struct net_device * result = __kernel_result;
  pnook_t nook = __driver->nook;

  if (nook->state != nook_state_running) {
      return(alloc_etherdev(sizeof_priv));
    }

  return(result);
}

/**********************************************************
 *
 * sknetops_register_netdev
 *
 * 
 *
 *********************************************************/


int sknetops_register_netdev(struct net_device *dev, int __kernel_result, pnook_driver_t __driver)
{
  int result = __kernel_result;
  pnook_t nook = __driver->nook;
  result = 0;
  //printk ("In sknetops_register_netdev.\n");
  if (nook->state != nook_state_running) {

    pnook_driver_t driver;

    if (strcmp(dev->name, "eth1") == 0)
	driver = shadow_find_driver(__driver->nook, netops_e, 1); 
    else
	driver = shadow_find_driver(__driver->nook, netops_e, NULL); // BUGBUG
    

    if (driver != NULL) {
      printk ("sknetops_register_netdev:shadow driver already exists.\n");
      driver->shadow_state = shadow_state_running;
      return(0);      
    }
    
    result = register_netdev(dev);
    printk ("register netdev skops returns with result %d.", result);
  }
  return(result);
}

int sknetops___pci_register_driver(struct pci_driver *dev, struct module * module, int __kernel_result, pnook_driver_t __driver)
{
  int result = __kernel_result;
  pnook_t nook = __driver->nook;
  pshadow_driver_t shadow = __driver->nook->shadow;
  //printk ("In sknetops_pci_register_driver.\n");
  if (nook->state != nook_state_running) {
    pshadow_pciops_info_t pci_info = NULL;
    pnook_driver_t driver;

    driver = shadow_find_driver(__driver->nook, pciops_e, NULL); // BUGBUG

    if (driver != NULL) {
      printk ("sknetops___pci_register_driver:shadow driver for pci already exists.\n");
      driver->shadow_state = shadow_state_running;
      result = nooks_find_in_hash(&shadow->kernel_objects,
                                  pcidev_e,
                                  dev,
                                  (void **) &pci_info );
      if (result >= 0) {
        result = dev->probe(pci_info->dev,
                             &pci_info->u.pci_id);
	printk ("Result is %d.\n", result);
     /* 	if (result >= 0) {
          	return(1);
	       } else {
        	  return(0);
	
          }
	*/
	 return (0);//hack for de4x5
        }
    }
   
    result = __pci_register_driver(dev, module);
    printk ("sknetops__pci register driver  returns with result %d.", result);
  }
  return(result);
}






void sknetops_unregister_netdev(struct net_device *dev, pnook_driver_t __driver)
{
  pnook_t nook = __driver->nook;

  if (nook->state != nook_state_running) {

    unregister_netdev(dev);
  }
  return;
}




int sknetops_register_chrdev(unsigned int arg0, const char * arg1, struct file_operations * arg2, int __kernel_result, pnook_driver_t __driver)
{
  int result = 0;
  pnook_t nook = __driver->nook;

  if (nook->state != nook_state_running) {
    return(result);
  } else {
    return(__kernel_result);
  }
}


int sknetops_pci_request_regions(struct pci_dev * arg0, char * arg1, int __kernel_result, pnook_driver_t __driver)
{
  int result = __kernel_result;
  pnook_t nook = __driver->nook;

  if (nook->state != nook_state_running) {
    //
    // Fake out success

    result = nooks_find_in_hash(&nook->shadow->kernel_objects,
				pciregion_e,
				arg0,
				NULL );
    if (result < 0) {
      result = pci_request_regions(arg0, arg1);
    }

    return(result);
  } else {
    return(__kernel_result);
  }
}

void sknetops_pci_release_regions(struct pci_dev * dev, pnook_driver_t __driver)
{
  pnook_t nook = __driver->nook;
  if (nook->state != nook_state_running) {
	
	return;
  }	  

  //pci_release_regions (dev);
  //
  // BUGBUG: what to do here?
  //
  return;
}

void sknetops_pci_enable_device(struct pci_dev * dev, pnook_driver_t __driver)
{
  //printk ("Inside sknetops_pci_enable_device.\n");
  pci_enable_device(dev);
  //
  // BUGBUG: what to do here?
  //
  return;
}

void sknetops_pci_set_master(struct pci_dev *dev, pnook_driver_t __driver)
{
  pnook_t nook = __driver->nook;
  //printk ("Inside sknetops_pci_set_master.\n");
  if (nook->state != nook_state_running) {
    //
    // Fake out success
    return;
   
  } else {
    return(pci_set_master(dev));
  }
}

void sknetops_pci_unregister_driver (struct pci_driver *dev, pnook_driver_t __driver)
{
	pnook_t nook = __driver->nook;
	printk ("sknetops_pci_unregister_driver called for device %s.\n", dev->name);
	if (nook->state != nook_state_running) {
		return;
	}
	else	{
		//DO NOTHING	
	}
}


sknetops_operations_t shadow_sknetops =  {
  __pci_register_driver	 : sknetops___pci_register_driver,
//  pci_alloc_consistent   : sknetops_pci_alloc_consistent,
//  pci_free_consistent    : sknetops_pci_free_consistent,
//  dma_alloc_coherent	 : sknetops_dma_alloc_coherent,
//  dma_free_coherent	 : sknetops_dma_free_coherent,
  pci_enable_device      : sknetops_pci_enable_device,
//  pci_find_class         : sknetops_pci_find_class,
  pci_set_master         : sknetops_pci_set_master,
  pci_request_regions    : sknetops_pci_request_regions,
  pci_release_regions    : sknetops_pci_release_regions,
  register_netdevice	 : sknetops_register_netdevice,
  free_irq	         : sknetops_free_irq,
//  alloc_etherdev	 : sknetops_alloc_etherdev,
  unregister_netdev	 : sknetops_unregister_netdev,
  request_nooks_irq	 : sknetops_request_nooks_irq,
  pci_unregister_driver	 : sknetops_pci_unregister_driver,
  register_netdev	 : sknetops_register_netdev,
//  request_irq   	 : sknetops_request_irq,
//  __request_region	 : sknetops___request_region,
//  __check_region	 : sknetops___check_region,

};

