/*
 * netshadow-recover.c --  Shadow driver recovery functions
 *
 * Copyright (C) 2004 Mike Swift
 */

#include "netshadow-i.h"
#include <linux/in.h>
#include <linux/if_arp.h>
#include <linux/timer.h>
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
#include "shadow-replug.h"
#endif
#ifdef CONFIG_NOOKS_PROC
#include "shadow-proc.h"
#endif

// #include "netshadowr-c-gen.h"

/*********************************/
#include <linux/ide.h>
struct hd_struct *part;
extern struct list_head vfsmntlist;
/*********************************/




/**********************************************************
 * driver cleanup routines
 * 
 *
 * 
 *
 *********************************************************/
int shadow_recover_irqops(pnook_t nook, pnook_driver_t driver, int pass)
{
  unsigned int irq = (int) driver->data;


  driver->shadow_state =  shadow_state_faulted;
  if (pass == NOOKS_RECOVERY_STOP_PASS) {
    int result;
    void * dev_id = NULL;
    
    result =  nooks_del_resource_elem(nook, 
				      irq_dev_e, 
				      (void *) irq, 
				      &dev_id);
    if (result >= 0) {
      //      irq_desc_t * desc = irq_desc + irq;
      //      free_irq(irq, dev_id);
      //
      // If there are any other actions, turn on polling
      //
      //      if (desc->action != NULL) {

      nooks_poll_device((int) irq);
      //      }
      //
      // BUGBUG: bogus hack for pcnet32
      //
      
      nooks_set_recovering_interrupt(irq);

    }
    return(-ENOTEMPTY);
      
  } else if (pass == NOOKS_RECOVERY_UNLOAD_PASS1) {
      shadow_add_driver(nook,
			(void * ) irq,
			driver );


    return(-ENOTEMPTY);
  } else if (pass >= NOOKS_RECOVERY_CLEANUP_PASS) {
    //    result =  nooks_del_resource_elem(nook, 
    //    				      irq_dev_e, 
    //    				      (void *) irq, 
    //    				      &dev_id);
    //    if (result >= 0) {
    //      free_irq(irq, dev_id);
    //    }
    
    //
    // Re-enable the irq (disabled by disable drivers)
    //
    
    nooks_unpoll_device(irq, 1);
    nooks_clear_recovering_interrupt(irq);
    return(-ENOTEMPTY);
  } else {
    return(-EAGAIN);
  }
  
}


int shadow_recover_pciops(pnook_t nook, pnook_driver_t driver, int pass)
{
  struct pci_driver * kernel_drv = NULL;

  driver->shadow_state =  shadow_state_faulted;
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }

  //
  // We keep the driver structure around. We'll use it later, and look it up
  // by the kernel_drv field
  //

  kernel_drv = (struct pci_driver *) driver->data;

  //
  // BUGBUG: what is key here?
  //
  shadow_add_driver(nook,
		    0,
		    driver );


  return(-ENOTEMPTY);
}


int shadow_recover_netops(pnook_t nook, pnook_driver_t driver, int pass)
{
  struct net_device * device;
  device = (struct net_device *) driver->data;

  //printk ("In shadow_recover_netops.\n");
  static int driver_count = 1;
  driver->shadow_state =  shadow_state_faulted;
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
      //
      // For net devices, stop the queue so no packets will be send.
      //
      netif_stop_queue(device);
      return(-EAGAIN);

  }    

  //
  // BUGBUG: what is key here?
  //
  if (pass == NOOKS_RECOVERY_UNLOAD_PASS2) {
   printk ("Adding shadow information of type %d to hash for device %p of count %d.\n",driver->driver_type, driver->data, driver_count); 
   shadow_add_driver(nook,
		      driver_count,  
		      driver);
   driver_count--;
  }
  
  return(-ENOTEMPTY);
}


int shadow_recover_mops(pnook_t nook, pnook_driver_t driver, int pass)
{
  struct module * module, * mod;
  int found = 0;
  int result = 0;
  void * module_data = NULL;
  void * bss_start = NULL;
  void * bss_end = NULL;

  //printk("shadow_recover_mops 1\n");

  driver->shadow_state =  shadow_state_faulted;
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS2) {
    return(-EAGAIN);
  }
  /*if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }*/
  module = (struct module *) driver->data;

  //printk("shadow_recover_mops 2\n");
  //
  // See if the module still exists
  //
//  lock_kernel();

  
  mod = (struct module *) find_module(module->name);
  if (mod != NULL)
        found = 1;
 // unlock_kernel(); 
  
  //printk("shadow_recover_mops 3\n");
  if (!found) {
    return(0);
  }

 #if 1 
  //
  // Find the module data
  //

  result = nooks_find_in_hash(&nook->shadow->kernel_objects,
			      module_data_e,
			      mod,
			      &module_data);
  if (result >= 0) {

    /* Copy the initalized data. */
    memcpy(mod->writeable_start,
	   module_data,
	   mod->writeable_end - mod->writeable_start);
  } else {
    // BUGBUG: what do we do here?
    //printk("In netshadow-recover: BUGBUGBUGBUG: couldn't find module\n");
  }

  /* Zero the bss. */
 

  result = nooks_find_in_hash(&nook->shadow->kernel_objects,
				bss_start_e,
				mod,
				&bss_start);
  if (result < 0)	{
	printk ("BSS Copy Failure.\n");
	goto BssEnd;
  }

  result = nooks_find_in_hash(&nook->shadow->kernel_objects,
                               bss_end_e,
                               mod,
                               &bss_end);

   if (result < 0)	{
	printk ("BSS Copy Failure(end).\n");
	goto BssEnd;
   }
	
   //printk ("Bss found during recovery from %p to %p.\n", bss_start, bss_end);
   memset (bss_start, 0, (bss_end - bss_start));
 
  #endif

BssEnd:
  //printk("shadow_recover_mops 4\n");



  shadow_add_driver(nook,
		    mod,
		    driver );


#ifdef CONFIG_SWIFI
  result = nooks_find_in_hash(&nook->shadow->kernel_objects,
			      module_code_e,
			      mod,
			      &module_data);
  33
  if (result >= 0) {
    memcpy((char *) mod + sizeof(struct module),
	   module_data,
	   mod->size - sizeof(struct module));
  }
  printk("shadow_recover_mops 5\n");
	   
#endif // CONFIG_SWIFI


  return(-ENOTEMPTY);

  
}

/**********************************************************
 *
 * shadow_recover_fops
 *
 * 
 *
 *********************************************************/

int shadow_recover_fops(pnook_t nook, pnook_driver_t driver, int pass)
{
    
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }

  //
  // This gets moved over during res_recover_fops
  return(-ENOTEMPTY);    
}


/**********************************************************
 * resource cleanup routines
 * 
 * add resource to kernel object hash
 * 
 *
 *********************************************************/

int shadow_res_recover_region(pnook_t nook, void * key, void * data, int pass)
{
  int result = 0;
  //  pnooks_resource_t resource = (pnooks_resource_t) data;

  if (pass >= NOOKS_RECOVERY_UNLOAD_PASS1) {
	
	/*
	  pnooks_resource_t kern_resource;
	  result = nooks_find_in_hash(nook,
	  region_e,
	  key, // same as resource->start
	  &kern_resource);*/
	
	result = nooks_add_to_hash(&nook->shadow->kernel_objects,
			       region_e,
			       key, // same as resource->start
			       data );
	
    return(-ENOTEMPTY);
  } else {
    return(-EAGAIN);
  }
}

int shadow_res_recover_pci_memory(pnook_t nook, void * key, void * data, int pass)
{
  if (pass >= NOOKS_RECOVERY_UNLOAD_PASS1) {
    nooks_mm_free_pages(nook, 
			(unsigned long) key, 
			get_order((unsigned long) data));
    return(0);
  }

  return(-EAGAIN);
}

int shadow_res_recover_netdev_stats(pnook_t nook, void * key, void * data, int pass)
{
  int result;
  if (pass >= NOOKS_RECOVERY_UNLOAD_PASS1) {
    result = nooks_add_to_hash(&nook->shadow->kernel_objects,
			       netdev_stats_e,
			       key,
			       data );
    return(-ENOTEMPTY);
  } else {
    return(-EAGAIN);
  }
}


int shadow_res_recover_file_ops(pnook_t nook, void * key, void * data, int pass)
{
  pnook_driver_t driver = (pnook_driver_t) data;
  struct file_operations * ops = (struct file_operations *) key;

  driver->shadow_state =  shadow_state_faulted;
  shadow_add_driver(nook, ops, driver);
  return(0);
}


int shadow_res_recover_pciregion(pnook_t nook, void * key, void * data, int pass)
{
  int result;
  if (pass <= NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }
  result = nooks_add_to_hash(&nook->shadow->kernel_objects,
			     pciregion_e,
			     key,
			     data );
  return(result);
}

int shadow_res_recover_vmalloc(pnook_t nook, void * key, void * data, int pass)
{
  int result;
  if (pass <= NOOKS_RECOVERY_UNLOAD_PASS1) {
    return(-EAGAIN);
  }
  result = nooks_add_to_hash(&nook->shadow->kernel_objects,
                             vmalloc_e,
                             key,
                             data );
  return(result);
}



int shadow_res_recover_dmamappage(pnook_t nook, void * key, void * data, int pass)
{
  int result;
  if (pass <= NOOKS_RECOVERY_UNLOAD_PASS1) {
	return (-EAGAIN);
  }

  result  = nooks_add_to_hash(&nook->shadow->kernel_objects,
                             dma_map_page_e,
                             key,
                             data);
  return(result);
}

int shadow_res_recover_dmamapsingle(pnook_t nook, void * key, void * data, int pass)
{
  int result;
  if (pass <= NOOKS_RECOVERY_UNLOAD_PASS1) {
        return (-EAGAIN);
  }

  result  = nooks_add_to_hash(&nook->shadow->kernel_objects,
                             dma_map_single_e,
                             key,
                             data);
  return(result);
}


int shadow_res_recover_dmamapalloc(pnook_t nook, void * key, void * data, int pass)
{
  int result;
  if (pass <= NOOKS_RECOVERY_UNLOAD_PASS1) {
        return (-EAGAIN);
  }

  result  = nooks_add_to_hash(&nook->shadow->kernel_objects,
                             dma_alloc_e,
                             key,
                             data);
  return(result);
}





int shadow_res_recover_skbuff_k(pnook_t nook, void * key, void * data, int pass)
{
  int result;
  
  //
  // We don't have to do anything - these are sk_buffs owned by the kernel.
  // Just free the socket buffer
  //

  result =  nooks_add_to_hash(&nook->shadow->kernel_objects,
			      skbuff_k_e,
			      key,
			      data);
  return(result);
}


/**********************************************************
 *
 * log replay routines 
 *
 * 
 *
 *********************************************************/

int shadow_extra_release = 1;

int
shadow_replay_ioctls(pnook_t nook) 
{
  pshadow_driver_t shadow = nook->shadow;
  int result = 0;
  nooks_hash_enum_t context = {0};
  pnooks_hash_entry_t entry = NULL;
  mm_segment_t fs;

  down(&shadow->recovery_lock);

  fs = get_fs();
  set_fs(KERNEL_DS);

  while (nooks_enumerate_in_hash(&shadow->kernel_objects,
				 &context,
				 &entry) == 0) {
      
      if (entry->entry_type == ioctl_e) {
	struct list_head * next, *safe;
	struct file * filp = (struct file *) entry->key;
	struct inode * inode = filp->f_dentry->d_inode;
	
	pshadow_ioctl_head_t head = (pshadow_ioctl_head_t) entry->data;

	//
	// Reset read offset
	//

	head->read_offset = head->bytes_read;


	//
	// BUGBUG: is this safe?
	//

	result = head->opener->wrapper_operations.fops.open(inode,filp);	
	if (result < 0) {
	  continue;
	}

	//
	// Replay the IOCTLs
	//
	list_for_each_safe(next, safe, &head->list) {
	  pshadow_ioctl_data_t data = list_entry(next,shadow_ioctl_data_t, next);
	  unsigned int ioctl_dir = 0;

	  ioctl_dir = _IOC_DIR(data->command);
	  //
	  // We only do IN and no-dir (command) type ioctl's
	  //

	  result = 0;
	  if (ioctl_dir & _IOC_WRITE) {
	    printk("Replaying data %p in ioctl command 0x%x, data 0x%lx\n",
		   data, data->command,   (unsigned long) data->data);
	    result = filp->f_op->ioctl(inode,
				 filp, 
				 data->command, 
				 (unsigned long) data->data);
	  } else if (ioctl_dir == 0) {
	    printk("Replaying data %p ioctl command 0x%x, data 0x%lx\n",
		   data, data->command,   (unsigned long) data->argument);
	    result = filp->f_op->ioctl(inode,
				       filp, 
				       data->command, 
				       data->argument);
	  } 	    
	  printk("ioctl result %d\n", result);
	  
	}
	
	if (shadow_extra_release) {
	  filp->f_op->release(inode, filp);

	result = head->opener->wrapper_operations.fops.open(inode,filp);	

	//
	// Replay the IOCTLs
	//
	list_for_each_safe(next, safe, &head->list) {
	  pshadow_ioctl_data_t data = list_entry(next,shadow_ioctl_data_t, next);
	  unsigned int ioctl_dir = 0;

	  ioctl_dir = _IOC_DIR(data->command);
	  //
	  // We only do IN and no-dir (command) type ioctl's
	  //

	  if (ioctl_dir & _IOC_WRITE) {
	    printk("Replaying data %p in ioctl command 0x%x, data 0x%lx\n",
		   data, data->command,   (unsigned long) data->data);
	    result = filp->f_op->ioctl(inode,
				 filp, 
				 data->command, 
				 (unsigned long) data->data);
	  } else if (ioctl_dir == 0) {
	    printk("Replaying data %p ioctl command 0x%x, data 0x%lx\n",
		   data, data->command,   (unsigned long) data->argument);
	    result = filp->f_op->ioctl(inode,
				       filp, 
				       data->command, 
				       data->argument);
	  }
	  printk("ioctl result %d\n", result);
	  
	}
	}
	
      }
  } 

  set_fs(fs);
  up(&shadow->recovery_lock);
  return(result);
}

/**********************************************************
 *
 * log replay routines 
 *
 * 
 *
 *********************************************************/

int
shadow_dump_ioctls(pnook_t nook) 
{
  pshadow_driver_t shadow = nook->shadow;
  nooks_hash_enum_t context = {0};
  pnooks_hash_entry_t entry = NULL;

  while (nooks_enumerate_in_hash(&shadow->kernel_objects,
				 &context,
				 &entry) == 0) {
      
      if (entry->entry_type == ioctl_e) {
	struct list_head * next, *safe;
	struct file * filp = (struct file *) entry->key;
	
	pshadow_ioctl_head_t head = (pshadow_ioctl_head_t) entry->data;

	//
	// Replay the IOCTLs
	//
	list_for_each_safe(next, safe, &head->list) {
	  pshadow_ioctl_data_t data = list_entry(next,shadow_ioctl_data_t, next);
	  unsigned int ioctl_dir = 0;

	  ioctl_dir = _IOC_DIR(data->command);
	  //
	  // We only do IN and no-dir (command) type ioctl's
	  //

	  printk("File %p:  ioctl dir %d: command 0x%x, data 0x%lx\n",
		 filp, ioctl_dir, data->command,  (ioctl_dir & _IOC_WRITE) ? (unsigned long) data->data : data->argument);
	  
	  
	}
	
	
      }
  } 

  return(0);
}



/**********************************************************
 *
 * nooks_mm_init_nook
 *
 * Initialize the slab cache for a nook
 *
 *********************************************************/

void
shadow_mm_init_nook(pnook_t nook) 
{
  int i = 0;
  size_t size = NOOKS_CACHE_FIRST_SIZE;
  
  NOOKS_ENTER();
  INIT_LIST_HEAD(&nook->resources.cache_list);
  INIT_LIST_HEAD(&nook->locals->cache_list);
  skb_queue_head_init(&nook->locals->skb_pool);
  skb_queue_head_init(&nook->locals->skb_rx_list);
  INIT_LIST_HEAD(&nook->locals->skb_free_list);


  memset(&nook->resources.cache_map, 0, sizeof(nooks_cache_map_t));
  memset(&nook->locals->cache_map, 0, sizeof(nooks_local_cache_map_t));

#if 0
  //
  // Set up the buckets for the slab allocator
  //

  for (i = 0; i < NOOKS_NUM_CACHE_SIZES - 1; i++) {
    nook->resources.cache_map.caches[i].cs_size = size;
    nook->locals->cache_map.caches[i].cs_size = size;
    size = size << 1;
  }
#endif
  nook->resources.cache_map.caches[i].cs_size = 0;
  nook->locals->cache_map.caches[i].cs_size = 0;
  
  NOOKS_EXIT();

}

/**********************************************************
 *
 * nooks_mm_cleanup_nook
 *
 * Clean up the memory management state of a nook
 *
 *********************************************************/

void
shadow_mm_cleanup_nook(pnook_t nook)
{
  int i;
  struct list_head * next, *spare;
  kmem_cache_t * temp;


  //printk("CLEANUP: Cleaning up shadow_mm_cleanup_nook.\n");
  //
  // Loop through  the kmalloc caches and free them
  //

#if 0
  We do not use cache map or caches. -- Asim
  for (i = 0; i < NOOKS_NUM_CACHE_SIZES; i++) {
    if ( nook->resources.cache_map.caches[i].cs_cachep != NULL) {
      kmem_cache_destroy(nook->resources.cache_map.caches[i].cs_cachep);
      nook->resources.cache_map.caches[i].cs_cachep = NULL;
    }
    if ( nook->resources.cache_map.caches[i].cs_dmacachep != NULL) {
      kmem_cache_destroy(nook->resources.cache_map.caches[i].cs_dmacachep);
      nook->resources.cache_map.caches[i].cs_dmacachep = NULL;
    }

  }

  //
  // Loop through the separately allocated caches and free them
  //
  list_for_each_safe(next, spare, &nook->resources.cache_list) {
    temp = list_entry(next,struct kmem_cache, next);
    kmem_cache_destroy(nook->resources.cache_map.caches[i].cs_dmacachep);
  }
#endif
  //
  // Clean up memory areas
  //

  //
  // 1. Free the vm area map
  //

  nooks_free_resource_hash(nook,
			   &nook->resources.area_map,
			   NOOKS_RECOVERY_CLEANUP_PASS);

  //
  // 2. Traverse the active skbs and free them now
  //

  nooks_free_resource_hash(nook,
			   &nook->resources.active_skbs,
			   NOOKS_RECOVERY_CLEANUP_PASS);

  //
  // 3. Free the work list -- N/A for us - Asim
  //

  //nooks_delete_work_list(nook);

}

long
shadow_clean_nook(pnook_t nook)
{
  struct list_head * next;
  struct list_head * spare;
  pnook_driver_t driver = NULL;
  
list_for_each_safe(next, spare, &nook->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_nook);
    
    if (driver->driver_type < last_ops_e) {
    
	nook->recovery_state.info.driver_recovery[driver->driver_type](nook, driver, NOOKS_RECOVERY_CLEANUP_PASS);
      
      //nooks_driver_recovery[driver->driver_type](nook, driver, 3);
      //
      // If this returns zero, we can free the driver
      //
      if ((driver->driver_type != mops_e)  && (driver->driver_type != netops_e) && (driver->driver_type != pciops_e))  //TBD Changed by Asim
	{
		printk("WARNING SOMETHING BEING RELEASED of type %d and netops is of type %d. \n", driver->driver_type, netops_e);
		nooks_free_driver(driver);
	}      

    }
  }  



  shadow_mm_cleanup_nook(nook);

  //
  // Reset all locals
  //

  memset(nook->locals, 0, PAGE_SIZE);

  spin_lock_init(&nook->locals->lock);
  INIT_LIST_HEAD(&nook->locals->free_work_list);

  nooks_mm_init_lookaside(&nook->locals->work_pool,
			  64,
			  sizeof(nooks_work_item_t),
			  NOOKS_LOOKASIDE_NOOK);

  nooks_mm_init_lookaside(&nook->locals->elem_pool,
			  20,
			  sizeof(nooks_list_elem_t),
			  NOOKS_LOOKASIDE_NOOK);

  nooks_mm_init_lookaside(&nook->locals->page_pool,
			  20,
			  0, 
			  NOOKS_LOOKASIDE_NOOK_PAGE);
  shadow_mm_init_nook(nook);

  return 0;

}

/**********************************************************
 *
 * management routines
 *
 * 
 *
 *********************************************************/

long
shadow_unload_nook(pnook_t nook) 
{
  long result = 0;


  result = nooks_do_unload_nook(nook);

  shadow_clean_nook(nook);

  nooks_reset_nook_state(nook);
  //printk ("CLEANUP COMPLETE. (shadow_unload_nook).\n");
  return(result);
}

/***************************************/
int shadow_recover_bdops(pnook_t nook, pnook_driver_t driver, int pass)
{
  //struct net_device * device;
  //device = (struct net_device *) driver->data;

  printk("Inside shadow_recover_bdops\n");
  
  driver->shadow_state =  shadow_state_faulted;
  if (pass < NOOKS_RECOVERY_UNLOAD_PASS1) {
      //
      // For net devices, stop the queue so no packets will be send.
      //
      //netif_stop_queue(device);
      return(-EAGAIN);

  }    

  //
  // BUGBUG: what is key here?
  //
  //if (pass == NOOKS_RECOVERY_UNLOAD_PASS2) {
  if (pass == NOOKS_RECOVERY_UNLOAD_PASS1) {
    	printk("Adding shadow bdops driver!!\n");
   	shadow_add_driver(nook,
			  (void *) driver->major,
			  driver );
  }
  
  return(-ENOTEMPTY);
}




/**************************************/

long
shadow_recover_netdev_nook(pnook_t nook) 
{
  long result = 0;
  //struct module * mod = NULL;

  //printk ("In netdev recovery function.\n");

  //mod = find_module ("e1000");

  nook_print_drivers(nook); 
  /* Releases all nook resources and deletes the nook. */
  result = shadow_unload_nook(nook);
  if (result < 0) {
    printk("failed to unload nook: %ld\n", result);
    goto Cleanup;
  }
 
  printk ("Nook released performing shadow recovery.\n");

  ENTER_SHADOW_RECOVERY();
  result = shadow_start_recovery(nook);
  if (result < 0) {
    printk("failed to start recovery: %ld\n", result);
    goto Cleanup;
  }


  //printk ("Performing netdev recovery now.\n");

 // result = shadow_netdev_recovery(nook);
  if (result < 0) {
    printk("failed netdev recovery: %ld\n", result);
    goto Cleanup;
  }

  //printk ("Performing finish recovery.\n");
  //result = shadow_finish_recovery(nook);
  if (result < 0) {
    printk("failed to start recovery(finish): %ld\n", result);
    goto Cleanup;
  }

  //
  // Reset the recovery function now that it worked
  //

  nook->recovery_state.info.recover = shadow_recover_netdev_nook;

 Cleanup:
  EXIT_SHADOW_RECOVERY();
  return(result);
}



long
shadow_recover_sound_nook(pnook_t nook) 
{
  long result = 0;

  result = shadow_unload_nook(nook);
  if (result < 0) {
    printk("failed to unload nook: %ld\n", result);
    goto Cleanup;
  }
  ENTER_SHADOW_RECOVERY();
  result = shadow_start_recovery(nook);
  if (result < 0) {
    printk("failed to start recovery: %ld\n", result);
    goto Cleanup;
  }

  result = shadow_sound_recovery(nook);
  if (result < 0) {
    printk("failed netdev recovery: %ld\n", result);
    goto Cleanup;
  }

  result = shadow_finish_recovery(nook);
  if (result < 0) {
    printk("failed to finish recovery: %ld\n", result);
    goto Cleanup;
  }

  //
  // Reset the recovery function now that it worked
  //

  nook->recovery_state.info.recover = shadow_recover_sound_nook;
  EXIT_SHADOW_RECOVERY();
 Cleanup:
  return(result);
}


int
shadow_build_recovery_block(pnook_t nook, nooks_recovery_fn * recovery_fn)
{
  int result = 0;
  nooks_driver_recovery_fn ** driver_fns = NULL;
  nooks_resource_recovery_fn ** resource_fns = NULL;
 
  //printk("Building shadow recovery block.\n");

  driver_fns = (nooks_driver_recovery_fn **) kmalloc(sizeof(nooks_driver_recovery_fn *) * (int) last_ops_e, GFP_KERNEL);
  if (driver_fns == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }

  memcpy(driver_fns, nook->recovery_state.info.driver_recovery, sizeof(nooks_driver_recovery_fn *) * (int) last_ops_e);

  driver_fns[irqops_e] = shadow_recover_irqops;
  driver_fns[pciops_e] = shadow_recover_pciops;
  driver_fns[netops_e] = shadow_recover_netops;
  driver_fns[mops_e]   = shadow_recover_mops;
  driver_fns[fops_e]   = shadow_recover_fops;
  driver_fns[bdops_e]  = shadow_recover_bdops;
  

  resource_fns = (nooks_resource_recovery_fn **) kmalloc(sizeof(nooks_resource_recovery_fn *) * (int) last_restype_e, GFP_KERNEL);
  if (resource_fns == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }

  memcpy(resource_fns, nook->recovery_state.info.resource_recovery, sizeof(nooks_resource_recovery_fn *) * (int) last_restype_e);

  
  resource_fns[region_e] = shadow_res_recover_region;
  resource_fns[pci_memory_e] = shadow_res_recover_pci_memory;
  resource_fns[netdev_stats_e] = shadow_res_recover_netdev_stats;
  resource_fns[file_ops_e] = shadow_res_recover_file_ops;
  resource_fns[pciregion_e] = shadow_res_recover_pciregion;
  resource_fns[skbuff_k_e] = shadow_res_recover_skbuff_k;
  resource_fns[dma_map_page_e] = shadow_res_recover_dmamappage;
  resource_fns[vmalloc_e] = shadow_res_recover_vmalloc;

  nook->recovery_state.info.driver_recovery = driver_fns;
  driver_fns = NULL;
  nook->recovery_state.info.resource_recovery = resource_fns;
  resource_fns = NULL;

  nook->recovery_state.info.recover = recovery_fn;

 Cleanup:
  if (driver_fns != NULL) {
    kfree(driver_fns);
  }

  if (resource_fns != NULL) {
    kfree(resource_fns);
  }
  return(result);
}


void
shadow_free_recovery_block(pnook_t nook)
{
  if (nook->recovery_state.info.driver_recovery != nooks_driver_recovery) {
    kfree(nook->recovery_state.info.driver_recovery);
    nook->recovery_state.info.driver_recovery = nooks_driver_recovery;
  }

  if (nook->recovery_state.info.resource_recovery != nooks_resource_recovery) {
    kfree(nook->recovery_state.info.resource_recovery);
    nook->recovery_state.info.resource_recovery = nooks_resource_recovery;
  }
  nook->recovery_state.info.recover = nooks_do_unload_nook;
}

#define SET_CURRENT_NOOK_DRIVER3(_p_) \
do { \
if ((_p_) != NULL) { \
   get_current()->nook_state.curr.current_driver  = (_p_);  \
   SET_CURRENT_NOOK((_p_)->nook); \
  } else {\
    SET_CURRENT_NOOK(NULL); \
  }  \
} while(0)


/**********************************************************
 *
 * trigger recovery by reloading drivers
 *
 * 
 *
 *********************************************************/

int
shadow_start_recovery(pnook_t nook)
{
  int result = 0;
  struct list_head * next;
  struct list_head * spare;
  pshadow_driver_t shadow = nook->shadow;
  pnook_driver_t driver = NULL;

  printk("In shadow_start_recovery.\n");
#ifdef CONFIG_NOOKS_PROC
  char *tmp;
 
  printk ("Processing proc information.\n"); 
  do_gettimeofday (&(nook->last_fault_time));
  nook->last_fault_reason = shadow_fault_type_to_string(nook->recovery_state.type);
  nook->last_fault_eip = nook->recovery_state.regs.eip;
  if (nook->last_fault_mod)
	  kfree (nook->last_fault_mod);
  tmp = shadow_proc_get_module_name(nook->last_fault_eip);
  nook->last_fault_mod = kmalloc(strlen(tmp)+1, GFP_KERNEL);
  if (nook->last_fault_mod)
	  strcpy(nook->last_fault_mod, tmp);
#endif

  //printk ("Post proc information here.\n");	
  list_for_each_safe(next, spare, &shadow->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_shadow);


    if (driver->driver_type == mops_e) {
      printk ("Module mops driver found for driver %s.\n", driver->name);
      struct module * module = (struct module *) driver->data;
 
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
      if (module->new_version) { /* there's a replugging module */
	  printk("shadow_start_recovery : replugging module\n");
	  result = replug_replace_module_in_list(module);
	  if (result >= 0) {
		  printk("shadow_start_recovery : replugging module succeed\n");
		  module = driver->data;
	  }
	  else if ((result != -EINVAL) && (result != -EBUSY) && (result != -ENOTSUPP)) {
		  printk("BUGBUGBUGBUG: unrecoverable error (%d) during recovery, nooks has gooonnnnnneeeee NUTS !\n", result);
		  goto Cleanup;
	  }
      }
#endif
      //
      // Move back to the nook
      //
      list_del(&driver->next_in_shadow);
      list_add(&driver->next_in_nook, &nook->drivers);
      printk ("Done till point with module as %p %p.\n",module,module->init);
       //nook_print_drivers(nook);
      if (module->init != NULL) {
	rtnl_lock();
	result = module->init();
	printk ("Done with init.\n");
	rtnl_unlock();
	if (result < 0) {
	  goto Cleanup;
	}
      }
      driver->shadow_state = shadow_state_running;
    }
  }
 Cleanup:
  return(result);
}

#ifdef CONFIG_NOOKS_SHADOW_REPLUG
void nooks_fault_start_recovery(void * u_nook);
#endif
	
int
shadow_finish_recovery(pnook_t nook)
{
  int result = 0;
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
  int compat = 1;
#endif
  struct list_head * next;
  struct list_head * spare;
  pshadow_driver_t shadow = nook->shadow;
  pnook_driver_t driver = NULL;
  list_for_each_safe(next, spare, &shadow->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_shadow);
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
    compat = compat && replug_compatibility_resource_used(nook, driver);
#endif
    list_del(&driver->next_in_shadow);
    driver->shadow_state = shadow_state_running;
    //printk ("Shadow running for device %s of data %p.\n", driver->name, driver->data);
  }

#ifdef CONFIG_NOOKS_SHADOW_REPLUG
  if (!compat) {
	nooks_fault_start_recovery((void*)nook);
	return -EFAULT;
  }
#endif

  nook->state = nook_state_running;
  //nook_print_drivers(nook);
  printk ("Nook is backup and running with result %d.\n", result);
  //
  // Wake up anyone who blocked for recovery
  //

  wake_up_all(&shadow->wait_for_recovery);
  return(result);
}


int shadow_delay_irq_recovery = 0;

int 
shadow_irq_recovery(pnook_t nook)
{
  int result = 0;
  struct list_head * next;
  struct list_head * spare;
  pshadow_driver_t shadow = nook->shadow;
  pnook_driver_t driver = NULL;
  //
  // Reenable interrupts (stop polling)
  //

  list_for_each_safe(next, spare, &shadow->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_shadow);


    if (driver->driver_type == irqops_e) {
      unsigned int irq = (unsigned int) driver->data;

      //
      // Allow external interrupts into the real driver
      //

      driver->shadow_state = shadow_state_running;
      list_add(&driver->next_in_nook, &nook->drivers);


      //
      // BUGBUG: test code
      //

      if (shadow_delay_irq_recovery) {
	if (irq != 10) {
	  nooks_clear_recovering_interrupt(irq);
	  nooks_unpoll_device(irq, 0);
	}
      } else {
	nooks_clear_recovering_interrupt(irq);
	nooks_unpoll_device(irq, 0);
      }
    }
  }

  return(result);
}



int
shadow_replay_skbuffs(pnook_t nook) 
{
  pshadow_driver_t shadow = nook->shadow;
  int result = 0;
  nooks_hash_enum_t context = {0};
  pnooks_hash_entry_t entry = NULL;


  while (nooks_enumerate_in_hash(&shadow->kernel_objects,
				 &context,
				 &entry) == 0) {
      
    if (entry->entry_type == skbuff_k_e) {
      struct sk_buff * skb = (struct sk_buff *) entry->key;
      printk ("Replaying skbuffs.\n");      
      if (skb->dev != NULL) {
	result = skb->dev->hard_start_xmit(skb, skb->dev);
      }
    }
  } 
  return(result);
}

void nooks_send_fake_arp(struct net_device *dev)
{
#ifdef CONFIG_INET
        struct sk_buff *skb;
        u32             src_ip, dst_ip;

        dst_ip = INADDR_BROADCAST;
        src_ip = inet_select_addr(dev, dst_ip, RT_SCOPE_LINK);

        /* No IP? Then nothing to do. */
        if (src_ip == 0)	{
                return;
	}

	//do
	{
	        skb = arp_create(ARPOP_REPLY, ETH_P_ARP,
        	                 dst_ip, dev, src_ip,
                	         /*dst_hw*/ NULL, /*src_hw*/ NULL,
                        	 /*target_hw*/ dev->dev_addr);
	        if (skb == NULL)
        	        return;
		//printk ("Sending fake arp in loop.\n");	
		//printk ("Fake arp sent as %d.\n", dev_queue_xmit(skb));
	}
	//while (dev_queue_xmit(skb) != 0);
#endif
}

void thread_nooks_send_fake_arp (struct net_device * dev)	
{
  //printk ("Spawned new thread to send arp.\n");
  while (netif_carrier_ok(dev))
	;
  //printk ("netifcarrier ok returns %d.\n", netif_carrier_ok(dev));
  while (!netif_running(dev))
  	;
  netif_wake_queue(dev);
  //netif_carrier_on(dev);
  //dev_watchdog(dev);
  //dev_init_scheduler(dev);
  //rtnl_lock();
  //netdev_state_change(dev);
  //rtnl_unlock();
  nooks_send_fake_arp(dev);
}

#ifdef NOOKS_MIGRATE
void arp_timer_function(unsigned long arg)	{

  struct timer_list * tmr = (struct timer *) arg;
  int result;
  result = shadow_ioctl("nook1", 13000, 0);
  if (result != 0)
	mod_timer(tmr, jiffies + 2 * HZ); 
  else
	kfree(tmr);
}

#endif
int 
shadow_netdev_recovery(pnook_t nook)
{
  int result = 0;
  struct list_head * next;
  struct list_head * spare;
  pshadow_driver_t shadow = nook->shadow;
  pnook_driver_t driver = NULL;
  struct net_device *ndev = NULL;
  struct timer_list *arp_timer = NULL;
#if 0
  arp_timer = (struct timer_list *) kzalloc(sizeof(struct timer_list), GFP_KERNEL);

  if (arp_timer != NULL)
  {
  	init_timer(arp_timer);
	arp_timer->function = &arp_timer_function;
	arp_timer->data = arp_timer;
  }
  else
	printk ("arp_timer memory allocation failed.\n");i
#endif
  //
  // Now need to call post-init routines: dev->open
  //

  down(&shadow->recovery_lock);

  list_for_each_safe(next, spare, &shadow->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_shadow);

   if (driver->driver_type == pciops_e) {
        //printk ("PCI driver found.\n");
	struct pci_driver * pci_dev = (struct pci_driver *)driver->data;
 	//printk ("PCI Device in shadow_netdev_recovery is %p.\n", driver->data);
      //
      // Move back to the nook
      //
      // list_del(&driver->next_in_shadow);
      list_add(&driver->next_in_nook, &nook->drivers);

      driver->shadow_state = shadow_state_recovering;

    }


    if (driver->driver_type == netops_e) {
      struct net_device * netdev = (struct net_device *) driver->data;
      //printk ("Network Device in shadow_netdev_recovery is %p.\n", driver->data);
      //
      // Move back to the nook
      //
      // list_del(&driver->next_in_shadow);
      list_add(&driver->next_in_nook, &nook->drivers);

      driver->shadow_state = shadow_state_recovering;
      //if (strcmp(driver->name,"eth%d") == 0)
	//	continue;
	

      if (netdev->open != NULL) {
	SET_CURRENT_NOOK(nook);
	result = netdev->open(netdev);
		if (result < 0) {
		  goto Cleanup;
		}
		//printk ("Done with open.\n");
 		//goto Cleanup;	
	
		netdev->flags |= IFF_UP;
		if (netdev->set_multicast_list != NULL)
			netdev->set_multicast_list(netdev);
		netdev->reg_state = NETREG_REGISTERED;
		//printk ("Activating netdev.\n");
		dev_activate(netdev);
		netif_schedule(netdev);
		netif_carrier_on(netdev);
		printk ("Done with bringing up the interfaces.\n");
		//printk("Sending fake arp.\n");
		//thread_nooks_send_fake_arp(netdev);
		//nooks_send_fake_arp(netdev);
		ndev = netdev;
      } // end open 
      
    } //end if
  }

  //shadow_replay_skbuffs(nook);
  shadow_irq_recovery(nook);
  //if (ndev != NULL)
	//thread_nooks_send_fake_arp(ndev);

  //mod_timer(arp_timer, jiffies); 

 Cleanup:
  up(&shadow->recovery_lock);
  return(result);
}

/**********************************************************
 *
 * shadow_sound_recovery
 *
 * 
 *
 *********************************************************/

int 
shadow_sound_recovery(pnook_t nook)
{
  int result = 0;
  struct list_head * next;
  struct list_head * spare;
  pshadow_driver_t shadow = nook->shadow;
  pnook_driver_t driver = NULL;



  //
  // Now need to call post-init routines: set to recovering
  //

  list_for_each_safe(next, spare, &shadow->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_shadow);


    if (driver->driver_type == fops_e) {
      list_add(&driver->next_in_nook, &nook->drivers);
      driver->shadow_state = shadow_state_recovering;
    }
  }


  //
  // Walk through IOCTLs
  //

  result = shadow_replay_ioctls(nook);
  if (result < 0) {
    goto Cleanup;
  }

  //
  // Reenable interrupts (stop polling)
  //

  shadow_irq_recovery(nook);
  
 Cleanup:
  return(result);
}

/************************************************/

int 
shadow_blkdev_recovery(pnook_t nook)
{
  int result = 0;
  struct list_head * next;
  struct list_head * spare;
  pshadow_driver_t shadow = nook->shadow;
  pnook_driver_t driver = NULL;


  //
  // Now need to call post-init routines: dev->open
  //

  list_for_each_safe(next, spare, &shadow->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_shadow);


    if (driver->driver_type == bdops_e) {
      //struct net_device * netdev = (struct net_device *) driver->data;

      //
      // Move back to the nook
      //
      // list_del(&driver->next_in_shadow);
      printk("Returning driver to nooks!\n");
      list_add(&driver->next_in_nook, &nook->drivers);

      driver->shadow_state = shadow_state_recovering;

      /*if (netdev->open != NULL) {
	result = netdev->open(netdev);
	if (result < 0) {
	  goto Cleanup;
	}
	netif_schedule(netdev);
      }*/
    }
  }
  shadow_irq_recovery(nook);
  return(result);


}

#if 0
long shadow_blkdev_unlog(pnook_t nook)
{
	struct gendisk *gp, *nook_gp;
	int result, result2, error = 0, major, *nook_major;
	struct nameidata nd;
  	struct list_head *next;
	char *dev_name = NULL;
	struct inode *bdev_inode;
	struct file fake_file={};
	struct dentry fake_dentry={};
	pnook_driver_t driver=NULL;

	printk("Going to walk the path\n");
	
	result = nooks_del_resource_elem(nook, major_e, 0,
					(void **) &nook_major);

	if(result<0)
	{
		printk("Failed in finding major in shadow_blkdev_unlog\n");
		return -1;
	}	
	
	printk("MAJOR FOUND %d\n", *nook_major);	
	
	major = *nook_major;		
	
	list_for_each(next, &vfsmntlist) 
    	{
    		struct vfsmount *tmp = list_entry(next, struct vfsmount, 
							mnt_list);
    		struct super_block * s = tmp->mnt_sb;
		
	    	if (MAJOR(s->s_dev)==major) 
		{
			printk("\n\n/*** Found ide driver ***/\n\n");  
			printk("Name: %s \n", tmp->mnt_devname);
	    		dev_name=tmp->mnt_devname;
		}
	    	else
		{
			continue;
		}
	    	
    	} 
  

	if(path_init(dev_name, LOOKUP_FOLLOW|LOOKUP_POSITIVE, &nd))
		error=path_walk(dev_name, &nd);

	if(error)
	{
		printk("ERROR IN PATH WALK\n");
		return -1;
	}

	if(!nd.dentry->d_inode->i_bdev)
		printk("WE NEED BD_ACQUIRE\n");		
	else
		printk("WE DONT NEED BD_ACQUIRE\n");		
	
	bdev_inode = nd.dentry->d_inode->i_bdev->bd_inode;		
	printk("Result of Walk Major: %d\n", major); 

	path_release(&nd);
	
	printk("Finished walking the path\n");

	driver = nooks_find_driver_in_nook(major, nook);
	
	if(driver==NULL)
	{
		printk("DRIVER NOT FOUND!!\n");	
		return -1;
	}
	
	printk("DRIVER FOUND!!\n");

	fake_dentry.d_inode=bdev_inode;
	fake_file.f_dentry=&fake_dentry;
	
	driver->wrapper_operations.bdops.open(bdev_inode, &fake_file);

	printk("Executing shadow_blkdev_unlog\n");	
	gp=get_gendisk(major<<MINORBITS);
	result2 = nooks_find_resource_elem_reverse(nook, 
						   gendisk_e, 
						   gp, 
						   (void **) &nook_gp); 

	if(result2>=0)
	{
		memcpy(&nook_gp->part[4], part, sizeof(struct hd_struct));	
		memcpy(&gp->part[4], part, sizeof(struct hd_struct));	
	
	}
	else
		printk("GENDISK NOT FOUND2\n");
	return 0;

}

long shadow_blkdev_log(pnook_t nook)
{
	struct gendisk *gp;
	struct gendisk *nook_gp;
	int result2;

	printk("Executing shadow_blkdev_log\n");	
	gp=get_gendisk(3<<MINORBITS);
	result2 = nooks_find_resource_elem_reverse(nook, 
						   gendisk_e, 
						   gp, 
						   (void **) &nook_gp); 

	if(result2>=0)
	{
		part = (struct hd_struct *)kmalloc(sizeof(struct hd_struct),GFP_KERNEL);
		memcpy(part, &nook_gp->part[4], sizeof(struct hd_struct));	
	}
	else
		printk("GENDISK NOT FOUND\n");

	return 0;

}

#endif
#if 0
long
shadow_recover_blkdev_nook(pnook_t nook) 
{
  long result = 0;

  printk("Recovery for block device triggered!!!\n");
  
  result = shadow_blkdev_log(nook);
  if (result < 0) {
    printk("failed to log nook: %ld\n", result);
    goto Cleanup;
  }
  result = shadow_unload_nook(nook);
  if (result < 0) {
    printk("failed to unload nook: %ld\n", result);
    goto Cleanup;
  }
  ENTER_SHADOW_RECOVERY();
  result = shadow_start_recovery(nook);
  if (result < 0) {
    printk("failed to start recovery: %ld\n", result);
    goto Cleanup;
  }
  result = shadow_blkdev_unlog(nook);
  if (result < 0) {
    printk("failed to unlog nook: %ld\n", result);
    goto Cleanup;
  }

  result = shadow_blkdev_recovery(nook);
  if (result < 0) {
    printk("failed netdev recovery: %ld\n", result);
    goto Cleanup;
  }
  result = shadow_finish_recovery(nook);
  if (result < 0) {
    printk("failed to start recovery: %ld\n", result);
    goto Cleanup;
  }

  //
  // Reset the recovery function now that it worked
  //

  nook->recovery_state.info.recover = shadow_recover_blkdev_nook;

 Cleanup:
  EXIT_SHADOW_RECOVERY();
  return(result);
  
}

#endif

