#ifndef _LINUX_NETSHADOW_H
#define _LINUX_NETSHADOW_H

#include "netshadowk-h-gen.h"
#include "netshadow-h-gen.h"

// #define NETSHADOW_DEBUG 1
#undef NETSHADOW_DEBUG


typedef union nooks_shadow_operations {
  void * sdummy_ops;
  struct smodule_operations * smops;
  struct sfile_operations * sfops;
  struct sblock_device_operations * sbdops;
  struct svm_operations_struct * svmops;
  struct sblk_operations * sblkops;
  struct sprocfs_operations * sprocops;
  struct snetdev_operations * snetops;
  struct sirq_operations * sirqops;
  struct spci_operations * spciops;
  struct scallback_operations * scallbackops;
  struct stimer_operations * stimerops;
  struct sinode_operations * sinodeops;
  struct ssysctl_operations * ssysctlops;
  struct sactor_operations * sactorops;
  struct saddress_space_operations * sasops;
  struct sdentry_operations * sdentryops;
  struct ssuper_operations * ssuperops;
  struct sgetblock_operations * sgetblockops;
  struct sfs_operations * sfsops;
  struct snotifier_operations * snotifierops;
  struct block_device_operations * bdops;
} nooks_shadow_operations_t, * pnooks_shadow_operations_t;

typedef union nooks_shadow_kernel_operations {
  void * skdummy_ops;
  sknetops_operations_t * sknetops;
} nooks_shadow_kernel_operations_t, * pnooks_shadow_kernel_operations_t;


typedef struct nooks_shadow_info {
  pnook_driver_t driver;
} nooks_shadow_info_t, *pnooks_shadow_info_t;



#ifdef NETSHADOW_DEBUG
void
shadow_log(pnook_driver_t driver, char * routine, int result);

#else 
static inline void
shadow_log(pnook_driver_t driver, char * routine, int result)
{
  return; 
}
#endif


#define NETSHADOW_IOCTL_TRIGGER_RECOVERY    10000
#define NETSHADOW_IOCTL_SET_SOUND_DRIVER    10004
#define NETSHADOW_IOCTL_SET_NETDEV_DRIVER   10005
#define NETSHADOW_IOCTL_SET_BLOCK_DRIVER    10007

#define NETSHADOW_IOCTL_DUMP_LOG            11000
#define NETSHADOW_IOCTL_CLEAR_LOG           11001
#define NETSHADOW_IOCTL_PRINT_DRIVERS       11002
#define NETSHADOW_IOCTL_DUMP_IOCTLS         11003

#define NETSHADOW_IOCTL_UNLOAD_NOOK         12000
#define NETSHADOW_IOCTL_START_RECOVERY      12001
#define NETSHADOW_IOCTL_FINISH_RECOVERY     12002
#define NETSHADOW_IOCTL_NETDEV_RECOVERY     12011
#define NETSHADOW_IOCTL_SOUND_RECOVERY      12012

#define NETSHADOW_SEND_ARP		    13000
#define NETSHADOW_INJECT_FAULT      14000

#define NETSHADOW_TASK_STATE_BLOCK          0x1
#define NETSHADOW_TASK_STATE_NOBLOCK        0x2


#define TASK_SHADOW_STATE(_t_) ((_t_)->nook_state.shadow_flags)
#define TASK_SHADOW_BLOCK_RETRY(_t_) ((TASK_SHADOW_STATE(_t_) & NETSHADOW_TASK_STATE_BLOCK) != 0)
#define SHADOW_BLOCK_RETRY() (TASK_SHADOW_BLOCK_RETRY(get_current()))
#define SET_SHADOW_BLOCK_RETRY() (TASK_SHADOW_STATE(get_current()) |= NETSHADOW_TASK_STATE_BLOCK)
#define DRIVER_RUNNING(_d_) (((_d_)->nook->state == nook_state_running) || ((_d_)->shadow_state == shadow_state_running))

typedef enum shadow_state {
  shadow_state_init = 1,
  shadow_state_running,
  shadow_state_faulted,
  shadow_state_recovering,
  shadow_state_cleanup
} shadow_state_t, *pshadow_state_t;

typedef struct shadow_log_entry {
  struct list_head next;
  unsigned long timestamp;
  char * name;
  int result;
} shadow_log_entry_t, *pshadow_log_entry_t;


typedef struct shadow_log {
  struct list_head log;
  unsigned int entries;
} shadow_log_t, *pshadow_log_t;
//
// Stored as pci_dev_e, indexed by driver name
//

typedef struct shadow_pciops_info {
  struct pci_driver * driver;
  struct pci_dev * dev;
  union {
    struct pci_device_id pci_id;
    struct isapnp_device_id isapnp_id;
  } u;
  int recover;
} shadow_pciops_info_t, *pshadow_pciops_info_t;

typedef struct shadow_netops_info {
  char name[IFNAMSIZ];
  int recover;
} shadow_netops_info_t, *pshadow_netops_info_t;

typedef struct shadow_irqops_info {
  unsigned int irq;
  unsigned int irqflags;
  
  int recover;
} shadow_irqops_info_t, *pshadow_irqops_info_t;



typedef struct shadow_driver_info {
  shadow_pciops_info_t pciops_info;
  shadow_netops_info_t netops_info;
  shadow_irqops_info_t irqops_info;
} shadow_driver_info_t, *pshadow_driver_into_t;

typedef struct shadow_region_info {
  struct list_head next;
  unsigned long start;
  unsigned long end;
  struct resource *parent;
  int recover;
} shadow_region_info_t, *pshadow_region_info_t;

typedef struct shadow_pci_memory_info {
  struct list_head next;
  unsigned long start;
  unsigned long end;
  struct resource *parent;
  int recover;
} shadow_pci_memory_info_t, *pshadow_pci_memory_info_t;

typedef struct shadow_resource_info {
  pshadow_region_info_t region_info;
  pshadow_pci_memory_info_t pci_memory_info;
} shadow_resource_info_t, *pshadow_resource_info_t;

typedef struct shadow_driver {
  struct list_head next;
  pnook_t nook;
  shadow_log_t log;
  //  shadow_state_t state;
  //  shadow_driver_info_t driver_info;
  //  shadow_resource_info_t resource_info;
  nooks_hash_table_t kernel_objects;
  nooks_hash_table_t kernel_drivers;
  struct list_head drivers;
  wait_queue_head_t wait_for_recovery;
  struct semaphore recovery_lock;
} shadow_driver_t, *pshadow_driver_t;



int
shadow_build_recovery_block(pnook_t nook, nooks_recovery_fn * recovery_fn);

void
shadow_free_recovery_block(pnook_t nook);


int
shadow_add_driver(pnook_t nook,
		  void * key, 
		  pnook_driver_t driver );

pnook_driver_t 
shadow_get_driver(pnook_t nook, 
		  nooks_iwrapper_type_t driver_type,
		  void * key);

#ifdef CONFIG_NOOKS_SHADOW_REPLUG
extern int replug_can_recover_again(pnook_t nook); //Could be FIXME Loading problems
#else
static inline int replug_can_recover_again(pnook_t nook)
{
	return 0;
}
#endif

#ifdef CONFIG_NOOKS_SHADOW
/*

int nooks_store_dma_info(pnook_t nook, struct device *dev, 
                            dma_addr_t dma_handle, 
			    size_t size,
			    enum dma_data_direction direction);
int nooks_transfer_dma_to_cpu(void);
int nooks_transfer_dma_to_device(void);
*/

long
shadow_ioctl(char * name, unsigned int cmd, unsigned long arg);


void shadow_del_interface(pnook_driver_t driver);

int shadow_add_interface(pnook_driver_t driver);


pnook_driver_t 
shadow_find_driver(pnook_t nook, 
		   nooks_iwrapper_type_t driver_type,
		   void * key);

pnook_driver_t
shadow_find_driver2(pnook_t nook,
		    nooks_iwrapper_type_t driver_type,
		    unsigned int key,
		    char *key2);

int
shadow_init_shadow_module(struct module * module, pnook_driver_t __driver);


#else // CONFIG_NOOKS_SHADOW (not defined)
static inline long
shadow_ioctl(char * name, unsigned int cmd, unsigned long arg) 
{
  return(0);
}

static inline void shadow_del_interface(pnook_driver_t driver)
{
  return;
}

static inline int shadow_add_interface(pnook_driver_t driver)
{
  return(0);
}


static inline pnook_driver_t 
shadow_find_driver(pnook_t nook, 
		   nooks_iwrapper_type_t driver_type,
		   void * key) 
{
  return(NULL);
}

static inline pnook_driver_t
shadow_find_driver2(pnook_t nook,
		    nooks_iwrapper_type_t driver_type,
		    unsigned int key,
		    char *key2)
{
	return NULL;
}

static inline int shadow_init_shadow_module(struct module * module, pnook_driver_t __driver)
{
  return(0);
}
#endif // CONFIG_NOOKS_SHADOW

int
shadow_start_recovery(pnook_t nook);

long
shadow_unload_nook(pnook_t nook);

int
shadow_finish_recovery(pnook_t nook);

int
shadow_netdev_recovery(pnook_t nook);

int
shadow_sound_recovery(pnook_t nook);

int
shadow_dump_ioctls(pnook_t nook);

typedef struct shadow_ioctl_head {
  struct list_head next;
  struct list_head list;
  pnook_driver_t opener;
  struct inode * inode;
  int samples_per_second;
  int sample_size;
  int channels;
  int bytes_read;
  int read_offset;
} shadow_ioctl_head_t, *pshadow_ioctl_head_t;

typedef struct shadow_ioctl_data {
  struct list_head next;
  unsigned int command;
  unsigned long argument;
  uint8_t data[0];
} shadow_ioctl_data_t, *pshadow_ioctl_data_t;


long
shadow_recover_netdev_nook(pnook_t nook);

long
shadow_recover_sound_nook(pnook_t nook);
/*
long
shadow_recover_blkdev_nook(pnook_t nook);
*/
#define ENTER_SHADOW_RECOVERY() \
do { \
get_current()->nook_state.curr.flags |= NOOK_STATE_SHADOW_RECOVERY; } while (0)

#define EXIT_SHADOW_RECOVERY() \
do { \
get_current()->nook_state.curr.flags &= ~NOOK_STATE_SHADOW_RECOVERY; } while (0)

#endif // LINUX_NETSHADOW_H
