/*
 * shadow-replug.c - functions for doing driver replugging
 *
 * 2004 - Damien Martin-Guillerez
 * work under Mike Swift & Hank Levy
 */

#include "shadow-replug.h"
#include <../nooks/nooks-iwrappers.h>
#include <../nooks/nooks-nwrappers.h>

extern struct module kernel_module;

int replug_can_recover_again(pnook_t nook)
{
	struct module *mod;
	pnook_driver_t driver;
	struct list_head *next;
	NOOKS_ENTER();
	list_for_each(next, &(nook->drivers)) {
		driver = list_entry(next, nook_driver_t, next_in_nook);
		if (driver->driver_type == mops_e) {
			mod = (struct module*)driver->data;
			if (mod->mod_nook_replug_flag & MOD_NOOK_REPLUG_OLD)
				return 0;
			if (mod->new_version)
				return 1;
		}
	}
	return 0;
}

void replug_delete_old_version(pnook_t nook)
{
	struct module *orig_mod, *mod;
	pnook_driver_t driver;
	struct list_head *next;
	
	NOOKS_ENTER();

	list_for_each(next, &(nook->drivers)) {
		driver = list_entry(next, nook_driver_t, next_in_nook);
		if (driver->driver_type == mops_e) {
			orig_mod = (struct module*)driver->data;
			mod = orig_mod->new_version;
			if (mod && (orig_mod->mod_nook_replug_flag & MOD_NOOK_REPLUG_OLD)) {
				orig_mod->mod_nook_replug_flag &= ~MOD_NOOK_REPLUG_OLD;
				vfree(mod);
				orig_mod->new_version = NULL;
			}
		}
	}
}

int replug_replace_op(struct module *mod, struct module *old_mod)
{
  struct list_head *next;
  void *data;
  pnook_driver_t driver;
  int result;
  
  NOOKS_ENTER();
  
  list_for_each(next, &(mod->nook->shadow->drivers)) {
	driver = list_entry(next, nook_driver_t, next_in_shadow);
	
	switch(driver->driver_type) {
	    case mops_e:
	      if (!strcmp(mod->name, driver->name)) {
		nooks_iwrapper_copy_mops(mod, &driver->operations.mops, 
				driver->iwrapper_operations.mops, mod);
		driver->data = mod;
	/*	
		if (mod_member_present(mod, writeable_end) 
				&& (mod->writeable_end != 0)) {
			nooks_mm_update_pagetable(driver->nook, NOOKS_WRITEABLE, 
				(unsigned long) mod->writeable_start, 
				(mod->writeable_end - mod->writeable_start
				   + PAGE_SIZE - 1) & PAGE_MASK);
		}
*/		
		result = shadow_add_smops(driver, mod);
		if (result < 0)
			goto err0;
		
		/* we need to free old infos before init the shadow */
		if (nooks_del_from_hash(
					&(driver->nook->shadow->kernel_objects),
					module_data_e,
					old_mod,
					&data) >= 0)
			vfree(data);
		
#ifdef CONFIG_SWIFI
		if (nooks_del_from_hash(
				&(driver->nook->shadow->kernel_objects),
				module_code_e,
				old_mod,
				&data) >= 0)
			vfree(data);
#endif
		result = shadow_init_shadow_module(mod, driver);
		
		if (result < 0)
			goto err0;
	      }
	
  	/* Stupid owners field to update : is that good ?
	 * Anyway, even if it's not good, shadow replay make things okay
	 */
	    case fops_e:
		// BUGBUG : is there another important owner field to update ?
		if (driver->wrapper_operations.fops.owner == old_mod)
			driver->wrapper_operations.fops.owner = mod;
		break;
	    case bdops_e:
		if (driver->wrapper_operations.bdops.owner == old_mod)
			driver->wrapper_operations.bdops.owner = mod;
		break;
	    default:
		break;
	}
  }
  result = 0;
  
err0:
  NOOKS_EXIT();
  return result;
}

int replug_replace_nook(struct module *mod)
{
	NOOKS_ENTER();
	/* just replace all the pointers of the nook */
	/* quick say for just ;-) it's the critical part */
	
	mod->new_version->nook = mod->nook;
	mod->new_version->nook_flags = mod->nook_flags;
	//mod->new_version->uc = mod->uc;
	/* mod->new_version->flags = (mod->new_version->flags && MOD_AUTOCLEAN)
		? (mod->flags | MOD_AUTOCLEAN)
		: (mod->flags & (~MOD_AUTOCLEAN));
		FIXME This will need a relook as per new module flags.
	*/
	// BUG(S BUNNY) what other thing we need to copy ?
	
	return replug_replace_op(mod->new_version, mod);

	NOOKS_EXIT();
}

#ifdef CONFIG_NOOKS_PROC
#define set_last_replug(m, s) m->last_replug = s
#else
#define set_last_replug(m,s)
#endif

int replug_replace_module_in_list(struct module *mod)
{
	struct module *m = (mod == NULL) ? NULL : mod->new_version;
	struct module *m1;
	struct module **r;
	struct module_ref *dep, *ref;
	int i;
	int error = -EBUSY; // EBUSY is to try to reload the previous mod
	int rollingback = 1;
	NOOKS_ENTER();
	
	/* some stupid test */
	if (mod == NULL) {
		printk(KERN_ERR "Nooks BUG\n");
		return -EINVAL;
	}
	if (m == NULL) {
		printk(KERN_ERR "Nooks BUG : no replugging module for module %s\n", mod->name);
		set_last_replug(mod, "replugging module not found");
		return -EINVAL;
	}
	if (!(mod->mod_nook_replug_flag & MOD_NOOK_REPLUG_OLD)) {
		rollingback = replug_save_old_version(mod);

		// Dependencies checks
		if (m->nrefs_replugged) {
			printk(KERN_DEBUG "Dependencies lack, not replugging current module %s\n", mod->name);
			set_last_replug(mod, "dependencies lack (ref)");
			return -EINVAL;
		}
	/* TBD FIXME
		for (dep = mod->deps, i = 0; i < mod->ndeps; i++, dep++) {
			if (dep->dep_replugged) {
				printk(KERN_DEBUG "Dependencies lack, not replugging current module %s\n", mod->name);
				set_last_replug(mod, "dependencies lack (dep)");
				return -EINVAL;
			}
		}
	*/
	}
	/*
	for (dep = mod->refs; dep; dep = dep->next_ref)
		dep->dep_replugged = 0;

	FIXME */	
	lock_kernel(); // Is that usefull now ?
	
	set_last_replug(mod, "unrecoverable error");
	set_last_replug(m, "unrecoverable error");
	/* First, replace that in the module list */
	//m->next = mod->next; FIXME


	m1 = find_module_by_ptr(mod);
	//FIXME do a struct copy to copy m inside m1
	if (m1 == NULL)
		printk ("Nooks BUG : unknow module %s to replug\n", mod->name);	
	/* FIXME Write code to do the below thing.
	if (mod == module_list) 
		r = &module_list;
	else {
		for (m1 = module_list; m1->next && (m1->next != mod); 
				m1 = m1->next);
		
		if (m1->next == NULL) {			
			printk(KERN_ERR "Nooks BUG : unknow module %s to replug\n", mod->name);
			return -ENOENT;
		}
		r = &(m1->next);
	}
	*r = m;
	// remove the replugging flag 
	m->flags &= ~MOD_NOOK_REPLUG;
	*/
	/* 
	 * now we are like after a create_module
	 * we can try the init_module missing part
	 * come mostly from init_module
	 */
	
        /* Sanity check the module's dependents */
#if 0
	FIXME
        for (i = 0, dep = m->deps; i < m->ndeps; ++i, ++dep) {
                struct module *o, *d = dep->dep;

		dep->dep_replugged = 0;

                /* Make sure the indicated dependencies are really modules.  */
                if (d == m) {
                        printk(KERN_ERR "nooks_replace_module_in_list: self-referential "
                                        "dependency in mod->deps.\n");
                        goto err3;
	        }

                /* Scan the current modules for this dependency */
                for (o = module_list; o != &kernel_module && o != d; o = o->next)
                        ;

                if (o != d) {
                        printk(KERN_ERR "init_module: found dependency that is "
                                "(no longer?) a module.\n");
                        goto err3;
                }
        }

        /* Update module references.  */
        for (i = 0, dep = m->deps; i < m->ndeps; ++i, ++dep) {
                struct module *d = dep->dep;

                dep->ref = m;
                dep->next_ref = d->refs;
                d->refs = dep;
                /* Being referenced by a dependent module counts as a
		   use as far as kmod is concerned.  */
                d->flags |= MOD_USED_ONCE;
	}
#endif
	/* now we have to get the nook for the module and to modify it */
	error = replug_replace_nook(mod);
	if(error < 0)
		goto err3;

#if 0 //FIXME delete module and its dependencies
	/* let's delete the module and his dependencies/refs */
        for (i = 0, dep = mod->deps; i < mod->ndeps; ++i, ++dep) {
		if (dep->dep) {
			if (dep->dep->refs == dep)
				dep->dep->refs = dep->dep->refs->next_ref;
			else if (dep->dep->refs) {
				for (ref = dep->dep->refs; ref->next_ref && 
						(ref->next_ref != dep);
						ref = ref->next_ref);
				if (ref->next_ref)
					ref->next_ref = ref->next_ref->next_ref;
			}
		}
	}

	if (rollingback) {
		m->new_version = NULL;
		for (ref = mod->refs; ref; ref = ref->next_ref)
			ref->dep = NULL;
		vfree(mod);
		m->flags &= ~MOD_NOOK_REPLUG_OLD;
	}
	else {
		mod->refs = NULL;
		m->flags |= MOD_NOOK_REPLUG_OLD;
		m->new_version = mod;
	}
#endif
#ifdef CONFIG_NOOKS_PROC
	do_gettimeofday(&(m->nook->last_replug_time));
	if(rollingback > 0)
		set_last_replug(m, "rolling back to old version (incompatibility)");
	else
		set_last_replug(m, "succeed");
#endif
	
	/* ok all done, let's the ends of the recovery process recall
	 * all the stuff like the init function
	 */

	error = 0;
	goto err0;
	

err3:   /* we just failed to replug the module so we try to reload the old one */
	*r = mod;
	/* and we delete the replugging module */
	mod->new_version = NULL;
	vfree(m);
err0:
	unlock_kernel();
	NOOKS_EXIT();
	return error;
}

int replug_save_old_version(struct module *mod)
{
	pnook_t nook = mod->nook;
	void *module_data = NULL;
	struct list_head *next;
	pnook_driver_t driver;
	int i;

	// restore the old datas
	i = nooks_find_in_hash(&nook->shadow->kernel_objects,
						module_data_e,
						mod,
						&module_data);
	if (i >= 0) {
		memcpy(mod->writeable_start,
			module_data,
			mod->writeable_end - mod->writeable_start);
        } else {
		// BUGBUG: what do we do here?
		printk("In replug_save_old_version: BUGBUGBUGBUG: couldn't find module\n");
		return -1;
        }
	// restore the old module operations pointers
	list_for_each(next, &(nook->shadow->drivers)) {
		driver = list_entry(next, nook_driver_t, next_in_shadow);
		if ((driver->driver_type == mops_e)
	             && (!strcmp(mod->name, driver->name))) {
			mod->init       = driver->operations.mops.init;
			mod->exit    = driver->operations.mops.exit;;
			//mod->can_unload = driver->operations.mops.can_unload;
			return 0;
		}
	}
	return -1;
}

extern struct list_head super_blocks;
extern spinlock_t sb_lock;

#if 0 //Seems not required for us
int replug_find_files(struct file_operations * fops)
{
	struct list_head * next, *spare;
	spin_lock(&sb_lock);

	list_for_each_safe(next, spare, &super_blocks) {
		struct super_block *sb = sb_entry(next);
		struct list_head *p;

		file_list_lock();
		for (p = sb->s_files.next; p != &sb->s_files; p = p->next) {
			struct file *file = list_entry(p, struct file, fu_list);
			struct inode *inode = file->f_dentry->d_inode;
			if (fops != &NOOKS_NWRAPPER(fops))
				if((file->f_op == fops) 
					|| (inode->i_fop == fops)) {
					file_list_unlock();
					spin_unlock(&sb_lock);
					return 1;
				}
		}
		file_list_unlock();
	}
	spin_unlock(&sb_lock);
	return 0;
}
#endif

int replug_compatibility_resource_used(pnook_t nook, pnook_driver_t driver)
{
	pnook_driver_t nd;
	struct list_head *next;
	if (nook->recovery_state.info.recover == NULL) return 1;
	if (driver->driver_type == pciops_e) return 1;
	list_for_each(next, &(nook->drivers)) {
		nd = list_entry(next, nook_driver_t, next_in_nook);
		if (nd == driver)
			return 1;
	}
	if (driver->driver_type != fops_e)
		return 0;
	return 0;
	//return !replug_find_files(&driver->wrapper_operations.fops);
}
